/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2013. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.filesystem.cli.core;

import java.io.File;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.MultiStatus;

import com.ibm.team.filesystem.cli.core.subcommands.CommonOptions;
import com.ibm.team.filesystem.cli.core.Constants;
import com.ibm.team.filesystem.cli.core.subcommands.IScmClientConfiguration;
import com.ibm.team.filesystem.cli.core.subcommands.IScmCommandLineArgument;
import com.ibm.team.filesystem.cli.core.internal.Activator;
import com.ibm.team.filesystem.cli.core.internal.ScmCommandLineArgument;
import com.ibm.team.filesystem.cli.core.util.CLIFileSystemClientException;
import com.ibm.team.filesystem.cli.core.internal.Messages;
import com.ibm.team.filesystem.cli.core.util.RepoUtil;
import com.ibm.team.filesystem.cli.core.util.StatusHelper;
import com.ibm.team.filesystem.cli.core.util.SubcommandUtil;
import com.ibm.team.filesystem.cli.core.subcommands.IUuidAliasRegistry.IUuidAlias;
import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.internal.FileSystemStatusUtil;
import com.ibm.team.rtc.cli.infrastructure.internal.core.CLIClientException;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IClientConfiguration;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.ICommandLine;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.IOptionKey;

/** Abstract base for subcommands. */
public abstract class AbstractSubcommand extends com.ibm.team.rtc.cli.infrastructure.internal.core.AbstractSubcommand {
    protected IScmClientConfiguration config = null;
    
    public abstract void run() throws FileSystemException;
    
    public int run(IClientConfiguration config) throws CLIClientException {
        this.config = (IScmClientConfiguration) config;
        processArguments();        
        
        try {
            run();
        } catch (Exception e) {
            throw StatusHelper.toCLIClientException(e);
        }

        return Constants.STATUS_OK;
    }
    
    private void processArguments() throws CLIClientException {
        try {
            // infer repos
            ICommandLine cmdLine = config.getSubcommandCommandLine();
            
            Set<String> repos = new HashSet<String>();
            List<IScmCommandLineArgument> argsWithoutRepo = new ArrayList<IScmCommandLineArgument>();
            
            for (IOptionKey option : cmdLine.getDefinedOpts()) {
                if (null != option.getDelimiter()) {
                    List<IScmCommandLineArgument> args = ScmCommandLineArgument.createList(cmdLine.getOptionValues(option), config);
                    if (args != null) {
                        for (IScmCommandLineArgument arg : args) {
                            if (null != arg.getRepositorySelector()) {
                                repos.add(arg.getRepositorySelector());
                            } else {
                                argsWithoutRepo.add(arg);
                            }
                        }
                    }
                }
            }
            
            // if no "-r" option specified and we encountered some selectors without any
            // repos being associated with them, then we need to infer repos from the sandbox
            // for these selectors
            if (!cmdLine.hasOption(CommonOptions.OPT_URI) && argsWithoutRepo.size() > 0) {                
                // check if this command is being run inside of CFA
                File cfaRootPath = SubcommandUtil.findAncestorCFARoot(config.getContext().getCurrentWorkingDirectory());
                
                // cfa is null && we have selectors without repos
                if (cfaRootPath == null) {                    
                    // if repos < 1, atleast one repo is needed to infer repos for args with absent repos
                    
                    // check for alias 
                    boolean isAlias = isAlias(argsWithoutRepo);
                    
                    // if no explicit repos were defined and we found selector names eg: ws1
                    // then we dont know which repo to infer it from
                    // probably the user has similar name in multiple repos
                    // so error out
                    if (repos.size() == 0 && !isAlias) {
                        throw StatusHelper.misconfiguredLocalFS(Messages.SubcommandDefinition_7);
                        
                      // repos > 1, then an ambiguous case
                      // eg:  lscm ws ac w1@r1 c1@r2 c3
                      // error out as repos r1 & r2 are present 
                    } else if (repos.size() > 1 && !isAlias) {
                        throw StatusHelper.misconfiguredLocalFS(Messages.SubcommandDefinition_MULTIPLE_REPOS);
                    } else if (repos.size() > 0) { // we need a single repo to infer                            
                        Iterator<String> it = repos.iterator();
                        String repo = it.next();                        
                        
                        // infer repos for remaining args
                        for (IScmCommandLineArgument arg : argsWithoutRepo) {
                            
                            String selector = arg.getItemSelector();
                            try {
                                Integer.parseInt(selector);
                            
                                // check for alias
                                IUuidAlias uuid = RepoUtil.lookupUuidAndAlias(selector);
                                
                                // if not an alias, probably a selector which is numeric
                                // eg: ws name "1917" which is numeric, assoc the repo with it
                                if (uuid == null || (uuid != null && uuid.getUuid() == null)) {
                                    
                                    if (repos.size() > 1) {
                                        throw StatusHelper.misconfiguredLocalFS(Messages.SubcommandDefinition_MULTIPLE_REPOS);
                                    }
                                    
                                    ((ScmCommandLineArgument) arg).setRepositorySelector(repo);
                                    
                                }  // else the alias matched in the registry
                                   // dont have to associate repo with the alias.
                                   // alias has info as to which repo it belongs to.
                                 
                                
                            } catch (NumberFormatException e) { // a name or a UUID, assoc repo with it.
                                ((ScmCommandLineArgument) arg).setRepositorySelector(repo);
                            }
                        }
                    }
                }
            }
        } catch (CLIFileSystemClientException e) {
            throw new CLIClientException(e.getStatus());
        } catch (FileSystemException e) {
            IStatus status = new MultiStatus(Activator.PLUGIN_ID, Constants.STATUS_INTERNAL_ERROR,
                    new IStatus[] {FileSystemStatusUtil.getStatusFor(e)}, Messages.Application_23, e);
            throw new CLIClientException(status);            
        }
    }

    private boolean isAlias(List<IScmCommandLineArgument> argsWithoutRepo) {        
        // check for a alphanumeric selector eg: ws1
        for (IScmCommandLineArgument arg : argsWithoutRepo) {
            try {
                Integer.parseInt(arg.getItemSelector());            
                
            } catch (NumberFormatException e) {
                return false;
            }
        }
        return true;
    }
}

