/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2005, 2016. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.filesystem.cli.core;

import java.net.SocketTimeoutException;
import java.util.Locale;

import org.eclipse.core.resources.IResourceStatus;
import org.eclipse.equinox.app.IApplication;

/**
 * Constants used by the CLI.
 */
public class Constants {

	/** Set to <code>true</code> when running on MS Windows, <code>false</code>
	 * otherwise.  
	 */
	public final static boolean ON_WINDOWS = System.getProperty("os.name").toLowerCase(Locale.ENGLISH).contains("windows");  //$NON-NLS-1$//$NON-NLS-2$

	/**
	 * Valid names of the jazz metadata directory. In order of preference.
	 */
	public final static String[] METADATA_ROOT_NAMES = {".jazz5", ".jazz4", ".jazz3"}; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
	
	/**
	 * Indicate that the app ran properly. 
	 */
	public final static Integer STATUS_OK = IApplication.EXIT_OK; 
	
    /** 
     * Indicates that there was a syntax error on the command-line. The user 
     * has provided the wrong number of required arguments, an argument that 
     * the given subcommand doesn't understand, or an argument that doesn't 
     * make sense in the given context. 
     * 
     * <p>
     * The user should modify the arguments and re-run.
     * 
     * @core
     */
    public final static int STATUS_ARGUMENT_SYNTAX_ERROR = 1;

    /**
	 * Indicates that the client couldn't log in with the given credentials. 
	 * 
	 * <p>
	 * The user should verify the credentials, correct any problems, and try 
	 * again.
	 * 
	 * @client
	 */
	public static final int STATUS_CONNECT_LOGIN_FAILURE = 2;


	/** Indicates that a failure occurred for an unknown reason.
	 * 
	 * @core
	 */
	public static final int STATUS_FAILURE = 3;


	/** An internal error indicating that the client was not properly 
	 * configured, specifically a subcommand couldn't be loaded, likely due 
	 * to a configuration error by the programmer.
	 * 
	 * @core
	 */
	public static final int STATUS_CONFIG_FAILURE = 4;
	
	
	/**
	 * Indicates that the local filesystem is misconfigured. At least one of 
	 * the following is true: 
	 * <ul>
	 * 		<li>a CFA root could not be found,
	 * 		<li>an item that the client expects to be shared is not shared,
	 *		<li>the configuration directory (usually ~/.jazz-scm) cannot 
	 *				be written to.  
	 * </ul>
	 * 
	 * @client
	 */
	public static final int STATUS_LOCAL_FS_MISCONFIGURED = 5;

	
	/** Indicate that a resource exists on the filesystem, preventing an 
	 * update/load from creating a new resource. 
	 * 
	 * @client
	 */
	public static final int STATUS_CFA_COLLISION = 6;
	
	
	/** An error of last resort. Indicates that condition that should have 
	 * been handled was not handled. A more severe version of 
	 * {@link #STATUS_FAILURE}.
	 * 
	 *  @core
	 */
	public static final int STATUS_INTERNAL_ERROR = 7;
	
	
	/** Indicate that an operation could not be completed because the filesystem 
	 * was out of sync with the remote workspace. The user should reload or 
	 * reshare.
	 * 
	 * @client
	 */
	public static final int STATUS_OUT_OF_SYNC = 8;

	
	/**
	 * Indicates that an operation could not be completed because a selector 
	 * supplied by the user either does not match enough items, or matches too 
	 * many. The user should determine which selector was at fault, and use 
	 * an alias or UUID to non-ambiguously refer to the requested items. 
	 * 
	 * @client
	 */
	public static final int STATUS_AMBIGUOUS_SELECTOR = 9;
	

	// 10 IS NOT USED. IT'S UP FOR RECLEMATION. 
	
	
	/**
	 * Indicates that there is a conflict in one or more remote workspaces 
	 * touched by the subcommand. The user should resolve those conflicts. 
	 * 
	 * @client
	 */
	public static final int STATUS_CONFLICT = 11;
	
	
	// NOTE! STATUS 13 IS USED BY OSGI TO INDICATE AN EXCEPTION. DO NOT USE.
	
	
	/** Indicates that move failed because the target exists, or the target 
	 * path is invalid. The user should correct the incorrect path(s).
	 * 
	 * @client
	 *
	 * @see IResourceStatus#PATH_OCCUPIED
	 * @see IResourceStatus#RESOURCE_EXISTS
	 */
	public static final int STATUS_MOVE_FAILED = 14;

	
	/**
	 * Indicates data changed remotely during execution. The command should be 
	 * rerun.
	 * 
	 * @client
	 */
	public static final int STATUS_STALE_DATA = 15;

	
	/**
	 * Indicates that the command couldn't complete because it would introduce 
	 * a gap into the change set history.
	 * 
	 * @client
	 */
	public static final int STATUS_GAP = 16;

	/**
	 * Indicates that an operation was blocked by process. The user should 
	 * follow the advice given to allow the operation to complete properly.
	 * 
	 * @client
	 */
	public static final int STATUS_PROCESS = 17;

	/**
	 * Indicates that an operation failed due to a permissions failure. The 
	 * message returned explains the nature of the failure. 
	 * 
	 * @client
	 */
	public static final int STATUS_PERMISSION_FAILURE = 18;

	/**
	 * The requested operation could not complete because preconditions were 
	 * not met. The user is attempting to do something that doesn't really make
	 * sense. 
	 * 
	 * @client
	 */
	public static final int STATUS_DISALLOWED = 19;


	/**
	 * Attempted to transfer a file that has improper line terminators or 
	 * encoding.
	 * 
	 * @client
	 */
    public static final int STATUS_MALFORMED_INPUT = 20;
    
    
	/**
	 * Indicates that the command couldn't complete because it would create an
	 * n-way conflict in the change set history.
	 * 
	 * @client
	 */
	public static final int STATUS_NWAY_CONFLICT = 21;
	
	/**
	 * Indicates that the command couldn't complete because it would create one 
	 * or more items that don't have a parent. 
	 * 
	 * @client
	 */
	public static final int STATUS_ORPHAN = 22;
	
	// NOTE! STATUS 23 IS USED BY OSGI TO INDICATE A FORCED APPLICATION RESTART. DO NOT USE.
	// NOTE! STATUS 24 IS USED BY OSGI TO INDICATE [SOMETHING]. DO NOT USE.
	
	/**
	 * Indicates a requested repository item was not found. The user should 
	 * verify that they have the correct selector/alias/UUID for the item they 
	 * want to deal with, and that the selector/alias are still valid (ie, 
	 * the name of the item hasn't changed since the last invocation or the 
	 * alias hasn't been aged out).
	 * 
	 * @client
	 */
	public static final int STATUS_ITEM_NOT_FOUND = 25;


	/**
	 * Indicates that something went wrong on the server. The user should 
	 * contact the server's administrator. 
	 * 
	 * @client
	 */
	public static final int STATUS_SERVER_ERROR = 26;

	
	/**
	 * Indicates that this CFA is in use by another RCP process (either 
	 * an eclipse GUI, or another CLI instance). The user should find that 
	 * process and terminate it. 
	 * 
	 * @client
	 */
	public static final int STATUS_WORKSPACE_IN_USE = 27;

	/**
     * Indicates that the remote filesystem is misconfigured: the client 
     * couldn't find an item in the remote filesystem. The user should verify 
     * the path to the given item, and ensure that the remote structure hasn't 
     * changed since the last invocation of the client.
     * 
     * @client
     */
    public static final int STATUS_REMOTE_FS_MISCONFIGURED = 28;

    /**
     * Indicates that an operation could not complete because of a license
     * limitation. The user should upgrade their license. 
     * 
     * @client
     */
    public static final int STATUS_LICENSE_DISALLOWED = 29;
	
    
    /**
     * Indicates that an operation could not complete because an argument is 
     * valid (it's type is okay, and the item could be found), but the item's 
     * state is inappropriate for the request.
     */
    public static final int STATUS_INAPPROPRIATE_ARGUMENT = 30;
    
	
    /**
     * Indicates that the filesystem daemon didn't receive an initial connection 
     * within the alloted amount of time. 
     */
    public static final int STATUS_INITIAL_CONNECTION_TIMEOUT = 31;
    
    
    /**
     * Indicates that communication with the repository timed out with a 
     * {@link SocketTimeoutException}. 
     */
    public static final int STATUS_REPOSITORY_TIMEOUT = 32;
    
    /**
     * Indicates that only some properties could be fetched
     * for the specified items.
     */
    public static final int STATUS_PROPERTIES_UNAVAILABLE = 33;
    

    /**
     * Indicates that an operation cannot continue because there
     * are files that are not checked in
     */
    public static final int STATUS_UNCHECKED_IN_CHANGE = 34;
    
    /**
     * Indicates that the property could not be set
     * for the specified file.
     */
    public static final int STATUS_PROPERTY_INVALID = 35;
    
    /**
     * Indicates that the repository information was incorrectly specified
     * with respect to an item.
     */
    public static final int STATUS_REPO_INCORRECTLY_SPECIFIED = 36;

    /**
     * Indicates that the change set cannot be completed.
     */
    public static final int STATUS_CHANGESET_INCOMPLETE = 37;
    
    /**
     * Indicates that the suspend is illegal. This happens when suspend cmd is executed on the same changeset twice.
     */
    public static final int STATUS_ILLEGAL_SUSPEND = 38;

    /**
     * Indicates IO exception.
     */
    public static final int STATUS_IO_EXCEPTION = 39;
    
    /**
     * Indicates that the changeset is not in history. Might have already been discarded.
     */    
    public static final int STATUS_DISCARD_NOT_IN_HISTORY = 40;
    
    /**
     * Indicates that there was a service failure when communicating with the repository
     */
    public static final int STATUS_SERVICE_FAILURE = 41;
    
    /**
     * Indicates that the mime type could not be set
     * for the specified file.
     */
    public static final int STATUS_MIME_TYPE_INVALID = 42;
    
    /**
     * Only used in the FrontEndClient. Indicates that the daemon closed the 
     * stderro before printing the connection coordinates. 
     * 
     * Note that it's used as a literal.
     */
    public static final int STATUS_DAEMON_FAILED_TO_START = 43;
    
   	/**
     * Indicates that the specified load rules cannot be processed because it has inaccessible items
     * or has invalid load rules. 
     */
    public static final int STATUS_LOAD_RULES_FAILURE = 44;
    
    /**
     * Indicates that delivering changes would create conflicts in the stream.
     */
    public static final int STATUS_DELIVERY_REQUIRES_HISTORY_REORDERING = 45;
    
    /**
     * Indicates that the requested state of the versionable is deleted.
     */
    public static final int STATUS_VERSIONED_CONTENT_DELETED = 46;

	/**
	 * Indicates that we could not determine the repository URI for the given
	 * repository id.
	 */
    public static final int STATUS_REPOSITORY_LOOKUP_FAILED = 47;

	/**
	 * Indicates that a server state mismatch has been detected. This can happen
	 * when we try to connect to the server from a daemon which has been active
	 * through the server rename process.
	 */
    public static final int STATUS_SERVER_STATE_MISMATCH = 48;
    
    /**
     * Indicates the FEC to rerun the command;
     */
    public static final int STATUS_REQUEST_RERUN = 49;
    
    /**
     * Indicates that one or more items are locked in the stream
     */
    public static final int STATUS_ITEM_LOCKED_IN_STREAM = 50;
    
    /**
     * Indicates that we have hit an unexpected condition
     */
    public static final int STATUS_UNEXPECTED_FAILURE = 51;

    /**
     * Indicates that the operation completed normally but
     * resulted in no workspace changes.
     */
    public static final int STATUS_WORKSPACE_UNCHANGED = 52;
    
    /**
     * Indicates that the operation completed normally but
     * there was nothing to deliver
     */
    public static final int STATUS_NOTHING_TO_DELIVER = 53;

	/**
	 * Indicates that some of the loaded workspaces are unreachable, hence the
	 * status command output does not contain the status of all the workspaces
	 * in the sandbox. Indicates that not all the items/arguments were
	 * successfully processed.
	 * 
	 * Note: In RTC 5.0, this field was renamed from
	 * "STATUS_PARTIAL_STATUS_OUTPUT".
	 * 
	 * @since 0.10.0 (RTC 5.0)
	 */
	public static final int STATUS_PARTIAL_SUCCESS = 54;

	/**
	 * Indicates that there is a 'port' in progress (ex: In the UI, the
	 * decorator on a component node would show "Pending Gap Resolution")
	 * 
	 * @since 0.9.0.5 (RTC 4.0.5)
	 */
	public static final int STATUS_PORTS_IN_PROGRESS = 55;

	/**
	 * Indicates that some of change sets into subcomponents are active and the
	 * option to close them was not specified.
	 *
	 * @since 0.11.0 (RTC 6.0)
	 */
	public static final int STATUS_ACTIVE_CHANGE_SETS = 56;

	/**
	 * Indicates that some hierarchies are not entirely in the snapshot.
	 *
	 * @since 0.11.0 (RTC 6.0)
	 */
	public static final int STATUS_ENTIRE_HIERARCHY_IN_SNAPSHOT = 57;

	/**
	 * Indicates that some subcomponents selected have conflicting flow targets
	 *
	 * @since 0.11.0 (RTC 6.0)
	 */
	public static final int STATUS_COMPONENTS_HAVE_CONFLICTING_TARGETS = 58;

	/**
	 * Indicates that some subcomponents selected in an accept affect other
	 * components not in the selection
	 *
	 * @since 0.11.0 (RTC 6.0)
	 */
	public static final int STATUS_COMPONENTS_HAVE_MULTIPLE_PARTICIPANTS = 59;

	/**
	 * Indicates that no backup baseline is being created during a component
	 * replacement or removal.
	 *
	 * @since 0.11.0 (RTC 6.0)
	 */
	public static final int STATUS_NO_BACKUP_BASELINE = 60;

	/**
	 * Indicates that one or more components belong to multiple hierarchies.
	 *
	 * @since 0.11.0 (RTC 6.0)
	 */
	public static final int STATUS_COMPONENTS_IN_MULTIPLE_HIERARCHIES = 61;

	/**
	 * Indicates that there are some local conflicts that need to be handled
	 * before the operation runs.
	 *
	 * @since 0.11.0.1 (RTC 6.0.1)
	 */
	public static final int STATUS_LOCAL_CONFLICTS = 62;

	/**
	 * Indicates that an attempt was made to lock a file when there are incoming
	 * change sets.
	 *
	 * @since 0.11.0.1 (RTC 6.0.1)
	 */
	public static final int STATUS_CONFLICT_WITH_INCOMING_CHANGE = 63;

	/**
	 * Indicates that one or more components are inaccessible and not in the
	 * seed workspace/stream.
	 *
	 * @since 0.11.0.1 (RTC 6.0.1)
	 */
	public static final int STATUS_INACCESSIBLE_COMPONENTS_NOT_IN_SEED = 64;
	
	/**
	 * Indicates that the stream is locked.
	 *
	 * @since 0.11.0.1 (RTC 6.0.1)
	 */
	public static final int STATUS_STREAM_LOCKED = 65;

	/**
	 * Indicates that an attempt to resolve a conflict was made while there are 
	 * some unresolved dependent changes present.
	 *
	 */
	public static final int STATUS_UNRESOLVED_DEPENDENT_CHANGES = 66;

	/**
	 * Indicates that there is no conflict in local change set
	 */
	public static final String HAS_NO_CONFLICT = "-"; //$NON-NLS-1$

	/**
	 * Indicates that there is/are conflict(s) in local change set
	 */
	public static final String HAS_CONFLICT = "#"; //$NON-NLS-1$

	/**
	 * Indicates that there is no remote changesets which has conflict with local changeset
	 */
	public static final String HAS_NO_COLLISION = "-"; //$NON-NLS-1$

	/**
	 * Indicates that there is/are remote changesets which has conflict(s) with local changesets 
	 */
	public static final String HAS_COLLISION = "!"; //$NON-NLS-1$

	/**
	 * Indicates that the changeset is active
	 */
	public static final String IS_CS_ACTIVE = "@"; //$NON-NLS-1$

	/**
	 * Indicates that the changeset is not active
	 */
	public static final String IS_CS_INACTIVE = "$"; //$NON-NLS-1$
	
	/** 
	 * Test whether or not a stack trace should be shown for the given error
	 * state. 
	 */
	public static boolean showStackTraceFor(int code) {
		return code == STATUS_INTERNAL_ERROR || code == STATUS_FAILURE;
	}

	/** Use this to inform the severity of the status message in the json output */
	public static enum Severity {error, warning, info, success};
	
	/** Name of the extension point used to add subcommands. */
	public static final String PT_SUBCOMMAND = "subcommand"; //$NON-NLS-1$

	/** Name of the "name" attribute. Human-meaningful name of subcommands. */
	public static final String PT_ATTR_NAME = "name"; //$NON-NLS-1$

	/** Name of the attribute providing a short, human-meaningful explanation of 
	 * the subcommand.
	 */
	public static final String PT_ATTR_SHORT_HELP = "shortHelp"; //$NON-NLS-1$

	/**
	 * Name of the attribute providing a long, human meaningful explanation of 
	 * subcommand.
	 */
	public static final String PT_ATTR_LONG_HELP = "longHelp"; //$NON-NLS-1$
	
	/** Name of the class providing the behaviour for the given subcommand. */
	public static final String PT_ATTR_CLASS = "class"; //$NON-NLS-1$

    /** Name of the Eclipse workspace initialization policy. */
	public static final String PT_ATTR_INIT_POLICY = "initpolicy"; //$NON-NLS-1$

	/** Name of the class providing the options for a subcommand. */
	public static final String PT_ATTR_OPTIONS = "options"; //$NON-NLS-1$

	/** Alternate names for the subcommand. Space separated list. */
	public static final String PT_ATTR_ALIASES = "aliases"; //$NON-NLS-1$

	/** Unique ID for the subcommand. */
    public static final String PT_ATTR_ID = "id"; //$NON-NLS-1$
	
	/** The ID of the parent. 
	 */
	public static final String PT_ATTR_PARENT = "parent"; //$NON-NLS-1$

	/**
	 * Needed for ISubcommandDefinition#showInHelpListing()
	 * @deprecated
	 */
	public static final String PT_ATTR_HIDE = "hideFromHelp"; //$NON-NLS-1$

	/**
	 * A grouping for the subcommand. 
	 */
	public static final String PT_ATTR_SCOPE = "scope"; //$NON-NLS-1$
	
	/**
	 * The "default" group that subcommands can fall into. 
	 * @see #PT_ATTR_SCOPE
	 */
	public static final String PT_ATTR_SCOPE_DEFAULT = "default"; //$NON-NLS-1$
	
	public static final String JAZZSCM_FORCE_GLOBBING = "JAZZSCM_FORCE_GLOBBING"; //$NON-NLS-1$
	
	public static final String JAZZSCM_FORCE_GLOB_VALUE = "1"; //$NON-NLS-1$
	
	// any string after this token is considered as repo in distributed SCM. 
	// basically used to specify the http urls following this token.
	public static final String URI_STR = "@uri:"; //$NON-NLS-1$

	// any string after this token is considered as repo in distributed SCM 
	public static final String AT_STR = "@";  //$NON-NLS-1$
	
	/** Name of the extension point used to provide information for customizing the product. */
	public static final String PT_CUSTOMIZE_PRODUCT = "customizeProduct"; //$NON-NLS-1$
	public static final String PT_ATTR_PRODUCT_NAME = "appName"; //$NON-NLS-1$
	
	public static final int MAX_CHANGES_DEFAULT = -1;
	
    public static final String WORKSPACE = "workspace"; //$NON-NLS-1$
    public static final String STREAM = "stream"; //$NON-NLS-1$
    public static final String COMPONENT = "component"; //$NON-NLS-1$
    public static final String CHANGESET = "changeset"; //$NON-NLS-1$
    public static final String CHANGE = "change"; //$NON-NLS-1$
    public static final String SNAPSHOT = "snapshot"; //$NON-NLS-1$
    public static final String BASELINE = "baseline"; //$NON-NLS-1$
    public static final String VERSIONABLE = "versionable"; //$NON-NLS-1$
    public static final String CONTRIBUTOR = "contributor"; //$NON-NLS-1$
    public static final String PROJECTAREA = "projectarea"; //$NON-NLS-1$
    public static final String TEAMAREA = "teamarea"; //$NON-NLS-1$
    public static final String WORKITEM = "workitem"; //$NON-NLS-1$
    public static final String ACCESSGROUP = "accessgroup"; //$NON-NLS-1$
}
