/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2005, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.filesystem.cli.core.internal;

import static com.ibm.team.filesystem.client.internal.daemon.trace.DaemonTraceConfigurator.DEFAULT_MAX_TRACE_DIR_SIZE;
import static com.ibm.team.filesystem.client.internal.daemon.trace.DaemonTraceConfigurator.DEFAULT_MAX_TRACE_FILE_SIZE;
import static com.ibm.team.filesystem.client.internal.daemon.trace.DaemonTraceConfigurator.DEFAULT_TRACE_COMPRESS;
import static com.ibm.team.filesystem.client.internal.daemon.trace.DaemonTraceConfigurator.DEFAULT_TRACE_DISABLE;
import static com.ibm.team.filesystem.client.internal.daemon.trace.DaemonTraceConfigurator.PREF_MAX_TRACE_DIR_SIZE;
import static com.ibm.team.filesystem.client.internal.daemon.trace.DaemonTraceConfigurator.PREF_MAX_TRACE_FILE_SIZE;
import static com.ibm.team.filesystem.client.internal.daemon.trace.DaemonTraceConfigurator.PREF_TRACE_COMPRESS;
import static com.ibm.team.filesystem.client.internal.daemon.trace.DaemonTraceConfigurator.PREF_TRACE_DIR;
import static com.ibm.team.filesystem.client.internal.daemon.trace.DaemonTraceConfigurator.PREF_TRACE_DISABLE;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Locale;
import java.util.Map;
import java.util.logging.Level;

import org.apache.commons.logging.Log;
import org.eclipse.osgi.util.NLS;

import com.ibm.team.filesystem.cli.core.internal.compare.CompareToolBeyondCompare;
import com.ibm.team.filesystem.cli.core.internal.compare.CompareToolClearCaseDiffMerge;
import com.ibm.team.filesystem.cli.core.internal.compare.CompareToolDiffMerge;
import com.ibm.team.filesystem.cli.core.internal.compare.CompareToolKDiff3;
import com.ibm.team.filesystem.cli.core.internal.compare.CompareToolP4Merge;
import com.ibm.team.filesystem.cli.core.internal.compare.CompareToolWinMerge;
import com.ibm.team.filesystem.cli.core.internal.compare.ExternalCompareToolsUtil;
import com.ibm.team.filesystem.cli.core.internal.compare.IExternalCompareTool;
import com.ibm.team.filesystem.cli.core.subcommands.ITypedPreferenceRegistry;
import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.internal.utils.PropertiesLoader;
import com.ibm.team.repository.common.LogFactory;
import com.ibm.team.repository.common.transport.TeamServerConfiguration;
import com.ibm.team.rtc.cli.infrastructure.internal.util.StringUtil;

@SuppressWarnings("restriction")
public class PreferenceRegistry implements ITypedPreferenceRegistry {
    private static final Log log = LogFactory.getLog(PreferenceRegistry.class.getName());
	
	public static final String PREF_REPO_CONNECTION_TIMEOUT = "repository.timeout"; //$NON-NLS-1$
    private static final String DEFAULT_REPO_CONNECTION_TIMEOUT = String
            .valueOf(TeamServerConfiguration.DEFAULT_SOCKET_TIMEOUT / 1000);

    public static final String PREF_MASK_CREDENTIALS = "password.mask"; //$NON-NLS-1$
    private static final boolean DEFAULT_MASK_CREDENTIALS = true;  
    
    public static final String PREF_CONTENT_TRANSFER_THREADS = "content.threads"; //$NON-NLS-1$
    
    /**
     * Deprecated don't use
     */
    public static final String PREF_AUTOREFRESH_POLICY = "local.collision.policy"; //$NON-NLS-1$
    
    public static final String PREF_COMMIT_ATOMIC_MAXIMUM = "commit.atomic.maximum"; //$NON-NLS-1$
    
    public static final String PREF_DUMP_ON_CONSOLE = "dump.exception"; //$NON-NLS-1$
    private static final boolean DEFAULT_DUMP_ON_CONSOLE = false;
    
    public static final String PREF_LOG_LEVEL = "log.level"; //$NON-NLS-1$
    public static final String DEFAULT_LOG_LEVEL = "SEVERE"; //$NON-NLS-1$
    
    public static final String PREF_JSON_OUTPUT = "json.output"; //$NON-NLS-1$
    public static final boolean DEFAULT_JSON_OUTPUT = false;
    
    public static final String PREF_REFRESH_LOCAL_CHANGES = "refresh.local.changes"; //$NON-NLS-1$
    private static final boolean DEFAULT_REFRESH_LOCAL_CHANGES = true;
    
    public static final String DATE_FORMAT_CHANGE = "date.format"; //$NON-NLS-1$
        
    /**
     * Autorefresh policy corresponding with always refreshing the filesystem 
     * before an op runs. 
     */
    public static final String AUTOREFRESH_POLICY_REFRESH = "refresh"; //$NON-NLS-1$

    private static final String NL = System.getProperty("line.separator"); //$NON-NLS-1$

    // Compare tool preferences key
    public static final String PREF_EXTERNAL_2WAY_COMPARE_TOOL = "external.compare.2way"; //$NON-NLS-1$
    public static final String PREF_EXTERNAL_3WAY_COMPARE_TOOL = "external.compare.3way"; //$NON-NLS-1$
    
    //auto-lock preferences key
    public static final String PREF_SUSPEND_RELEASE_LOCK = "suspend.release.locks"; //$NON-NLS-1$
    private static final boolean DEFAULT_SUSPEND_RELEASE_LOCK = false;
    
    public static final String PREF_UNDO_RELEASE_LOCK = "undo.release.locks"; //$NON-NLS-1$
    private static final boolean DEFAULT_UNDOD_RELEASE_LOCK = false;
    
    public static final String PREF_DISCARD_RELEASE_LOCK = "discard.release.locks"; //$NON-NLS-1$
    private static final boolean DEFAULT_DISCARD_RELEASE_LOCK = false;
    
    public static final String PREF_RESUME_ACQUIRE_LOCK = "resume.acquire.locks"; //$NON-NLS-1$
    private static final boolean DEFAULT_RESUME_ACQUIRE_LOCK = false;
    
    public static final String PREF_DELIVER_RELEASE_LOCK = "deliver.release.locks"; //$NON-NLS-1$
    private static final boolean DEFAULT_DELIVER_RELEASE_LOCK = false;

    private static Map<String, PreferenceModel> defaultPreferences = new HashMap<String, PreferenceModel>();
    {
        defaultPreferences.put(
                PREF_REPO_CONNECTION_TIMEOUT,
                new PreferenceModel(PREF_REPO_CONNECTION_TIMEOUT, DEFAULT_REPO_CONNECTION_TIMEOUT,
                        "seconds", NLS.bind(Messages.Application_CONCATENATE, //$NON-NLS-1$
                                Messages.PreferenceRegistry_5, Messages.PreferenceRegistry_6)));

        defaultPreferences.put(
                PREF_MASK_CREDENTIALS,
                new PreferenceModel(PREF_MASK_CREDENTIALS,
                        String.valueOf(DEFAULT_MASK_CREDENTIALS), "", NLS.bind( //$NON-NLS-1$
                                Messages.Application_CONCATENATE, Messages.PreferenceRegistry_9,
                                NLS.bind(Messages.PreferenceRegistry_10, "true")))); //$NON-NLS-1$

        defaultPreferences.put(
                PREF_CONTENT_TRANSFER_THREADS,
                new PreferenceModel(PREF_CONTENT_TRANSFER_THREADS, String
                        .valueOf(DEFAULT_CONTENT_TRANSFER_THREADS), "", NLS.bind( //$NON-NLS-1$
                        Messages.Application_CONCATENATE, Messages.PreferenceRegistry_13,
                        Messages.PreferenceRegistry_14)));

        defaultPreferences.put(
                PREF_COMMIT_ATOMIC_MAXIMUM,
                new PreferenceModel(PREF_COMMIT_ATOMIC_MAXIMUM, String
                        .valueOf(UNKNOWN_ATOMIC_COMMIT_MAXIMUM), "files", NLS.bind( //$NON-NLS-1$
                        Messages.Application_CONCATENATE, Messages.PreferenceRegistry_21,
                        Messages.PreferenceRegistry_22)));

        defaultPreferences.put(
                PREF_DUMP_ON_CONSOLE,
                new PreferenceModel(PREF_DUMP_ON_CONSOLE,
                        Boolean.toString(DEFAULT_DUMP_ON_CONSOLE), "", NLS.bind( //$NON-NLS-1$
                                Messages.Application_CONCATENATE, Messages.PreferenceRegistry_25,
                                NLS.bind(Messages.Application_CONCATENATE,
                                        Messages.PreferenceRegistry_26,
                                        Messages.PreferenceRegistry_27))));

        StringBuilder str = new StringBuilder(Messages.PreferenceRegistry_29);
        str.append(Messages.PreferenceRegistry_30);
        str.append(" "); //$NON-NLS-1$
        str.append(Messages.PreferenceRegistry_31);
        str.append(", "); //$NON-NLS-1$
        str.append(Messages.PreferenceRegistry_32);
        str.append(", "); //$NON-NLS-1$
        str.append(Messages.PreferenceRegistry_33);
        str.append(", "); //$NON-NLS-1$
        str.append(Messages.PreferenceRegistry_34);
        str.append(", "); //$NON-NLS-1$
        str.append(Messages.PreferenceRegistry_35);
        str.append(", "); //$NON-NLS-1$
        str.append(Messages.PreferenceRegistry_36);
        str.append(", "); //$NON-NLS-1$
        str.append(Messages.PreferenceRegistry_37);
        str.append(". "); //$NON-NLS-1$
        str.append(Messages.PreferenceRegistry_38);
        defaultPreferences.put(PREF_LOG_LEVEL, new PreferenceModel(PREF_LOG_LEVEL,
                DEFAULT_LOG_LEVEL, "", str.toString())); //$NON-NLS-1$ 

        defaultPreferences.put(
                PREF_TRACE_DISABLE,
                new PreferenceModel(PREF_TRACE_DISABLE, String.valueOf(DEFAULT_TRACE_DISABLE),
                        "", NLS.bind(Messages.Application_CONCATENATE, //$NON-NLS-1$
                                Messages.PreferenceRegistry_40,
                                NLS.bind(Messages.Application_CONCATENATE, NLS.bind(
                                        Messages.PreferenceRegistry_41, PREF_TRACE_DISABLE), NLS
                                        .bind(Messages.PreferenceRegistry_54, "true", "false"))))); //$NON-NLS-1$  //$NON-NLS-2$

        defaultPreferences.put(
                PREF_TRACE_DIR,
                new PreferenceModel(PREF_TRACE_DIR, null, "", NLS.bind( //$NON-NLS-1$
                        Messages.Application_CONCATENATE, Messages.PreferenceRegistry_45,
                        Messages.PreferenceRegistry_46)));

        defaultPreferences.put(
                PREF_MAX_TRACE_FILE_SIZE,
                new PreferenceModel(PREF_MAX_TRACE_FILE_SIZE, String
                        .valueOf(DEFAULT_MAX_TRACE_FILE_SIZE), "bytes", NLS.bind( //$NON-NLS-1$
                        Messages.Application_CONCATENATE, Messages.PreferenceRegistry_49,
                        Messages.PreferenceRegistry_50)));

        defaultPreferences.put(
                PREF_TRACE_COMPRESS,
                new PreferenceModel(PREF_TRACE_COMPRESS, String.valueOf(DEFAULT_TRACE_COMPRESS),
                        "", NLS.bind(Messages.Application_CONCATENATE, //$NON-NLS-1$
                                Messages.PreferenceRegistry_53,
                                NLS.bind(Messages.PreferenceRegistry_54, "true", "false")))); //$NON-NLS-1$ //$NON-NLS-2$

        defaultPreferences.put(
                PREF_MAX_TRACE_DIR_SIZE,
                new PreferenceModel(PREF_MAX_TRACE_DIR_SIZE, String
                        .valueOf(DEFAULT_MAX_TRACE_DIR_SIZE), "bytes", NLS.bind( //$NON-NLS-1$
                        Messages.Application_CONCATENATE, Messages.PreferenceRegistry_57, NLS.bind(
                                Messages.Application_CONCATENATE, Messages.PreferenceRegistry_58,
                                Messages.PreferenceRegistry_59))));

        defaultPreferences.put(PREF_JSON_OUTPUT,
                new PreferenceModel(PREF_JSON_OUTPUT, String.valueOf(DEFAULT_JSON_OUTPUT),
                        "", Messages.PreferenceRegistry_62)); //$NON-NLS-1$

        defaultPreferences.put(
                PREF_REFRESH_LOCAL_CHANGES,
                new PreferenceModel(PREF_REFRESH_LOCAL_CHANGES, String
                        .valueOf(DEFAULT_REFRESH_LOCAL_CHANGES), "", NLS.bind( //$NON-NLS-1$
                        Messages.Application_CONCATENATE, Messages.PreferenceRegistry_65, NLS.bind(
                                Messages.Application_CONCATENATE, Messages.PreferenceRegistry_66,
                                Messages.PreferenceRegistry_67))));

        defaultPreferences.put(DATE_FORMAT_CHANGE,
                new PreferenceModel(DATE_FORMAT_CHANGE, "", Messages.PreferenceRegistry_DATE_FORMAT_DEFAULT, //$NON-NLS-1$
                        "", Messages.PreferenceRegistry_DATE_FORMAT)); //$NON-NLS-1$

        defaultPreferences.put(PREF_EXTERNAL_2WAY_COMPARE_TOOL,
                new PreferenceModel(PREF_EXTERNAL_2WAY_COMPARE_TOOL, "", Messages.PreferenceRegistry_74, "", //$NON-NLS-1$ //$NON-NLS-2$
                        new String[] {Messages.PreferenceRegistry_68, Messages.PreferenceRegistry_69,
                        Messages.PreferenceRegistry_70, Messages.PreferenceRegistry_71,
                        getExternalCompareToolExample(CompareToolBeyondCompare.EXTERNAL_TOOL_NAME, false),
                        getExternalCompareToolExample(CompareToolClearCaseDiffMerge.EXTERNAL_TOOL_NAME, false),
                        getExternalCompareToolExample(CompareToolDiffMerge.EXTERNAL_TOOL_NAME, false),
                        getExternalCompareToolExample(CompareToolKDiff3.EXTERNAL_TOOL_NAME, false),
                        getExternalCompareToolExample(CompareToolP4Merge.EXTERNAL_TOOL_NAME, false),
                        getExternalCompareToolExample(CompareToolWinMerge.EXTERNAL_TOOL_NAME, false)}));

        defaultPreferences.put(PREF_EXTERNAL_3WAY_COMPARE_TOOL,
                new PreferenceModel(PREF_EXTERNAL_3WAY_COMPARE_TOOL, "", Messages.PreferenceRegistry_74, "", //$NON-NLS-1$ //$NON-NLS-2$
                        new String[] {Messages.PreferenceRegistry_72, Messages.PreferenceRegistry_69,
                        Messages.PreferenceRegistry_73, Messages.PreferenceRegistry_70,
                        Messages.PreferenceRegistry_71,
                        getExternalCompareToolExample(CompareToolBeyondCompare.EXTERNAL_TOOL_NAME, true),
                        getExternalCompareToolExample(CompareToolClearCaseDiffMerge.EXTERNAL_TOOL_NAME, true),
                        getExternalCompareToolExample(CompareToolDiffMerge.EXTERNAL_TOOL_NAME, true),
                        getExternalCompareToolExample(CompareToolKDiff3.EXTERNAL_TOOL_NAME, true),
                        getExternalCompareToolExample(CompareToolP4Merge.EXTERNAL_TOOL_NAME, true),
                        getExternalCompareToolExample(CompareToolWinMerge.EXTERNAL_TOOL_NAME, true)}));
        
        defaultPreferences.put(PREF_SUSPEND_RELEASE_LOCK,
                new PreferenceModel(PREF_SUSPEND_RELEASE_LOCK, String.valueOf(DEFAULT_SUSPEND_RELEASE_LOCK),
                        "", Messages.PreferenceRegistry_75)); //$NON-NLS-1$
        defaultPreferences.put(PREF_UNDO_RELEASE_LOCK,
                new PreferenceModel(PREF_UNDO_RELEASE_LOCK, String.valueOf(DEFAULT_UNDOD_RELEASE_LOCK),
                        "", Messages.PreferenceRegistry_76)); //$NON-NLS-1$
        defaultPreferences.put(PREF_DISCARD_RELEASE_LOCK,
                new PreferenceModel(PREF_DISCARD_RELEASE_LOCK, String.valueOf(DEFAULT_DISCARD_RELEASE_LOCK),
                        "", Messages.PreferenceRegistry_77)); //$NON-NLS-1$
        defaultPreferences.put(PREF_RESUME_ACQUIRE_LOCK,
                new PreferenceModel(PREF_RESUME_ACQUIRE_LOCK, String.valueOf(DEFAULT_RESUME_ACQUIRE_LOCK),
                        "", Messages.PreferenceRegistry_78)); //$NON-NLS-1$
        defaultPreferences.put(PREF_DELIVER_RELEASE_LOCK,
                new PreferenceModel(PREF_DELIVER_RELEASE_LOCK, String.valueOf(DEFAULT_DELIVER_RELEASE_LOCK),
                        "", Messages.PreferenceRegistry_79)); //$NON-NLS-1$
        
    }

    public static class PreferenceStreamSource {
    	private final File file;
    	public PreferenceStreamSource(File f) {
    		file = f;
    	}
    	
    	public boolean exists() {
    		return file.exists();
    	}
    	
    	public InputStream getInputStream() throws IOException {
    		return new FileInputStream(file);
    	}
    	public OutputStream getOutputStream() throws IOException {
    		return new FileOutputStream(file);
    	}
    }
    
	final PreferenceStreamSource source;
	boolean loaded = false;
	
	Map<String, String> currentPreferences = new HashMap<String, String>();
    
    public PreferenceRegistry(PreferenceStreamSource source) {
        this.source = source;
    }
    
    /**
     * @param externalToolName
     * @return expression that has the tool executable and arguments
     */
    private String getExternalCompareToolExpression(String externalToolName, boolean compare3way) {
        IExternalCompareTool tool = ExternalCompareToolsUtil.getToolByName(externalToolName);
        String path = null, arguments = null;
        
        if (compare3way) {
            path = tool.getDefault3WayMergeExecutablePath();
            arguments = tool.getDefault3WayMergeArguments();
        } else {
            path = tool.getDefault2WayCompareExecutablePath();
            arguments = tool.getDefault2WayLocalCompareArguments();
        }
        
        if (path == null || path.length() == 0) {
            return ""; //$NON-NLS-1$
        }
        
        if (ExternalCompareToolsUtil.isMSWindows()) {
            path = ExternalCompareToolsUtil.addQuotes(path);
        }
        
        return path + " " + arguments; //$NON-NLS-1$
    }

    private String getExternalCompareToolExample(String externalToolName, boolean compare3way) {
        return getExternalCompareToolExample(externalToolName, null, compare3way);
    }
    
    private String getExternalCompareToolExample(String externalToolName, String prefix,
            boolean compare3way) {
        String toolExpression = getExternalCompareToolExpression(externalToolName, compare3way);
        return NLS.bind(Messages.PendingChangesUtil_PREFIX_CHANGESET_INFO,
                (prefix != null) ? prefix : externalToolName,
                toolExpression);
    }
    

    /** Load the prefs from the stream. 
     * @throws FileSystemException 
     * @throws IOException 
     */
    private void load() {
    	if (loaded) {
    	    return;
    	}

        loaded = true;
        
        if (!source.exists()) {
        	writePreferences();
        }

        InputStream in;
		try {
			in = source.getInputStream();
		} catch (IOException e) {
			// Something went wrong opening the file. We don't care. 
            return;
		}
        
        if (in == null) {
            return;
        }

        try {
			PropertiesLoader.load(currentPreferences, in, "UTF-8", PropertiesLoader.UNKNOWN_LENGTH, null); //$NON-NLS-1$
		} catch (IOException e) {
			// The file was unparsable. Again: we don't care
			currentPreferences.clear();
		    return;
		} finally {
		    try {
				in.close();
			} catch (IOException e) {
			}
		}
    }
    
    class Preferences {
        ArrayList<String> prefLines = new ArrayList<String>();

        private String makeComment(String line) {
            return "# " + line; //$NON-NLS-1$
        }
        
        private void addLine(String line, boolean isComment) {
            addLine(line, isComment, true);
        }
        
        private void addLine(String line, boolean isComment, boolean trimLine) {
            if (trimLine) {
                String[] prefLineArr = StringUtil.splitString(line, " ", 80); //$NON-NLS-1$
                for (String prefLine : prefLineArr) {
                    prefLines.add(isComment ? makeComment(prefLine) : prefLine);
                }
            } else {
                prefLines.add(isComment ? makeComment(line) : line);
            }
        }
        
        public void addLine(String line) {
            addLine(line, false, true);
        }
        
        public void addComment() {
            prefLines.add(makeComment("")); //$NON-NLS-1$
        }
        
        public void addComment(String line) {
            addLine(line, true, true);
        }
        
        public void addComment(String line, boolean trimLine) {
            addLine(line, true, false);
        }

        public void addKeyValue(final String key, final String value) {
            prefLines.add(key + ": " + value); //$NON-NLS-1$
        }
        
        public void addKeyValueAsComment(final String key, final String value) {
            prefLines.add(makeComment(key + ": " + value)); //$NON-NLS-1$
        }

        public void addNewLine() {
            prefLines.add(""); //$NON-NLS-1$
        }
        
        public void print(OutputStream out) throws UnsupportedEncodingException, IOException {
            byte[] newLine = System.getProperty("line.separator").getBytes("UTF-8"); //$NON-NLS-1$ //$NON-NLS-2$

            try {
                for (String prefLine : prefLines) {
                    out.write(prefLine.getBytes("UTF-8")); //$NON-NLS-1$
                    out.write(newLine);
                }
            } finally {
                out.close();
            }
        }
    }
    
	/**
	 * Create a default preference file.
	 */
	private void writePreferences() {
	    Preferences preferences = new Preferences();
	    preferences.addComment(Messages.PreferenceRegistry_0);
	    preferences.addComment(Messages.PreferenceRegistry_1);
	    preferences.addComment(Messages.PreferenceRegistry_2);
	    preferences.addComment(Messages.PreferenceRegistry_3);
	    preferences.addNewLine();
	    preferences.addNewLine();	            
	    preferences.addComment(Messages.PreferenceRegistry_4);
	    preferences.addComment();
	    preferences.addComment(Messages.PreferenceRegistry_5);
	    preferences.addComment();
	    preferences.addComment(Messages.PreferenceRegistry_6);
	    preferences.addComment();
	    preferences.addComment(Messages.PreferenceRegistry_7);
	    preferences.addKeyValueAsComment(PREF_REPO_CONNECTION_TIMEOUT, "600"); //$NON-NLS-1$
        String value = getPreferenceValue(PREF_REPO_CONNECTION_TIMEOUT, null);
        if (value != null) {
            preferences.addKeyValue(PREF_REPO_CONNECTION_TIMEOUT, value);
        }
	    preferences.addNewLine();
	    preferences.addNewLine();
	    preferences.addComment(Messages.PreferenceRegistry_8);
	    preferences.addComment();
	    preferences.addComment(Messages.PreferenceRegistry_9);
	    preferences.addComment();
	    preferences.addComment(NLS.bind(Messages.PreferenceRegistry_10, "true")); //$NON-NLS-1$
	    preferences.addComment();
	    preferences.addComment(Messages.PreferenceRegistry_11);
	    preferences.addKeyValueAsComment(PREF_MASK_CREDENTIALS, "false"); //$NON-NLS-1$
        value = getPreferenceValue(PREF_MASK_CREDENTIALS, String.valueOf(DEFAULT_MASK_CREDENTIALS));
        preferences.addKeyValue(PREF_MASK_CREDENTIALS, value);
	    preferences.addNewLine();
	    preferences.addNewLine();
	    preferences.addComment(Messages.PreferenceRegistry_12);
	    preferences.addComment();
	    preferences.addComment(Messages.PreferenceRegistry_13);
	    preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_14);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_15);
        preferences.addKeyValueAsComment(PREF_CONTENT_TRANSFER_THREADS, "25"); //$NON-NLS-1$
        value = getPreferenceValue(PREF_CONTENT_TRANSFER_THREADS, Integer.toString(DEFAULT_CONTENT_TRANSFER_THREADS));
        preferences.addKeyValue(PREF_CONTENT_TRANSFER_THREADS, value);
        preferences.addNewLine();
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_20);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_21);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_22);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_23);
        preferences.addKeyValueAsComment(PREF_COMMIT_ATOMIC_MAXIMUM, "10");  //$NON-NLS-1$
        value = getPreferenceValue(PREF_COMMIT_ATOMIC_MAXIMUM, null);
        if (value != null) {
            preferences.addKeyValue(PREF_COMMIT_ATOMIC_MAXIMUM, value);
        }
        preferences.addNewLine();
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_24);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_25);
        preferences.addComment(Messages.PreferenceRegistry_26);
        preferences.addComment(Messages.PreferenceRegistry_27);
        value = getPreferenceValue(PREF_DUMP_ON_CONSOLE, Boolean.toString(DEFAULT_DUMP_ON_CONSOLE));
        preferences.addKeyValue(PREF_DUMP_ON_CONSOLE, value);
        preferences.addNewLine();
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_28);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_29);
        preferences.addComment(Messages.PreferenceRegistry_30);
        preferences.addComment(Messages.PreferenceRegistry_31);
        preferences.addComment(Messages.PreferenceRegistry_32);
        preferences.addComment(Messages.PreferenceRegistry_33);
        preferences.addComment(Messages.PreferenceRegistry_34);
        preferences.addComment(Messages.PreferenceRegistry_35);
        preferences.addComment(Messages.PreferenceRegistry_36);
        preferences.addComment(Messages.PreferenceRegistry_37);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_38);
        preferences.addKeyValueAsComment(PREF_LOG_LEVEL, DEFAULT_LOG_LEVEL);
        value = getPreferenceValue(PREF_LOG_LEVEL, null);
        if (value != null) {
            preferences.addKeyValue(PREF_LOG_LEVEL, value);
        }
        preferences.addNewLine();
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_39);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_40);
        preferences.addComment(NLS.bind(Messages.PreferenceRegistry_41, PREF_TRACE_DISABLE));
        preferences.addComment(NLS.bind(Messages.PreferenceRegistry_54, "true", "false")); //$NON-NLS-1$ //$NON-NLS-2$
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_43);
        preferences.addKeyValueAsComment(PREF_TRACE_DISABLE, "true"); //$NON-NLS-1$
        value = getPreferenceValue(PREF_TRACE_DISABLE, null);
        if (value != null) {
            preferences.addKeyValue(PREF_TRACE_DISABLE, value);
        }
        preferences.addNewLine();
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_44);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_45);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_46);
        preferences.addKeyValueAsComment(PREF_TRACE_DIR, Messages.PreferenceRegistry_47);
        value = getPreferenceValue(PREF_TRACE_DIR, null);
        if (value != null) {
            preferences.addKeyValue(PREF_TRACE_DIR, value);
        }
        preferences.addNewLine();
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_48);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_49);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_50);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_51);
        preferences.addKeyValueAsComment(PREF_MAX_TRACE_FILE_SIZE, "50000"); //$NON-NLS-1$
        value = getPreferenceValue(PREF_MAX_TRACE_FILE_SIZE, String.valueOf(DEFAULT_MAX_TRACE_FILE_SIZE));
        preferences.addKeyValue(PREF_MAX_TRACE_FILE_SIZE, value); 
        preferences.addNewLine();
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_52);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_53);
        preferences.addComment();
        preferences.addComment(NLS.bind(Messages.PreferenceRegistry_54, "true", "false"));  //$NON-NLS-1$//$NON-NLS-2$
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_55);
        preferences.addKeyValueAsComment(PREF_TRACE_COMPRESS, "false"); //$NON-NLS-1$
        value = getPreferenceValue(PREF_TRACE_COMPRESS, String.valueOf(DEFAULT_TRACE_COMPRESS));
        preferences.addKeyValue(PREF_TRACE_COMPRESS, value); 
        preferences.addNewLine();
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_56);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_57);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_58);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_59);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_60);
        preferences.addKeyValueAsComment(PREF_MAX_TRACE_DIR_SIZE, "800000"); //$NON-NLS-1$
        value = getPreferenceValue(PREF_MAX_TRACE_DIR_SIZE, String.valueOf(DEFAULT_MAX_TRACE_DIR_SIZE));
        preferences.addKeyValue(PREF_MAX_TRACE_DIR_SIZE, value); 
        preferences.addNewLine();
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_61);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_62);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_63);
        preferences.addKeyValueAsComment(PREF_JSON_OUTPUT, "true"); //$NON-NLS-1$
        value = getPreferenceValue(PREF_JSON_OUTPUT, null);
        if (value != null) {
            preferences.addKeyValue(PREF_JSON_OUTPUT, value);
        }
        preferences.addNewLine();
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_64);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_65);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_66);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_67);
        value = getPreferenceValue(PREF_REFRESH_LOCAL_CHANGES,
                String.valueOf(DEFAULT_REFRESH_LOCAL_CHANGES));
        preferences.addKeyValue(PREF_REFRESH_LOCAL_CHANGES, value);
        preferences.addNewLine();
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_DATE);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_DATE_FORMAT);
        value = getPreferenceValue(DATE_FORMAT_CHANGE, ""); //$NON-NLS-1$
        if (value != null && value.length() > 0) {
            preferences.addKeyValue(DATE_FORMAT_CHANGE, value);
        }
        preferences.addNewLine();
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_68);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_69);
        preferences.addComment(Messages.PreferenceRegistry_70);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_71);
        preferences.addComment(CompareToolBeyondCompare.EXTERNAL_TOOL_NAME);
        preferences.addComment(getExternalCompareToolExample(
                CompareToolBeyondCompare.EXTERNAL_TOOL_NAME, PREF_EXTERNAL_2WAY_COMPARE_TOOL, false),
                false);
        preferences.addComment(CompareToolClearCaseDiffMerge.EXTERNAL_TOOL_NAME);
        preferences.addComment(getExternalCompareToolExample(
                CompareToolClearCaseDiffMerge.EXTERNAL_TOOL_NAME, PREF_EXTERNAL_2WAY_COMPARE_TOOL, false),
                false);
        preferences.addComment(CompareToolDiffMerge.EXTERNAL_TOOL_NAME);
        preferences.addComment(getExternalCompareToolExample(
                CompareToolDiffMerge.EXTERNAL_TOOL_NAME, PREF_EXTERNAL_2WAY_COMPARE_TOOL, false),
                false);
        preferences.addComment(CompareToolKDiff3.EXTERNAL_TOOL_NAME);
        preferences.addComment(getExternalCompareToolExample(
                CompareToolKDiff3.EXTERNAL_TOOL_NAME, PREF_EXTERNAL_2WAY_COMPARE_TOOL, false),
                false);
        preferences.addComment(CompareToolP4Merge.EXTERNAL_TOOL_NAME);
        preferences.addComment(getExternalCompareToolExample(
                CompareToolP4Merge.EXTERNAL_TOOL_NAME, PREF_EXTERNAL_2WAY_COMPARE_TOOL, false),
                false);
        preferences.addComment(CompareToolWinMerge.EXTERNAL_TOOL_NAME);
        preferences.addComment(getExternalCompareToolExample(
                CompareToolWinMerge.EXTERNAL_TOOL_NAME, PREF_EXTERNAL_2WAY_COMPARE_TOOL, false),
                false);
        value = getPreferenceValue(PREF_EXTERNAL_2WAY_COMPARE_TOOL, ""); //$NON-NLS-1$
        if (value != null && value.length() > 0) {
            preferences.addKeyValue(PREF_EXTERNAL_2WAY_COMPARE_TOOL, value);
        }
        preferences.addNewLine();
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_72);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_69);
        preferences.addComment(Messages.PreferenceRegistry_73);
        preferences.addComment(Messages.PreferenceRegistry_70);
        preferences.addComment();
        preferences.addComment(Messages.PreferenceRegistry_71);
        preferences.addComment(CompareToolBeyondCompare.EXTERNAL_TOOL_NAME);
        preferences.addComment(getExternalCompareToolExample(
                CompareToolBeyondCompare.EXTERNAL_TOOL_NAME, PREF_EXTERNAL_3WAY_COMPARE_TOOL, true),
                false);
        preferences.addComment(CompareToolClearCaseDiffMerge.EXTERNAL_TOOL_NAME);
        preferences.addComment(getExternalCompareToolExample(
                CompareToolClearCaseDiffMerge.EXTERNAL_TOOL_NAME, PREF_EXTERNAL_3WAY_COMPARE_TOOL, true),
                false);
        preferences.addComment(CompareToolDiffMerge.EXTERNAL_TOOL_NAME);
        preferences.addComment(getExternalCompareToolExample(
                CompareToolDiffMerge.EXTERNAL_TOOL_NAME, PREF_EXTERNAL_3WAY_COMPARE_TOOL, true),
                false);
        preferences.addComment(CompareToolKDiff3.EXTERNAL_TOOL_NAME);
        preferences.addComment(getExternalCompareToolExample(
                CompareToolKDiff3.EXTERNAL_TOOL_NAME, PREF_EXTERNAL_3WAY_COMPARE_TOOL, true),
                false);
        preferences.addComment(CompareToolP4Merge.EXTERNAL_TOOL_NAME);
        preferences.addComment(getExternalCompareToolExample(
                CompareToolP4Merge.EXTERNAL_TOOL_NAME, PREF_EXTERNAL_3WAY_COMPARE_TOOL, true),
                false);
        preferences.addComment(CompareToolWinMerge.EXTERNAL_TOOL_NAME);
        preferences.addComment(getExternalCompareToolExample(
                CompareToolWinMerge.EXTERNAL_TOOL_NAME, PREF_EXTERNAL_3WAY_COMPARE_TOOL, true),
                false);
        value = getPreferenceValue(PREF_EXTERNAL_3WAY_COMPARE_TOOL, ""); //$NON-NLS-1$
        if (value != null && value.length() > 0) {
            preferences.addKeyValue(PREF_EXTERNAL_3WAY_COMPARE_TOOL, value);
        }
        preferences.addNewLine();
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_suspend);
        preferences.addComment(Messages.PreferenceRegistry_75);
        preferences.addKeyValue(PREF_SUSPEND_RELEASE_LOCK, getPreferenceValue(PREF_SUSPEND_RELEASE_LOCK, String.valueOf(DEFAULT_SUSPEND_RELEASE_LOCK)));
        
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_undo);
        preferences.addComment(Messages.PreferenceRegistry_76);
        preferences.addKeyValue(PREF_UNDO_RELEASE_LOCK, getPreferenceValue(PREF_UNDO_RELEASE_LOCK, String.valueOf(DEFAULT_UNDOD_RELEASE_LOCK)));
        
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_discard);
        preferences.addComment(Messages.PreferenceRegistry_77);
        preferences.addKeyValue(PREF_DISCARD_RELEASE_LOCK, getPreferenceValue(PREF_DISCARD_RELEASE_LOCK, String.valueOf(DEFAULT_DISCARD_RELEASE_LOCK)));
        
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_resume);
        preferences.addComment(Messages.PreferenceRegistry_78);
        preferences.addKeyValue(PREF_RESUME_ACQUIRE_LOCK, getPreferenceValue(PREF_RESUME_ACQUIRE_LOCK, String.valueOf(DEFAULT_RESUME_ACQUIRE_LOCK)));
        
        preferences.addNewLine();
        preferences.addComment(Messages.PreferenceRegistry_deliver);
        preferences.addComment(Messages.PreferenceRegistry_79);
        preferences.addKeyValue(PREF_DELIVER_RELEASE_LOCK, getPreferenceValue(PREF_DELIVER_RELEASE_LOCK, String.valueOf(DEFAULT_DELIVER_RELEASE_LOCK)));
        
        try {
            preferences.print(source.getOutputStream());
        } catch (IOException e) {
            log.warn("IOException when generating preferences file", e); //$NON-NLS-1$
        }
	}
	
	
	@Override
    public boolean getBoolean(String name, boolean def) {
	    load();
	    
	    String text = currentPreferences.get(name);
	    if (text == null) {
	        return def;
	    }
	    
	    return text.trim().equalsIgnoreCase("true"); //$NON-NLS-1$
	}
	
	@Override
    public int getInteger(String name, int def) {
		load();
		
        String text = currentPreferences.get(name);
        if (text == null) {
            return def;
        }
        
        try {
        	return Integer.parseInt(text.trim());
        } catch (NumberFormatException e) {
        	return def;
        }
	}

	@Override
    public int getConnectionTimeout() {
		return getInteger(PREF_REPO_CONNECTION_TIMEOUT, UNKNOWN_TIMEOUT);
	}

	@Override
    public boolean getMaskPassword() {
		return getBoolean(PREF_MASK_CREDENTIALS, DEFAULT_MASK_CREDENTIALS);
	}

	@Override
    public int getContentTransferThreads() {
		return getInteger(PREF_CONTENT_TRANSFER_THREADS, DEFAULT_CONTENT_TRANSFER_THREADS);
	}

	@Override
    public boolean getAutoRefresh() {
	    load();
	    return getBoolean(PREF_REFRESH_LOCAL_CHANGES, DEFAULT_REFRESH_LOCAL_CHANGES);
	}
	
	@Override
    public int getAtomicCommitMaximum() {
		return getInteger(PREF_COMMIT_ATOMIC_MAXIMUM, UNKNOWN_ATOMIC_COMMIT_MAXIMUM);
	}
	
	@Override
    public boolean getDumpOnConsole() {
		return getBoolean(PREF_DUMP_ON_CONSOLE, DEFAULT_DUMP_ON_CONSOLE);
	}
	
	@Override
    public Level getLogLevel() {
		
		load();

		Map<String, Level> mp = new HashMap<String, Level>() {
								  {
									put("severe", Level.SEVERE); //$NON-NLS-1$
									put("warning", Level.WARNING);   //$NON-NLS-1$
									put("info",  Level.INFO);//$NON-NLS-1$
									put("config", Level.CONFIG); //$NON-NLS-1$
									put("fine", Level.FINE); //$NON-NLS-1$
									put("finer", Level.FINER); //$NON-NLS-1$
									put("finest", Level.FINEST);  //$NON-NLS-1$
								  }
								 };

		String level = currentPreferences.get(PREF_LOG_LEVEL);
	    
		if (level == null) {
	        return Level.SEVERE;
	    } else {
	    	level = level.trim().toLowerCase(Locale.ENGLISH);
	    }

	    for (String txtLevel : mp.keySet()) {
	    	if (txtLevel.equals(level)) {
	    		return mp.get(txtLevel);
	    	}
	    }

	    return Level.SEVERE;	    
	}
	
	@Override
    public boolean getJSONOutput() {
		return getBoolean(PREF_JSON_OUTPUT, DEFAULT_JSON_OUTPUT);
	}
	
	@Override
	public String getDateFormat() {
	    return currentPreferences.get((PreferenceRegistry.DATE_FORMAT_CHANGE));
	}
	
	@Override
	public String getExternal2wayCompareTool() {
        String externalCompareTool = currentPreferences.get(PREF_EXTERNAL_2WAY_COMPARE_TOOL);
        if (externalCompareTool == null || externalCompareTool.length() == 0) {
            return null;
        }
        return externalCompareTool;
    }
	
    @Override
    public String getExternal3wayCompareTool() {
        String externalCompareTool = currentPreferences.get(PREF_EXTERNAL_3WAY_COMPARE_TOOL);
        if (externalCompareTool == null || externalCompareTool.length() == 0) {
            return null;
        }
        return externalCompareTool;
    }

    @Override
    public Map<String, String> getRegistryContents() {
		return currentPreferences;
	}
	
	public static Map<String, PreferenceModel> getDefaultPreferences() {
	    return defaultPreferences;
	}

    public static class PreferenceModel {
        private final String name;
        private final String defaultValue;
        private final String defaultDisplayValue;
        private final String units;
        private final String[] description;

        public PreferenceModel(String name, String defaultValue, String units, String description) {
            this(name, defaultValue, null, units,
                    (description != null) ? new String[] {description} : new String[] {""}); //$NON-NLS-1$
        }
        
        public PreferenceModel(String name, String defaultValue, String defaultDisplayValue,
                String units, String description) {
            this(name, defaultValue, defaultDisplayValue, units,
                    (description != null) ? new String[] {description} : new String[] {""}); //$NON-NLS-1$
        }
        
        public PreferenceModel(String name, String defaultValue, String units, String[] description) {
            this(name, defaultValue, null, units, description);
        }
        
        public PreferenceModel(String name, String defaultValue, String defaultDisplayValue,
                String units, String[] description) {
            this.name = name;
            this.defaultValue = (defaultValue != null) ? defaultValue : ""; //$NON-NLS-1$
            this.defaultDisplayValue = (defaultDisplayValue != null) ? defaultDisplayValue : ""; //$NON-NLS-1$
            this.units = units;
            if (description == null) {
                this.description = new String[] {""}; //$NON-NLS-1$
            } else {
                this.description = description;
            }
        }

        public String getName() {
            return this.name;
        }

        public String getDefaultValue() {
            return this.defaultValue;
        }

        public String getDefaultDisplayValue() {
            return this.defaultDisplayValue;
        }

        public String getUnits() {
            return this.units;
        }
        
        public String[] getDescription() {
            return this.description;
        }
    }

    @Override
    public void setPreference(String name, String value) {
        load();
        currentPreferences.put(name, value);
        writePreferences();
    }

    private String getPreferenceValue(String key, String defaultValue) {
        String value = currentPreferences.get(key);
        value = value != null ? value : defaultValue;
        return value;
    }
    
    @Override
    public boolean getReleaseLockForSuspend() {
	    load();
	    return getBoolean(PREF_SUSPEND_RELEASE_LOCK, DEFAULT_SUSPEND_RELEASE_LOCK);
	}
    
    @Override
    public boolean getReleaseLockForUndo() {
	    load();
	    return getBoolean(PREF_UNDO_RELEASE_LOCK, DEFAULT_UNDOD_RELEASE_LOCK);
	}
    
    @Override
    public boolean getReleaseAutoLockForDiscard() {
	    load();
	    return getBoolean(PREF_DISCARD_RELEASE_LOCK, DEFAULT_DISCARD_RELEASE_LOCK);
	}
    
    @Override
    public boolean getAcquireLockForResume() {
	    load();
	    return getBoolean(PREF_RESUME_ACQUIRE_LOCK, DEFAULT_RESUME_ACQUIRE_LOCK);
	}
    
    @Override
    public boolean getReleaseLockForDeliver() {
	    load();
	    return getBoolean(PREF_DELIVER_RELEASE_LOCK, DEFAULT_DELIVER_RELEASE_LOCK);
	}
}
