/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2016. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.filesystem.cli.core.internal;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;

import com.ibm.team.filesystem.cli.core.util.ISandboxWorkspace;
import com.ibm.team.filesystem.cli.core.util.SandboxWorkspace;
import com.ibm.team.filesystem.cli.core.util.SubcommandUtil;
import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.ILocation;
import com.ibm.team.filesystem.client.internal.PathLocation;


/**
 * Stores the information regarding the workspaces that were loaded in a sandbox
 * as well as workspaces that were tracked in the cli. If the sandbox path is
 * empty then the workspace is being tracked.
 * 
 * Format:
 * <sandboxHistory lastCleanupTime="">
 *   <sandbox path="">
 *     <repositoryWorkspace repositoryId="" workspaceItemId=""/>
 *     <repositoryWorkspace repositoryId="" workspaceItemId=""/>
 *   </sandbox>
 *   <sandbox path=""/>
 * </sandboxHistory>
 * 
 * lastCleanupTime is used to run the cleanup routine which will check if the 
 * sandbox path exists and is a valid sandbox (check for the existence of the
 * meta data directory).
 */

@XmlRootElement
public class SandboxHistory {

	private static class RepositoryWorkspace {
		private String repositoryId;
		private String workspaceItemId;
		
		public RepositoryWorkspace() {
		}

		public RepositoryWorkspace(String repositoryId, String workspaceItemId) {
			this.repositoryId = repositoryId;
			this.workspaceItemId = workspaceItemId;
		}

		@XmlAttribute(name="repositoryId")
		public String getRepositoryId() {
			return repositoryId;
		}

		public void setRepositoryId(String repositoryId) {
			this.repositoryId = repositoryId;
		}

		@XmlAttribute(name="workspaceItemId")
		public String getWorkspaceItemId() {
			return workspaceItemId;
		}

		public void setWorkspaceItemId(String workspaceItemId) {
			this.workspaceItemId = workspaceItemId;
		}
	}

	private static class Sandbox {
		private String path;
		private List<RepositoryWorkspace> repositoryWorkspaces;

		public Sandbox() {		
		}
		
		public Sandbox(String path, List<RepositoryWorkspace> repositoryWorkspaces) {
			this.path = path;
			this.repositoryWorkspaces = repositoryWorkspaces;
		}

		@XmlAttribute(name="path")
		public String getPath() {
			return path;
		}

		public void setPath(String path) {
			this.path = path;
		}

		@XmlElement(name="repositoryWorkspace")
		public List<RepositoryWorkspace> getRepositoryWorkspaces() {
			return repositoryWorkspaces;
		}

		public void setRepositoryWorkspaces(List<RepositoryWorkspace> repositoryWorkspaces) {
			this.repositoryWorkspaces = repositoryWorkspaces;
		}

		public void addRepositoryWorkspace(RepositoryWorkspace repositoryWorkspace) {
			if (repositoryWorkspaces == null) {
				repositoryWorkspaces = new ArrayList<RepositoryWorkspace>();
			}
			repositoryWorkspaces.add(repositoryWorkspace);
		}
	}

	private List<Sandbox> sandboxes;
	private long lastCleanupTime;

	@XmlAttribute(name="lastCleanupTime")
	public long getLastCleanupTime() {
		return lastCleanupTime;
	}

	public void setLastCleanupTime(long lastCleanupTime) {
		this.lastCleanupTime = lastCleanupTime;
	}

	@XmlElement(name="sandbox")
	public List<Sandbox> getSandboxes() {
		return sandboxes;
	}

	public void setSandboxes(List<Sandbox> sandboxes) {
		this.sandboxes = (sandboxes != null) ? sandboxes : new ArrayList<Sandbox>();
	}

	public SandboxHistory() {
		sandboxes = new ArrayList<Sandbox>();
		lastCleanupTime = System.currentTimeMillis();
	}

	/**
	 * Adds a repository workspace to the sandbox
	 * 
	 * @param path The path to the sandbox
	 * @param repositoryId The repository uuid
	 * @param workspaceItemId The workspace item id
	 */
	public boolean addWorkspace(ILocation path, String repositoryId, String workspaceItemId) {
		boolean update = true;
		for (Sandbox sandbox : sandboxes) {
			ILocation sandboxPath = new PathLocation(sandbox.path);
			if (!sandboxPath.equals(path)) {
				continue;
			}
			if (sandbox.getRepositoryWorkspaces() != null) {
				for (RepositoryWorkspace repositoryWorkspace : sandbox.getRepositoryWorkspaces()) {
					if (repositoryWorkspace.getRepositoryId().equals(repositoryId)
							&& repositoryWorkspace.getWorkspaceItemId().equals(workspaceItemId)) {
						return false;
					}
				}
			}
			if (update) {
				RepositoryWorkspace newRepositoryWorkspace = new RepositoryWorkspace(repositoryId, workspaceItemId);
				sandbox.addRepositoryWorkspace(newRepositoryWorkspace);
				update = false;
				break;
			}
		}
		if (update) {
			List<RepositoryWorkspace> repoWorkspacesList = new ArrayList<RepositoryWorkspace>(1);
			RepositoryWorkspace repoWorkspace = new RepositoryWorkspace(repositoryId, workspaceItemId);
			repoWorkspacesList.add(repoWorkspace);

			Sandbox newSandbox = new Sandbox(path.toOSString(), repoWorkspacesList);
			sandboxes.add(newSandbox);
		}

		return true;
	}

	/**
	 * Removes a workspace from the sandbox
	 * 
	 * @param path The path to the sandbox
	 * @param workspaceItemId The workspace item id to remove from the sandbox
	 */
	public boolean removeWorkspace(ILocation path, String workspaceItemId) {
		boolean removed = false;
		Iterator<Sandbox> iteratorSb = sandboxes.iterator();
		while (iteratorSb.hasNext()) {
			Sandbox sandbox = iteratorSb.next();
			ILocation sandboxPath = new PathLocation(sandbox.path);
			if (!sandboxPath.equals(path)) {
				continue;
			}
			if (sandbox.getRepositoryWorkspaces() != null) {
				Iterator<RepositoryWorkspace> iteratorWs = sandbox.getRepositoryWorkspaces().iterator();
				while (iteratorWs.hasNext()) {
					RepositoryWorkspace repositoryWorkspace = iteratorWs.next();
					if (repositoryWorkspace.getWorkspaceItemId().equals(workspaceItemId)) {
						iteratorWs.remove();
						removed = true;
						break;
					}
				}			
				if (removed && sandbox.getRepositoryWorkspaces().size() == 0) {
					iteratorSb.remove();
				}
			}
		}

		return removed;
	}

	/**
	 * Returns the list of repository workspaces in the sandbox
	 * 
	 *  @param sandboxPath The path to the sandbox
	 */
	public List<ISandboxWorkspace> getWorkspaces(ILocation path) {
		return getWorkspaces(path, null);
	}
	
	/**
	 * Returns the list of repository workspaces in the sandbox
	 * 
	 *  @param sandboxPath The path to the sandbox
	 *  @param repositoryId The repository uuid
	 */
	public List<ISandboxWorkspace> getWorkspaces(ILocation path, String repositoryId) {
		List<ISandboxWorkspace> repoWorkspaces = new ArrayList<ISandboxWorkspace>();
		for (Sandbox sandbox : sandboxes) {
			ILocation sandboxPath = new PathLocation(sandbox.path);
			if (!sandboxPath.equals(path)) {
				continue;
			}
			for (RepositoryWorkspace repositoryWorkspace : sandbox.getRepositoryWorkspaces()) {
				if (repositoryId != null && !repositoryWorkspace.getRepositoryId().equals(repositoryId)) {
					continue;
				}
				
				ISandboxWorkspace repoWorkspace = new SandboxWorkspace(repositoryWorkspace.getWorkspaceItemId(),
						repositoryWorkspace.getWorkspaceItemId(), repositoryWorkspace.getRepositoryId());
				repoWorkspaces.add(repoWorkspace);
			}
			break;
		}
		return repoWorkspaces;
	}

	public void cleanup() {
		Iterator<Sandbox> sandboxIterator = sandboxes.iterator();
		while (sandboxIterator.hasNext()) {
			Sandbox sandbox = sandboxIterator.next();
			try {
				SubcommandUtil.findNamedAncestorCFARoot(sandbox.path);
			} catch (FileSystemException e) {
				sandboxIterator.remove();
			}
		}
	}
}
