/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2016. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.filesystem.cli.core.internal;

import java.io.File;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.osgi.util.NLS;

import com.ibm.team.filesystem.cli.core.util.ISandboxWorkspace;
import com.ibm.team.filesystem.cli.core.util.SubcommandUtil;
import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.ILocation;
import com.ibm.team.filesystem.client.internal.FileSystemStatusUtil;
import com.ibm.team.filesystem.client.internal.LoggingHelper;
import com.ibm.team.filesystem.client.internal.utils.FilesystemLock;

/**
 * Manages the list of sandboxes along with the repository workspaces
 * in the sandboxes. 
 */
public class SandboxHistoryRegistry implements ISandboxHistoryRegistry {
	private File sandboxHistoryFile = null;
	public static final String FILENAME = "sandboxHistory.txt"; //$NON-NLS-1$
	private static long lastFileModTime = 0;
	private static long lastFileSize = 0;
	private static final int CLEANUP_INTERVAL = 10; // in days 
	
	private FilesystemLock lock = null;
	private static final String FILE_LOCK = "sandboxHistory.lck"; //$NON-NLS-1$
    private static final long LOCK_WAIT = 20000;

	private SandboxHistory sandboxHistory;
	private boolean loaded = false;
	
	/**
     * Base class for exceptions that deal with the {@link SandboxHistoryRegistry}. 
     */
    @SuppressWarnings("serial")
    public static class SandboxHistoryRegistryException extends FileSystemException {
        public SandboxHistoryRegistryException(String message, Throwable cause) {
            super(message, cause);
        }
        
        public SandboxHistoryRegistryException(String msg) {
            super(msg);
        }
    }

    public SandboxHistoryRegistry(File file) {
		if (file == null) {
			throw new IllegalArgumentException();
		}
			
		sandboxHistoryFile = file;
		lock = new FilesystemLock(new File(sandboxHistoryFile.getParentFile(), FILE_LOCK));
		sandboxHistory = new SandboxHistory();
	}

	private void load(boolean shouldLock) {
		if (!sandboxHistoryFile.exists()) {
			return;
		}
		
		if (loaded && 
				(sandboxHistoryFile.lastModified() != lastFileModTime || sandboxHistoryFile.length() != lastFileSize)) {
			loaded = false;
		}

		if (loaded) {
			return;
		}
		
		if (shouldLock) {
		    if (!lock.acquire(LOCK_WAIT, null).isOK()) {
		        throw new IllegalStateException("Could not acquire lock"); //$NON-NLS-1$
		    }
	    }

		try {
			JAXBContext jaxbContext = JAXBContext.newInstance(SandboxHistory.class);
			Unmarshaller jaxbUnmarshaller = jaxbContext.createUnmarshaller();
			sandboxHistory = (SandboxHistory) jaxbUnmarshaller.unmarshal(sandboxHistoryFile);
			lastFileModTime = sandboxHistoryFile.lastModified();
			lastFileSize = sandboxHistoryFile.length();
			loaded = true;
		} catch (JAXBException e) {
			LoggingHelper.log(FileSystemStatusUtil.getStatusFor(IStatus.WARNING, 
                    NLS.bind(Messages.SandboxHistoryRegistry_LOAD_FAILURE, sandboxHistoryFile.getPath()), e));
		} finally {
			if (shouldLock) {
				lock.release(null);
			}
		}
	}

	private void cleanup() {
		long diffTime = System.currentTimeMillis() - sandboxHistory.getLastCleanupTime();
		long diffDays = diffTime / (24 * 60 * 60 * 1000);
		if (diffDays < CLEANUP_INTERVAL) {
			return;
		}

		sandboxHistory.cleanup();
}

	private void save() throws SandboxHistoryRegistryException {
		cleanup();
		
		try {
			JAXBContext jaxbContext = JAXBContext.newInstance(SandboxHistory.class);
			Marshaller jaxbMarshaller = jaxbContext.createMarshaller();
			jaxbMarshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);

			jaxbMarshaller.marshal(sandboxHistory, sandboxHistoryFile);
			lastFileModTime = sandboxHistoryFile.lastModified();
			lastFileSize = sandboxHistoryFile.length();
		} catch (JAXBException e) {
			throw new SandboxHistoryRegistryException(NLS.bind(Messages.SandboxHistoryRegistry_WRITE_FAILURE, sandboxHistoryFile.getAbsolutePath()), e);
		}
	}

	/**
	 * Adds a repository workspace to the sandbox
	 * 
	 * @param sandboxPath The path to the sandbox
	 * @param repositoryUUid The repository uuid
	 * @param workspaceItemId The workspace item id
	 */
	public synchronized void addWorkspace(ILocation sandboxPath, String repositoryUuid, String workspaceItemId) throws SandboxHistoryRegistryException {
		if (!lock.acquire(LOCK_WAIT, null).isOK()) {
			throw new IllegalStateException("Could not acquire lock"); //$NON-NLS-1$
		}
		
		try {
			load(false);
			
			if (sandboxHistory.addWorkspace(sandboxPath, repositoryUuid, workspaceItemId)) {
				save();
			}
		} finally {
			lock.release(null);
		}
	}

	/**
	 * Removes a workspace from the sandbox
	 * 
	 * @param sandboxPath The path to the sandbox
	 * @param workspaceItemId The workspace item id to remove from the sandbox
	 */
	public synchronized void removeWorkspace(ILocation sandboxPath, String workspaceItemId) throws SandboxHistoryRegistryException {
		removeWorkspaces(sandboxPath, Collections.singletonList(workspaceItemId));
	}

	/**
	 * Removes the workspaces from the sandbox
	 * 
	 * @param sandboxPath The path to the sandbox
	 * @param workspaceItemIds The list of workspace item id's to remove from the sandbox
	 */
	public synchronized void removeWorkspaces(ILocation sandboxPath, List<String> workspaceItemIds) throws SandboxHistoryRegistryException {
		if (!lock.acquire(LOCK_WAIT, null).isOK()) {
			throw new IllegalStateException("Could not acquire lock"); //$NON-NLS-1$
		}
		
		try {
			load(false);
			
			boolean save = false;
			for (String workspaceItemId : workspaceItemIds) {
				if (sandboxHistory.removeWorkspace(sandboxPath, workspaceItemId)) {
					save = true;
				}
			}
		
			if (save) {
				save();
			}
		} finally {
			lock.release(null);
		}
	}

	/**
	 * Returns the list of repository workspaces in the sandbox
	 * 
	 *  @param sandboxPath The path to the sandbox
	 */
	public synchronized List<ISandboxWorkspace> getWorkspaces(ILocation sandboxPath) {
		return getWorkspaces(sandboxPath, null);
	}

	/**
	 * Returns the list of repository workspaces in the sandbox
	 * 
	 *  @param sandboxPath The path to the sandbox
	 *  @param repositoryUuid The repository uuid
	 */
	public synchronized List<ISandboxWorkspace> getWorkspaces(ILocation sandboxPath, String repositoryUuid) {
		load(true);
		return sandboxHistory.getWorkspaces(sandboxPath, repositoryUuid);
	}
}
