/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2016. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.filesystem.cli.core.internal;

import static com.ibm.team.filesystem.cli.core.subcommands.CommonOptions.OPT_CERTIFICATE_FILE;
import static com.ibm.team.filesystem.cli.core.subcommands.CommonOptions.OPT_PASSWORD;
import static com.ibm.team.filesystem.cli.core.subcommands.CommonOptions.OPT_SMART_CARD;
import static com.ibm.team.filesystem.cli.core.subcommands.CommonOptions.OPT_KERBEROS;
import static com.ibm.team.filesystem.cli.core.subcommands.CommonOptions.OPT_INTEGRATED_WINDOWS;
import static com.ibm.team.filesystem.cli.core.subcommands.CommonOptions.OPT_URI;
import static com.ibm.team.filesystem.cli.core.subcommands.CommonOptions.OPT_USERNAME;
import static com.ibm.team.filesystem.client.internal.EnvironmentVariables.SCM_ALLOW_INSECURE;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.logging.ConsoleHandler;
import java.util.logging.FileHandler;
import java.util.logging.Handler;
import java.util.logging.LogManager;
import java.util.logging.Logger;

import org.apache.log4j.PropertyConfigurator;

import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Status;
import org.eclipse.equinox.app.IApplicationContext;
import org.eclipse.osgi.util.NLS;

import com.ibm.team.filesystem.cli.core.internal.Messages;
import com.ibm.team.filesystem.cli.core.CommandLineCore;
import com.ibm.team.filesystem.cli.core.Constants;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IExecutionContext;
import com.ibm.team.filesystem.cli.core.subcommands.CommonOptions;
import com.ibm.team.filesystem.cli.core.subcommands.IScmClientConfiguration;
import com.ibm.team.rtc.cli.infrastructure.internal.core.CLIClientException;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ClientConfiguration;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IApplicationEndListener;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IClientConfiguration;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommandDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.core.LocalContext;
import com.ibm.team.rtc.cli.infrastructure.internal.core.SubcommandLoader;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.ICommandLine;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.IOptionKey;
import com.ibm.team.filesystem.cli.core.subcommands.ITypedPreferenceRegistry;
import com.ibm.team.filesystem.cli.core.util.AliasUtil.IAliasOptions;
import com.ibm.team.filesystem.cli.core.util.CLIFileSystemClientException;
import com.ibm.team.filesystem.cli.core.util.ConnectionInfo;
import com.ibm.team.filesystem.cli.core.util.LocalAppDataMigrator;
import com.ibm.team.filesystem.cli.core.util.RepoUtil;
import com.ibm.team.filesystem.cli.core.util.StatusHelper;
import com.ibm.team.filesystem.cli.core.util.SubcommandUtil;
import com.ibm.team.filesystem.cli.core.util.SubcommandUtil.IPropertyChangeEvent;
import com.ibm.team.filesystem.cli.core.subcommands.IUuidAliasRegistry;
import com.ibm.team.filesystem.cli.core.internal.aliases.UuidAliasRegistry;
import com.ibm.team.filesystem.client.FileSystemCore;
import com.ibm.team.filesystem.client.FileSystemCore.UserConfigDir;
import com.ibm.team.filesystem.client.CopyFileAreaLockedByOtherProcess;
import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.FileSystemStatusException;
import com.ibm.team.filesystem.client.ILocation;
import com.ibm.team.filesystem.client.ISandbox;
import com.ibm.team.filesystem.client.daemon.events.ILightweightEventListener;
import com.ibm.team.filesystem.client.internal.FileSystemStatusUtil;
import com.ibm.team.filesystem.client.internal.HardcodedPermissionUtils;
import com.ibm.team.filesystem.client.internal.LoggingHelper;
import com.ibm.team.filesystem.client.internal.PathLocation;
import com.ibm.team.filesystem.client.internal.Sandbox;
import com.ibm.team.filesystem.client.internal.SharingManager;
import com.ibm.team.filesystem.client.internal.PlatformDetector.LauncherType;
import com.ibm.team.filesystem.client.internal.utils.IPermissionUtil;
import com.ibm.team.filesystem.client.internal.utils.RepositoryRegistry;
import com.ibm.team.filesystem.client.internal.utils.IPermissionUtil.CannotCreateException;
import com.ibm.team.filesystem.client.internal.utils.IPermissionUtil.CannotModifyException;
import com.ibm.team.filesystem.client.internal.utils.IPermissionUtil.PermissionUtilsException;
import com.ibm.team.filesystem.client.util.IRepositoryRecord;
import com.ibm.team.filesystem.client.util.IRepositoryRegistry;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.UUID;

public class ScmClientConfiguration extends ClientConfiguration
                                    implements IScmClientConfiguration, IAliasOptions {
    
    /** Debug server trips. */
    public static final boolean DEBUG_TRIPS =
        Activator.getDefault().isDebugging() &&
        "true".equalsIgnoreCase(Platform.getDebugOption("com.ibm.team.filesystem.cli.core/debug/trips")); //$NON-NLS-1$ //$NON-NLS-2$
	
    /** Default name of the preference properties file. */
	public static final String PREFERENCES_FILENAME = "preferences.properties"; //$NON-NLS-1$
	
	/** Default name of the alias registry. */
    public static final String ALIAS_FILENAME = "aliases"; //$NON-NLS-1$

    /** Default name of the repository registry. */
    public static final String REPOREG_FILENAME = RepositoryRegistry.REPOSITORIES_FILE_v2;
    
    public static final String PREFERENCES_RCP_CORE_PREFS =
            ".metadata/.plugins/org.eclipse.core.runtime/.settings/com.ibm.team.filesystem.rcp.core.prefs"; //$NON-NLS-1$
    
    public static final String PREFERENCES_ECLIPSE_CORE_RESOURCES =
            ".metadata/.plugins/org.eclipse.core.resources"; //$NON-NLS-1$

    public static final String PROP_LOG = "scm_log"; //$NON-NLS-1$
    public static final String PROP_LOG4J = "log4j_properties"; //$NON-NLS-1$
    public static final String PROP_LOG_JAVA_UTIL = "java_util_logging"; //$NON-NLS-1$

    /** Log settings */
    public static final String LOG_PATTERN = "scm%g.log"; //$NON-NLS-1$
    public static final int LOG_LIMIT = 5000000; // 5MB
    public static final int NUM_LOG_FILES = 10; // will rollover after 10 files
    public static final String LOG_DIR = "logs"; //$NON-NLS-1$

    /** Synchronization object for log files */
    private static Object logObj = new Object();

    private static Logger logger = null;
    private static FileHandler logFileHandler = null;

    private RepositoryRegistry repoReg;
	
	private UuidAliasRegistry aliasReg;
	
	private File cfaRoot; 
	
	private boolean runningInScratch;
	
	private HashMap<String, ConnectionInfo> cachedConnectInfo = 
	        new HashMap<String, ConnectionInfo>();
	
	private HashMap<UUID, String> repoUuidToUriMap = new HashMap<UUID, String>();

	private ITypedPreferenceRegistry prefs;

	private boolean isLoggingEnabled;
	
	private boolean isJSONEnabled;
	
	private static File configLocation = null;
	
	private SandboxHistoryRegistry sandboxHistoryReg = null;
	
    public static class PathPair {
        final File cfaRoot, cwd;
        final boolean isScratch;

        public PathPair(File cfaRoot, File cwd, boolean isScratch) {
            this.cfaRoot = cfaRoot;
            this.cwd = cwd;
            this.isScratch = isScratch;
        }
    }

    public void initialize(SubcommandLoader loader, IExecutionContext environment,
            IApplicationContext context, ISubcommandDefinition defn,
            List<ISubcommandDefinition> parents, ICommandLine args, ICommandLine subargs)
            throws CLIClientException {
	    super.initialize(loader, environment, context, defn, parents, args, subargs);	    
	    CommandLineCore.setConfig(this);
	    
	    // If the init policy is 'none' then the base initialisation is good enough
	    Policy policy = null;
	    try {
	        policy = getInitPolicy(); 
        } catch (FileSystemException e) {
            throw complain(Constants.STATUS_CONFIG_FAILURE, Messages.Application_23, e.getLocalizedMessage());
        }
	    
	    if (policy == Policy.NONE) {
	        return;
	    }
	    
        IExecutionContext ctx = getContext();

        // Get our configuration root
        UserConfigDir configRoot;
        try {
            configRoot = findConfigDirectory();
        } catch (FileSystemException e) {
            e.printStackTrace(ctx.stderr());
            throw complain(Constants.STATUS_CONFIG_FAILURE, Messages.Application_23, e.getLocalizedMessage());
        }

        // Import pre-3.0 config into 3.0. 
        if (configRoot.isUsingLocalAppData()) {
            LocalAppDataMigrator.migrate(configRoot);
        }
        
        // Secure the root of our configuration
        HardcodedPermissionUtils permUtils = new HardcodedPermissionUtils(LauncherType.CLI,
                getContext().environment());
        secureConfigLocation(ctx, permUtils, configRoot.getConfigDirAsFile(), getCommandLine());
        
        // Assign the eclipse/CFA root
        PathPair pair;
        try {
            pair = initializeDataArea(ctx, configRoot);            
        } catch (FileSystemException e) {
            int code = 0;
            if (e instanceof FileSystemStatusException) {
                code = ((FileSystemStatusException) e).getStatus().getCode();
            }
            throw complain(code, Messages.Application_42, e.getLocalizedMessage());
        }
        
        if (pair == null) {
			this.cfaRoot = null;
			this.cwd = null;
			this.runningInScratch = true;
		} else {
			this.cfaRoot = pair.cfaRoot;
			if (cfaRoot == null) {
				SubcommandUtil.addPropertyChangeListener(
				        new ILightweightEventListener<IPropertyChangeEvent>() {
					public void handleEvent(IPropertyChangeEvent event) {
						if (SubcommandUtil.PROPERTY_WORKSPACE_ROOT.equals(event.getPropertyName())) {
							ScmClientConfiguration.this.cfaRoot = (File)event.getPropertyValue();
						}
					}
				});
			}
			
			this.cwd = pair.cwd;
			this.runningInScratch = pair.isScratch;
		}
		
		this.repoReg = new RepositoryRegistry(permUtils, new File(configRoot.getConfigDirAsFile(), REPOREG_FILENAME));
		this.repoReg.setAllowMigrationForRepositoryId(true);
		
		this.aliasReg = new UuidAliasRegistry(new File(configRoot.getConfigDirAsFile(), ALIAS_FILENAME), 1000);
		
		prefs = new PreferenceRegistry(new PreferenceRegistry.PreferenceStreamSource(new File(configRoot.getConfigDirAsFile(), PREFERENCES_FILENAME)));

        setShowUuids(computeShowPolicy(CommonOptions.OPT_SHOW_UUID, ShowPolicy.NEVER));
        setShowAliases(computeShowPolicy(CommonOptions.OPT_SHOW_ALIAS, ShowPolicy.ALWAYS));
        setEnableJSON(getPersistentPreferences().getJSONOutput());
        
        if (environment instanceof LocalContext) {
            FileSystemCore.enableHeadlessProxies();
        }

        try {
            initLoggers(configLocation, configRoot.getConfigDirAsFile());
        } catch (FileSystemException e) {
            throw new CLIClientException(e);
        }
        
        configLocation = configRoot.getConfigDirAsFile();
        
        this.sandboxHistoryReg = new SandboxHistoryRegistry(new File(configLocation, SandboxHistoryRegistry.FILENAME));
	}
	
    /** Get the platform specific user configuration directory. The given 
     * directory may not exist.
     */
    public static UserConfigDir getDefaultConfigDirectory() throws IOException {
        return FileSystemCore.getUserConfigDir();
    }
    
    private UserConfigDir findConfigDirectory() throws FileSystemException {
        String requestedPath = null;
        if (args != null && args.hasOption(CommonOptions.OPT_CONFIG_LOCATION)) {
            requestedPath = args.getOption(CommonOptions.OPT_CONFIG_LOCATION);
        }
            
        return SubcommandUtil.getConfigRoot(requestedPath);
    }
    
    /**
     * Secure the config directory.
     * @param ctx 
     */
    private void secureConfigLocation(IExecutionContext ctx, IPermissionUtil permUtils,
            File configRoot, ICommandLine args) throws CLIClientException {
        try {
            boolean secured = permUtils.mkRoot(configRoot.getCanonicalPath());
            
            if (permUtils.shouldRunOnPlatform() && !secured) {
                if (!ctx.environment().containsKey(SCM_ALLOW_INSECURE)) {
                    throw complain(Constants.STATUS_CONFIG_FAILURE,
                            NLS.bind(
                                    Messages.SubcommandLauncher_COULD_NOT_CREATE_DIRECTORY_FULL_STRING,
                                    configRoot.toString()),
                            NLS.bind(
                                    Messages.SubcommandLauncher_FAILED_TO_SECURE_DIRECTORY,  
                                    new String[] {
                                            SCM_ALLOW_INSECURE, 
                                            com.ibm.team.filesystem.client.internal.PlatformDetector.Platform.PRGPATH, 
                                            com.ibm.team.filesystem.client.internal.PlatformDetector.Platform.RTC_SCRIPT_BASE
                                    }));
                }
            }
            
        } catch (CannotCreateException e) {
            throw complain(Constants.STATUS_LOCAL_FS_MISCONFIGURED, NLS.bind(Messages.SubcommandLauncher_COULD_NOT_CREATE_DIRECTORY, configRoot.getName()), NLS.bind(Messages.SubcommandLauncher_COULD_NOT_CREATE_DIRECTORY_FULL_STRING, configRoot.toString()));
            
        } catch (CannotModifyException e) {
            throw complain(Constants.STATUS_LOCAL_FS_MISCONFIGURED, NLS.bind(Messages.SubcommandLauncher_COULD_NOT_SECURE_DIRECTORY, configRoot.getName()), NLS.bind(Messages.SubcommandLauncher_COULD_NOT_SET_PERMISSIONS_ON_DIRECTORY, configRoot.toString()));
            
        } catch (PermissionUtilsException e) {
            LoggingHelper.log(FileSystemStatusUtil.getStatusFor(e));
            throw complain(Constants.STATUS_LOCAL_FS_MISCONFIGURED, NLS.bind(Messages.SubcommandLauncher_ERROR_SECURING_DIRECTORY, configRoot.getName()), NLS.bind(Messages.SubcommandLauncher_ERROR_SECURING_DIRECTORY_LONG, configRoot.toString(), e.getLocalizedMessage()));
            
        } catch (InterruptedException e) {
            throw complain(Constants.STATUS_LOCAL_FS_MISCONFIGURED, NLS.bind(Messages.SubcommandLauncher_INTERRUPTION_SECURING_DIRECTORY, configRoot.getName()), NLS.bind(Messages.SubcommandLauncher_INTERRUPTION_SECURING_DIRECTORY_LONG, configRoot.toString()));
            
        } catch (IOException e) {
            LoggingHelper.log(FileSystemStatusUtil.getStatusFor(e));
            throw complain(Constants.STATUS_LOCAL_FS_MISCONFIGURED, NLS.bind(Messages.SubcommandLauncher_EXCEPTION_SECURING_DIRECTORY, configRoot.getName()), NLS.bind(Messages.SubcommandLauncher_EXCEPTION_SECURING_DIRECTORY_WITH_MESSAGE, configRoot.toString(), e.getLocalizedMessage()));
        }
    }

    private static enum Policy {
        NONE, DEFAULT, ANCESTOR, ARGUMENT, URI_ARG_ANC, ARG_ANC, SCRATCH, ARGUMENT_CREATE
    }
    
    private Policy getInitPolicy() throws FileSystemException {
        String policy = getSubcommandDefinition().getAttr(Constants.PT_ATTR_INIT_POLICY);
        
        if ("none".equals(policy)) { //$NON-NLS-1$
            return Policy.NONE;
        }
        else if ("default".equals(policy)) { //$NON-NLS-1$
            return Policy.DEFAULT;
        }
        else if ("ancestor".equals(policy)) { //$NON-NLS-1$
            return Policy.ANCESTOR;
        }
        else if ("argument".equals(policy)) { //$NON-NLS-1$
            return Policy.ARGUMENT;
        }
        else if ("argument/ancestor".equals(policy)) { //$NON-NLS-1$
            return Policy.ARG_ANC;
        }
        else if ("uri/argument/ancestor".equals(policy)) { //$NON-NLS-1$
            return Policy.URI_ARG_ANC;
        }
        else if ("scratch".equals(policy)) { //$NON-NLS-1$
            return Policy.SCRATCH;
        }
        else if ("argument-create".equals(policy)) { //$NON-NLS-1$
            return Policy.ARGUMENT_CREATE;
        }
        else {
            IConfigurationElement configElem = getSubcommandDefinition().getConfigurationElement();
            throw StatusHelper.internalError(NLS.bind(Messages.SubcommandDefinition_1,
                    new Object[] {getSubcommandDefinition().getFullSubcommandName(),
                    configElem.getContributor().getName(), policy}));
        }
    }

    /**
     * Initialise the OSGi data area (aka the Eclipse workspace root).
     *  
     * @param ctx Set to <code>true</code> when the subcommand is running 
     *      within a daemon. 
     * 
     * @return The CFA and CWD. 
     */
    public PathPair initializeDataArea(IExecutionContext ctx, UserConfigDir configRoot)
            throws FileSystemException  {
        Policy policy = getInitPolicy();
        
        // vvv Note - argPath is treated as a flag, since we trust the 
        //            launcher to examine the arguments and repoint the 
        //            CWD to use this value. 
        String argPath = subargs.getOption(CommonOptions.OPT_DIRECTORY, null);
        
        File cfaRoot = null;
        File cwd = null;

        boolean registerCfa = false;
        boolean isScratch = false;

        switch (policy) {
        case NONE:
            return null;
            
        case DEFAULT:
            return new PathPair(null, null, false);

        case ANCESTOR:
            String userDir = ctx.getCurrentWorkingDirectory();
            cfaRoot = SubcommandUtil.findAncestorCFARoot(userDir);
            if (cfaRoot == null) {
                throw StatusHelper.misconfiguredLocalFS(NLS.bind(Messages.SubcommandDefinition_2, userDir));
            }
            
            cwd = new File(ctx.getCurrentWorkingDirectory());
            registerCfa = true;
            break;
        
                
        case ARGUMENT_CREATE:
        case ARGUMENT: {
            if (argPath == null) {
                verifyArg(getSubcommandDefinition());
                throw StatusHelper.argSyntax(Messages.SubcommandDefinition_3);
            }
                
            cwd = SubcommandUtil.canonicalize(new File(ctx.getCurrentWorkingDirectory()));
            cfaRoot = new File(SubcommandUtil.canonicalize(ctx.getCurrentWorkingDirectory()));
                
            if (policy == Policy.ARGUMENT_CREATE) {
                // Verify that we don't have an ancestor CFA root
                File ancestorRoot = SubcommandUtil.findAncestorCFARoot(cfaRoot.getAbsolutePath());
                if (ancestorRoot == null) {
                    // There is no ancestor cfaRoot. Ensure we have one. 
                    if (!cfaRoot.exists() && !cfaRoot.mkdirs()) {
                        throw StatusHelper.misconfiguredLocalFS(
                                NLS.bind(Messages.SubcommandDefinition_4, cfaRoot.getAbsolutePath()));
                    }
                    
                }
                else if (!cfaRoot.equals(ancestorRoot)) {
                    // We have an ancestor that is also a root
                    throw StatusHelper.misconfiguredLocalFS(
                            NLS.bind(Messages.SubcommandDefinition_5, ancestorRoot.getAbsolutePath()));
                }
                
                // We have a valid CFA root.

            }
            else {
                if (!cfaRoot.exists()) {
                    throw StatusHelper.misconfiguredLocalFS(NLS.bind(Messages.SubcommandDefinition_6, argPath));
                }
            }
            
        } break;
            
        case URI_ARG_ANC: 
        case ARG_ANC: {
            // Find a root
            if (argPath == null) {
                cfaRoot = SubcommandUtil.findAncestorCFARoot(ctx.getCurrentWorkingDirectory());
                
                if (cfaRoot == null) {
                    
                    // If we have a URI, fallback to scratch or
                    // we encountered some explicit repos
                    cwd = cfaRoot = configRoot.getConfigDirAsFile();
                    isScratch = true;
                    break;

                } 
                else {
                    registerCfa = true;
                }
                
                cwd = SubcommandUtil.canonicalize(new File(ctx.getCurrentWorkingDirectory()));
            } else {
                cfaRoot = SubcommandUtil.findNamedAncestorCFARoot(ctx.getCurrentWorkingDirectory());
                cwd = SubcommandUtil.canonicalize(new File(ctx.getCurrentWorkingDirectory()));
                if (!cwd.exists()) {
                    throw StatusHelper.misconfiguredLocalFS(NLS.bind(Messages.SubcommandDefinition_8, cwd.getAbsolutePath()));
                }
                
                registerCfa = true;
            }
            
        } break;
        
        case SCRATCH: {
            cwd = cfaRoot = configRoot.getConfigDirAsFile();
            isScratch = true;
        } break;
        }

        Assert.isNotNull(cfaRoot);
        File desiredRoot = cfaRoot;
        if (isScratch) {
            File owned = null;
            
            if (false == ctx.isLocal()) {
                owned = SubcommandUtil.getEclipseWorkspaceRoot();
            } else {
                // If we're scratch, we'll walk workspace slots until we get to 
                //  something we can lock.
                final File scratchRoot = new File(cfaRoot, "scratch"); //$NON-NLS-1$
                scratchRoot.mkdirs();
                
                for (int i = 0; i < Integer.MAX_VALUE; i++) {
                    File attempted = new File(scratchRoot, Integer.toString(i));
                    attempted.mkdirs();
                    
                    if (!attempted.exists()) {
                        throw StatusHelper.misconfiguredLocalFS(NLS.bind(Messages.SubcommandDefinition_SCRATCH_CREATION_FAILURE, scratchRoot.getAbsolutePath()));
                    }
                    
                    try {
                        SubcommandUtil.initializeEclipseWorkspaceRoot(attempted);
                        owned = attempted;
                        break;
                    } catch (CLIFileSystemClientException e) {
                        if (e.getStatus().getCode() == Constants.STATUS_WORKSPACE_IN_USE) {
                            continue;
                        }
                        
                        throw e;
                    }
                }
    
                if (owned == null) {
                    throw StatusHelper.cfaInUse(NLS.bind(Messages.SCRATCH_UNAVAILABLE10, cfaRoot.getAbsolutePath())); 
                }
                
                clearPreferences(owned);
            }

            cwd = cfaRoot = owned;
        } 
        else {
            if (true == ctx.isLocal()) {
                SubcommandUtil.initializeEclipseWorkspaceRoot(cfaRoot);
            }
        }
        
        if (registerCfa) {
            try {
                SharingManager.getInstance().register(new Sandbox(new PathLocation(desiredRoot.getAbsolutePath())), true, null);
            } catch (CopyFileAreaLockedByOtherProcess e) {
                    throw StatusHelper.cfaInUse(NLS.bind(Messages.SubcommandDefinition_ALREADY_LOCKED, desiredRoot.getAbsolutePath()));
            } // other exceptions are still thrown.
        }
        
        return new PathPair(cfaRoot, cwd, isScratch);
    }

    /**
     * Note. Daemon may reuse temp workspace so daemon plugins cannot rely on instance scope
     * preferences in such case. However those same plugins rely on preferences when running inside
     * the eclipse client. Workaround is to delete the instance scope preference file for
     * the plugins that need a fresh environment. Daemon users are responsible for restoring
     * the desired state through daemon api's.
     */
    private static void clearPreferences(File cfa) throws FileSystemException {    	
        
    	//Cleanup the files under scratch directory.    	
		delete(new File(cfa, PREFERENCES_RCP_CORE_PREFS));  
		delete(new File(cfa, PREFERENCES_ECLIPSE_CORE_RESOURCES));    	
    }
    
    private static void delete(File f) throws FileSystemException {
    	if (!f.exists()) {
    		return;
    	}
    	if (f.isDirectory()) {
    		for (File c : f.listFiles()) {
    			delete(c);
    		}
		}
		if (!f.delete()) {
			throw StatusHelper.preferencesNotCleared(f);
		}
	}

    /** Ensure that the {@link CommonOptions#OPT_DIRECTORY} argument is defined 
     * for the given command-line. 
     * @param iSubcommandDefinition 
     * 
     * @param subargs
     * @throws FileSystemException
     */
    public void verifyArg(ISubcommandDefinition defn) throws FileSystemException {
    	if (!subargs.getDefinition().hasOption(CommonOptions.OPT_DIRECTORY)) {
            throw StatusHelper.configurationFailure(
                    NLS.bind(Messages.SubcommandDefinition_9, defn.getFullSubcommandName()));
        }
    }
    
    private static String[] ALWAYS_VALUES = new String[] {"y", "yes"};  //$NON-NLS-1$ //$NON-NLS-2$
    private static String[] NEVER_VALUES = new String[] {"n", "no"};  //$NON-NLS-1$ //$NON-NLS-2$
    
    private ShowPolicy computeShowPolicy(IOptionKey key, ShowPolicy defaultValue)
            throws CLIClientException {
        if (!args.hasOption(key)) {
            return defaultValue;
        }
        
        String value = args.getOption(key);
        
        for (String always : ALWAYS_VALUES) {
            if (always.equals(value)) {
                return ShowPolicy.ALWAYS;
            }
        }
        
        for (String never : NEVER_VALUES) {
            if (never.equals(value)) {
                return ShowPolicy.NEVER;
            }
        }

        // Build the list.
        throw complain(Constants.STATUS_ARGUMENT_SYNTAX_ERROR, 
                NLS.bind(Messages.Application_53, args.getDefinition().getOption(key).getName()), 
                NLS.bind(Messages.Application_54, new Object[] {ALWAYS_VALUES[0], ALWAYS_VALUES[1], NEVER_VALUES[0], NEVER_VALUES[1]})); 
    }

    private CLIClientException complain(int errorCode, String errorTitle, String errorDescription) {
        return new CLIClientException(
                new Status(IStatus.ERROR, Activator.PLUGIN_ID, errorCode,
                        errorTitle + ". " + errorDescription, null)); //$NON-NLS-1$
    }
    
    private void initLoggers(File oldConfigLoc, File currConfigLoc) throws FileSystemException {
        Map<String, String> env = getContext().environment();
        String propLog4j = env.get(PROP_LOG4J);
        String propJavaUtil = env.get(PROP_LOG_JAVA_UTIL);
                
        if (propLog4j == null && propJavaUtil == null) {

            synchronized (logObj) {
                // if there's a logger with our file handler and config path hasn't changed, then return
                if (logger != null) {
                    if (oldConfigLoc != null && oldConfigLoc.equals(currConfigLoc)) {
                        // set the log level
                        logger.setLevel(prefs.getLogLevel());
                        return;                 
                    }
                    
                    // remove the old handler
                    if (logFileHandler != null) {
                        logger.removeHandler(logFileHandler);
                        logFileHandler = null;
                    }
                } else {
                    // get the root logger
                    logger = Logger.getLogger(""); //$NON-NLS-1$
                    logger.setUseParentHandlers(false);

                    // remove the console (default) handler
                    for (Handler h : logger.getHandlers()) {
                        if (h instanceof ConsoleHandler) {
                            logger.removeHandler(h);
                        }
                    }
                }
                
                // initiate logging
                File logDir = new File(currConfigLoc.getAbsolutePath() + File.separator + LOG_DIR);

                // create log dir
                if (!logDir.exists()) {
                    logDir.mkdir();
                }

                // configure our log file handler
                String logFile = logDir.getAbsolutePath() + File.separator + LOG_PATTERN;

                try {
                    logFileHandler = new FileHandler(logFile, LOG_LIMIT, NUM_LOG_FILES);
                    logger.addHandler(logFileHandler);
                } catch (IOException e) {
                    throw StatusHelper.failure(Messages.Application_57, e);
                }

                // set the log level
                logger.setLevel(prefs.getLogLevel());
                setEnableLogging(true);
            }
            
        } else if (propLog4j != null) {
            
            synchronized (logObj) {
                // Configure log4j
                if (isLoggingEnabled()) {
                    return;
                }

                InputStream in = null;
                try {               
                    in = new FileInputStream(new File(propLog4j));
                } catch (FileNotFoundException e1) {
                    throw StatusHelper.failure(NLS.bind(Messages.RepoUtil_FILE_PATH_NOT_FOUND, propLog4j), null);
                }

                Properties props = new Properties();

                try {
                    props.load(in);
                } catch (IOException e1) {
                    throw StatusHelper.failure(Messages.Application_57, e1);
                } finally {
                    try {
                        in.close();
                    } catch (IOException e) {
                        StatusHelper.failure(Messages.Application_57, e);
                    }
                }

                PropertyConfigurator.configure(props);
                setEnableLogging(true);
            }
            
        } else if (propJavaUtil != null) {

            synchronized (logObj) {
                // Configure java.util.logging
                if (isLoggingEnabled()) {
                    return;
                }

                InputStream in = null;

                try {
                    in = new FileInputStream(new File(propJavaUtil));
                    LogManager.getLogManager().readConfiguration(in);

                } catch (FileNotFoundException e) {
                    throw StatusHelper.failure(NLS.bind(Messages.RepoUtil_FILE_PATH_NOT_FOUND, propJavaUtil), null);

                } catch (IOException e) {
                    throw StatusHelper.failure(Messages.Application_57, e);

                } finally {
                    try {
                        if (in != null)
                            in.close();
                    } catch (IOException e) {
                        StatusHelper.failure(Messages.Application_57, e);
                    }
                }
                setEnableLogging(true);
            }
        }       
    }

    /** Our policy for showing system-internal information to the user. 
     * At some point in the future we should display an "AUTO" that will 
     * display identifiers only if the selector is non-unique.
     */
	public enum ShowPolicy {
		ALWAYS, NEVER
	}
	
	/** Our policy on showing aliases. 
     * @see IUuidAlias#getMonicker()
     */
	private ShowPolicy showAlias;
	private ShowPolicy showUuids;

	public File getEclipseWorkspaceRoot() {
		return cfaRoot;
	}
	
	/**
	  * Get the connection information by querying the url from the command line arguments
	  * 
	  * @return
	  */
	public ConnectionInfo getConnectionInfo() throws FileSystemException {
		return getConnectionInfo(null, OPT_URI, true, true);
	}
	
	/**
	  * Get the connection information for the passed repo.
	  *  
	  * @param repoURI The URL of the repository to get credentials for. 
	  * 
	  * @return
	  */
	public ConnectionInfo getConnectionInfo(String repoUri) throws FileSystemException {
        ICommandLine subargs = getSubcommandCommandLine();

        // if we have credentials specified on the command line we need to use it to login
        if (subargs.hasOption(OPT_USERNAME) || subargs.hasOption(OPT_CERTIFICATE_FILE) || subargs.hasOption(OPT_SMART_CARD)) {
        	// if the repo option is specified on the command line, use the credentials if uri matches 
        	if (subargs.hasOption(CommonOptions.OPT_URI)) {
	            String uri = subargs.getOption(CommonOptions.OPT_URI);
	        
	            if (RepoUtil.isRepoUriSame(uri, repoUri, this)) {
	                return getConnectionInfo(repoUri, null, true, true);                
	            }
        	} else { // if no repo option specified, use the credentials
        		return getConnectionInfo(repoUri, null, true, true);
        	}
        }
        
		return getConnectionInfo(repoUri, null, false, true);
	}
	
	public ConnectionInfo getConnectionInfo(String defaultURI, IOptionKey repoOpt, boolean fromCmdArgs, boolean fromCache) throws FileSystemException {
		return getConnectionInfo(defaultURI, repoOpt, fromCmdArgs, fromCache, true);
	}
	
	/**
	  * Get the connection information
	  *  
	  * @param defaultURI The URL of the repository to get credentials for if repoOpt is null 
	  * @param repoOpt Get the URL of the repository by reading this option from the command line
	  * @param fromCmdArgs If set to true, will query the command line arguments for user's credentials.
	  *                    If not it checks in the cache or prompts the user for the credentials.
	  * @param fromCache If set to true, will use the default values that matched the repo uri from the stored repository
	  * @param prompt If set to true, will prompt for the credentials if the credential information is not available.
	  * 
	  * @return
	  */
	public ConnectionInfo getConnectionInfo(String defaultURI, IOptionKey repoOpt, boolean fromCmdArgs, boolean fromCache, boolean prompt)
			throws FileSystemException {
        boolean hasUsername = subargs.hasOption(OPT_USERNAME);
        boolean hasPassword = subargs.hasOption(OPT_PASSWORD);
        boolean hasCertFile = subargs.hasOption(OPT_CERTIFICATE_FILE);
        boolean hasSmartCard = subargs.hasOption(OPT_SMART_CARD);
        boolean hasKerberos = subargs.hasOption(OPT_KERBEROS);
        boolean hasIntegratedWindows = subargs.hasOption(OPT_INTEGRATED_WINDOWS);

        if (fromCmdArgs) {
	        if ((hasCertFile && (hasUsername || hasSmartCard || hasKerberos)) ||
	             (hasSmartCard && (hasCertFile || hasPassword || hasKerberos)) ||
	             (hasKerberos && (hasUsername || hasPassword))) {
	                throw StatusHelper.argSyntax(NLS.bind(Messages.ClientConfiguration_INVALID_ARGS, new String[] { OPT_USERNAME.getName(), OPT_PASSWORD.getName(),
	                        OPT_CERTIFICATE_FILE.getName(), OPT_PASSWORD.getName(), OPT_SMART_CARD.getName(), OPT_USERNAME.getName(), OPT_KERBEROS.getName()}));
	        }
	        if (!hasUsername && !hasPassword && !hasCertFile && !hasSmartCard && !hasKerberos && !hasIntegratedWindows) {
	        	// There were no connection arguments so let's set the variable to 
	        	// false so we use the values from the cache below
	        	fromCmdArgs = false;
	        }
	    }
	    
	    IRepositoryRecord loggedInto = null;
		
		// Get the URI
		String repoURI = SubcommandUtil.getRepoUri(subargs, defaultURI, repoOpt);
		{
	        // There may be a logged-in repository with this URI/nickname.
	        IRepositoryRegistry reg = getRepositoryRegistry();

        	loggedInto = reg.getRecord(repoURI);
	        if (loggedInto != null) {
	        	repoURI = loggedInto.getUrl();
	        }

	        // If we have logged in, get the exact URI represented in the repository  
            try {
                ITeamRepository repo = RepoUtil.getSharedRepository(repoURI, false);
                repoURI = repo.getRepositoryURI();
            } catch (Exception e) {
                ; // continue
            }
        }

        // Get the credentials
        ConnectionInfo info = null;
		String usernameCache = null;
		String passwordCache = null;
		UUID userIdCache = null;
		String certificateFileCache = null;
		boolean smartCardCache = false;
		boolean kerberosCache = false;
		boolean integratedWindowsCache = false;
		
		if (fromCache) {
            // check if we have in the cache
            info = cachedConnectInfo.get(repoURI);
            if (info != null) {
                usernameCache = info.getUsername();
                passwordCache = info.getPassword();
                if (info.getCertificateFile() != null) {
                    certificateFileCache = info.getCertificateFile().toOSString();
                }               
                userIdCache = info.getUserUUID();
                smartCardCache = info.isSmartCard();
                kerberosCache = info.isKerberos();
                integratedWindowsCache = info.isIntegratedWindows();
            } else if (loggedInto != null) {
                usernameCache = loggedInto.getUsername();
                passwordCache = loggedInto.getPassword();
                certificateFileCache = loggedInto.getCertficiateLocation();
                smartCardCache = loggedInto.isSmartCard();
                userIdCache = loggedInto.getUserId();
                kerberosCache = loggedInto.isKerberos();
                integratedWindowsCache = loggedInto.isIntegratedWindows();
            }
        }
		
		// if we are not getting info from command line, reset with what we have in our cache
		if (!fromCmdArgs) {
	        hasUsername = (usernameCache == null) ? false : true;
	        hasPassword = (passwordCache == null) ? false : true;
	        hasCertFile = (certificateFileCache == null) ? false : true;
	        hasSmartCard = smartCardCache;
	        hasKerberos = kerberosCache;
	        hasIntegratedWindows = integratedWindowsCache;
		}
		
        String username = null;
        String password = null;
        ILocation certificateLocation = null;

        if (hasCertFile) {
            certificateLocation = SubcommandUtil.getCertificateFile(subargs, certificateFileCache, this, fromCmdArgs);
			password = SubcommandUtil.getPassword(subargs, passwordCache, null, repoURI, this, fromCmdArgs, mayInteract() && prompt);
		} else if (hasSmartCard) {
			username = SubcommandUtil.getUsername(subargs, usernameCache, repoURI, this, fromCmdArgs, mayInteract() && prompt);
		} else if (!hasKerberos && !hasIntegratedWindows) {
			username = SubcommandUtil.getUsername(subargs, usernameCache, repoURI, this, fromCmdArgs, mayInteract() && prompt);			
			password = SubcommandUtil.getPassword(subargs, passwordCache, username, repoURI, this, fromCmdArgs, mayInteract() && prompt);
		}
	
        info = new ConnectionInfo(repoURI, username, password, certificateLocation, hasSmartCard, hasKerberos, hasIntegratedWindows);
        if (userIdCache != null) {
            info.setUserUUID(userIdCache);
        }

        // Cache the connect info
        cachedConnectInfo.put(repoURI, info);

        return info;
	}
	
	/**
     * Check if the connection info has been cached in memory or in the repository file.
     *  
     * @param repoURI The URL of the repository. 
     * 
     * @return
     */
    public boolean isConnectionInfoCached(String repoURI) {
		ConnectionInfo cachedInfo = cachedConnectInfo.get(repoURI);
		if (cachedInfo != null) {
		    return true;
		}

		IRepositoryRegistry reg = getRepositoryRegistry();        
        IRepositoryRecord loggedInto = reg.getRecordByURL(repoURI);
        if (loggedInto != null) {
        	return true;
        }
        
        return false;
    }
    
    public void updateConnectionInfo(ConnectionInfo info, ITeamRepository repo) {
    	String repoURI = repo.getRepositoryURI();
        if (!repoURI.equals(info.getURI())) {
            if (cachedConnectInfo.get(info.getURI()) != null) {
                cachedConnectInfo.remove(info.getURI());
            }
            
            info.setURI(repoURI);
            cachedConnectInfo.put(repoURI, info);
        }
        
        repoUuidToUriMap.put(repo.getId(), repo.getRepositoryURI());
    }
    
    public void removeConnectionInfo(String repoURI) {
        if (cachedConnectInfo.get(repoURI) != null) {
            cachedConnectInfo.remove(repoURI);
        }
        
        if(repoUuidToUriMap.values().contains(repoURI)) {
        	UUID repositoryId = null;
        	for(Map.Entry<UUID, String> entry : repoUuidToUriMap.entrySet()) {
        		if(entry.getValue()!=null && entry.getValue().equals(repoURI)) {
        			repositoryId = entry.getKey();
        			break;
        		}
        	}
        	repoUuidToUriMap.remove(repositoryId);
        }
    }

	public IRepositoryRegistry getRepositoryRegistry() {
		return repoReg;
	}
	
    public IAliasOptions getAliasConfig() {
        return this;
    }

    public IUuidAliasRegistry getAliasRegistry() {
		return aliasReg;
	}

    public boolean mayInteract() {
        return !getCommandLine().hasOption(CommonOptions.OPT_DISALLOW_INTERACTION);
    }

    /**
	 * Notify the receiver that it's life cycle is at an end. 
	 * Runs the {@link IApplicationEndListener}s and cleans up anything that 
	 * needs a cleaning. 
	 */
	public void deInitialize() {
	    super.deInitialize();
	    
		CommandLineCore.setConfig(null);
	}

	public void setShowUuids(ShowPolicy policy) {
		showUuids = policy;
	}

	public void setShowAliases(ShowPolicy policy) {
		showAlias = policy;
	}
	
	public IClientConfiguration getClientConfiguration() {
		return this;
	}

	public boolean showAlias() {
		return showAlias == ShowPolicy.ALWAYS;
	}

	public boolean showUuid() {
		return showUuids == ShowPolicy.ALWAYS;
	}

	public File getConfigDirectory() throws FileSystemException {
	    return findConfigDirectory().getConfigDirAsFile();
	}

	public ISandbox[] getSandboxes() {
		if (runningInScratch || cfaRoot == null) {
			return new ISandbox[0];
		}
		
		return new ISandbox[] {new Sandbox(new PathLocation(cfaRoot.getAbsolutePath()))};
	}


	public boolean mayMaskPassword() {
	    ICommandLine args = getCommandLine();
		if (args != null && args.hasOption(CommonOptions.OPT_NO_MASK)) {
			return false;
		}
		
		return prefs.getMaskPassword();
	}

	public ITypedPreferenceRegistry getPersistentPreferences() {
		return prefs;
	}

	public void setEnableLogging(boolean enableLogging) {
		isLoggingEnabled = enableLogging;
	}
	
	public boolean isLoggingEnabled() {
		return isLoggingEnabled;
	}
	
	public void setEnableJSON(boolean enableJSON) {
		if (enableJSON) { // need to do this, otherwise cmd line arg will override preference.properties value
			isJSONEnabled = true;
		}
	}
	
	public boolean isJSONEnabled() {
		return isJSONEnabled;
	}

	public String getRepositoryURI(String repositoryId, String defaultURI)
			throws FileSystemException {
		return getRepositoryURI(UUID.valueOf(repositoryId), defaultURI);
	}
	
	/**
	 * Determine the repository URI for the given repository Id. The order of
	 * lookup is
	 * <ul>
	 * <li>Lookup the TeamPlatform and return the repository URI from the
	 * ITeamRepository instance that corresponds to the given repository id.</li>
	 * <li>If the above lookup fails, lookup the repository registry with the
	 * given repository id and return the repository URI associated with the
	 * corresponding repository record.</li>
	 * <li>If the above lookup fails, return the default URI if it is not null.</li>
	 * </ul>
	 * 
	 * 
	 * @param repositoryId
	 * @param defaultURI
	 * 
	 * @return repository URI.
	 * 
	 * @throws FileSystemException
	 *             If the repository URI can not be determined and the default
	 *             URI is null.
	 */
	public String getRepositoryURI(UUID repositoryId, String defaultURI) throws FileSystemException{
		
		if (repositoryId == null) {
			throw new IllegalArgumentException();
		}
		// see if the mapping is cached.
		// this cache will also be updated when we login through RepoUtil.login.
		String repositoryURI = repoUuidToUriMap.get(repositoryId);
		if(repositoryURI != null) {
			return repositoryURI;
		}
		ITeamRepository repo = RepoUtil.getTeamRepository(repositoryId);
		if (repo != null) {
			// we are logged, this repository URI should always be up to date.
			repoUuidToUriMap.put(repositoryId, repo.getRepositoryURI());
			return repo.getRepositoryURI();
		}
		IRepositoryRecord repoRecord = repoReg.getRecordByRepositoryId(repositoryId);
		if (repoRecord != null) {
			// we are not logged in, but we have the repository entry in repository.txt file.
			// this repository URI will be stale after renaming the server,
			// this stale URI will be fixed when the user runs the login command
			// with the new repository URI.
			repoUuidToUriMap.put(repositoryId, repoRecord.getUrl());
			return repoRecord.getUrl();
		}
		// at any case we should have a default URI, else error out
		// this might occur when we run status on a loaded workspace without logging into the server 
		if (defaultURI == null) {
			throw StatusHelper.repositoryLookupFailed(NLS.bind(Messages.ClientConfiguration_COULD_NOT_DETERMINE_REPOSITORY, repositoryId.getUuidValue()));
		}
		// we are not logged in and we don't have the repository entry in repository.txt file.
		// this can occur when we are using aliases in the command line.
		// this repository URI will be stale after renaming the server.
		// the stale repository URI in alias record will be fixed when we lookup
		// for the alias after the user has logged in with the new repository
		// URI.
		return defaultURI;
	}
	
	public ISandboxHistoryRegistry getSandboxHistoryRegistry() {
		return this.sandboxHistoryReg; 
	}
}
