/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2005, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.filesystem.cli.core.internal.aliases;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.Enumeration;
import java.util.Locale;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.ibm.team.filesystem.cli.core.util.RepoUtil.ItemType;
import com.ibm.team.repository.common.UUID;

/**
 * Responsible for loading and saving aliases. This is a stub for a 
 * version-specific alias file loader.
 */
public class AliasLoader {
	
	private static String ENCODING = "UTF-8"; //$NON-NLS-1$
	
	/** The structure of the line indicating the last allocated alias. */
	private static String LAST_ALLOCATED_LINE = "lastAllocated="; //$NON-NLS-1$

	/** A regex to determine the last allocated number. */
	private static Pattern LAST_ALLOCATED_REGEX = Pattern.compile("lastAllocated\\s*=\\s*(\\d+)"); //$NON-NLS-1$
	
	/** Structure of the line indicating the number of the last run. */
	private static String LAST_RUN_LINE = "lastRun="; //$NON-NLS-1$
	
	/** The structure of the line indicating the last run alias. */
	private static Pattern LAST_RUN_REGEX = Pattern.compile("lastRun\\s*=\\s*(\\d+)"); //$NON-NLS-1$
	
	final static String VERSION_PREAMBLE = "### Jazz UUID Aliases "; //$NON-NLS-1$

	/**
	 * Populate lookup with the aliases found in file.
	 */
	public static void load(UuidAliasRegistry lookup, File file) throws IOException {
		// Open the monicker listing
		BufferedReader reader = new BufferedReader(new InputStreamReader(
				new FileInputStream(file), ENCODING));

		boolean corrupted = false;

		try {
			boolean hasLastAllocated = false;
			boolean hasLastRun = false;
			
			// Load the aliases
			String line;
			while (null != (line = reader.readLine())) {
				line = line.trim();
				if (line.length() == 0) {
					continue;
				}
				if (line.startsWith("#")) { //$NON-NLS-1$
					continue;
				}
				
				{
					// Read our state
					long num;
					
					if (!hasLastAllocated && -1 != (num = readNumber(line, LAST_ALLOCATED_REGEX))) {
						lookup.setLastAllocated(num);
						
						hasLastAllocated = true;
						continue;
					}
					
					if (!hasLastRun && -1 != (num = readNumber(line, LAST_RUN_REGEX))) {
						lookup.setLastRun(num);
						
						hasLastRun = true;
						continue;
					}
				}

				// Parse an ordinary line.
				corrupted |= parseLine(line, lookup);
			}
		} finally {
			reader.close();
		}
		// if corrupted, we can save what we could
		if (corrupted) {
			lookup.save();
		}
	}
	
	private static long readNumber(String line, Pattern regex) {
		Matcher matcher = regex.matcher(line.trim());
		if (matcher.matches()) {
			String number = matcher.group(1);
			if (number != null) {
				try {
					return Long.valueOf(number);
				} catch (NumberFormatException e) {
					if (AliasLoader.class.getClass().desiredAssertionStatus()) {
						e.printStackTrace();
					}
				}
			}
		}
		
		return -1;
	}

	/**
	 * Returns {@code true} if the line is corrupted, {@code false} otherwise
	 * 
	 * @return {@code true} if the line is corrupted, {@code false} otherwise
	 */
	private static boolean parseLine(String line, UuidAliasRegistry lookup) {
		line = line.trim();
		
		String[] matches = line.split("\\s+"); //$NON-NLS-1$
		
		switch(matches.length) {
		case 3: // old format
		case 5: // new format
		case 6: // new format
			break;
		default:
			// any other value is a sign of corruption
			return true;
		}
		
		String repositoryUrl;
		UUID repositoryId;
		long lastRun;
		ItemType itemType = ItemType.UNKNOWN;
		if (matches.length == 3) {
			repositoryId = UuidAliasRegistry.UNKNOWN_REPOSITORY_ID;
			repositoryUrl = UuidAliasRegistry.UNKNOWN_REPOSITORY_URL;
			try {
				lastRun = Long.valueOf(matches[2]);
			} catch (NumberFormatException ex) {
				return true;
			}
		} else {
			repositoryId = UUID.valueOf(matches[2]);
			repositoryUrl = matches[3];
			try {
				lastRun = Long.valueOf(matches[4]);
			} catch (NumberFormatException ex) {
				return true;
			}

			if (matches.length == 6) {
				try {
					itemType = ItemType.valueOf(matches[5].toUpperCase(Locale.ENGLISH));
				} catch (IllegalArgumentException ex) {
					// Does not match any value in the ItemType enum...
					// If this happens, we do not want to fail the parsing by throwing an IAE.
					// This likely will cause the CLI command to fail as well (See Defect 366479).
					// Since we already have a concept of an 'UNKNOWN' item type, we can use
					// this instead.
					itemType = ItemType.UNKNOWN;
				}
			}
		}
		
		lookup.createAliasFor(UUID.valueOf(matches[0]), matches[1], repositoryUrl, repositoryId, lastRun, itemType, false);
		return false;
	}

	/**
	 * Save the given aliases. It's the caller's responsibility to age aliases 
	 * out of the list.  
	 */
	public static void save(Enumeration<UuidAlias> iterable, long lastAllocated, long lastRun, File file) throws IOException {
		PrintWriter writer = null;
		try {
			writer = new PrintWriter(new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), ENCODING)));
			// Write a version preamble
			writer.print(VERSION_PREAMBLE);
			writer.println(1);
			
			writer.print(LAST_ALLOCATED_LINE);
			writer.println(lastAllocated);
			
			writer.print(LAST_RUN_LINE);
			writer.println(lastRun);
			
			// Write the data
			while (iterable.hasMoreElements()) {
				UuidAlias alias = iterable.nextElement();
			
				StringBuffer buf = new StringBuffer(/* UUID */24
						+ /* monicker */UuidAliasRegistry.WIDTH + /* sep */2 + /* count */4 + /* UUID */ 24 + /* url length */255
						+ 12 /* item type */ + /* fudgefactor */8);

				buf.append(alias.uuid.getUuidValue());
				buf.append(" "); //$NON-NLS-1$
				buf.append(alias.getMonicker());
				buf.append(" "); //$NON-NLS-1$
				buf.append(alias.getRepositoryId().getUuidValue());
				buf.append(" "); //$NON-NLS-1$
				buf.append(alias.getRepositoryUrl());
				buf.append(" "); //$NON-NLS-1$
				buf.append(alias.getLastRun());
				buf.append(" "); //$NON-NLS-1$
				buf.append(alias.getItemType().toString());

				writer.println(buf.toString());
			}
			writer.flush();
		} finally {
			if (writer != null) {
				writer.close();
			}
		}
	}
}
