/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2005, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.filesystem.cli.core.internal.aliases;

import java.io.File;
import java.io.IOException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.SortedSet;
import java.util.TreeSet;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.osgi.util.NLS;

import com.ibm.team.filesystem.cli.core.CommandLineCore;
import com.ibm.team.filesystem.cli.core.internal.Messages;
import com.ibm.team.filesystem.cli.core.internal.ScmClientConfiguration;
import com.ibm.team.filesystem.cli.core.subcommands.IScmClientConfiguration;
import com.ibm.team.filesystem.cli.core.subcommands.IUuidAliasRegistry;
import com.ibm.team.filesystem.cli.core.util.RepoUtil;
import com.ibm.team.filesystem.cli.core.util.RepoUtil.ItemType;
import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.internal.FileSystemStatusUtil;
import com.ibm.team.filesystem.client.internal.LoggingHelper;
import com.ibm.team.filesystem.client.internal.utils.FilesystemLock;
import com.ibm.team.repository.common.UUID;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IClientConfiguration;
import com.ibm.team.rtc.cli.infrastructure.internal.util.StringUtil;

public class UuidAliasRegistry implements IUuidAliasRegistry {

	/** Tokens we use to generate monickers. */
	public static final char[] ALPHABET = new char[] { '0', '1', '2', '3', '4', '5',
			'6', '7', '8', '9'};

	/** Maximum token width in tokens. */
	public static final int WIDTH = 4;

	/** Character that unusable aliases are constructed out of. An alias is 
	 * unusable if we know that it won't make it into the cache. Unusable 
	 * aliases may be passed back to the caller on alias creation, but should 
	 * never be passed into {@link UuidAliasRegistry#findAliasByName(String)}. 
	 */
	public static final char UNUSABLE_CHAR = '-';

	
	static {
		// Verify that our alphabet doesn't contain our unusable character.
		assert !new String(ALPHABET).contains(new String(new char [] {UNUSABLE_CHAR}));
	}

	/** Number of UUIDs to store. */
	private int cacheSize = 3000;

	/** Maximum number of UUIDs in the cache. */
	private int maxCacheSize;

	/** The file containing our aliases. */
	private final File file;

	/** The id of the last allocated alias. We default to the first thousand, 
	 * to train our users that {@link UuidAliasRegistry#ALPHABET}[0] is 
	 * significant.
	 */
	private long lastAllocated = (long)Math.pow(ALPHABET.length, WIDTH - 1);
	
	/**
	 * The number of the last run. If this number is negative during a run, 
	 * the number has rolled over. 
	 */
	private long lastRun;
	
	/** Help us handle rollover. During rollover, set to the least value of
	 * {@link UuidAlias#lastRun}. 
	 */
	private long lowestLastRun = Long.MAX_VALUE;

	/** Set to true after {@link #load()} has been called. Does not indicate 
	 * that a load has been successful, only that a load has been performed.
	 */
	private boolean loaded = false;
	
	public static final String UNKNOWN_REPOSITORY_URL = "<unknown>"; //$NON-NLS-1$
	public static final UUID UNKNOWN_REPOSITORY_ID = UUID.valueOf("_R9t0ULhfEduBzK-rvZth-A"); //$NON-NLS-1$

	private HashMap<UUID, HashMap<UUID, UuidAlias>> repoUuidMap = new HashMap<UUID, HashMap<UUID, UuidAlias>>();
	private HashMap<String, UuidAlias> nameMap = new HashMap<String, UuidAlias>();
	private HashMap<String, UUID> repoUriToUuidMap = new HashMap<String, UUID>();

	/** The sorted version of the monickers. */
	private final SortedSet<UuidAlias> sorted = new TreeSet<UuidAlias>(
			new AliasComparator());
	
	private FilesystemLock lock = null;
	private static final String ALIASES_FILE_LOCK = "aliases.lck"; //$NON-NLS-1$
	private static final long LOCK_WAIT = 20000;
	private long aliasFileLastModTime = 0;
	private long aliasFileSize = 0;
	
	private long aliasesGeneratedInCurrentRun = 0;

	public UuidAliasRegistry(File file, int cacheSize) {
		this.file = file;
		setCacheSize(cacheSize);
		
		if (file != null) {
			lock = new FilesystemLock(new File(file.getParentFile(), ALIASES_FILE_LOCK));
		}
	}

	public void discard(IUuidAlias alias) {
		synchronized (sorted) {
			if (!lock.acquire(LOCK_WAIT, null).isOK()) {
				throw new IllegalStateException("Could not acquire lock"); //$NON-NLS-1$
			}
			
			try {
				load();
				
			    IUuidAlias fromUuid = null;
			    if (repoUuidMap.containsKey(alias.getRepositoryId())) {
				    fromUuid = repoUuidMap.get(alias.getRepositoryId()).remove(alias.getUuid());
			    }
			    IUuidAlias fromAlias = nameMap.remove(alias.getMonicker());
			    boolean inSorted = sorted.remove(alias);
			
			    if (fromUuid != null || fromAlias != null || inSorted) {
				    save();
			    }
			} finally {
				lock.release(null);
			}
		}
	}

	/**
	 * Initialize the monicker cache. Can be called repeatedly.
	 */
	private void load() {
		if (loaded && (file.lastModified() != aliasFileLastModTime || file.length() != aliasFileSize)) {
			loaded = false;
		}
		
		if (loaded) {
			return;
		}
		
		loaded = true; // Even if we fail out, we don't want to retry loading the file

		synchronized (sorted) {
			try {
				AliasLoader.load(this, file);
				aliasFileLastModTime = file.lastModified();
				aliasFileSize = file.length();
			} catch (IOException e) {
				// We can't read the file. Silently fail.
				aliasFileLastModTime = 0;
				aliasFileSize = 0;
			}
		}
	}
	
	public void save() {
		final Iterator<UuidAlias> iter = sorted.iterator();
		final int count = getCacheSize();
		Enumeration<UuidAlias> enumeration = new Enumeration<UuidAlias>() {
			int i = 0;
			public boolean hasMoreElements() {
				return i < count && iter.hasNext();
			}

			public UuidAlias nextElement() {
				UuidAlias alias;
				do {
					alias = iter.next();
				} while (!alias.persist);
				
				i++;
				
				if (lastRun < 0) {
					// Handle rollover 
					alias = alias.createCopyForSave();
					if (lowestLastRun < alias.lastRun) {
						alias.lastRun -= lowestLastRun; 
					}
					else {
						alias.lastRun = 0;
					}
				}
				
				return alias;
			}
		};
		
		synchronized (sorted) {
			try {
				AliasLoader.save(enumeration, lastAllocated, lastRun < 0 ? Long.MAX_VALUE - lowestLastRun : lastRun, file);
				aliasFileLastModTime = file.lastModified();
				aliasFileSize = file.length();
			} catch (IOException e) {
				// just log it
				LoggingHelper.log(FileSystemStatusUtil.getStatusFor(IStatus.WARNING, 
			    		NLS.bind(Messages.UuidAliasRegistry_CouldNotWriteToFile, ScmClientConfiguration.ALIAS_FILENAME), e));
			}
		}
	}

	/**
	 * Create the given alias with the suggested values. 
	 * @param uuid
	 * @param alias
	 * @param aliasLastRun Note that this will be manipulated when  
	 * 		{@link #lastRun} rolls over. 
	 * @return
	 */
	public IUuidAlias createAliasFor(UUID uuid, String alias, String repositoryUrl, UUID repositoryId, long aliasLastRun,
	        ItemType itemType, boolean save) {
		if (lastRun < 0) {
			// Handle rollover
			if (aliasLastRun < 0) {
				// We're being called from createAliasFor(UUID)
				aliasLastRun = Long.MAX_VALUE; // We know that #lowestLastRun will be subtracted from this 
			}
			else {
				lowestLastRun = Math.min(lowestLastRun, aliasLastRun);
			}
		}
		
		UuidAlias moni = new UuidAlias(alias, uuid, repositoryUrl, repositoryId, aliasLastRun, itemType);
		if (!repoUriToUuidMap.values().contains(repositoryId)) {
		    repoUriToUuidMap.put(repositoryUrl, repositoryId);
		}

		synchronized (sorted) {
			assert !hasUuid(uuid, repositoryId);
			assert !nameMap.containsKey(alias);
			
			HashMap<UUID, UuidAlias> uuidMap = repoUuidMap.get(repositoryId);
			if (uuidMap == null) {
				uuidMap = new HashMap<UUID, UuidAlias>();
				repoUuidMap.put(repositoryId, uuidMap);
			}
			
			uuidMap.put(uuid, moni);
			nameMap.put(alias, moni);
			sorted.add(moni);
			
			assert nameMap.size() == sorted.size();
			
			if (save) {
				save();
	            aliasesGeneratedInCurrentRun++;
			}
		}

		return moni;
	}

	public IUuidAlias createAliasFor(UUID uuid, String repositoryUrl, ItemType itemType) {
		synchronized (sorted) {
			if (!lock.acquire(LOCK_WAIT, null).isOK()) {
				throw new IllegalStateException("Could not acquire lock"); //$NON-NLS-1$
			}
			
			try {
		        UUID repositoryId = RepoUtil.getSharedRepository(repositoryUrl, false).getId();
		
		        load();

		        // Ensure this monicker doesn't already exist
		        if (hasUuid(uuid, repositoryId)) {
			        throw new IllegalArgumentException("Monicker for " //$NON-NLS-1$
					        + uuid.getUuidValue() + " @ " + repositoryUrl + " already exists: " //$NON-NLS-1$ //$NON-NLS-2$
					        + repoUuidMap.get(repositoryId).get(uuid).getMonicker());
		        }

				// we allow a max of 1000 aliases per run, check if we have exceeded that
				if (aliasesGeneratedInCurrentRun >= 1000) {
					// Create a new (non-persisted) alias. 
					UuidAlias alias = (UuidAlias)createAliasFor(uuid, 
							StringUtil.pad(new StringBuffer(), WIDTH, UNUSABLE_CHAR).toString(), repositoryUrl,
							repositoryId, lastRun, itemType, false);
					alias.persist = false;
					
					return alias;
				}
		
				// if we have exceeded our max cache capacity, trim the cache to make way for new aliases
				if (sorted.size() >= maxCacheSize) {
				    HashMap<UUID, UuidAlias> uuidMap = repoUuidMap.get(repositoryId);

					// trim by cachSize number of aliases
					for (int i = 0; i < cacheSize; i++) {
				        IUuidAlias lastAlias = sorted.last();
				        
				        if (uuidMap != null) {
				            uuidMap.remove(lastAlias.getUuid());
				        }
			            nameMap.remove(lastAlias.getMonicker());
				        sorted.remove(lastAlias);
				    }
				    
					// report rollover
				    LoggingHelper.log(FileSystemStatusUtil.getStatusFor(IStatus.WARNING, Messages.UuidAliasRegistry_RolloverWarning, null));
				}

				// Generate a new name
				String candidate = null;
				
	            // try to find any unused alias within our permitted aliases (0000 and 9999)
	            long maxTries = (long)Math.pow(ALPHABET.length, WIDTH);

	            for (int count = 0; count < maxTries; count++) {
				    candidate = generateAlias();
				    
				    if (!nameMap.containsKey(candidate)) {
				        break;
				    }
				}
				
				return createAliasFor(uuid, candidate, repositoryUrl, repositoryId, lastRun, itemType, true);
			} finally {
				lock.release(null);
			}
		}
	}
	
	private boolean hasUuid(UUID uuid, UUID repositoryId) {
		HashMap<UUID, UuidAlias> uuidMap = new HashMap<UUID, UuidAlias>();
		
		// add the UUID's where the repository info was not available, for backwards compatibility 
		if (repoUuidMap.containsKey(UNKNOWN_REPOSITORY_ID)) {
			uuidMap.putAll(repoUuidMap.get(UNKNOWN_REPOSITORY_ID));
		}
		
		if (repoUuidMap.containsKey(repositoryId)) {
			uuidMap.putAll(repoUuidMap.get(repositoryId));
		}
		
		if (uuidMap.containsKey(uuid)) {
			return true;
		}
		
		return false;
	}

	/**
	 * Generate the next alias
	 */
	private String generateAlias() {
        long acc = lastAllocated++ % (long)Math.pow(ALPHABET.length, WIDTH);
        int SIZE = WIDTH - 1;

        char newName[] = new char[WIDTH];
        
        for (int i = SIZE; i >= 0; i--) {
            long power = (long) Math.pow(ALPHABET.length, i);

            newName[SIZE - i] = ALPHABET[(int) (acc / power)];
            acc = acc % power;
        }
        
        // reset lastAllocated if it reached the max value
        if (lastAllocated == Long.MAX_VALUE) {
            lastAllocated = 1000;
        }
        
        return new String(newName);
	}


	public IUuidAlias findAliasByName(String name) throws NoSuchAliasException {
		synchronized (sorted) {
			if (!lock.isAcquired() && !lock.acquire(LOCK_WAIT, null).isOK()) {
				throw new IllegalStateException("Could not acquire lock"); //$NON-NLS-1$
			}
			
			try {
				load();
				
			    UuidAlias alias = nameMap.get(name);
			    if (alias == null) {
				    throw new NoSuchAliasException();
			    }
			    
			    // updatedAlias is not null when the repository URI has to be updated
			    UuidAlias updatedAlias = validateAndUpdateRepositoryUrl(alias);
				if (updatedAlias != null) {
			    	updateAndIncrement(alias, updatedAlias);
			    	alias = updatedAlias;
				} else {
					increment(alias);
				}
		
			    return alias;
		    } finally {
				lock.release(null);
			}
		}
	}

	/**
	 * Up the usage count associated with this alias.
	 */
	private void increment(UuidAlias alias) {
		synchronized (sorted) {
			long aliasLastRun = lastRun < 0 ? 0 : lastRun;
			
			if (alias.lastRun != aliasLastRun) {
				sorted.remove(alias);
			
				alias.lastRun = aliasLastRun;
			
				sorted.add(alias);
				
				save();
			}
		}
	}

    /**
     * Update alias and up the usage count associated with this alias.
     */
    private void updateAndIncrement(UuidAlias alias, UuidAlias updatedAlias) {
        synchronized (sorted) {
            assert (updatedAlias != null);
            long aliasLastRun = lastRun < 0 ? 0 : lastRun;
            
            if (alias.lastRun != aliasLastRun) {
                sorted.remove(alias);
           
                updatedAlias.lastRun = aliasLastRun;
            
                sorted.add(updatedAlias);
                
                save();
            }
        }
    }
    
	/**
	 * Validate and update repository URI in the alias.
	 * 
	 * After a server has been renamed, repository URI in the alias is stale.
	 * Hence lookup for the new repository URI and update the alias.
	 */
    private UuidAlias validateAndUpdateRepositoryUrl(UuidAlias alias) {
    	
    	UuidAlias updatedAlias = null;
    	
	    IScmClientConfiguration config = CommandLineCore.getConfig();
	    String updatedRepositoryUrl = null; 
		if (config != null) {
			try {
				updatedRepositoryUrl = config.getRepositoryURI(alias.getRepositoryId(),	alias.getRepositoryUrl());
			} catch (FileSystemException ex) {
				// do nothing
				// we will never reach here as alias.getRepositoryUrl is always not null
			}
		}
		if (updatedRepositoryUrl != null && !updatedRepositoryUrl.equals(alias.getRepositoryUrl())) {
			updatedAlias = new UuidAlias(alias.getMonicker(), alias.getUuid(), updatedRepositoryUrl,
					alias.getRepositoryId(), alias.getLastRun(), alias.getItemType());
			repoUriToUuidMap.put(updatedRepositoryUrl, alias.getRepositoryId());
		}
    	return updatedAlias;
    }

    public IUuidAlias findAliasByUuid(String uuidText, String repositoryUrl) throws NoSuchAliasException {
		return findAliasByUuid(UUID.valueOf(uuidText), repositoryUrl);
	}
	
	public IUuidAlias findAliasByUuid(UUID uuid, String repositoryUrl) throws NoSuchAliasException {
		synchronized (sorted) {
			if (!lock.acquire(LOCK_WAIT, null).isOK()) {
				throw new IllegalStateException("Could not acquire lock"); //$NON-NLS-1$
			}
			
			try {
                load();

                // get the repository id
		        UUID repositoryId = null;
		        if (repositoryUrl != null) {
		            try {
		                repositoryId = RepoUtil.getSharedRepository(repositoryUrl, false).getId();
		            } catch (Exception e) {
		                ; // not to worry, lets see if we can match the url
		            }
		            if (repositoryId == null) {
		                repositoryId = repoUriToUuidMap.get(repositoryUrl);
		            }
		        }
		        if (repositoryId == null) {
		            repositoryId = UNKNOWN_REPOSITORY_ID;
		        }
		
		        // get the UUID's for the repo
			    HashMap<UUID, UuidAlias> uuidMap = new HashMap<UUID, UuidAlias>();
		
			    if (repositoryId == UNKNOWN_REPOSITORY_ID) {
			        // if we do not know which repository then add all the UUID's
			        for (UUID uuidKey : repoUuidMap.keySet()) {
			            uuidMap.putAll(repoUuidMap.get(uuidKey));
			        }			        
			    } else {
    			    // add the UUID's where the repository info was not available, for backwards compatibility 
    			    if (repoUuidMap.containsKey(UNKNOWN_REPOSITORY_ID)) {
    				    uuidMap.putAll(repoUuidMap.get(UNKNOWN_REPOSITORY_ID));
    			    }

    			    // add the UUID's for the specified repository
    			    if (repoUuidMap.containsKey(repositoryId)) {
    				    uuidMap.putAll(repoUuidMap.get(repositoryId));
    			    }
			    }
			
			    UuidAlias alias = null;
			    if (uuidMap.containsKey(uuid)) {
				    alias = uuidMap.get(uuid);
			    }
			
			    if (alias == null) {
				    throw new NoSuchAliasException();
			    }
			    
			    // Update the repo information if the alias has unknown repo.
			    // This may happen because of the migration from 2.x to future versions
			    if (repositoryUrl != null && alias.getRepositoryId().equals(UNKNOWN_REPOSITORY_ID)) {
			        UuidAlias updatedAlias = new UuidAlias(alias.getMonicker(), alias.getUuid(), repositoryUrl,
			                repositoryId, alias.getLastRun(), alias.getItemType());
	                updateAndIncrement(alias, updatedAlias);
	                alias = updatedAlias;
			    } else {
			    	UuidAlias updatedAlias = validateAndUpdateRepositoryUrl(alias);
			    	// updatedAlias is not null when the repository URI has to be updated
					if (updatedAlias != null) {
						updateAndIncrement(alias, updatedAlias);
						alias = updatedAlias;
					} else {
						increment(alias);
					}
			    }
			
			    return alias;
		    } finally {
				lock.release(null);
			}
		}
	}

	public int getAliasCount() {
		synchronized (sorted) {
		    load();
			return sorted.size();
		}
	}

	
	SortedSet<UuidAlias> getSorted() {
		synchronized (sorted) {
			load();
			return sorted;
		}
	}

	/** Find the number of monickers to store. */
	public int getCacheSize() {
		return cacheSize;
	}

	/** Set the number of monickers to store. */
	public void setCacheSize(int cacheSize) {
		this.cacheSize = cacheSize;
		maxCacheSize = (int) Math.pow(ALPHABET.length, WIDTH);
	}
	
	public long getLastAllocated() {
		return lastAllocated;
	}
	
	/**
	 * Update the identity of the last allocated alias. 
	 */
	public void setLastAllocated(long last) {
		lastAllocated = last;
	}
	
	/** For debugging only. Checks to see if the given alias exists in the 
	 * registry.
	 */
	public boolean containsAlias(String alias) {
		synchronized (sorted) {
			load();
			return nameMap.containsKey(alias);
		}
	}

	/**
	 * @param run
	 */
	public void setLastRun(long run) {
		lastRun = run + 1;
	}
}
