/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2009, 2013. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use,
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.cli.minimal.client;

import static com.ibm.team.filesystem.cli.minimal.protocol.MessageType.EXIT_CODE;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintStream;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.nio.channels.ServerSocketChannel;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Map;

import org.eclipse.core.runtime.Path;

import com.ibm.team.filesystem.cli.core.Constants;
import com.ibm.team.filesystem.cli.minimal.client.HttpSwitchingClient.HttpSwitchResult;
import com.ibm.team.filesystem.cli.minimal.client.HttpSwitchingClient.ProtocolNegotiationException;
import com.ibm.team.filesystem.cli.minimal.protocol.Message;
import com.ibm.team.filesystem.cli.minimal.protocol.MessageType;
import com.ibm.team.filesystem.cli.minimal.protocol.ProtocolUtil;
import com.ibm.team.filesystem.cli.minimal.protocol.Reader;
import com.ibm.team.filesystem.client.FileSystemCore;
import com.ibm.team.filesystem.client.internal.EnvironmentVariables;
import com.ibm.team.filesystem.client.internal.utils.IDaemonLocation;
import com.ibm.team.filesystem.client.internal.utils.IDaemonRegistry;
import com.ibm.team.filesystem.client.internal.utils.IDaemonRegistry.IDaemonSandbox;
import com.ibm.team.filesystem.client.internal.utils.IDaemonRegistry.IRegistryEntry;
import com.ibm.team.filesystem.client.restproxy.DaemonRegistry;
import com.ibm.team.rtc.cli.infrastructure.internal.util.PromptUtil;
import com.ibm.team.rtc.cli.infrastructure.internal.util.StringUtil;

/**
 * Entry point for the lightweight CLI. 
 */
public class FrontEndClient {
	

	public static final String REQUEST_URI = "/com.ibm.team.filesystem.cli.client.internal.daemon.CommandLineClient";

	/**
     * Port of the daemon we're supposed to attach to.
     */
    private final int port;
    
    /**
     * Secret key protecting the daemon we're supposed to connect to.
     */
    private final String secretKey;

    /**
     * Arguments passed to the daemon.
     */
	private final String[] args;

	/**
	 * Environment we're to pass to the daemon. 
	 */
	private final Map<String, String> env;
	
	/**
	 * Current working directory. 
	 */
	private final String cwd;

	private final InputStream stdin;
	
	private final PrintStream stdout;

	private final PrintStream stderr;
	
	private int exitCode = -1;

	private final String appName;

	
	public static int run(String[] args, Map<String, String> env, InputStream in, PrintStream out, PrintStream err) {
		File config;
		try {
			config = findConfigDirectory(args, env);
		} catch (IOException e1) {
			System.err.println("Could not resolve configuration directory: " + e1.getMessage());
			return 13;
		}
		
		String cwd = findEffectiveCwd(args);
		
		// Get the registry
		DaemonRegistry registry;
		if (config == null) {
			try {
				registry = (DaemonRegistry) DaemonRegistry.getInstance();
			} catch (IOException e) {
				e.printStackTrace();
				System.err.println("Could not canonicalize path to daemon registry");
				return 13;
			}
		} else {
			registry = new DaemonRegistry(config);
		}
		
		if (args.length < 1) {
			throw new IllegalArgumentException("Requires at least one argument: name-of-executable");
		}
		String[] argsToPass = new String[args.length - 1];
		System.arraycopy(args, 1, argsToPass, 0, argsToPass.length);
		
		// Find the daemon to talk to
		IDaemonLocation loc = findDaemon(registry, cwd);
		
		int exit;
		while (true) {
			if (loc == null) {
				// There is no daemon - start a new one
				try {
					loc = startNewDaemon(config);
				} catch (IOException e) {
					e.printStackTrace(err);
					System.err.println("Could not start daemon");
					return 13;
				}
			}
			// Connect and run
			FrontEndClient fec = new FrontEndClient(loc.getKey(),
					loc.getPort(), new File(args[0]).getName(), argsToPass,
					env, cwd, in, out, err);

			try {
				fec.run();
				exit = fec.exitCode;
				// start a new daemon and rerun the command
				if (exit == Constants.STATUS_REQUEST_RERUN) {
					loc = null;
					continue;
				}
			} catch (ProtocolNegotiationException e) {
				e.printStackTrace(err);
				exit = 13;
			} catch (IOException e) {
				e.printStackTrace(err);
				exit = 13;
			} catch (MessageProtocolException e) {
				e.printStackTrace(err);
				exit = 13;
			}
			
			return exit;
		}
	}

	/**
	 * Start a new daemon.
	 * 
	 * @param config Config directory to use. 
	 */
	private static IDaemonLocation startNewDaemon(File config) throws IOException {
		// Guess the name of the daemon executable
		String pathToDaemon = System.getenv("SCM_DAEMON_PATH");
		if (pathToDaemon == null) {
			pathToDaemon = "scm";
		}
		
		String connTimeOut = System.getenv(EnvironmentVariables.SCM_DAEMON_CONNECTION_TIME_OUT);
		if (connTimeOut == null) {
			connTimeOut = "120000";
		}
		String inactivTimeOut = System.getenv(EnvironmentVariables.SCM_DAEMON_INACTIVE_TIME_OUT);
		if (inactivTimeOut == null) {
			inactivTimeOut = "7200000";
		}

		// Create the argument list
		ArrayList<String> args = new ArrayList<String>(13);
		Collections.addAll(args,				
				pathToDaemon, "--config", config.getAbsolutePath(), 
				"daemon", "start", 
				"--connection-timeout", connTimeOut, 
				"--inactive-timeout", inactivTimeOut
		);
		
		
		if (System.getenv().containsKey(EnvironmentVariables.SCM_DAEMON_START_DEBUG)) {
			// We're asked to start in a debug mode. This is dangerous, as it 
			// opens the debug port, which is accessible to any user.
			
			// Find a port. There's a race condition here. But that's the standard. =(
			ServerSocketChannel sc = ServerSocketChannel.open();
			sc.socket().bind(new InetSocketAddress(InetAddress.getByName(null), 0));
			
			int localPort;
			try {
				localPort = sc.socket().getLocalPort();
			} finally {
				sc.close();
			}

			args.add("-vmargs");
			args.add("-Xrunjdwp:transport=dt_socket,server=y,suspend=n,address=" + localPort);
		}
		
		// Start the daemon
		ProcessBuilder pb = new ProcessBuilder();
		pb.command(args);
		
		Process proc = pb.start();
		
		// Read the connection coordinates
		InputStream in = proc.getInputStream();
		
		BufferedReader reader = new BufferedReader(new InputStreamReader(in, Charset.defaultCharset().name()));

		String portLine = reader.readLine();
		if (portLine == null) {
			throw handleDaemonExecFailure(proc);
		}
		
		if (portLine.startsWith("Listening for transport dt_socket")) {
			// If we're debugging the daemon, then the VM will helpfully 
			// pollute our output. Ignore the pollution. 
			portLine = reader.readLine();
			if (portLine == null) {
				throw handleDaemonExecFailure(proc);
			}
		}
		
		String keyLine = reader.readLine();
		if (keyLine == null) {
			throw handleDaemonExecFailure(proc);
		}
		
		final String port = portLine.substring("Port: ".length()).trim();
		final String key = keyLine.substring("Key: ".length()).trim();
		
		return new IDaemonLocation() {
			public String getKey() {
				return key;
			}

			public int getPort() {
				return Integer.parseInt(port);
			}
		};
	}

	/**
	 * The daemon truncated its stdout text. Emit the contents of stderr
	 * and die. 
	 * @throws IOException 
	 */
	private static IOException handleDaemonExecFailure(Process proc) throws IOException {
		InputStream err = proc.getErrorStream();
		
		BufferedReader reader = new BufferedReader(new InputStreamReader(err, Charset.defaultCharset().name()));
		
		String s;
		try {
			while (null != (s = reader.readLine())) {
				System.err.println(s);
			}
		} catch (IOException e) {
			e.printStackTrace();
		}

		int exitVal = waitForDaemonToTerminate(proc, 60000);
		if (exitVal == -1) {
			System.err.println("Daemon closed stdout before printing connection coordinates.");
			proc.destroy();
			exitVal = 43; //< See com.ibm.team.filesystem.cli.core.Constants.STATUS_DAEMON_FAILED_TO_START
		}

		System.exit(exitVal);
		
		return new IOException("Unreachable");
	}

	private static int waitForDaemonToTerminate(Process proc, long waitMillis) {
		long stopTime = System.currentTimeMillis() + waitMillis;
		
		int exitValue = -1;
		while (System.currentTimeMillis() < stopTime) {
			try {
				exitValue = proc.exitValue();
				break;
			} catch (IllegalThreadStateException e) {
				try {
					Thread.sleep(100);
				} catch (InterruptedException e1) {
				}
			}
		}
		
		return exitValue;
	}

	public static void main(String[] args) {
		System.exit(run(args, System.getenv(), System.in, System.out, System.err));
	}
	
    
	public FrontEndClient(String secretKey, int port, String appName, String[] args, Map<String, String> env, String cwd, InputStream stdin, PrintStream stdout, PrintStream stderr) {
	    this.port = port;
	    this.secretKey = secretKey;
		this.appName = appName;
	    this.args = args;
	    this.env = env;
	    this.cwd = cwd;
	    this.stdin = stdin;
	    this.stdout = stdout;
	    this.stderr = stderr;
	}
	
	/**
	 * Calculate the daemon registry directory, either as the default location, 
	 * or as the result of a "--config" argument.
	 *   
	 * @return A canonical file representing the cwd. 
	 */
	public static File findConfigDirectory(String[] args, Map<String, String> env) throws IOException {
		String path = StringUtil.getPathArgument(args, null, "--config");
		
		if (path == null) {
			path = FileSystemCore.getUserConfigDir(env).getConfigDir();
		}

		File result = new File(path);
		try {
			//This use of getCanonicalFile is probably correct, as we would want to expand
			//symlink targets if the user specified symlinks in --config
			result = result.getCanonicalFile();
		} catch (IOException e) {
		}
		
		return result;
	}
	
	
	
	/**
	 * Calculate the current working directory, either as the actual CWD, or 
	 * as the result of a "-d" or "--directory" argument.  
	 * 
	 * @return A canonical file representing the cwd. 
	 */
	public static String findEffectiveCwd(String[] args) {
		return StringUtil.findEffectiveCwd(args);
	}
	
	/**
	 * Connect to the daemon and perform initial handshake. 
	 *  
	 * @throws ProtocolNegotiationException If the client could not negotiate 
	 * 			a connection with the daemon. 
	 * 
	 * @throws IOException If there was a transport layer problem with the 
	 * 			connection to the daemon.
	 * 
	 * @return An integer return value that should be dumped as an exit code. 
	 */
	public int run() throws ProtocolNegotiationException, IOException, MessageProtocolException {
		HttpSwitchResult conn = HttpSwitchingClient.negotiate("localhost", port, secretKey, 
				REQUEST_URI, ProtocolUtil.PROTOCOL_NAME);

		try {
			ProtocolUtil.writeSubcommandDescription(conn, appName, args, env, cwd);

			while (true) {
				Message msg = Reader.readMessage(conn.getInputStream());
				
				if (msg == null) {
					conn.getInputStream().close();
					conn.close();
					return exitCode;
				}
				
				switch (msg.getMessageType()) {
				case STDOUT: stdout.write(msg.getFields()[0]); stdout.flush(); break;
				case STDERR: stderr.write(msg.getFields()[0]); stderr.flush(); break;
				
				case PASSWORD_READ: 
					password_read(conn.getOutputStream(), msg);
					break;

				case STDIN_READ:
					stdin_read(conn.getOutputStream(), msg);
					break;
					
				case EXIT_CODE:
					if (msg.getFields()[0].length != 1) {
						throw new MessageProtocolException("Require exactly one byte in " + EXIT_CODE);
					}
					
					exitCode = msg.getFields()[0][0];
					break;
					
				default:
					throw new RuntimeException("Unknown message: " + msg);
				}
			}
		} finally {
			conn.close();
		}
	}
	
	
	private void stdin_read(OutputStream outputStream, Message msg) throws IOException {
		int size = Integer.parseInt(new String(msg.getFields()[0], "UTF-8"));

		byte[] buf = new byte[size];
		int read = System.in.read(buf);
		
		if (read == -1) {
			ProtocolUtil.writeMessage(outputStream, MessageType.STDIN_CLOSED);
		}
		else {
			ProtocolUtil.writeMessage(outputStream, MessageType.STDIN_RESPONSE, buf);
		}
		
		outputStream.flush();
	}

	private void password_read(OutputStream outputStream, Message msg) throws IOException {
		boolean mask = Boolean.parseBoolean(new String(msg.getFields()[1], "UTF-8"));
		
		String password = PromptUtil.prompt(new String(msg.getFields()[0], "UTF-8"), mask, stdin, stdout);
		
		ProtocolUtil.writeMessage(outputStream, MessageType.PASSWORD_RESPONSE, password.getBytes("UTF-8"));
		outputStream.flush();
	}

	/**
	 * Locate a daemon to use for this session. Selection algorithm:
	 * <ul>
	 *  <li>Walk the daemon registry. Accept the first daemon that has a CFA 
	 *    registered above the CWD.
	 *  <li>(the CLI is outside of a CFA) Connect to the first daemon without a CFA.
	 *  <li>(the CLI is outside of a CFA, and there are no free daemons) Start a new daemon.
	 * </ul>
	 * 
	 * @param registry The registry we should be reading from. 
	 * 
	 * @param cwd The current working directory. Must be canonical. 
	 * 
	 * @return The daemon we're supposed to connect to, or <code>null</code>
	 * 		if there is no appropriate daemon.
	 */
	public static IRegistryEntry findDaemon(IDaemonRegistry registry, String cwd) {
		Collection<IRegistryEntry> entries = registry.listDaemons(null);
		
		Path cwdPath = new Path(cwd);
		
		IRegistryEntry uncommitted = null;
		
		for (IRegistryEntry entry : entries) {
			boolean hasSandbox = false;
			for (IDaemonSandbox sandbox : entry.getRegisteredSandboxes()) {
				hasSandbox = true;
				File f = sandbox.getSandboxRoot();
				try {
					f = f.getCanonicalFile();
				} catch (IOException e) {
				}
				
				Path sandboxPath = new Path(f.getAbsolutePath());
				
				if (sandboxPath.isPrefixOf(cwdPath)) {
					return entry;
				}
			}
			
			if (!hasSandbox) {
				uncommitted = entry;
			}
		}
		
		return uncommitted;
	}
}
