/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2009. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use,
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.cli.minimal.client;

import static com.ibm.team.filesystem.cli.minimal.protocol.ProtocolUtil.readLine;
import static com.ibm.team.filesystem.cli.minimal.protocol.ProtocolUtil.writeLine;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;

import com.ibm.team.filesystem.cli.minimal.protocol.ProtocolUtil;
import com.ibm.team.filesystem.client.internal.http.constants.Header;
import com.ibm.team.filesystem.client.internal.http.constants.ResponseCode;

/**
 * Manages an HTTP request that switches protocols while using our FSD security 
 * mechanism (ie, the {@link Header#SECRET_KEY}). 
 */
public class HttpSwitchingClient {
	
	/** The result of a switch request. */
	public static class HttpSwitchResult {
		private final Socket socket;
		private final InputStream in;
		private final OutputStream out;
		
		private HttpSwitchResult(Socket socket, InputStream in, OutputStream out) {
			this.socket = socket;
			this.in = in;
			this.out = out;
		}
		
		/**
		 * Get a stream that can be used by the client to read from the server. 
		 */
		public InputStream getInputStream() {
			return in;
		}

		/**
		 * Get a stream that the client can use to write to the server. 
		 */
		public OutputStream getOutputStream() {
			return out;
		}

		/**
		 * Free the resources associated with this connection. This MUST be 
		 * called 
		 */
		public void close() throws IOException {
			socket.close();
		}
	}

	/**
	 * Indicates that the server sent an unexpected response. 
	 */
	public static class ProtocolNegotiationException extends Exception {

		private static final long serialVersionUID = -8913782636727288881L;
		
		public ProtocolNegotiationException(String message) {
			super(message);
		}
	}

	/**
	 * Negotiate a non-HTTP protocol using the FSD HTTP protocol.
	 * 
	 * @param host The host to connect to. 
	 * @param port The port to connect to. 
	 * @param key The secret key to pass to the server (see IDaemonHandle#getKey()).
	 * @param path The URL to get.  
	 * @param targetProtocol The desired protocol to switch to.
	 * 
	 * @return An object containing streams to connect to.  
	 */
	public static HttpSwitchResult negotiate(String host, int port, String key, String path, String targetProtocol) throws ProtocolNegotiationException, IOException {
        Socket s = new Socket(host, port);

        OutputStream out = s.getOutputStream();
        
        // Request
        writeLine(out, "GET " + path + " HTTP/1.1");
        writeLine(out, Header.SECRET_KEY + ": " + key);
        writeLine(out, Header.CONNECTION + ": " + Header.UPGRADE);
        writeLine(out, Header.UPGRADE + ": " + targetProtocol);
        ProtocolUtil.writeHttpBlockTerminator(out);
        
        out.flush();
        
        // Read the response
        InputStream in = s.getInputStream();
        
        int state = 0;
        while (true) {
            String str = readLine(in);

            if (state == 0) {
                // In initial response line
                if (!str.contains(Integer.toString(ResponseCode.SWITCHING_PROTOCOLS.getCode()))) {
                	throw new ProtocolNegotiationException("Expected protocol switch. Got: " + str);
                }
                state = 1;
                
                continue;
            }
            
            if (state == 1) {
                // In headers, but not upgrade type
                if (str.startsWith(Header.UPGRADE)) {
                	if (!str.contains(targetProtocol)) {
                		throw new ProtocolNegotiationException("Unexpected upgrade header: " + str);
                	}
                    state = 2;
                }
                
                continue;
            }
            
            if ("".equals(str.trim())) {
            	if (state != 2) {
            		throw new ProtocolNegotiationException("Unexpected response termination");
            	}
                break;
            }
        }
        
        return new HttpSwitchResult(s, in, out);
	}
}
