/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2009, 2010. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use,
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.cli.minimal.protocol;


/**
 * Type of message thrown across the wire.
 */
public enum MessageType {
    /**
     * A message indicating the name of the application that is being run. 
     * 
     * @toServer
     */
    APP_NAME("app", 1),
    
	/** 
	 * A message describing a command-line argument passed to the running 
	 * app. The single field is the string passed in.
	 * 
	 * @toServer
	 */
	ARGUMENT("arg", 1),
	
	/**
	 * A message describing an environment variable held by the client. The
	 * first field is the name of the variable, the second is the value of the 
	 * variable.
	 * 
	 * @toServer
	 */
	ENVIRONMENT("env", 2),
	
	/**
	 * A flag passed from the app to the server indicating that the arguments 
	 * and environment variables have been sent, and that the server may start 
	 * processing.
	 * 
	 * @toServer
	 */
	PREAMBLE_DONE("preamble-done", 0),
	
	/**
	 * A message containing text to write to the standard out. The single 
	 * field should be written out.
	 * 
	 * @toClient
	 */
	STDOUT("stdout", 1),
	
	/**
	 * A message containing text to write to the standard error. The single 
	 * field should be written out.
	 * 
	 * @toClient
	 */
	STDERR("stderr", 1),
	
	/**
	 * The return code the process should exit with. The single field should 
	 * be interpreted as an integer.
	 * 
	 * @toClient
	 */
	EXIT_CODE("exit-code", 1),
	
	/**
	 * Ask the client to read a password. The first argument is the prompt to 
	 * display, the second is a string indicating whether the input should be 
	 * masked ("true" if so, encoded in UTF-8). 
	 * 
	 * <p>
	 * The server expects a {@link #PASSWORD_RESPONSE} to be sent back. 
	 * 
	 * @toClient
	 */
	PASSWORD_READ("password-read", 2),
	
	/**
	 * Sent by the client in response to a {@link #PASSWORD_READ}. The single 
	 * argument is the password entered by the user. 
	 * 
	 * @toServer
	 */
	PASSWORD_RESPONSE("password-entered", 1), 

	/**
	 * Ask the client to read bytes from the standard in. The first argument 
	 * is the number of bytes to read, encoded as a decimal string in UTF-8.
	 * Expects either {@link #STDIN_RESPONSE} or {@link #STDIN_CLOSED} in 
	 * return. 
	 * 
	 * @toClient
	 */
	STDIN_READ("stdin-request", 1), 
	
	/**
	 * Bytes read from the client's stdin. Sent in response to {@link #STDIN_READ}.
	 * 
	 * @toServer
	 */
	STDIN_RESPONSE("stdin-response", 1),
	
	/**
	 * Indicates that the stdin has been closed. Sent in response to 
	 * {@link #STDIN_READ}.
	 * 
	 * @toServer
	 */
	STDIN_CLOSED("stdin-closed", 0), 
	
	/**
	 * Set the current working directory of the client. Argument is a UTF-8 
	 * encoded string, representing the full path of the CWD. 
	 */
	CWD("cwd", 1), 
	;
	
	
	/** Text to send over the wire. */
	private final String text;
	
	/** Number of arguments. */
	private final int cardinality;
	
	private MessageType(String text, int cardinality) {
		this.text = text;
		this.cardinality = cardinality;
	}
	
	/**
	 * @return The bytes passed across the stream as the message header. 
	 */
	public String getWireText() {
		return text;
	}
	
	/**
	 * Get the number of fields this message may have. 
	 */
	public int getCardinality() {
		return cardinality;
	}
	
	static MessageType fromWireFormat(String s) {
		for (MessageType t : values()) {
			if (t.getWireText().equals(s)) {
				return t;
			}
		}
		
		throw new IllegalArgumentException("Unknown message type \"" + s + "\"");
	}
}
