/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2009, 2013. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use,
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.cli.minimal.protocol;

import static com.ibm.team.filesystem.cli.minimal.protocol.MessageType.PREAMBLE_DONE;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.nio.charset.Charset;
import java.util.Map;

import com.ibm.team.filesystem.cli.minimal.client.HttpSwitchingClient.HttpSwitchResult;

/**
 * Utilities for reading and writing our thin client protocol. 
 */
public abstract class ProtocolUtil {

	/**
	 * Name of the protocol that we speak. 
	 */
	public static final String PROTOCOL_NAME = "vnd.ibm.jazz.cli.minimal-1.0";


	/**
	 * Delimiters that appear on either side of a string.
	 */
	static final byte STRING_DELIM = '"';

	/**
	 * Separators between strings in an argument list. 
	 */
	static final byte FIELD_SEPARATOR = ' ';
	
	static final byte MESSAGE_TYPE_TERMINATOR = '.';
	
	/** Newline for use in our minimal protocol. */
	static byte MINIMAL_NEWLINE = '\n';
	
	/** Newline for use when speaking HTTP. */
	static byte[] HTTP_NEWLINE = new byte[] {'\r', '\n'};
	
	public static final String CHARSET_NAME_UTF8 = "UTF-8";
	
	public static boolean isCharsetUtf8(Charset charset)
	{
		// Sanity check
		if ((charset == null) || (!Charset.isSupported(charset.name())))
		{
			return false;
		}
		
		// Compare the names itself, first
		if (CHARSET_NAME_UTF8.equalsIgnoreCase(charset.name()))
		{
			return true;
		}
		
		// Names are not matching, compare with one of the aliases.
		for (String alias : charset.aliases())
		{
			if (CHARSET_NAME_UTF8.equalsIgnoreCase(alias))
			{
				return true;
			}
		}
		
		// No match so far; return false
		return false;
	}
	
	public static byte[] transcodeToUtf8(byte[] input, Charset srcCharset) throws UnsupportedEncodingException
	{
		// Sanity check
		if ((input == null) || (input.length < 0) || (srcCharset == null))
		{
			throw new IllegalArgumentException();
		}
		
		if (isCharsetUtf8(srcCharset))
		{
			// Already in UTF-8 format. No need to transcode; return as is
			return input;
		}
		
		// Check if the input is an empty array... 
		if (input.length == 0)
		{
			// Input is an empty byte array; we just return it as is; since there is nothing to convert. An empty array is valid in all encodings! :)
			return input;
		}

		byte[] utf8Bytes = null;

		// First, create a string object from the bytes, using the appropriate encoding.
		String srcStr = new String(input, srcCharset);

		// Second, try encoding to UTF-8 by extracting UTF-8 bytes from the string object
		try
		{
			utf8Bytes = srcStr.getBytes(CHARSET_NAME_UTF8);
		}
		catch (Exception e)
		{
			// In case of failure, we could fallback to default charset,
			// but lets be strict to catch any unsupported encoding. Hence, we propagate this
			// error via an exception.
			throw new UnsupportedEncodingException(e.getMessage());
		}
		
		return utf8Bytes;
	}

	/**
	 * Write a CRLF terminated line to the given stream. 
	 * 
	 * @param out The stream to write to.
	 * @param toWrite The string to write. Should not be newline terminated. 
	 * 
	 * @throws IOException
	 */
	public static void writeLine(OutputStream out, String toWrite) throws IOException {
		out.write(toWrite.getBytes("UTF-8"));
		
		out.write(HTTP_NEWLINE);
	}
	
	public static void writeHttpBlockTerminator(OutputStream out) throws IOException {
		out.write(HTTP_NEWLINE);
	}
	
	/**
	 * Read a CRLF terminated line from the given stream.
	 *  
	 * @param in The stream to read from. The stream is not closed under
	 * 		any circumstance.  
	 * 
	 * @return The line (minus terminating CRLF), or <code>null</code> if the 
	 * 		other end properly terminated the connection. 
	 * 
	 * @throws IOException
	 */
	public static String readLine(InputStream in) throws IOException {
		StringBuilder b = new StringBuilder();
		
		boolean inText = true;
		while (true) {
			int read;
			
			read = in.read();

			if (read == -1) {
				if (b.length() == 0 && inText) {
					//client closed the connection
					return null;
				}
				
				// partial request
				throw new IOException("Client closed connection. Received: " + b);
			}

			if (read == '\r' && inText) {
				inText = false;
			}
			else if (read == '\n' && !inText) {
				return b.toString();
			}
			else if (inText) {
				b.append((char)read);
			} 
			else {
				throw new IOException("Unexpected state/input. State: "
						+ (inText ? "inText" : "!inText") 
						+ " input: " + read + "  current line: " + b);
			}
		}
	}

	static final byte ESCAPE_CHAR = '\\';
	private static final byte[] ESCAPE_ESCAPE = new byte[] {ESCAPE_CHAR, '5', 'c'};
	private static final byte[] ESCAPE_QUOTE = new byte[] {ESCAPE_CHAR, '2', '2'};
	private static final byte[] ESCAPE_CR = new byte[] {ESCAPE_CHAR, '0', 'd'};
	private static final byte[] ESCAPE_LF = new byte[] {ESCAPE_CHAR, '0', 'a'};
	
	private static void writeEscaped(OutputStream out, byte[] arg) throws IOException {
		out.write(STRING_DELIM);
		
		for (byte b : arg) {
			switch (b) {
				case '\\': out.write(ESCAPE_ESCAPE); break;
				case '"':  out.write(ESCAPE_QUOTE); break; 
				case '\r': out.write(ESCAPE_CR); break;
				case '\n': out.write(ESCAPE_LF); break;
			
				default: out.write(b);
			}
		}
		
		out.write(STRING_DELIM);
	}

	public static void writeMessage(OutputStream out, MessageType msg, byte[] ... args) throws IOException {
		if (args.length != msg.getCardinality()) {
			throw new IllegalArgumentException();
		}

		// Writing to a buffer significantly increases our comm speed
		ByteArrayOutputStream toWrite = new ByteArrayOutputStream();
		
		toWrite.write(msg.getWireText().getBytes("UTF-8"));
		toWrite.write(MESSAGE_TYPE_TERMINATOR);

		for (byte[] arg : args) {
			toWrite.write(FIELD_SEPARATOR);
			writeEscaped(toWrite, arg);
		}
		
		toWrite.write(MINIMAL_NEWLINE);
		
		out.write(toWrite.toByteArray());
	}
	
	/**
	 * Writes the given commandline context to the server.
	 * 
	 * @param conn The connection to write on.
	 * @param appName The name of the application the FEC is running as.  
	 * @param args The command line arguments.
	 * @param env The environment variables to write.
	 * @param cwd The current working directory we're to report over the wire. 
	 * 
	 * @throws IOException 
	 */
	public static void writeSubcommandDescription(HttpSwitchResult conn,
			String appName, String[] args, Map<String, String> env, String cwd) throws IOException {
		OutputStream out = conn.getOutputStream();

		final String UTF8 = "UTF-8";
		
		writeMessage(out, MessageType.APP_NAME, appName.getBytes(UTF8));
		
		for (String arg : args) {
			writeMessage(out, MessageType.ARGUMENT, arg.getBytes(UTF8));
		}
		
		for (Map.Entry<String, String> entry : env.entrySet()) {
			writeMessage(out, MessageType.ENVIRONMENT, 
					entry.getKey().getBytes(UTF8), entry.getValue().getBytes(UTF8));
		}
		
		writeMessage(out, MessageType.CWD, cwd.getBytes(UTF8));
		
		writeMessage(out, PREAMBLE_DONE);
		
		out.flush();
	}

	public static void writeMessage(OutputStream out, MessageType type, String[] strings) throws IOException {
		byte[][] arrays = new byte[strings.length][];
		
		for (int i = 0; i < strings.length; i++) {
			arrays[i] = strings[i].getBytes("UTF-8");
		}
		
		writeMessage(out, type, arrays);
	}
	
}
