/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2009, 2012. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use,
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.cli.minimal.protocol;

import static com.ibm.team.filesystem.cli.minimal.protocol.ProtocolUtil.ESCAPE_CHAR;
import static com.ibm.team.filesystem.cli.minimal.protocol.ProtocolUtil.FIELD_SEPARATOR;
import static com.ibm.team.filesystem.cli.minimal.protocol.ProtocolUtil.MESSAGE_TYPE_TERMINATOR;
import static com.ibm.team.filesystem.cli.minimal.protocol.ProtocolUtil.MINIMAL_NEWLINE;
import static com.ibm.team.filesystem.cli.minimal.protocol.ProtocolUtil.STRING_DELIM;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;


/**
 * Reads messages from the wire. 
 */
public class Reader {

    private enum MessageReadState {
        DONE_FIELD, IN_FIELD_START, IN_FIELD, IN_FIELD_ESCAPE_0, IN_FIELD_ESCAPE_1
    }

    public static Message readMessage(InputStream in) throws IOException {
        // Read the MessageType
        StringBuilder msgBuf = new StringBuilder();
        while (true) {
            int c = in.read();
            
            if (c == -1) {
                return null;
            }
            
            // See if we have the field terminator
            if (c == MESSAGE_TYPE_TERMINATOR) {
                break;
            }
            
            if (c >= '!' && c <= 'z') {
            	msgBuf.append((char)c);
                continue;
            }
            
            throw new IOException("Unexpected message character: \"" + (char)c + "\" (" + c + ")");
        }
        
        MessageType type = MessageType.fromWireFormat(msgBuf.toString());
        
        // Read the fields
        MessageReadState state = MessageReadState.DONE_FIELD;
        
        byte[] hexAccumulator = {0, 0}; 
        ByteArrayOutputStream fieldAccumulator = null;
        ArrayList<byte[]> fields = new ArrayList<byte[]>(type.getCardinality());
        
        READ: while (true) {
            int c = in.read();

            switch (state) {
            case DONE_FIELD:
                if (c == MINIMAL_NEWLINE) {
                    break READ;
                }
                
                if (c == FIELD_SEPARATOR) {
                    state = MessageReadState.IN_FIELD_START;
                    break;
                }
                
                throw new IOException("Unexpected byte: " + (byte)c + " (" + c + ") in state " + state);
                
                
            case IN_FIELD_START:
                if (c == STRING_DELIM) {
                    state = MessageReadState.IN_FIELD;
                    fieldAccumulator = new ByteArrayOutputStream();
                    break;
                }
                throw new IOException("Unexpected byte: " + (byte)c + " (" + c + ") in state " + state);
                
                
            case IN_FIELD:
                if (c == STRING_DELIM) {
                    state = MessageReadState.DONE_FIELD;
        
                    assert fieldAccumulator != null;
                    fields.add(fieldAccumulator.toByteArray());
                    fieldAccumulator = null;
                    break;
                }
                
                if (c == ESCAPE_CHAR) {
                    state = MessageReadState.IN_FIELD_ESCAPE_0;
                    hexAccumulator[0] = 0;
                    hexAccumulator[1] = 0;
                    break;
                }
                
                assert fieldAccumulator != null;
                fieldAccumulator.write(c);
                break;
                
                
            case IN_FIELD_ESCAPE_0:
                if ((c >= '0' && c <= '9')) {
                    hexAccumulator[0] = (byte)c;
                    state = MessageReadState.IN_FIELD_ESCAPE_1;
                    break;
                }
                
                if ((c >= 'a' && c <= 'f')) {
                    hexAccumulator[0] = (byte)c;
                    state = MessageReadState.IN_FIELD_ESCAPE_1;
                    break;
                }

                throw new IOException("Unexpected byte: " + (byte)c + " (" + c + ") in state " + state);
                
            case IN_FIELD_ESCAPE_1:
                assert fieldAccumulator != null;
                if ((c >= '0' && c <= '9')) {
                    hexAccumulator[1] = (byte)c;
                    fieldAccumulator.write(Integer.parseInt(new String(hexAccumulator, "UTF-8"), 16));
                    
                    state = MessageReadState.IN_FIELD;
                    break;
                }
                
                if ((c >= 'a' && c <= 'f')) {
                    hexAccumulator[1] = (byte)c;
                    fieldAccumulator.write(Integer.parseInt(new String(hexAccumulator, "UTF-8"), 16));
                    
                    state = MessageReadState.IN_FIELD;
                    break;
                }

                throw new IOException("Unexpected byte: " + (byte)c + " (" + c + ") in state " + state);
            }
        }
        
        if (fields.size() != type.getCardinality()) {
            throw new IOException("Field cardinality incorrect for " + type);
        }
        
        // Generate the message
        return new Message(type, fields.toArray(new byte[fields.size()][]));
    }

}
