/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2009, 2013. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use,
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.cli.minimal.util;

import java.io.File;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.nio.charset.Charset;
import java.util.Collections;
import java.util.SortedSet;
import java.util.TreeSet;

import com.ibm.team.filesystem.client.internal.EnvironmentVariables;

/**
 * Generate a shell script that will run the FEC
 */
public class GenerateShellScript {

	private static String[] requiredPaths = new String[] {
		"com.ibm.team.filesystem.cli.minimal",
		"com.ibm.team.rtc.cli.infrastructure",
		"com.ibm.team.filesystem.cli.core",//< For PromptUtil
		"com.ibm.team.filesystem.client.daemon",
		"org.eclipse.equinox.common",
		"com.ibm.team.filesystem.client",
		"org.eclipse.osgi",
		"com.ibm.team.repository.common"
	};
	
	private static final String DEFAULT_JAVA_EXECUTABLE = "../../client/eclipse/jdk/jre/bin/java";
	private static boolean DEBUG = false;
	private static final String LSCM_USE_JAVA_FEC = "LSCM_USE_JAVA_FEC";

	public static void main(String[] args) throws IOException {
		if (args.length != 4 && args.length != 5) {
			System.err.println("Require 4 or 5 arguments: path-to-plugins-dir, output-path-directory, filename, daemon-command-name, optional-java-path");
			System.exit(1);
		}
		
		File pluginsDir = new File(args[0]).getCanonicalFile();
		File outputPath = new File(args[1]);
		String suffix = args[2];
		String daemonCommandName = args[3];
		
		String javaPath = args.length == 4 ? null : args[4];
		
		if( !outputPath.exists() || ! outputPath.isDirectory() ){
			System.err.println("output-path-directory must be an existing directory");
			System.exit(1);
		}
		
		// Generate the list of classpaths
		String[] pluginPaths = new String[requiredPaths.length];
		
		for (int i = 0; i < requiredPaths.length; i++) {
			final String prefix = requiredPaths[i];
			
			// Find the matching files/dirs
			String[] matches = pluginsDir.list(new FilenameFilter() {
				public boolean accept(File dir, String name) {
					return name.startsWith(prefix);
				}
			});
			
			// Get the greatest match
			SortedSet<String> set = new TreeSet<String>(String.CASE_INSENSITIVE_ORDER);
			
			Collections.addAll(set, matches);
			
			pluginPaths[i] = set.last();
		}
		
		File batchFile = new File(outputPath + File.separator + suffix + ".bat");
		// Write the script
		writeBatchFile(javaPath, batchFile, pluginPaths);
		
		File shellFile = new File(outputPath + File.separator + suffix);
		writeScript(javaPath, shellFile, pluginPaths, daemonCommandName);
		
		File macShellFile = new File(outputPath + File.separator + suffix + ".sh.mac");
		writeMacScript(macShellFile, pluginPaths, daemonCommandName);
		
		if (!DEBUG)
		{
			return;
		}
		
		String vars = "@echo off\r\n" +
				"SET PRGPATH=%~dp0\r\n" +
				"SET SCM_DAEMON_PATH=%PRGPATH%scm.exe\r\n";
		String fecSelectionLogic = "SET USE_NATIVE=1\r\n" +
				"\r\nIF NOT \"%" + LSCM_USE_JAVA_FEC + "%\"==\"\" (" +  " \r\n" +
				"    SET USE_NATIVE=0\r\n" +
				")\r\n" +
				"\r\nIF NOT \"%" + EnvironmentVariables.SCM_DAEMON_START_DEBUG + "%\"==\"\" (" +  " \r\n" +
				"    SET USE_NATIVE=0\r\n" +
				")\r\n" +
				"\r\nIF NOT EXIST \"%PRGPATH%fec.exe\" (\r\n" +
				"    SET USE_NATIVE=0\r\n" +
				")\r\n";
		String conditionLine = "\r\nIF \"%USE_NATIVE%\"==\"1\" (\r\n";
		String exePath = "\"%PRGPATH%fec.exe\"";
		String elseLine = ") ELSE (";
		String endLine =")\r\n\r\nSET PRGPATH=\r\nSET SCM_DAEMON_PATH=\r\nSET USE_NATIVE=\r\n";
		write('\\', ';', "\r\n", "%PRGPATH%", vars, fecSelectionLogic, conditionLine, exePath, elseLine, javaPath, endLine, batchFile, pluginPaths, "%*");
			
		vars = "#!/bin/sh\n" +
				"PRGPATH=\"`dirname \"$0\"`\"\n" +
				"SCM_DAEMON_PATH=\"${PRGPATH}/" + daemonCommandName + "\"\n" +
				"export SCM_DAEMON_PATH\n";
		fecSelectionLogic = "\nUSE_NATIVE=\"1\"\n\n" +
				"if [ -n \"$" + LSCM_USE_JAVA_FEC + "\" ]; then\n" +
				"    USE_NATIVE=\"0\"\n" +
				"elif [ -n \"$" + EnvironmentVariables.SCM_DAEMON_START_DEBUG + "\" ]; then\n" +
				"    USE_NATIVE=\"0\"\n" +
				"elif [ ! -x \"${PRGPATH}/fec\" ]; then\n" +
				"    USE_NATIVE=\"0\"\n" +
				"fi\n\n";
		conditionLine = "\nif [ \"$USE_NATIVE\" = \"1\" ]; then\n";
		exePath = "\"${PRGPATH}/fec\"";
		elseLine = "else";
		endLine ="fi";
		write('/', ':', "\n", "${PRGPATH}", vars, fecSelectionLogic, conditionLine, exePath, elseLine, javaPath, endLine, shellFile, pluginPaths, "\"$@\"");
	}

	private static void writeBatchFile(String javaPath, File outputPath, String[] pluginPaths) throws IOException {
		FileOutputStream out = new FileOutputStream(outputPath);

		if (javaPath == null) {
		    javaPath = "%PRGPATH%/" + DEFAULT_JAVA_EXECUTABLE;
		}
		
        if (javaPath.charAt(0) == '/') {
            javaPath = javaPath.substring(1);
        }
        
        javaPath = javaPath.replace('/', '\\');
        
        // surround it with quotes so that it can handle spaces in directory paths
        if (javaPath.charAt(0) != '\"') {
            javaPath = "\"" + javaPath + "\"";
        }
        
		try {
			OutputStreamWriter writer = new OutputStreamWriter(out, Charset.defaultCharset());
			
			writer.write("@echo off\r\n");
			writer.write("SET PRGPATH=%~dp0\r\n");
			writer.write("SET JAVA=");
			writer.write(javaPath);
			writer.write("\r\n");
			writer.write("SET SCM_DAEMON_PATH=\"%PRGPATH%scm.exe\"\r\n");
			
			String fecSelectionLogic = "SET USE_NATIVE=1\r\n" +
					"\r\nIF NOT \"%" + LSCM_USE_JAVA_FEC + "%\"==\"\" (" +  " \r\n" +
					"    SET USE_NATIVE=0\r\n" +
					")\r\n" +
					"\r\nIF NOT \"%" + EnvironmentVariables.SCM_DAEMON_START_DEBUG + "%\"==\"\" (" +  " \r\n" +
					"    SET USE_NATIVE=0\r\n" +
					")\r\n" +
					"\r\nIF NOT EXIST \"%PRGPATH%fec.exe\" (\r\n" +
					"    SET USE_NATIVE=0\r\n" +
					")\r\n";

			writer.write(fecSelectionLogic);
			writer.write("\r\nIF \"%USE_NATIVE%\"==\"1\" (\r\n");
			writer.write("    \"%PRGPATH%fec.exe\" %*\r\n");
			writer.write(") ELSE (\r\n");
			writer.write("    %JAVA% -Xquickstart -Xshareclasses -classpath \"");
			
			boolean needsSep = false;
			for (String plugin : pluginPaths) {
				if (needsSep) {
					writer.write(";");
				}
				needsSep = true;
				
				writer.write("%PRGPATH%\\plugins\\" + plugin);
			}
			
			String appName = guessAppName(outputPath);
			
			writer.write("\" com.ibm.team.filesystem.cli.minimal.client.FrontEndClient \"" + appName + "\" %*\r\n)\r\n");		
			
			writer.close();
		} finally {
			out.close();
		}
	}

	/** Determine the name of the executable under windows */
	private static String guessAppName(File outputPath) {
		String name = outputPath.getName();
		
		int lastDot = name.lastIndexOf('.');
		if (lastDot == -1) {
			return name;
		}
		
		return name.substring(0, lastDot);
	}

	private static void writeScript(String javaPath, File outputPath, String[] pluginPaths, String daemonCommandName) throws IOException {
		FileOutputStream out = new FileOutputStream(outputPath);
		
		if (javaPath == null) {
		    javaPath = "${PRGPATH}/" + DEFAULT_JAVA_EXECUTABLE;
		}
		
		try {
			OutputStreamWriter writer = new OutputStreamWriter(out, Charset.defaultCharset());
			
			writer.write("#!/bin/sh\n");
			writer.write("PRGPATH=\"`dirname \"$0\"`\"\n");
			writer.write("export PRGPATH\n");
			writer.write("SCM_DAEMON_PATH=\"${PRGPATH}/" + daemonCommandName + "\"\n");
			writer.write("export SCM_DAEMON_PATH\n");
			writer.write("if [ `uname` = SunOS ];\n");
			writer.write("then\n");
			writer.write("\tEXTRA_JAVA_OPTS=\"\"\n");
			writer.write("else\n");
			writer.write("\tEXTRA_JAVA_OPTS=\"-Xquickstart -Xshareclasses\"\n");
			writer.write("fi\n");
			
			String fecSelectionLogic = "\nUSE_NATIVE=\"1\"\n\n" +
					"if [ -n \"$" + LSCM_USE_JAVA_FEC + "\" ]; then\n" +
					"    USE_NATIVE=\"0\"\n" +
					"elif [ -n \"$" + EnvironmentVariables.SCM_DAEMON_START_DEBUG + "\" ]; then\n" +
					"    USE_NATIVE=\"0\"\n" +
					"elif [ ! -x \"${PRGPATH}/fec\" ]; then\n" +
					"    USE_NATIVE=\"0\"\n" +
					"fi\n\n";
			writer.write(fecSelectionLogic);
			
			writer.write("\nif [ \"$USE_NATIVE\" = \"1\" ]; then\n");
			writer.write("    \"${PRGPATH}/fec\" \"$@\"\n");
			writer.write("else\n");
			writer.write("    " +javaPath);
			writer.write(" $EXTRA_JAVA_OPTS -classpath \"");
			
			boolean needsSep = false;
			for (String plugin : pluginPaths) {
				if (needsSep) {
					writer.write(":");
				}
				needsSep = true;
				
				writer.write( "${PRGPATH}/plugins/" + plugin);
			}
			
			writer.write("\" com.ibm.team.filesystem.cli.minimal.client.FrontEndClient \"$0\" \"$@\"");
			writer.write("\nfi\n");

			writer.close();
		} finally {
			out.close();
		}
	}
	
	private static void writeMacScript(File outputPath, String[] pluginPaths, String daemonCommandName) throws IOException {
		FileOutputStream out = new FileOutputStream(outputPath);
		
		try {
			OutputStreamWriter writer = new OutputStreamWriter(out, Charset.defaultCharset());
			
			writer.write("#!/bin/sh\n");
			writer.write("PRGPATH=\"`dirname \"$0\"`\"\n");
			writer.write("export PRGPATH\n");
			writer.write("SCM_DAEMON_PATH=\"${PRGPATH}/" + daemonCommandName + "\"\n");
			writer.write("export SCM_DAEMON_PATH\n");

			String fecSelectionLogic = "\nUSE_NATIVE=\"1\"\n\n" +
					"if [ -n \"$" + LSCM_USE_JAVA_FEC + "\" ]; then\n" +
					"    USE_NATIVE=\"0\"\n" +
					"elif [ -n \"$" + EnvironmentVariables.SCM_DAEMON_START_DEBUG + "\" ]; then\n" +
					"    USE_NATIVE=\"0\"\n" +
					"elif [ ! -x \"${PRGPATH}/fec\" ]; then\n" +
					"    USE_NATIVE=\"0\"\n" +
					"fi\n\n";
			writer.write(fecSelectionLogic);
			writer.write("\nif [ \"$USE_NATIVE\" = \"1\" ]; then\n");
			writer.write("    \"${PRGPATH}/fec\" \"$@\"\n");
			writer.write("else\n");
			writer.write("    java -classpath \"");
			
			boolean needsSep = false;
			for (String plugin : pluginPaths) {
				if (needsSep) {
					writer.write(":");
				}
				needsSep = true;
				
				writer.write( "${PRGPATH}/plugins/" + plugin);
			}
			
			writer.write("\" com.ibm.team.filesystem.cli.minimal.client.FrontEndClient \"$0\" \"$@\"");
			writer.write("\nfi\n");
			
			writer.close();
		} finally {
			out.close();
		}
	}
	
	private static void write(
			char separatorChar,
			char pathSeparatorChar,
			String newLine,
			String prgPath,
			String vars,
			String fecSelectionLogic,
			String conditionLine,
			String exePath,
			String elseLine,
			String javaPath,
			String endLine,
			File outputPath,
			String[] pluginPaths,
			String forwardVars) throws IOException
	{
		FileOutputStream out = new FileOutputStream(outputPath);
		
		if (javaPath == null) {
		    javaPath = prgPath + separatorChar + DEFAULT_JAVA_EXECUTABLE;
		}
		
		if ('/' != separatorChar) {
			if (javaPath.charAt(0) == '/') {
	            javaPath = javaPath.substring(1);
	        }
			javaPath = javaPath.replace('/', separatorChar);
		}
		
		try {
			OutputStreamWriter writer = new OutputStreamWriter(out, Charset.defaultCharset());
			
			writer.write(vars);

			writer.write(newLine + fecSelectionLogic + newLine + newLine + conditionLine + newLine);
			writer.write("    " + exePath + " " + forwardVars + newLine);
			writer.write(elseLine + newLine);
			writer.write("    %JAVA% -Xquickstart -Xshareclasses -classpath \"");
			
			writer.write(javaPath);
			if (DEBUG) {
				writer.write(" -Xdebug -Xrunjdwp:transport=dt_socket,server=y,suspend=y,address=8000");
			} else {
				writer.write(" -Xquickstart");
			}
			writer.write(" -Xshareclasses -classpath \"");
			
			boolean needsSep = false;
			for (String plugin : pluginPaths) {
				if (needsSep) {
					writer.write(pathSeparatorChar);
				}
				needsSep = true;
				
				writer.write(prgPath);
				writer.write("/plugins/");
				writer.write(plugin);
			}
			
			writer.write("\" com.ibm.team.filesystem.cli.minimal.client.FrontEndClient ");
			writer.write(forwardVars);

			writer.write(endLine + newLine);

			writer.close();
		} finally {
			out.close();
		}
	}
}
