/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.filesystem.cli.tools;

import java.io.BufferedWriter;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.ibm.team.filesystem.cli.core.util.ConnectionInfo;
import com.ibm.team.filesystem.cli.core.util.RepoUtil;
import com.ibm.team.filesystem.cli.core.util.SubcommandUtil;
import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.process.client.IProcessClientService;
import com.ibm.team.process.client.IProcessItemService;
import com.ibm.team.process.common.IProcessArea;
import com.ibm.team.process.common.IProcessAreaHandle;
import com.ibm.team.process.common.IProjectArea;
import com.ibm.team.process.common.ITeamArea;
import com.ibm.team.process.common.ITeamAreaHierarchy;
import com.ibm.team.repository.client.IItemManager;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.UUID;
import com.ibm.team.repository.common.json.JSONArray;
import com.ibm.team.repository.common.json.JSONObject;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IOptionSource;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.ICommandLine;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.Options;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.PositionalOptionDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.ConflictingOptionException;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.client.IWorkspaceManager;
import com.ibm.team.scm.client.SCMPlatform;
import com.ibm.team.scm.client.internal.WorkspaceManager;
import com.ibm.team.scm.common.IBaseline;
import com.ibm.team.scm.common.IBaselineHandle;
import com.ibm.team.scm.common.IComponent;
import com.ibm.team.scm.common.IComponentHandle;
import com.ibm.team.scm.common.IScmService;
import com.ibm.team.scm.common.IWorkspace;
import com.ibm.team.scm.common.IWorkspaceHandle;
import com.ibm.team.scm.common.dto.IWorkspaceSearchCriteria;
import com.ibm.team.scm.common.internal.ComponentEntry;
import com.ibm.team.scm.common.internal.HistoricBasis;

public class DumpStreamsAndComponentsCmd extends com.ibm.team.filesystem.cli.client.AbstractSubcommand implements IOptionSource {

	private static final PositionalOptionDefinition OPT_PROJECT_AREA = new PositionalOptionDefinition("pa", 0, 1); //$NON-NLS-1$
	
	private static final int MY_FETCH_PAGE_SIZE = 500;
	
	@Override
	public void run() throws FileSystemException {
		ICommandLine subargs = config.getSubcommandCommandLine();
		
		String projectAreaName = subargs.getOption(OPT_PROJECT_AREA, null);
		
		ConnectionInfo connection = config.getConnectionInfo();

		
		ITeamRepository repo = RepoUtil.login(config, connection);
		
		BufferedWriter processAreas;
		BufferedWriter streams;
		BufferedWriter baselines;
		try {
			processAreas = new BufferedWriter(new OutputStreamWriter(new FileOutputStream("processAreas.json")));
			streams = new BufferedWriter(new OutputStreamWriter(new FileOutputStream("streams.json")));
			baselines = new BufferedWriter(new OutputStreamWriter(new FileOutputStream("baselines.json")));
		} catch (FileNotFoundException e) {
			throw new FileSystemException(e);
		}

		try {
			List<IProcessAreaHandle> owners;
			try {
				owners = fetchProjectAreas(processAreas, projectAreaName, repo);
			} catch (RuntimeException e) {
				// We use this to indicate we can't find the owner
				System.out.println(e.getMessage());
				return;
			}
			
			fetchStreamsAndBaselines(streams, baselines, owners, repo);
		} catch (Exception e) {
			throw new FileSystemException(e);
		} finally {
			try {
				processAreas.close();
			} catch (IOException e) {
				throw new FileSystemException(e);
			}
			
			try {
				streams.close();
			} catch (IOException e) {
				throw new FileSystemException(e);
			}
			
			try {
				baselines.close();
			} catch (IOException e) {
				throw new FileSystemException(e);
			}

		}
		
	}

	private List<IProcessAreaHandle> fetchProjectAreas(BufferedWriter processAreas, String rootProcessAreaName, ITeamRepository repo) throws TeamRepositoryException, IOException {
		System.out.println("Fetch project areas");
		IProcessItemService process = (IProcessItemService) repo.getClientLibrary(IProcessItemService.class);

        List<IProjectArea> allProjectAreas = process.findAllProjectAreas(
                IProcessClientService.ALL_PROPERTIES,
                null);

        
        if (rootProcessAreaName != null) {
	    	IProjectArea foundRoot = null;
	        for (IProjectArea pa : allProjectAreas) {
	        	if (rootProcessAreaName.equals(pa.getName())) {
	        		foundRoot = pa;
	        		break;
	        	}
	        }
	        
	        if (foundRoot == null) {
	        	System.out.println("Known project areas:");
		        for (IProcessArea pa : allProjectAreas) {
		        	System.out.println("  " + pa.getName());
		        }
	        	
		        throw new RuntimeException("Unable to find project area named " + rootProcessAreaName); //$NON-NLS-1$

	        }
	        allProjectAreas = Collections.singletonList(foundRoot);

	        System.out.println("Found project area " + rootProcessAreaName);
        }

        
        System.out.println("Recording project areas");
        
        ArrayList<IProcessAreaHandle> owners = new ArrayList<IProcessAreaHandle>(); 
        
        for (IProjectArea pa : allProjectAreas) {
        	List<ITeamArea> teamAreas = repo.itemManager().fetchCompleteItems(pa.getTeamAreas(), IItemManager.DEFAULT, null);

        	{
	    		JSONObject paJSON = new JSONObject();
	    		
	    		paJSON.put("paName", pa.getName());
	    		paJSON.put("paUuid", pa.getItemId().getUuidValue());
	    		
	    		paJSON.serialize(processAreas);
	    		processAreas.write("\n");
        	}

        	owners.add(pa);

        	ITeamAreaHierarchy hier = pa.getTeamAreaHierarchy();
        	for (ITeamArea ta : teamAreas) {
        		owners.add(ta);
        		
        		JSONObject taJSON = new JSONObject();
        		
        		taJSON.put("taName", ta.getName());
        		taJSON.put("taUuid", ta.getItemId().getUuidValue());
        		if (hier.getParent(ta) != null) {
        			taJSON.put("taParent", hier.getParent(ta).getItemId().getUuidValue());
        		}
        		taJSON.put("paUuid", pa.getItemId().getUuidValue());
        		
        		taJSON.serialize(processAreas);
        		processAreas.write("\n");
        	}
        }
        
        // If the user ctrl-c's before the subcmd is done, the processAreas
        // may not get written. Force them to write now. 
        processAreas.close();
        
        return owners;
	}

	private void fetchStreamsAndBaselines(BufferedWriter streams, BufferedWriter baselines, List<IProcessAreaHandle> owners, ITeamRepository repo) throws TeamRepositoryException, IOException {
		// Fetch the set of streams
		IWorkspaceSearchCriteria search = IWorkspaceSearchCriteria.FACTORY.newInstance();
		
		search.setKind(IWorkspaceSearchCriteria.STREAMS);
		
		if (owners != null) {
			search.getFilterByOwnerOptional().addAll(owners);
		}
		
		Map<UUID, IComponent> comps = new HashMap<UUID, IComponent>();
		Set<UUID> baselineUuids = new HashSet<UUID>();

		System.out.println("Fetching stream list");
		
		IWorkspaceManager manager = SCMPlatform.getWorkspaceManager(repo);
		List<IWorkspaceHandle> workspaces = manager.findWorkspaces(search, Integer.MAX_VALUE, null);
		
		System.out.println(workspaces.size() + " streams to query");

		IScmService scmService = ((WorkspaceManager)manager).getServerConfigurationService();
		int cur = 0;
		int written = 0;
        BaselinePagerAndWriter blPagerAndWriter = new BaselinePagerAndWriter(repo, baselines);
		while (cur < workspaces.size()) {
			ArrayList<IWorkspaceHandle> currentPage = new ArrayList<IWorkspaceHandle>(MY_FETCH_PAGE_SIZE);
			int endOfPage = cur + MY_FETCH_PAGE_SIZE;
			for ( ; cur < endOfPage && cur < workspaces.size(); cur++) {
				currentPage.add(workspaces.get(cur));
			}
			
			List<IWorkspace> fullWorkspaces = repo.itemManager().fetchCompleteItems(currentPage, IItemManager.DEFAULT, null);
			
			for (IWorkspace ws : fullWorkspaces) {
				if (ws == null) {
					continue;
				}
				
				Collection<ComponentEntry> entries = ((WorkspaceManager)manager).getComponentEntries(ws, null);
				
				updateComponentList(repo, entries, comps);
				
				System.out.println("Dumping stream " + ws.getName() + " [" + written++ + "/" + workspaces.size() + "]");
				IWorkspaceConnection conn = manager.getWorkspaceConnection(ws, null);
				dump(streams, blPagerAndWriter, repo, scmService, ws, conn, entries, baselineUuids, comps);
				
				((WorkspaceManager)manager).clearCaches();// evict the stream
			}
		}		
		blPagerAndWriter.doFetch();
	}

	private void updateComponentList(ITeamRepository repo, Collection<ComponentEntry> entries,
			Map<UUID, IComponent> comps) throws TeamRepositoryException {
		List<IComponentHandle> toFetch = new ArrayList<IComponentHandle>(entries.size());
		for (ComponentEntry ce : entries) {
			IComponentHandle comp = ce.getComponent();
			if (comps.containsKey(comp.getItemId())) {
				continue;
			}
			
			toFetch.add(comp);
		}
		
		List<IComponent> fetched = repo.itemManager().fetchCompleteItems(toFetch, IItemManager.DEFAULT, null);
		for (IComponent f : fetched) {
			if (f == null) {
				continue;
			}
			
			comps.put(f.getItemId(), f);
		}
	}

	private void dump(Writer stream, BaselinePagerAndWriter blPagerAndWriter, ITeamRepository repo, IScmService scmService, IWorkspace ws, 
			IWorkspaceConnection conn, Collection<ComponentEntry> entries, Set<UUID> baselineUuids, Map<UUID, IComponent> comps) throws TeamRepositoryException, IOException {
		int ct = 0;
		for (ComponentEntry ce : entries) {
			IComponent comp = comps.get(ce.getComponent().getItemId());
			
			System.out.println("  Dumping " + ws.getName() + " / " + comp.getName() + " [" + ct++ + "/" + entries.size() + "]");
			// Record the baselines in the stream/comp pair
			JSONArray compBaselines = new JSONArray();
			int componentOffset = 0;
            long totalBasis = conn.getComponentInfo(ce.getComponent()).numBasisInHistory();
			while (compBaselines.size() < totalBasis) {
				HistoricBasis[] historicBasis = scmService.getComponentHistory(ws, ce.getComponent(), componentOffset, totalBasis - componentOffset, null, null);
				if (historicBasis.length == 0) {
					break;
				}
				
				componentOffset += historicBasis.length;
				
				for (HistoricBasis hb : historicBasis) {
					JSONObject hbJSON = new JSONObject();
					
					hbJSON.put("deliveredBy", hb.getDeliveredBy().getItemId().getUuidValue());
					if (hb.getBasis() == null) {
					    System.out.println("null basis");
					} else {
						hbJSON.put("basis", hb.getBasis().getItemId().getUuidValue());
						
						if (!baselineUuids.contains(hb.getBasis().getItemId())) {
							blPagerAndWriter.record(hb.getBasis());
							
							baselineUuids.add(hb.getBasis().getItemId());
						}
					}
					hbJSON.put("deliveredDate", hb.getDeliveryDate().toGMTString());
					hbJSON.put("isReplace", hb.isReplace());
					
					compBaselines.add(hbJSON);
				}
								
				System.out.println("    Recorded " + componentOffset + "/" + totalBasis + " baselines");
			}

			// Record the workspace
			JSONObject wsJSON = new JSONObject();
			wsJSON.put("type", "compHistory");
			
			wsJSON.put("workspaceName", ws.getName());
			wsJSON.put("workspaceUuid", ws.getItemId().getUuidValue());
			
			wsJSON.put("ownerUuid", ws.getOwner().getItemId().getUuidValue());
			
			wsJSON.put("componentUuid", ce.getComponent().getItemId().getUuidValue());
			wsJSON.put("componentName", comp.getName());
			
			wsJSON.put("baselines", compBaselines);
			
			wsJSON.serialize(stream);
			
			stream.write("\n");
		}
	}

	@Override
	public Options getOptions() throws ConflictingOptionException {
		Options options = new Options(false, true);

		SubcommandUtil.addRepoLocationToOptions(options);
		
		options.addOption(OPT_PROJECT_AREA, "Optional project area name. If specified, only streams in that project area will be queried.");

		return options;
	}

	public static class BaselinePagerAndWriter {
		private final ITeamRepository repo;
		private final Writer stream;
		
		private List<IBaselineHandle> toFetch;

		public BaselinePagerAndWriter(ITeamRepository repo, Writer stream) {
			super();
			this.repo = repo;
			this.stream = stream;
			this.toFetch = new ArrayList<IBaselineHandle>((int)(MY_FETCH_PAGE_SIZE * 1.2));
		}
		
		public void record(IBaselineHandle toWrite) throws IOException, TeamRepositoryException {
			toFetch.add(toWrite);
			
			if (toFetch.size() >= MY_FETCH_PAGE_SIZE) {
				doFetch();
			}
		}
		
		public void doFetch() throws IOException, TeamRepositoryException {
			List<IBaseline> fullBaselines = repo.itemManager().fetchCompleteItems(toFetch, IItemManager.DEFAULT, null);
			for (IBaseline bl : fullBaselines) {
				if (bl == null) {
					// Unexpected
					continue;
				}
				
				
				JSONObject blJSON = new JSONObject();
				
				blJSON.put("blNumber", bl.getId());
				blJSON.put("blName", bl.getName());
				blJSON.put("blUuid", bl.getItemId().getUuidValue());
				blJSON.put("componentUuid", bl.getComponent().getItemId().getUuidValue());
				blJSON.put("historyUuid", ((com.ibm.team.scm.common.internal.Baseline)bl).getHistory().getItemId().getUuidValue());
				blJSON.put("graphNode", ((com.ibm.team.scm.common.internal.Baseline)bl).getHistoryGraphNode());
				
				blJSON.serialize(stream);
				stream.write("\n");
			}

			toFetch = new ArrayList<IBaselineHandle>((int)(MY_FETCH_PAGE_SIZE * 1.2));
			
			System.out.println("    Wrote " + fullBaselines.size() + " baselines");
		}
	}
}
