/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.filesystem.cli.tools;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.osgi.util.NLS;

import com.ibm.team.filesystem.cli.core.util.CLIFileSystemClientException;
import com.ibm.team.filesystem.cli.core.util.StatusHelper;
import com.ibm.team.filesystem.cli.core.util.SubcommandUtil;
import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.ILocation;
import com.ibm.team.rtc.cli.infrastructure.internal.Messages;
import com.ibm.team.rtc.cli.infrastructure.internal.core.CLIClientException;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IClientConfiguration;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IOptionSource;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommand;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommandDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommandDefinition.LoadOptionsException;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.ICommandLine;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.INamedOptionDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.IPositionalOptionDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.NamedOptionDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.Options;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.PositionalOptionDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.ConflictingOptionException;
import com.ibm.team.rtc.cli.infrastructure.internal.util.OptionGroupUtil;

/**
 * Provide online help: a listing of available subcommands and basic options.
 */
public class GenerateHelpCmd implements ISubcommand, IOptionSource {
    
    IClientConfiguration config = null;
    String lang = null;

    public static final String NO_HELP_STRING = Messages.HelpCmd_0;
    public static final String HELP_FOLDERNAME = "scmclihelp";
    public static final String HELP_FILENAME = "scm_cli.html";
	
    public static PositionalOptionDefinition OPT_DIRECTORY = new PositionalOptionDefinition("directory", 1, 1); //$NON-NLS-1$
    
	private static final class NamedOptionComparator implements
			Comparator<INamedOptionDefinition> {
		public int compare(INamedOptionDefinition o1, INamedOptionDefinition o2) {
			String s1 = o1.getShortOpt();
			String s2 = o2.getShortOpt();
			if (s1 == null && s2 == null) {
				s1 = o1.getLongOpt();
				s2 = o2.getLongOpt();
			}
			if (s1 != null && s2 != null) {
				return s1.compareToIgnoreCase(s2);
			} else if (s1 != null) {
				return 1;
			} else if (s2 != null) {
				return -1;
			}
			return o1.getArgCount() - o2.getArgCount();
		}
	}


	public static class SubcommandNameCompator implements Comparator<ISubcommandDefinition> {
		public int compare(ISubcommandDefinition o1, ISubcommandDefinition o2) {
			return o1.getFullSubcommandName().compareTo(o2.getFullSubcommandName());
		}}

    public static final NamedOptionDefinition OPT_LANG = new NamedOptionDefinition("l", "lang", 1); //$NON-NLS-1$ //$NON-NLS-2$

    public Options getOptions() throws ConflictingOptionException {
		Options options = new Options(false);
		
		options.setLongHelp("Generate html help for all the subcommands.");		
		options.addOption(OPT_DIRECTORY, "The directory where the help files are written.");
		options.addOption(OPT_LANG, "The language to be set in the HTML tag.");
		return options;
	}
    
    public void initialize(IClientConfiguration config) {
        
    }
    
	public int run(IClientConfiguration config) throws CLIClientException {
	    this.config = config;
		ICommandLine subargs = config.getSubcommandCommandLine();
		
		// get the language
		if (subargs.hasOption(OPT_LANG)) {
		    this.lang = subargs.getOption(OPT_LANG);
		}
		
		// get the folder location to write the help files
		String helpFolderPath = subargs.getOption(OPT_DIRECTORY);
		
		// make it absolute path
		ILocation helpFolderLoc;
        try {
            helpFolderLoc = SubcommandUtil.makeAbsolutePath(config.getContext().getCurrentWorkingDirectory(), helpFolderPath);
        } catch (FileSystemException e) {
            throw new CLIClientException(e);
        }
		File helpFolder = new File (helpFolderLoc.toOSString(), HELP_FOLDERNAME);
		
		// create dirs
		helpFolder.mkdirs();
		
		// generate help
		try {
            generateHelp(helpFolder);
        } catch (FileSystemException e) {
            throw new CLIClientException(e);
        }
		
		return 0;
	}
	
	private void generateHelp(File helpFolder) throws FileSystemException {
	    File helpFilePath = new File(helpFolder, HELP_FILENAME);
	    
	    BufferedWriter writer = null;
        try {
            // create a writer for the main help file
            writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(helpFilePath), "UTF-8"));
            
            // write header
            writeHTMLTopPart(writer);
            
            writer.write("<body>");
            writer.newLine();
        
            // go over the subcomands
            List<ISubcommandDefinition> subcommands = new ArrayList<ISubcommandDefinition>(config.getSubcommandDefinitions());
    		Collections.sort(subcommands, new SubcommandNameCompator());

            List<FileSystemException> exceptions = new LinkedList<FileSystemException>();
            StringBuffer erroringNames = new StringBuffer();

            for (ISubcommandDefinition defn : subcommands) {
                if (!defn.showInHelpListing()) {
                    continue;
                }
                
                if(defn.isParent()) {
                	if(defn.isDeprecated() && !hasVisibleChildren(defn)) {
                		continue;
                	}
                }
                else {
                	if (defn.isDeprecated()) {
                		continue;
                	}
                }

                try {
                    generateSubcommandHelp(config, defn, helpFolder, writer);
                } catch (FileSystemException e) {
                    config.getContext().stderr().println(NLS.bind(Messages.HelpCmd_9, defn.getFullSubcommandName()));
                    exceptions.add(e);
                    
                    if (erroringNames.length() > 0) {
                        erroringNames.append(Messages.HelpCmd_10);
                    }
                    erroringNames.append(defn.getFullSubcommandName());
                }           
            }
            
            if (exceptions.size() > 0) {
                throw StatusHelper.internalError(NLS.bind(Messages.HelpCmd_11, erroringNames.toString()), exceptions);
            }
        } catch (IOException e) {
            config.getContext().stderr().println("IOException when creating a writer for " + helpFilePath.getAbsolutePath());
        } finally {
            try {
                if (writer != null) {
                    writer.write("</body></html>");
                    writer.flush();
                    writer.close();
                }
            } catch (IOException e) {
            }
        }	    
	}
	
	private void writeHTMLTopPart(BufferedWriter writer) throws IOException {
        if (lang != null) {
            writer.write("<html lang=\"" + lang + "\">");
        } else {
            writer.write("<html>");
        }
        writer.newLine();
        writer.write("<head>");
        writer.newLine();
        writer.write("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\" />");
        writer.newLine();
        writer.write("<link rel=\"stylesheet\" type=\"text/css\" href=\"PLUGINS_ROOT/com.ibm.help.common.resources.doc/css/help.css\" />");
        writer.newLine();
        writer.write("</head>");
        writer.newLine();
	}
	
	private Options getOptions(ISubcommandDefinition defn) throws CLIFileSystemClientException {
	    String commandName = defn.getFullSubcommandName();
        Options options;
        try {
            options = defn.getOptions();
        } catch (LoadOptionsException e) {
            throw StatusHelper.failure(NLS.bind(Messages.HelpCmd_16, commandName, e.getCause().getMessage()), e);           
        } catch (ConflictingOptionException e) {
            throw StatusHelper.failure(NLS.bind(Messages.HelpCmd_17, commandName, e.getMessage()), e);
        }
        
        return options;
	}
	

	/** Display an individual subcommand. 
	 * @param config 
	 * @param writer 
	 * @param helpFolder 
	 * @throws IOException */
	private void generateSubcommandHelp(IClientConfiguration config, ISubcommandDefinition defn, File helpFolder,
	        BufferedWriter writer) throws FileSystemException, IOException {
        List<ISubcommandDefinition> ancestors = defn.getAncestors();

        // create the command name hierarchy
        StringBuffer commandNameHierarchy = new StringBuffer();
        for (ISubcommandDefinition ancestor : ancestors) {
            commandNameHierarchy.append(ancestor.getFullSubcommandName());
            commandNameHierarchy.append(" "); //$NON-NLS-1$
        }
        commandNameHierarchy.append(defn.getFullSubcommandName());
        String commandName = commandNameHierarchy.toString();
        
        // write the command name & description into the main help file
        if (ancestors == null || ancestors.size() == 0) {
        	boolean isLink = true;
        	if (defn.isParent()) {
        		// By default, we make parent commands to be non link
        		isLink = false;

        		// However if the command is runnable and not deprecated
        		// it has to be a link
        		if (defn.isRunnable() && !defn.isDeprecated()) {
        			isLink = true;
        		}
        	}
        		
    		writeSubcommandName(writer, commandName, null, isLink);
    		
    		String help = defn.getShortHelp();
    		if (help != null && help.length() > 0) {
        		writeDescription(writer, help);
    		}
    		writeNewLine(writer);
        }
		
		// write the children of the parent
		if (defn.isParent())
		{
		    List<ISubcommandDefinition> children = new ArrayList<ISubcommandDefinition>(defn.getChildren());
            Collections.sort(children, new SubcommandNameCompator());
            
            writer.write("<table border=\"0\">");
            writer.newLine();
            for (ISubcommandDefinition child : children) {
                writer.write("<tr>");
                writer.newLine();
                writer.write("<td style=\"width: 50px\">");
                writer.write("</td>");
                writer.newLine();
                writer.write("<td>");
                writer.write("<p>");
    		    writeSubcommandName(writer, child.getFullSubcommandName(), commandName, (child.isParent()) ? false : true);
                String help = child.getShortHelp();
                if (help != null && help.length() > 0) {
//                    writeDescription(writer, help);
                    writer.write(help);
                }
                writer.write("</p>");
    		    writer.write("</td>");
    		    writer.newLine();
    		    writer.write("</tr>");
    		    writer.newLine();
            }
            writer.write("</table>");
            writer.newLine();

            // Generate help content if the subcommand is runnable, eventhough it is a parent command
            // It is possible that we are here to write a top level command which is now deprecated
            // but it has visible children. So ensure that the command is runnable and it is not deprecated
            // to actually continue creating a separate help file for it
            if (!defn.isRunnable()) {
            	return;
            }
            else { // the definition is runnable, check whether it is deprecated 
            	if (defn.isDeprecated()) { // never create a html file for it because it is deprecated 
            		return;
            	}
			}
		}
		
		// write the sub-command help file
        String subcommandHelpFileName = commandName.replace(' ', '_') + ".html";
		File subcommandHelpFile = new File(helpFolder, subcommandHelpFileName);
        generateSubcommandHelp(defn, commandName, subcommandHelpFile);
	}


    private void generateSubcommandHelp(ISubcommandDefinition defn, String commandName, File subcommandHelpFile)
            throws IOException, FileSystemException {
        Options options = getOptions(defn);
        String commandHelp = options.getLongHelp();
        if (commandHelp == null || commandHelp.length() == 0) {
            commandHelp = defn.getShortHelp();
        }

        BufferedWriter subCmdWriter = null;
        try {
            // create a writer for the subcommand help file
            subCmdWriter = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(subcommandHelpFile), "UTF-8"));
            
            // write header
            writeHTMLTopPart(subCmdWriter);            
            subCmdWriter.write("<body>");
            subCmdWriter.newLine();
            
            // write name
            writeSubcommandName(subCmdWriter, commandName, null, false);
            
            // write description
            if (commandHelp != null && commandHelp.length() > 0) {
                writeDescription(subCmdWriter, commandHelp);
            }
            
            // Collect the list of parent command  names
            // We want to print it before the list of aliases.
            // For eg, if remove merge-queue is the command
            // and mq is the alias for merge-queue, then
            // the alias wil be remove mq
            List<ISubcommandDefinition> ancestors = defn.getAncestors();
            StringBuffer parentCommands = new StringBuffer();
            if (ancestors != null) {
    			boolean trailing = false;
    			for (ISubcommandDefinition ancestor : ancestors) {
    				if(trailing)
    				{
    					parentCommands.append(Messages.HelpCmd_13);
    				}
    				trailing = true;
    				parentCommands.append(ancestor.getFullSubcommandName());
    			}
            }

            // aliases
            String[] aliases = defn.getAliases();
            if (aliases.length > 0) {
                writeText(subCmdWriter, Messages.HelpCmd_21, true);
                
                boolean trailing = false;
                StringBuffer aliasesStr = new StringBuffer();
                for (String alias : aliases) {
                    if (trailing) {
                        aliasesStr.append(Messages.HelpCmd_10);
                    }
                    trailing = true;
                    aliasesStr.append(parentCommands.toString().isEmpty()? alias : (parentCommands.toString() + Messages.HelpCmd_13 + alias));
                }
                
                writeText(subCmdWriter, aliasesStr.toString(), false);
            }
            
            if (options.getSupportsNewHelp()) {
            	showNewUsage(subCmdWriter, options, commandName);
            }
            else {
            	// usage
            	showUsage(subCmdWriter, options, commandName);
            }

            // example
            writeNewLine(subCmdWriter);
            subCmdWriter.write("<a href=\"");
            String exampleLink = subcommandHelpFile.getName().replaceAll(".html", "_ex.html");
            subCmdWriter.write(exampleLink);
            subCmdWriter.write("\">");
            writeText(subCmdWriter, Messages.HelpCmd_Example,true);
            subCmdWriter.write("</a>");            
        } catch (IOException e) {
            
        } finally {
            if (subCmdWriter != null) {
                subCmdWriter.write("</body></html>");
                subCmdWriter.flush();
                subCmdWriter.close();
            }
        }
    }
    
	/** 
	 * Print short usage help with mutually exclusive, optional options.
	 * Supports printing group of options on separate lines , along with common options for each group
	 */
	private void showNewUsage(BufferedWriter subCmdWriter, Options options, String fullSubcommandName) throws IOException {
		/**
		 * Logic 
		 * First create a structure with top level options
		 * This should include named, mutually exclusive groups and positional options
		 * Get the usage help for this top level options
		 * First look for any continuous groups
		 *   If there are no continuous groups , then print the top level options as is
		 *   Else for each continuous group, get the usage from each of its inner group
		 *   or positional. Finally print this along with the top level options.  
		 *   Then print a new line.
		 */

		
		// Includes named,simple,mutually exclusive, continuous and positional options
		List<StringBuffer> resultOptStrings = OptionGroupUtil.getShortUsageHelpForHtml(options);
		StringBuffer usageString = new StringBuffer();
		if (resultOptStrings.size() > 1 ) {
		    // Print out "Usage :" on a separate line  if short help is spread across 
		    // separate lines.   
            subCmdWriter.write("<b>");
			subCmdWriter.write(NLS.bind(Messages.HelpCmd_23, ""));
			subCmdWriter.write("</b>");
			subCmdWriter.newLine();
		}
		else { // Otherwise Usage : will appear on the same line as short help
			usageString = new StringBuffer(NLS.bind(Messages.HelpCmd_23, config.getContext().getAppName()));
			usageString.append(" ");
		}
		usageString.append(fullSubcommandName);
		
		for(StringBuffer resultOptString : resultOptStrings) {
			subCmdWriter.write("<p>");
			// Create a table
			subCmdWriter.write("<table border=\"0\">");
			subCmdWriter.newLine();
            subCmdWriter.write("<tr>");
            subCmdWriter.newLine();
            subCmdWriter.write("<td nowrap valign=\"top\">");
            subCmdWriter.write("<b>");
            subCmdWriter.write(usageString.toString());
            subCmdWriter.write("</b>");
            subCmdWriter.write("</td>");
            subCmdWriter.newLine();
            subCmdWriter.write("<td valign=\"top\">");
            subCmdWriter.write("<b>");
            subCmdWriter.write(resultOptString.toString());
            subCmdWriter.write("</b>");
            subCmdWriter.write("</td>");
            subCmdWriter.newLine();
            subCmdWriter.write("</tr>");
            subCmdWriter.newLine();
			subCmdWriter.write("</table>");
			subCmdWriter.write("</p>");
		}
		
		showUsageCommon(subCmdWriter, options, true);
	}

    /** Show the "usage" block for non-parent subcommands (old style)
     * @param subCmdWriter 
     * @param fullSubcommandName 
     * @throws IOException */
	private void showUsage(BufferedWriter subCmdWriter, Options options, String fullSubcommandName) throws IOException {
	    // show positional options
        StringBuffer usageLine = new StringBuffer(NLS.bind(Messages.HelpCmd_23, config.getContext().getAppName()));
        usageLine.append(" ");
        usageLine.append(fullSubcommandName);
		
		if (options.getNamedOptions().size() > 0) {
		    usageLine.append(" ").append(Messages.HelpCmd_29);
		}
		
		for (IPositionalOptionDefinition opt : options.getPositionalOptions()) {
			boolean isRequired = opt.getMinCount() > 0;

			if (isRequired) {
			    usageLine.append("&lt;"); //$NON-NLS-1$
			} else {
			    usageLine.append("["); //$NON-NLS-1$
			}

			usageLine.append(opt.getName());

			if (opt.getMaxCount() == Options.ARG_INFINITE) {
			    usageLine.append("..."); //$NON-NLS-1$
			}

			if (isRequired) {
			    usageLine.append("&gt;"); //$NON-NLS-1$
			} else {
			    usageLine.append("]"); //$NON-NLS-1$
			}
		}
		
		writeText(subCmdWriter, usageLine.toString(), true);
		
		showUsageCommon(subCmdWriter, options, false);
	}
	
	private void showUsageCommon(BufferedWriter subCmdWriter, Options options, boolean isNewStyle) throws IOException {
		// show positional options description
		subCmdWriter.write("<table border=\"0\">");
		subCmdWriter.newLine();
		for (IPositionalOptionDefinition opt : options.getPositionalOptions()) {
            subCmdWriter.write("<tr>");
            subCmdWriter.newLine();
            subCmdWriter.write("<td valign=\"top\">");
            subCmdWriter.write(opt.getName());
            subCmdWriter.write("</td>");
            subCmdWriter.newLine();
            subCmdWriter.write("<td valign=\"top\">");
			subCmdWriter.write(options.getHelpFor(opt.getId()));
			subCmdWriter.write("</td>");
			subCmdWriter.newLine();
			subCmdWriter.write("</tr>");
			subCmdWriter.newLine();
		}
		subCmdWriter.write("</table>");
		subCmdWriter.newLine();
	
		// Show named options
		if (options.getNamedOptions().size() > 0) {
			writeText(subCmdWriter, Messages.HelpCmd_31, true);
		}

		List<INamedOptionDefinition> namedOptions = new ArrayList<INamedOptionDefinition>(options.getNamedOptions());
		if (!isNewStyle) { // Sort if we are in old style, otherwise don't
			Collections.sort(namedOptions, new NamedOptionComparator());
		}
		
        subCmdWriter.write("<table border=\"0\">");
        subCmdWriter.newLine();
		for (INamedOptionDefinition opt : namedOptions) {
			StringBuffer buf = new StringBuffer();
	
	        // Encapsulates the printing style for named options.
			getFormattedNamedOption(opt, buf, isNewStyle);
			
            subCmdWriter.write("<tr>");
            subCmdWriter.newLine();
            subCmdWriter.write("<td valign=\"top\">");
            subCmdWriter.write(buf.toString());
            subCmdWriter.write("</td>");
            subCmdWriter.newLine();
            subCmdWriter.write("<td valign=\"top\">");
            subCmdWriter.write(options.getHelpFor(opt.getId()));
            subCmdWriter.write("</td>");
            subCmdWriter.newLine();
            subCmdWriter.write("</tr>");
            subCmdWriter.newLine();
		}		
        subCmdWriter.write("</table>");
        writeNewLine(subCmdWriter);
	}
	
	private void getFormattedNamedOption(INamedOptionDefinition opt, StringBuffer buf, boolean isNewStyle) {
		if (!isNewStyle) {
			// show the shortopt
			boolean hasShortOpt = false;
			if (opt.getShortOpt() != null) {
				buf.append("-" + opt.getShortOpt()); //$NON-NLS-1$
				hasShortOpt = true;
			}
	
			// Show the longopt
			if (opt.getLongOpt() != null) {
				if (hasShortOpt) {
					buf.append(" "); //$NON-NLS-1$
				}
	
				if (hasShortOpt) {
					buf.append("["); //$NON-NLS-1$
				}
	
				buf.append("--" + opt.getLongOpt()); //$NON-NLS-1$
	
				if (hasShortOpt) {
					buf.append("]"); //$NON-NLS-1$
				}
	
			}
		}
		else {
			// Show the shortopt
			boolean hasShortOpt = false;
			if (opt.getShortOpt() != null) {
				buf.append("-" + opt.getShortOpt()); //$NON-NLS-1$
				hasShortOpt = true;
			}

			// Show the longopt
			if (opt.getLongOpt() != null) {
				
				String commaSequence  = ", ";
				if (hasShortOpt) {
					buf.append(commaSequence); //$NON-NLS-1$
				}
				else { // Append ' ' multiplied by the number of for a shortOpt + commaSequence 
					int whiteSpaceLength = "-".length() + 1 + commaSequence.length();
					for (int i = 0 ; i < whiteSpaceLength; i++) {
						buf.append("&nbsp;");
					}
				}

				buf.append("--" + opt.getLongOpt()); //$NON-NLS-1$
			}

			// Show the args
			if (opt.getArgCount() == Options.ARG_INFINITE) {
				buf.append(OptionGroupUtil.HelpCmdHtmlArgsInfinite);
			}
			else {
				int minArgCount = opt.getMinArgCount();
				int argCount = opt.getArgCount();
				for (int i = 0; i < argCount; i++) {
					if (i < minArgCount) {
						buf.append(OptionGroupUtil.HelpCmdHtmlArgsFinite);
					}
					else {
						buf.append(Messages.HelpCmd_38);
					}
				}
			}
		}
	}	
	
    private boolean hasVisibleChildren(ISubcommandDefinition defn) {
        List<ISubcommandDefinition> children = new ArrayList<ISubcommandDefinition>(defn.getChildren());
        for (ISubcommandDefinition child : children) {
            if (!child.showInHelpListing() || child.isDeprecated())
                continue;
            
            return true;
        }
        return false;
    }

	private void writeSubcommandName(BufferedWriter writer, String name, String ancestorName, boolean asLink) throws IOException {
	    writer.write("<h3>");
	    
	    if (asLink) {
            String subcommandLink = name;
            if (ancestorName != null) {
                subcommandLink = ancestorName + " " + name;
            }
            subcommandLink = subcommandLink.replace(' ', '_') + ".html";
	        writer.write("<a href=\"");
	        writer.write(subcommandLink);
	        writer.write("\">");
	    }
	    
	    writer.write(name);
	    
	    if (asLink) {
	        writer.write("</a>");
	    }
	    
	    writer.write("</h3>");
	    writer.newLine();
	}

    private void writeDescription(BufferedWriter writer, String help) throws IOException {
        writer.write("<p>");
        writer.write(help);
        writer.write("</p>");
        writer.newLine();
    }
    
    private void writeText(BufferedWriter writer, String text, boolean isBold) throws IOException {
        writer.write("<p>");
        if (isBold) {
            writer.write("<b>");
        }
        writer.write(text);
        if (isBold) {
            writer.write("</b>");
        }
        writer.write("</p>");
        writer.newLine();
    }
    
    private void writeNewLine(BufferedWriter writer) throws IOException {
        writer.write("<br/>");
        writer.newLine();
    }
}
