/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2013. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.filesystem.cli.tools;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Map;
import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

import com.ibm.team.rtc.cli.infrastructure.internal.core.IOptionSource;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.ICommandLine;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.Options;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.PositionalOptionDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.ConflictingOptionException;
import com.ibm.team.filesystem.cli.core.AbstractSubcommand;
import com.ibm.team.filesystem.cli.tools.dump.MetadataDumpParser;
import com.ibm.team.filesystem.cli.tools.dump.MetadataDumpVisitor;
import com.ibm.team.filesystem.cli.tools.dump.blocks.ComponentMetadataBlock;
import com.ibm.team.filesystem.cli.tools.dump.blocks.ForwardMetadataBlock;
import com.ibm.team.filesystem.cli.tools.dump.blocks.ForwardSharingMetadataBlock;
import com.ibm.team.filesystem.cli.tools.dump.blocks.IScopedByComponentMetadataBlock;
import com.ibm.team.filesystem.cli.tools.dump.blocks.InverseMetadataBlock;
import com.ibm.team.filesystem.cli.tools.dump.blocks.InverseMetadataBlock.RemoteChild;
import com.ibm.team.filesystem.cli.tools.dump.blocks.InverseSharingMetadataBlock;
import com.ibm.team.filesystem.cli.tools.dump.blocks.MetametaMetadataBlock;
import com.ibm.team.filesystem.client.FileSystemException;

/**
 * Postprocesses a metadata dump. At this point that just means sorting the 
 * fields, but in future it could include computing paths or anonymizing the 
 * data. 
 */
public class MetadataDumpPostProcessCmd extends AbstractSubcommand implements IOptionSource {

	public static PositionalOptionDefinition OPT_METADATA = new PositionalOptionDefinition("metadata", 1, 1); //$NON-NLS-1$
	
	public Options getOptions() throws ConflictingOptionException {
		Options options = new Options(false);
		
		options.setLongHelp("Sort metadata sections generated with 'validate'. ");		
		options.addOption(OPT_METADATA, "Path to the metadata to sort. MAY ONLY INCLUDE INVERSE METADATA SECTION. ");
		return options;
	}

	private final static Comparator<InverseMetadataBlock.RemoteChild> COMPARATOR_REMOTE_CHILD = new Comparator<RemoteChild>() {
		@Override
		public int compare(RemoteChild r1, RemoteChild r2) {
			return r1.getId().compareTo(r2.getId());
		}
	};
	
	
	public static class SortingInverseMetadataDumpParserConsumer extends MetadataDumpVisitor {
		SortedSet<ForwardSharingMetadataBlock> forwardSharingMetadataBlocks = new TreeSet<ForwardSharingMetadataBlock>(ForwardSharingMetadataBlock.COMPARATOR);
        SortedSet<InverseSharingMetadataBlock> inverseSharingMetadataBlocks = new TreeSet<InverseSharingMetadataBlock>(InverseSharingMetadataBlock.COMPARATOR);
	    SortedSet<ComponentMetadataBlock> componentMetadataBlocks = new TreeSet<ComponentMetadataBlock>(IScopedByComponentMetadataBlock.COMPARATOR);
	    SortedSet<ForwardMetadataBlock> fowardMetadataBlocks = new TreeSet<ForwardMetadataBlock>(ForwardMetadataBlock.COMPARATOR);
		SortedSet<InverseMetadataBlock> inverseMetadataBlocks = new TreeSet<InverseMetadataBlock>(InverseMetadataBlock.COMPARATOR);
		SortedSet<MetametaMetadataBlock> metametaMetadataBlocks = new TreeSet<MetametaMetadataBlock>(MetametaMetadataBlock.COMPARATOR);
		
		@Override
		public void consumeInverseMetadataBlock(InverseMetadataBlock block) {
			Collection<RemoteChild> children = block.getChildren();
			if (children != null) {
				TreeSet<RemoteChild> kids = new TreeSet<RemoteChild>(COMPARATOR_REMOTE_CHILD);
				kids.addAll(children);
				block.rewriteChildren(new ArrayList<RemoteChild>(kids));
			}
			
			{
    			Map<String, String> original = block.getOriginalProps();
    			if (original != null && original.size() > 0) {
    			    block.rewriteOriginalProps(sort(original));
    			}
			}
			
			{
                Map<String, String> changed = block.getChangedProps();
                if (changed != null && changed.size() > 0) {
                    block.rewriteChangedProps(sort(changed));
                }
			}
            
			inverseMetadataBlocks.add(block);
		}
		
		@Override
		public void consumeComponentMetadataBlock(ComponentMetadataBlock cmb) {
		    componentMetadataBlocks.add(cmb);
		}
		
		@Override
		public void consumeForwardMetadataBlock(ForwardMetadataBlock forwardMetadataBlock) {
		    fowardMetadataBlocks.add(forwardMetadataBlock);
		}
		
		@Override
		public void consumeForwardSharingMetadataBlock(ForwardSharingMetadataBlock forwardMetadataBlock) {
		    forwardSharingMetadataBlocks.add(forwardMetadataBlock);
		}
		
		@Override
		public void consumeInverseSharingMetadataBlock(InverseSharingMetadataBlock inverseSharingMetadataBlock) {
		    inverseSharingMetadataBlocks.add(inverseSharingMetadataBlock);
		}
		
		@Override
		public void consumeMetametaMetadataBlock(MetametaMetadataBlock metametaMetadataBlock) {
		    metametaMetadataBlocks.add(metametaMetadataBlock);
		}
		
		private Map<String, String> sort(Map<String, String> original) {
		    SortedMap<String, String> sorted = new TreeMap<String, String>();
		    sorted.putAll(original);
		    
		    return sorted;
		}
	}
	
	@Override
	public void run() throws FileSystemException {
		ICommandLine cli = config.getSubcommandCommandLine();
		
		String metadataFile = cli.getOption(OPT_METADATA);
		
		SortingInverseMetadataDumpParserConsumer c = new SortingInverseMetadataDumpParserConsumer();
		
		try {
			FileInputStream inputStream = new FileInputStream(metadataFile);
			try {
				MetadataDumpParser.parse(inputStream, c);
			} finally {
				inputStream.close();
			}
		} catch (IOException e) {
			throw new FileSystemException(e);
		}
		
		PrintWriter out = new PrintWriter(new OutputStreamWriter(config.getContext().stdout()));
		
		try {
            for (ComponentMetadataBlock cmb : c.componentMetadataBlocks) {
                cmb.dump(out);
            }           
            
		    for (ForwardSharingMetadataBlock fsmb : c.forwardSharingMetadataBlocks) {
				fsmb.dump(out);
			}

		    for (InverseSharingMetadataBlock ismb : c.inverseSharingMetadataBlocks) {
		        ismb.dump(out);
		    }
		    
		    for (ForwardMetadataBlock fmb : c.fowardMetadataBlocks) {
		        fmb.dump(out);
		    }
		    
		    for (InverseMetadataBlock imb : c.inverseMetadataBlocks) {
		    	imb.dump(out);
		    }

		    for (MetametaMetadataBlock mmb : c.metametaMetadataBlocks) {
				mmb.dump(out);
			}
		} finally {
			out.close();
		}
	}

}
