/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2013. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.filesystem.cli.tools.dump;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.Map;

import com.ibm.team.filesystem.cli.tools.dump.blockparsers.BlockParser;
import com.ibm.team.filesystem.cli.tools.dump.blockparsers.ComponentMetadataDumpParser;
import com.ibm.team.filesystem.cli.tools.dump.blockparsers.ForwardMetadataDumpParser;
import com.ibm.team.filesystem.cli.tools.dump.blockparsers.ForwardSharingMetadataDumpParser;
import com.ibm.team.filesystem.cli.tools.dump.blockparsers.InverseMetadataDumpParser;
import com.ibm.team.filesystem.cli.tools.dump.blockparsers.InverseSharingMetadataDumpParser;
import com.ibm.team.filesystem.cli.tools.dump.blockparsers.MetametaDumpParser;
import com.ibm.team.filesystem.client.internal.copyfileareas.metadata.validator.MetaDatasCollector;
import com.ibm.team.filesystem.client.operations.IMetadataValidateOperation;

/**
 * Parser for metadata dumps generated by {@link IMetadataValidateOperation}.
 * The parser takes a visitor through the file.  
 */
public class MetadataDumpParser {

	/* Base class for metadata blocks that allows us to share some code. 
	 */
	public abstract static class MetadataBlock {
		protected String self;
		
		public MetadataBlock(String self) {
			this.self = self;
		}

		protected void dumpContent(PrintWriter w) {
			w.print(self);
		}
		
		protected abstract void dumpHeading(PrintWriter w);
		
		public void dump(PrintWriter w) {
			dumpHeading(w);
			
			dumpContent(w);
			
			w.println(MetaDatasCollector.TERMINATOR);
			w.println();
		}
	}
	
	public static void parse(InputStream in, MetadataDumpVisitor c) throws IOException {
		LineNumberReader lnr = new LineNumberReader(new InputStreamReader(in));
		
		Map<String, BlockParser> parsers = new HashMap<String, BlockParser>();
		
		addParser(parsers, new ComponentMetadataDumpParser());
		addParser(parsers, new ForwardMetadataDumpParser());
		addParser(parsers, new ForwardSharingMetadataDumpParser());
		addParser(parsers, new InverseSharingMetadataDumpParser());
		addParser(parsers, new InverseMetadataDumpParser());
		addParser(parsers, new MetametaDumpParser());
		
		while (true) {
		    String line = lnr.readLine();
		    if (line == null) {
		        break;
		    }
		    
		    SectioningLineNumberReader reader = new SectioningLineNumberReader(lnr);

		    BlockParser parser = parsers.get(line);
		    if (parser == null) {
		    	throw new RuntimeException("No parser for \"" + line + "\" on line " + lnr.getLineNumber());
		    }
		    parser.parse(reader, c);
		    
		    if (!reader.foundSectionTerminator()) {
		        throw new IllegalStateException("Failed to terminate on line " + lnr.getLineNumber());
		    }
		}
	}


	private static void addParser(Map<String, BlockParser> parsers, BlockParser blockParser) {
		parsers.put(blockParser.getSectionHeading(), blockParser);
	}
}