/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2013. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.filesystem.cli.tools.dump.blocks;

import java.io.PrintWriter;
import java.util.Comparator;
import java.util.List;
import java.util.Map;

import com.ibm.team.filesystem.cli.tools.dump.MetadataDumpParser.MetadataBlock;
import com.ibm.team.filesystem.client.internal.copyfileareas.metadata.validator.MetaDatasCollector;
import com.ibm.team.repository.common.UUID;

/**
 * Dump analogue for 
 * {@link com.ibm.team.filesystem.client.internal.InverseFileItemInfo}. 
 */
@SuppressWarnings("restriction")
public class InverseMetadataBlock extends MetadataBlock implements IScopedByComponentMetadataBlock {


    public static Comparator<InverseMetadataBlock> COMPARATOR = new Comparator<InverseMetadataBlock>() {
        @Override
        public int compare(InverseMetadataBlock b1, InverseMetadataBlock b2) {
            int r = IScopedByComponentMetadataBlock.COMPARATOR.compare(b1, b2);
            
            if (r == 0) {
                r = b1.getItemId().compareTo(b2.getItemId());
            }
            
            return r;
        }
    };

    
    public static class RemoteChild {
        final String self;
        private final UUID id;
        
        private final String type;
        
        public RemoteChild(String line, UUID id, String type) {
            this.self = line;
            this.id = id;
            this.type = type;
        }
        
        public void dump(PrintWriter w) {
            w.println(self);
        }
        
        public UUID getId() {
            return id;
        }
    
        public String getType() {
            return type;
        }
    }
    
	private List<RemoteChild> logicalChildren;
	
	private final UUID itemId;
	private final UUID componentId;
	private final UUID connectionId;

	private Map<String, String> originalProps;
	private Map<String, String> changedProps;
	
	public InverseMetadataBlock(String self, 
    		Map<String, String> originalProps,
    		Map<String, String> changedProps,
    		List<RemoteChild> children,
    		UUID itemId,
    		UUID componentId,
    		UUID connectionId
			) {
		super(self);
		this.originalProps = originalProps;
		this.changedProps = changedProps;
		this.itemId = itemId;
		this.componentId = componentId;
		this.connectionId = connectionId;
		
		this.originalProps = originalProps;
		this.changedProps = changedProps;
		this.logicalChildren = children;
	}

	public List<RemoteChild> getChildren() {
		return logicalChildren;
	}

	public void rewriteChildren(List<RemoteChild> newValue) {
		this.logicalChildren = newValue;
	}
	
	public UUID getItemId() {
		return itemId;
	}

	@Override
	protected void dumpContent(PrintWriter w) {
		super.dumpContent(w);

		writeProps(w, MetaDatasCollector.SUBSECTION_ORIGINAL_PROPERTIES, originalProps);
		writeProps(w, MetaDatasCollector.SUBSECTION_CHANGED_PROPERTIES, changedProps);
		
		if (logicalChildren != null && logicalChildren.size() > 0) {
			w.println(MetaDatasCollector.SUBSECTION_REMOTE_CHILDREN);
			for (RemoteChild rc : logicalChildren) {
				w.println(rc.self);
			}
		}
	}

	private void writeProps(PrintWriter w, String heading, Map<String, String> props) {
		if (props != null && props.size() > 0) {
			w.println(heading);
			for (Map.Entry<String, String> entry : props.entrySet()) {
				w.println("   " + entry.getKey() + " : " + entry.getValue());  //$NON-NLS-1$//$NON-NLS-2$
			}
		}
	}

	public Map<String, String> getOriginalProps() {
		return originalProps;
	}

	public void rewriteOriginalProps(Map<String, String> rewritten) {
		this.originalProps = rewritten;
	}

	
	public Map<String, String> getChangedProps() {
		return changedProps;
	}

	public void rewriteChangedProps(Map<String, String> rewritten) {
		this.changedProps = rewritten;
	}

	@Override
	protected void dumpHeading(PrintWriter w) {
		w.println(MetaDatasCollector.SECTION_INVERSE_INFO);
	}
	
	@Override
	public UUID getConnectionId() {
	    return connectionId;
	}
	
	@Override
	public UUID getComponentId() {
		return componentId;
	}
}
