/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2005, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:  Use, 
 * duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp.
 *******************************************************************************/
package com.ibm.team.filesystem.cli.tools.verify;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import org.eclipse.core.filesystem.EFS;
import org.eclipse.core.filesystem.IFileInfo;
import org.eclipse.core.filesystem.IFileStore;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.SubMonitor;

import com.ibm.team.filesystem.client.FileSystemCore;
import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.IFileContentManager;
import com.ibm.team.filesystem.client.ILocation;
import com.ibm.team.filesystem.client.IRelativeLocation;
import com.ibm.team.filesystem.client.IShare;
import com.ibm.team.filesystem.client.IShareable;
import com.ibm.team.filesystem.client.ISharingDescriptor;
import com.ibm.team.filesystem.client.ResourceType;
import com.ibm.team.filesystem.client.internal.FileItemInfo;
import com.ibm.team.filesystem.client.internal.FileOptions;
import com.ibm.team.filesystem.client.internal.FileSystemManager;
import com.ibm.team.filesystem.client.internal.FileSystemServiceProxy;
import com.ibm.team.filesystem.client.internal.IFileStorage;
import com.ibm.team.filesystem.client.internal.IShareableVisitor;
import com.ibm.team.filesystem.client.internal.LinkInfo;
import com.ibm.team.filesystem.client.internal.LinkType;
import com.ibm.team.filesystem.client.internal.PathLocation;
import com.ibm.team.filesystem.client.internal.RelativeLocation;
import com.ibm.team.filesystem.client.internal.Shareable;
import com.ibm.team.filesystem.client.internal.SharingDescriptor;
import com.ibm.team.filesystem.client.internal.TempHelper;
import com.ibm.team.filesystem.client.internal.copyfileareas.ICopyFileAreaManager;
import com.ibm.team.filesystem.client.internal.localchanges.LocalChangeManager;
import com.ibm.team.filesystem.common.IFileContent;
import com.ibm.team.filesystem.common.IFileItem;
import com.ibm.team.filesystem.common.IFileItemHandle;
import com.ibm.team.filesystem.common.ISymbolicLink;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.client.TeamPlatform;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.filesystem.cli.core.util.SubcommandUtil;
import com.ibm.team.filesystem.cli.tools.CliFilesystemToolsException;
import com.ibm.team.scm.client.IBaselineConnection;
import com.ibm.team.scm.client.IConfiguration;
import com.ibm.team.scm.client.IConnection;
import com.ibm.team.scm.client.IWorkspaceConnection;
import com.ibm.team.scm.client.SCMPlatform;
import com.ibm.team.scm.common.IBaselineHandle;
import com.ibm.team.scm.common.IComponentHandle;
import com.ibm.team.scm.common.IFolder;
import com.ibm.team.scm.common.IFolderHandle;
import com.ibm.team.scm.common.IVersionable;
import com.ibm.team.scm.common.IVersionableHandle;
import com.ibm.team.scm.common.IWorkspaceHandle;
import com.ibm.team.scm.common.dto.IAncestorReport;
import com.ibm.team.scm.common.dto.IComponentStateSummary;
import com.ibm.team.scm.common.dto.INameItemPair;
import com.ibm.team.scm.common.dto.ISyncTime;

/**
 * Base comparator class containing general functionality for comparing a copy
 * file area (either workspace based or local file system based) and the
 * repository.
 */
public class FileSystemComparator {
    
    private final IShareable startingPoint;

    private final IFileStorage anchorRoot;

    private ILocation copyFileAreaRoot;

    public FileSystemComparator(Shareable startingPoint) {
        this.startingPoint = startingPoint;
        this.anchorRoot = startingPoint.getFileStorage();
        this.copyFileAreaRoot = startingPoint.getCopyFileAreaRoot();
    }
    
    public static void fail(String message) throws CliFilesystemToolsException {
        throw new CliFilesystemToolsException(message);
    }
    
    public static void assertTrue(String message, boolean condition) throws CliFilesystemToolsException {
        if (!condition) {
            fail(message);
        }
    }
    
    public static void assertFalse(String message, boolean condition) throws CliFilesystemToolsException {
        assertTrue(message, !condition);
    }
    
    public static void assertNotNull(String message, Object obj) throws CliFilesystemToolsException {
        if (obj == null) {
            fail(message);
        }
    }
    
    public static void assertEquals(String message, Object expected, Object actual) throws CliFilesystemToolsException {
        if (expected == null && actual == null)
            return;
        if (expected != null && expected.equals(actual))
            return;
        fail(message);
    }

    /**
     * Determine if the workspace matches the repository contents.
     * 
     * They are the same if: 1) For every file / folder in the repository, there
     * is a corresponding file / folder in the workspace. 2) For every file in
     * the repository, the contents in the repository matchs the contents in the
     * workspace. 3) For every file / folder in the reposistory, there is a
     * FileItemInfo record (whose content agrees with the repository info). 4)
     * For every file / folder in the reposistory, there is not entry in the
     * exclusion list for it. 5) For every file / folder in the workspace, there
     * is a corresponding file / folder in the repository or in the exclusions
     * list. 6) For every file / folder excluded, there is a corresponding file /
     * folder in the workspace. 7) For every FileItemInfo record, there is a
     * corresponding file / folder in the Repository.
     * 
     * @param connection
     *            The connection to the workspace or baseline to verify against.
     * @param component
     *            The component containing the remote file system item
     * @param remoteHandle
     *            The remote filesystem item that should be in the copy file
     *            area along with all its children.
     * @param local
     *            The local workarea that should be in sync with the repository
     * @param excludedItems
     *            A list of name of files and folders that should be in the work
     *            area but not in the repository. This is a set of Strings. The
     *            name should start with a "/" separator and end with one only
     *            if it is a folder.
     * @throws TeamRepositoryException
     * @throws FileNotFoundException
     * @throws IOException
     * @throws CoreException
     * @throws CliFilesystemToolsException 
     */
    public void assertEquals(IConnection connection, IComponentHandle component, IVersionableHandle remoteHandle, Set excludedItems) throws TeamRepositoryException,
            FileNotFoundException, IOException, CoreException, CliFilesystemToolsException {

        IVersionable remote = null;
        try {
            remote = fetch(connection, component, remoteHandle);
        } catch (TeamRepositoryException e) {
            fail("Could not find remote file system item for " + startingPoint.getLocalPath());
        }
        IRelativeLocation remoteName = locate(connection, component, remoteHandle);
        
        validateSharingDescriptor(connection, component, remoteName);

        Shareable shareable = new Shareable(copyFileAreaRoot, startingPoint.getLocalPath().append(remoteName), ResourceType.getResourceType(remoteHandle));
        LocalChangeManager.getInstance().computeChanges(shareable, null);
        LocalChangeManager.getInstance().syncPendingChanges(connection.getContextHandle(), component, shareable.getSandbox(),
        		null);

        // walk through the elements in the remote repository:
        // verifying 1) through 4)
        HashSet itemsSeen = new HashSet();
        walkRemoteRepo(connection, component, remote, remoteName, excludedItems, itemsSeen, true);

        // 5) For every file / folder in the workspace, there is a corresponding
        // file /
        // folder in the repository or in the exclusions list.
        assertTrue("Local root resource does not exist", rootExists());
        walkFileSystem(new FileSystemValidator(excludedItems, itemsSeen));

        // 6) For every file / folder excluded, there is a corresponding file /
        // folder in the workspace.
        for (Iterator iExcludes = excludedItems.iterator(); iExcludes.hasNext();) {
            String name = (String) iExcludes.next();
            ILocation path = startingPoint.getFullPath().append(name);
            assertTrue("Excluded item missing" + name, SubcommandUtil.exists(path, TempHelper.MONITOR));

        }

        // 7) For every FileItemInfo record, there is a corresponding file /
        // folder
        // in the Repository.
        Shareable localShareable = anchorRoot.getShareable();
        assertNotNull("Missing Shareable information", localShareable);
        localShareable.accept(new ShareableVisitor(itemsSeen), IShareable.DEPTH_INFINITE, true, false, null);
    }

    private void validateSharingDescriptor(IConnection connection, IComponentHandle component, IRelativeLocation remoteName) throws TeamRepositoryException, FileSystemException, CliFilesystemToolsException {
        Shareable shareable = new Shareable(copyFileAreaRoot, startingPoint.getLocalPath().append(remoteName), ResourceType.FOLDER);
        assertTrue("Can't validate share", shareable.exists(null));
        IShare share = shareable.getShare(null);
        assertNotNull("Item is not shared with copy file area " + remoteName, share);
        SharingDescriptor sharingDescriptor = (SharingDescriptor) share.getSharingDescriptor();
        assertNotNull("Item is not shared with copy file area " + remoteName, sharingDescriptor);
        assertEquals("mis-matched share component", component.getItemId(), sharingDescriptor.getComponent().getItemId());
        // don't verify connection because some tests use a different connection to verify against..
        ISyncTime expected = null;

        FileSystemServiceProxy fss = ((FileSystemManager) FileSystemCore.getFileSystemManager(connection.teamRepository())).getFileSystemService();
        List<IComponentStateSummary> summaries = fss.getComponentStateSummaries(connection,null);
        for (IComponentStateSummary summary : summaries) {
            if (summary.getComponent().sameItemId(component)) {
                expected = summary.getConfigurationTime();
                break;
            }
        }
        assertEquals("Configuration state mismatch", expected, sharingDescriptor.getConfigurationState());
    }

    protected boolean rootExists() throws FileSystemException {
    	return startingPoint.exists(null); 
    }

    /**
     * Same as
     * {@linkplain FileSystemComparator#assertEquals(IConnection, IComponentRootHandle, IFileSystemItemHandle, IResource, Set, boolean)
     * no excluded items (excludedItems == Collections.EMPTY_SET)

     * @throws CliFilesystemToolsException      */

    public void assertEquals(IConnection connection, IComponentHandle component, IVersionableHandle remoteHandle) throws TeamRepositoryException, IOException, CoreException, CliFilesystemToolsException {
        assertEquals(connection, component, remoteHandle, Collections.EMPTY_SET);
    }

    public void assertEquals(ISharingDescriptor descriptor, Set<String> excludedItems, String repositoryUri) throws TeamRepositoryException, FileNotFoundException, IOException, CoreException, CliFilesystemToolsException {
        ITeamRepository repo = getRepo(repositoryUri);
        repo.login(new NullProgressMonitor());
        IWorkspaceConnection connection = SCMPlatform.getWorkspaceManager(repo).getWorkspaceConnection((IWorkspaceHandle) descriptor.getConnectionHandle(), null);

        assertEquals(connection, descriptor.getComponent(), descriptor.getRootVersionable(), excludedItems);
        repo.logout(null);        
    }
    
    public void assertEquals(ISharingDescriptor descriptor, String repositoryUri) throws TeamRepositoryException, IOException, CoreException, CliFilesystemToolsException {
        assertEquals(descriptor, Collections.EMPTY_SET, repositoryUri);
//        ITeamRepository repo = getRepo(descriptor.getRepositoryUri());
//        repo.login(new NullProgressMonitor());
//        IWorkspaceConnection connection = SCMPlatform.getWorkspaceManager(repo).getWorkspaceConnection((IWorkspaceHandle) descriptor.getConnectionHandle(), null);
//
//        assertEquals(connection, descriptor.getComponent(), descriptor.getRootFolder(), Collections.EMPTY_SET);
//        repo.logout(null);
    }

    private ITeamRepository getRepo(String uri) {
        ITeamRepository repo = TeamPlatform.getTeamRepositoryService()
                .getUnmanagedRepository(uri);
        repo.registerLoginHandler(new ITeamRepository.ILoginHandler() {
            public ILoginInfo challenge(ITeamRepository repo) {
                return new ILoginInfo() {
                    public String getUserId() {
                        return "ADMIN";
                    }

                    public String getPassword() {
                        return "ADMIN";
                    }
                };
            }
        });
        return repo;
    }
    
    private void walkRemoteRepo(IConnection connection, IComponentHandle component, IVersionable remote, IRelativeLocation remoteName, Set excludedItems, Set itemsSeen, boolean verifyMetaData)
            throws FileNotFoundException, TeamRepositoryException, IOException, CoreException, CliFilesystemToolsException {
        // 1) For every file / folder in the repository, there is a
        // corresponding
        // file / folder in the workspace.
        Shareable shareable = new Shareable(copyFileAreaRoot, remoteName, ResourceType.getResourceType(remote));
        IFileStorage fileStorage = shareable.getFileStorage();
        assertNotNull("Remote resource missing in copy file area " + remoteName, fileStorage);

        // verify the meta data is right
        if (verifyMetaData) {
            verifyMetaData(remote, fileStorage);
        }

        // verify the content matches
        assertEquals(connection, remote, shareable);

        boolean isFolder = remote instanceof IFolder;

        // 4) For every file / folder in the reposistory, there is not entry in
        // the
        // exclusion list for it.
        String stdName = ResourceTestUtils_standardizePath(remoteName.toString(), isFolder);
        assertFalse("Excluded item in repo " + remoteName, excludedItems.contains(stdName));

        itemsSeen.add(stdName);

        if (isFolder) {
            // for every item in the folder repeat the comparison...
            Map<String, IVersionableHandle> children;
            if (connection instanceof IWorkspaceConnection) {
                children = ((IWorkspaceConnection) connection).configuration(component).childEntries((IFolderHandle) remote, null);
            } else {
                children = ((IBaselineConnection) connection).configuration().childEntries((IFolderHandle) remote, null); 
            }
            for (Entry<String, IVersionableHandle> entry : children.entrySet()) {
                IVersionable child = null;
                try {
                    child = fetch(connection, component, entry.getValue());
                } catch (TeamRepositoryException e) {
                    e.printStackTrace();
                    fail("Could not find remote file system item for " + remoteName);
                }
                walkRemoteRepo(connection, component, child, remoteName.append(entry.getKey()), excludedItems, itemsSeen, verifyMetaData);
            }
        }
    }

    private static String ResourceTestUtils_standardizePath(String path, boolean isFolder) {
		if (path.equals("")) {
			return "/";
		}
		if (path.charAt(0) != '/') {
			path = "/" + path;
		}
		if (isFolder && path.charAt(path.length() - 1) != '/') {
			path = path + "/";
		}
		return path;
	}

	private static IVersionable fetch(IConnection connection, IComponentHandle component, IVersionableHandle handle) throws TeamRepositoryException {
        IConfiguration config;
        if (connection instanceof IWorkspaceConnection) {
            config = ((IWorkspaceConnection) connection).configuration(component);
        } else {
            config = ((IBaselineConnection) connection).configuration();
        }
        IVersionable child = (IVersionable) config.fetchCompleteItem(handle, null);
        return child;
    }

    /**
     * validate that the meta data matches the remote information
     * @throws FileSystemException 
     * @throws CliFilesystemToolsException 
     */
    private void verifyMetaData(IVersionable remote, IFileStorage fileStorage) throws FileSystemException, CliFilesystemToolsException {
        String name = fileStorage.getShareable().getLocalPath().toString();
        // 3) For every file / folder in the reposistory, there is a
        // FileItemInfo
        // record (whose content agrees with the repository info).

        // verify FileItemInfo
        FileItemInfo itemInfo = ICopyFileAreaManager.instance.createCopyFileArea(copyFileAreaRoot, null).getItemInfo(fileStorage.getShareable().getLocalPath());
        assertNotNull("Missing FileItemInfo for " + name, itemInfo);
        assertEquals("FileItemInfo " + name, remote.getItemId(), itemInfo.getVersionableHandle().getItemId());
        assertEquals("FileItemInfo " + name, remote instanceof IFolder, itemInfo.isFolder());

        // TODO Can't trust state id of folders since we work off of
        // IFileChanges
        if (remote instanceof IFileItem || remote instanceof ISymbolicLink) {
            assertEquals("FileItemInfo " + name, remote.getStateId(), itemInfo.getVersionableHandle().getStateId());
        }

        // verify IShareable
        IShareable shareable = fileStorage.getShareable();
        assertNotNull("Missing ISharable for " + name, shareable);
        assertFalse("IShareable " + name + " is dirty", isDirty(shareable));

        // verify local & remote
        ResourceType remoteType = ResourceType.getResourceType(remote);
        assertEquals("FileItemInfo " + name, remoteType, ResourceType.getResourceType(itemInfo.getVersionableHandle()));
    }

    private boolean isDirty(IShareable shareable) throws FileSystemException {
        return LocalChangeManager.getInstance().isContentDirty(shareable, null);
//        IShare share = shareable.getShare();
//        if (share != null) {
//            IVersionableHandle handle = shareable.getVersionable();
//            if (handle != null) {
//                ILocalChange pendingChange = LocalChangeManager.getInstance().getPendingChange(share, handle);
//                return pendingChange.getType() != ILocalChange.NO_CHANGE;
//            }
//        }
//        return false;
    }

    private static void assertEquals(IConnection connection, IVersionable remote, Shareable shareable) throws FileNotFoundException, TeamRepositoryException, IOException,
            CoreException, CliFilesystemToolsException {
        
        IFileStorage fileStorage = shareable.getFileStorage();
        String name = shareable.getLocalPath().toString();

        // verify local & remote
        ResourceType remoteType = ResourceType.getResourceType(remote);       
        ResourceType localType = shareable.getResourceType(null);

        assertEquals("Not same type " + name, remoteType, localType);

        if (remoteType == ResourceType.FILE) {
            // 2) For every file in the repository, the contents in the
            // repository
            // matchs the contents in the workspace.
            IFileContentManager contentManager = FileSystemCore.getContentManager(connection.teamRepository());

            IFileContent remoteContents = ((IFileItem) remote).getContent();
            ByteArrayOutputStream remoteBytes = new ByteArrayOutputStream((int) remoteContents.getEstimatedConvertedLength());
            contentManager.retrieveContent((IFileItemHandle)remote, remoteContents, remoteBytes, null);
            ByteArrayOutputStream localBytes = new ByteArrayOutputStream();
            InputStream fis = fileStorage.getContents(new FileOptions(false, (IFileItem) remote));
            ResourceTestUtils_transferStreams(fis, localBytes);

            assertEquals(name, remoteBytes.toByteArray(), localBytes.toByteArray());
        } else if (remoteType == ResourceType.SYMBOLIC_LINK) {
        	ISymbolicLink remoteLink = (ISymbolicLink)remote;
        	
        	String remoteTarget = remoteLink.getTarget();
            remoteTarget = shareable.getFileStorage().convertTargetIntoLocalForm(remoteTarget);
            boolean remoteIsDirectoryLink = remoteLink.isDirectoryLink();
            
        	LinkInfo info = shareable.getFileStorage().getLinkInfo(null);
        	
        	String localTarget = info.getTarget();
        	boolean localIsDirectoryLink = info.getType() == LinkType.NONE
        									? shareable.getFileItemInfo(null).isDirectoryLink()
        									: info.getType() == LinkType.DIRECTORY;
        	
        	assertTrue("Symbolic link targets don't match", remoteTarget.equals(localTarget) && remoteIsDirectoryLink == localIsDirectoryLink);
        }
    }

    private static void ResourceTestUtils_transferStreams(InputStream source, OutputStream destination) throws IOException {
        byte[] buffer = new byte[8192];
        try {
            while (true) {
                int bytesRead = -1;
                bytesRead = source.read(buffer);
                if (bytesRead == -1)
                    break;
                destination.write(buffer, 0, bytesRead);
            }
        } finally {
            try {
                source.close();
            } catch (IOException e) {
                // ignore
            } finally {
                try {
                    destination.close();
                } catch (IOException e) {
                    // ignore
                }
            }
        }
    }


	public static void assertEquals(String path, byte[] expected, byte[] actual) throws CliFilesystemToolsException {
        assertEquals("content lengths do not match for " + path, expected.length, actual.length);
        for (int i = 0; i < actual.length; i++)
            assertEquals("content does not match (" + i + "/" + actual.length + ") for " + path, expected[i], actual[i]);
    }

    public void assertEquals(ISharingDescriptor descriptor, IVersionableHandle remoteHandle, String repositoryUri) throws TeamRepositoryException, IOException, CoreException, CliFilesystemToolsException {
        ITeamRepository repo = getRepo(repositoryUri);
        repo.login(new NullProgressMonitor());
        // IRepositoryConnection repository = repoHandle.connect();
        IConnection connection;
        if (descriptor.associatedWithWorkspace()) {
            connection = SCMPlatform.getWorkspaceManager(repo).getWorkspaceConnection((IWorkspaceHandle) descriptor.getConnectionHandle(), null);
        } else {
            connection = SCMPlatform.getWorkspaceManager(repo).getBaselineConnection((IBaselineHandle) descriptor.getConnectionHandle(), null);
        }

        assertEquals(connection, descriptor.getComponent(), remoteHandle, new HashSet());
        repo.logout(null);
    }

    protected final ILocation getCopyFileAreaRoot() {
        return copyFileAreaRoot;
    }

    protected static class FileSystemValidator {

        private Set excludedItems;

        private Set itemsSeen;

        public FileSystemValidator(Set excludedItems, Set itemsSeen) {
            this.excludedItems = excludedItems;
            this.itemsSeen = itemsSeen;
        }

        public void validate(String relativePath, ResourceType type) throws CliFilesystemToolsException {
            String stdName = ResourceTestUtils_standardizePath(relativePath, type == ResourceType.FOLDER);

            assertTrue("Not seen nor excluded " + stdName, excludedItems.contains(stdName) || itemsSeen.contains(stdName));
        }

    }

    private static class ShareableVisitor implements IShareableVisitor {

        private final Set itemsSeen;

        public ShareableVisitor(Set itemsSeen) {
            this.itemsSeen = itemsSeen;
        }

        public boolean visit(IShareable shareable, IProgressMonitor monitor) {

            // if it is remote, verify that we saw the file
            // when we traversed the remote side.
        	try {
	        	SubMonitor progress = SubMonitor.convert(monitor, 100);
	            if ((shareable.getState(progress.newChild(50)) & IShareable.REMOTE) != 0) {
	                String stdName = ResourceTestUtils_standardizePath(shareable.getLocalPath().toString(),
	                		shareable.getRemote(progress.newChild(50)) instanceof IFolderHandle);
	                return false;
	            }
        	} catch (FileSystemException e) {
        		throw new RuntimeException(e);
        	}
            return true;
        }
        @Override
        public IComponentHandle getComponent() {
        	return null;
        }
    }

    /**
     * Determine if the workspace matches the repository contents.
     * 
     * They are the same if: 1) For every file / folder in the repository, there
     * is a corresponding file / folder in the workspace. 2) For every file in
     * the repository, the contents in the repository matchs the contents in the
     * workspace. 4) For every file / folder in the reposistory, there is not
     * entry in the exclusion list for it. 5) For every file / folder in the
     * workspace, there is a corresponding file / folder in the repository or in
     * the exclusions list. 6) For every file / folder excluded, there is a
     * corresponding file / folder in the workspace. 8) Ensure that there is no
     * Sharing or FileItem info for the workspace
     * 
     * @param connection
     *            The connection to the workspace or baseline to verify against.
     * @param component
     *            The component containing the remote file system item
     * @param remoteHandle
     *            The remote filesystem item that should be in the copy file
     *            area along with all its children.
     * @param local
     *            The local workarea that should be in sync with the repository
     * @param excludedItems
     *            A list of name of files and folders that should be in the work
     *            area but not in the repository. This is a set of Strings. The
     *            name should start with a "/" separator and end with one only
     *            if it is a folder.
     * @throws TeamRepositoryException
     * @throws CoreException
     * @throws IOException
     * @throws FileNotFoundException
     * @throws CliFilesystemToolsException 
     */
    public void assertLoaded(IWorkspaceConnection connection, IComponentHandle component, IVersionableHandle remoteHandle, HashSet excludes) throws TeamRepositoryException,
            FileNotFoundException, IOException, CoreException, CliFilesystemToolsException {

        IVersionable remote = null;
        try {
            remote = fetch(connection, component, remoteHandle);
        } catch (TeamRepositoryException e) {
            e.printStackTrace();
            fail("Could not find remote file system item for " + startingPoint.getLocalPath());
        }
        IRelativeLocation remoteName = locate(connection, component, remoteHandle);

        Shareable shareable = new Shareable(copyFileAreaRoot, remoteName, ResourceType.getResourceType(remote));
        LocalChangeManager.getInstance().computeChanges(shareable, null);
        LocalChangeManager.getInstance().syncPendingChanges(connection.getContextHandle(), component, shareable.getSandbox(), null);

        // walk through the elements in the remote repository:
        // verifying 1) through 4)
        HashSet itemsSeen = new HashSet();
        walkRemoteRepo(connection, component, remote, remoteName, excludes, itemsSeen, false);

        // 5) For every file / folder in the workspace, there is a corresponding
        // file /
        // folder in the repository or in the exclusions list.
        assertTrue("Local root resource does not exist", rootExists());
        walkFileSystem(new FileSystemValidator(excludes, itemsSeen));

        // 6) For every file / folder excluded, there is a corresponding file /
        // folder in the workspace.
        for (Iterator iExcludes = excludes.iterator(); iExcludes.hasNext();) {
            String name = (String) iExcludes.next();
            ILocation path = startingPoint.getFullPath().append(name);
            assertTrue("Excluded item missing" + name, SubcommandUtil.exists(path, TempHelper.MONITOR));
        }

        // 8) Ensure that there is no Sharing or FileItem info for the workspace
        // do not use CopyFileAreaStore.getCopyFileArea because that creates a
        // copy file area by default.
        assertFalse(null, ICopyFileAreaManager.instance.copyFileAreaExists(startingPoint.getFullPath(), ICopyFileAreaManager.NONE));
    }

    private IRelativeLocation locate(IConnection connection, IComponentHandle component, IVersionableHandle handle) throws TeamRepositoryException {
        IConfiguration configuration;
        if (connection instanceof IWorkspaceConnection) {
            configuration = ((IWorkspaceConnection) connection).configuration(component);
        } else {
            configuration = ((IBaselineConnection) connection).configuration();
        }
        List ancestors = configuration.locateAncestors(Collections.singletonList(handle), null);

        IRelativeLocation relativeLocation = RelativeLocation.EMPTY_LOCATION;
        // go through all the names but skip the root folder
        Iterator iAncestors = ((IAncestorReport) ancestors.get(0)).getNameItemPairs().iterator();
        iAncestors.next();
        for (; iAncestors.hasNext();) {
            INameItemPair pair = (INameItemPair) iAncestors.next();
            relativeLocation = relativeLocation.append(pair.getName());
        }
        return relativeLocation;
    }

    protected void walkFileSystem(FileSystemValidator validator) throws IOException, CoreException, CliFilesystemToolsException {
        traverse((File) startingPoint.getFullPath().getAdapter(File.class), validator);
    }

    /**
     * walks the file system to ensure all the items in the file system are
     * either in the repo or have been excluded (i.e. are local only)
     * 
     * @throws CoreException
     * @throws CliFilesystemToolsException 
     */
    private void traverse(File localFile, FileSystemValidator validator) throws IOException, CoreException, CliFilesystemToolsException {
        ResourceType type = null;
        IFileStore fileStore = EFS.getLocalFileSystem().getStore(new Path(localFile.getAbsolutePath()));
        IFileInfo info = fileStore.fetchInfo(EFS.NONE, null);
        if (info.getAttribute(EFS.ATTRIBUTE_SYMLINK)) {
            type = ResourceType.SYMBOLIC_LINK;
        } else if (localFile.isDirectory()) {
            type = ResourceType.FOLDER;
        } else if (info.exists()) {
            type = ResourceType.FILE;
        } else {
            fail("Unexpected resource " + localFile.getCanonicalPath());
        }

        ILocation local = new PathLocation(localFile.getAbsolutePath());
        IRelativeLocation path = local.getLocationRelativeTo(copyFileAreaRoot);

        validator.validate(path.toString(), type);

        // repeat for each of its children
        if (type == ResourceType.FOLDER) {
            File[] members = localFile.listFiles();
            for (int i = 0; i < members.length; i++) {
                traverse(members[i], validator);
            }
        }
    }
}
