/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2013. All Rights Reserved.
 *
 * Note to U.S. Government Users Restricted Rights:  
 * Use, duplication or disclosure restricted by GSA ADP Schedule 
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.filesystem.cli.tools.verify;
import static com.ibm.team.filesystem.cli.core.subcommands.CommonOptions.OPT_URI;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Date;
import java.util.List;

import org.eclipse.core.filesystem.EFS;
import org.eclipse.core.filesystem.IFileStore;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;

import com.ibm.team.filesystem.cli.core.AbstractSubcommand;
import com.ibm.team.filesystem.cli.core.internal.ScmCommandLineArgument;
import com.ibm.team.filesystem.cli.core.util.ConnectionInfo;
import com.ibm.team.filesystem.cli.core.util.RepoUtil;
import com.ibm.team.filesystem.cli.core.util.StatusHelper;
import com.ibm.team.filesystem.cli.core.util.SubcommandUtil;
import com.ibm.team.filesystem.client.FileSystemException;
import com.ibm.team.filesystem.client.internal.ManagedFileStore;
import com.ibm.team.filesystem.client.rest.IFilesystemRestClient;
import com.ibm.team.filesystem.client.rest.parameters.ParmsActivate;
import com.ibm.team.filesystem.client.rest.parameters.ParmsAutoMerge;
import com.ibm.team.filesystem.client.rest.parameters.ParmsBackupDilemmaHandler;
import com.ibm.team.filesystem.client.rest.parameters.ParmsCheckInChanges;
import com.ibm.team.filesystem.client.rest.parameters.ParmsCheckInShareablesRequest;
import com.ibm.team.filesystem.client.rest.parameters.ParmsCompleteChangeSets;
import com.ibm.team.filesystem.client.rest.parameters.ParmsConflictsToResolve;
import com.ibm.team.filesystem.client.rest.parameters.ParmsDeletedContentDilemmaHandler;
import com.ibm.team.filesystem.client.rest.parameters.ParmsGetWorkspace;
import com.ibm.team.filesystem.client.rest.parameters.ParmsGetWorkspaceDetails;
import com.ibm.team.filesystem.client.rest.parameters.ParmsPendingChangesDilemmaHandler;
import com.ibm.team.filesystem.client.rest.parameters.ParmsPostWorkspace;
import com.ibm.team.filesystem.client.rest.parameters.ParmsPreOperationRefresh;
import com.ibm.team.filesystem.client.rest.parameters.ParmsResumeChangeSets;
import com.ibm.team.filesystem.client.rest.parameters.ParmsSandboxUpdateDilemmaHandler;
import com.ibm.team.filesystem.client.rest.parameters.ParmsShare;
import com.ibm.team.filesystem.client.rest.parameters.ParmsShareRequest;
import com.ibm.team.filesystem.client.rest.parameters.ParmsSuspendChangeSets;
import com.ibm.team.filesystem.client.rest.parameters.ParmsWorkspace;
import com.ibm.team.filesystem.common.internal.rest.client.changeset.CheckInResultDTO;
import com.ibm.team.filesystem.common.internal.rest.client.core.WorkspaceComponentDTO;
import com.ibm.team.filesystem.common.internal.rest.client.core.WorkspaceDetailsDTO;
import com.ibm.team.filesystem.common.internal.rest.client.share.ShareResultDTO;
import com.ibm.team.filesystem.common.internal.rest.client.sync.ComponentSyncDTO;
import com.ibm.team.filesystem.common.internal.rest.client.sync.ConflictSyncDTO;
import com.ibm.team.filesystem.common.internal.rest.client.sync.SyncViewDTO;
import com.ibm.team.filesystem.common.internal.rest.client.sync.UnresolvedFolderSyncDTO;
import com.ibm.team.filesystem.common.internal.rest.client.sync.WorkspaceSyncDTO;
import com.ibm.team.filesystem.common.internal.rest.client.workspace.GetWorkspaceDetailsResultDTO;
import com.ibm.team.repository.client.ITeamRepository;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.UUID;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IOptionSource;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.ICommandLine;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.NamedOptionDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.Options;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.ConflictingOptionException;
import com.ibm.team.rtc.cli.infrastructure.internal.util.IndentingPrintStream;


public class PackagingTestCmd extends AbstractSubcommand implements IOptionSource {
	private static NamedOptionDefinition SANDBOX_LOC = new NamedOptionDefinition("s", "sandbox", 1);
	
	public Options getOptions() throws ConflictingOptionException {
		Options opts = new Options(false);
		
		opts.setLongHelp("This test command will test out the packaging of this install to\nensure that certain known areas are working.\nNote: this command should be run in an empty directory.");

		SubcommandUtil.addRepoLocationToOptions(opts);
		
		opts.addOption(SANDBOX_LOC, "Temporary sandbox location. Sandbox must be empty and must be canonical.");
	        
		return opts;
	}
	
	public PackagingTestCmd() {
		super();
	}

	@Override
	public void run() throws FileSystemException {
		// Find the credentials
		ICommandLine subargs = config.getSubcommandCommandLine();
		
		String url = subargs.getOption(OPT_URI, null);
		
		if (url == null) {
			throw StatusHelper.argSyntax("Must define a repository url");
		}

		ConnectionInfo info = config.getConnectionInfo(url, null, true, false);
		
		// Attempt to login
        IFilesystemRestClient client = SubcommandUtil.setupDaemon(config);
        ITeamRepository repo = RepoUtil.login(config, client, info);

        if (repo.getErrorState() != ITeamRepository.ERROR_STATE_NONE) {
        	repo.getLastError().printStackTrace(config.getContext().stderr());
        	return;
        }
        
		config.getContext().stdout().println("Logged in");
		
		// create a repo workspace
		ParmsPostWorkspace parms = new ParmsPostWorkspace();
		parms.name = "CLI Sniff Test "+new Date().toString();
		parms.description = "";
		parms.repositoryUrl = repo.getRepositoryURI();
		parms.isStream = false;
		
		WorkspaceDetailsDTO wsDetails;
		try {
			wsDetails = client.postWorkspace(parms, null);
		} catch (TeamRepositoryException e) {
			throw StatusHelper.wrap("Could not create repository workspace", e, new IndentingPrintStream(config.getContext().stderr()), parms.repositoryUrl);
		}
		
		config.getContext().stdout().println("Repository workspace created: "+parms.name);
		config.getContext().stdout().println("Note: A new component is created each time that this command is executed.");
		config.getContext().stdout().println("      This command should be run against a test server and not a production one.");
		
		ParmsWorkspace parmsWorkspace = new ParmsWorkspace();
		parmsWorkspace.repositoryUrl = repo.getRepositoryURI();
		parmsWorkspace.workspaceItemId = wsDetails.getItemId();
		
        ParmsActivate activateParms = new ParmsActivate();
        activateParms.workspaces = new ParmsWorkspace[] {parmsWorkspace};
        try {
			client.postActivate(activateParms, null);
		} catch (TeamRepositoryException e) {
			throw StatusHelper.wrap("Could not activate workspace", e, new IndentingPrintStream(config.getContext().stderr()), parms.repositoryUrl);
		}
		
		config.getContext().stdout().println("Repository workspace activated");
		
		IPath sandboxPath = null;
		File sandbox = null;
		
		if (config.getSubcommandCommandLine().hasOption(SANDBOX_LOC)) {
		    sandboxPath = new Path(ScmCommandLineArgument.create(config.getSubcommandCommandLine().getOptionValue(SANDBOX_LOC), config).getStringValue());
			sandbox = sandboxPath.toFile();
			if (sandbox.isFile()) {
				config.getContext().stderr().println("ERROR: Sandbox is a file, not a directory. Aborting...");
				return;
			}
			
			String[] children = sandbox.list();
			if (children != null && children.length != 0) {
				config.getContext().stderr().println("ERROR: Sandbox is not completely empty. Aborting...");
				return;
			}
		} else {
			sandboxPath = new Path(System.getProperty("java.io.tmpdir")).append(UUID.generate().getUuidValue());
			sandbox = sandboxPath.toFile();
			
			if (!sandbox.mkdirs()) {
				config.getContext().stderr().println("ERROR: Could not create sandbox in "+sandbox+ ". Try again using the sandbox option (-s) to provide the location of an empty sandbox.");
				return;
			}
		}
		
		// Check canonical path
		String resolvedPath;
		try {
			resolvedPath = ManagedFileStore.getResolvedPath(sandboxPath.toOSString());
			
			if (resolvedPath == null || !sandboxPath.toOSString().equals(resolvedPath)) {
				config.getContext().stderr().println("ERROR: Sandbox path is not in canonical form. Try again using the sandbox option (-s) to provide the location of an empty sandbox in canonical form.");
				return;
			}
		} catch (CoreException e1) {
			e1.printStackTrace(config.getContext().stderr());
			config.getContext().stderr().println("ERROR: Sandbox path is not in a correct form. Try again using the sandbox option (-s) to provide the location of an empty sandbox in canonical form.");
			return;
		}
		
		// Create a simple sandbox, file, folder structure
		File folder = new File(sandbox, "folder");
		folder.mkdirs();
		
		try {
			PrintStream os = new PrintStream(new File(folder, "file.txt"));
			writeFileContents(os, false, false);
			os.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace(config.getContext().stderr());
			config.getContext().stderr().println("ERROR: Could not write new file contents. Aborting...");
			return;
		}
		
		config.getContext().stdout().println("Sample files created in "+sandbox.getAbsolutePath());
		
		// Share the folder to the repository workspace
        ParmsShare parmsShare = new ParmsShare();
        
        ParmsShareRequest shareRequest = new ParmsShareRequest();
        shareRequest.workspace = parmsWorkspace;

        shareRequest.componentName = parms.name;
        shareRequest.sandboxPath = sandbox.getAbsolutePath();
        shareRequest.itemsToShare = new String[] {"folder"};
        
        parmsShare.shareRequests = new ParmsShareRequest[] {shareRequest};
        parmsShare.overwrite = false;
        
        parmsShare.sandboxUpdateDilemmaHandler = new ParmsSandboxUpdateDilemmaHandler();
        parmsShare.sandboxUpdateDilemmaHandler.backupDilemmaHandler = new ParmsBackupDilemmaHandler();
        parmsShare.sandboxUpdateDilemmaHandler.backupDilemmaHandler.backupEnabled = false;
        
        parmsShare.sandboxUpdateDilemmaHandler.deletedContentDilemmaHandler = new ParmsDeletedContentDilemmaHandler();
        parmsShare.sandboxUpdateDilemmaHandler.deletedContentDilemmaHandler.deletedContentDirection = IFilesystemRestClient.FAIL;

        ShareResultDTO shareResult = null;
        try {
            shareResult = client.postShare(parmsShare, null);
        } catch (TeamRepositoryException e) {
            throw StatusHelper.wrap("Share failed", e, new IndentingPrintStream(config.getContext().stderr()), "");
        }
        
        config.getContext().stdout().println("Shared files to repository workspace");
        
        // Complete the initial change set
        String cs1Id = (String)shareResult.getChangeSetsCommitted().get(0);
        
        ParmsCompleteChangeSets parmsComplete = new ParmsCompleteChangeSets();
        parmsComplete.changeSetItemIds = new String [] {cs1Id};
        parmsComplete.workspace = parmsWorkspace;
        
        try {
        	client.postCompleteChangeSets(parmsComplete, null);
        } catch (TeamRepositoryException e) {
        	throw StatusHelper.wrap("Complete failed", e, new IndentingPrintStream(config.getContext().stderr()), "");
        }
        
        config.getContext().stdout().println("Completed initial change set");
        
        // Make changes on disk and check into a new change set
		try {
			PrintStream os = new PrintStream(new File(folder, "file.txt"));
			
			writeFileContents(os, true, false);
			
			os.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace(config.getContext().stderr());
			config.getContext().stderr().println("ERROR: Could not write file contents to disk. Aborting...");
			return;
		}
		
		boolean linkCreated = true;
		
		try {
			IFileStore linkStore = EFS.getStore(new File(folder, "link").toURI());
			ManagedFileStore.createSymlink(linkStore, "file.txt", false, null);
		} catch (CoreException e) {
			config.getContext().stderr().println("ERROR: Could not create symbolic link.");
			config.getContext().stderr().println("       On Windows make sure to run the command prompt \"As Administrator\" and try again.");
			config.getContext().stderr().println("       Check that SCM filesystem libraries are available for this platform.");
			
			e.printStackTrace(config.getContext().stderr());
			
			linkCreated = false;
		}
		
		ParmsCheckInChanges parmsCheckIn = new ParmsCheckInChanges();
		ParmsCheckInShareablesRequest parmsCIShareable = new ParmsCheckInShareablesRequest();
		parmsCIShareable.sandboxPath = sandbox.getAbsolutePath();
		if (linkCreated) {
			parmsCIShareable.resourcesToCheckIn = new String[] { "folder/file.txt", "folder/link" };
		} else {
			parmsCIShareable.resourcesToCheckIn = new String[] { "folder/file.txt" };
		}
		parmsCheckIn.paths = new ParmsCheckInShareablesRequest[] {parmsCIShareable};
        parmsCheckIn.preoperationRefresh = new ParmsPreOperationRefresh();
        parmsCheckIn.preoperationRefresh.refreshMode = IFilesystemRestClient.PREOP_REFRESH_COMPLETE;
		
		CheckInResultDTO checkInResult = null;
		try {
			checkInResult = client.postCheckInChanges(parmsCheckIn, null);
        } catch (TeamRepositoryException e) {
        	throw StatusHelper.wrap("Checkin failed", e, new IndentingPrintStream(config.getContext().stderr()), "");
        }
        
        config.getContext().stdout().println("Checked in changes to the file");
        
        String cs2Id = (String) checkInResult.getChangeSetsCommitted().get(0);
        
        // Complete and suspend the new change set
        
        parmsComplete = new ParmsCompleteChangeSets();
        parmsComplete.changeSetItemIds = new String [] {cs2Id};
        parmsComplete.workspace = parmsWorkspace;
        
        try {
        	client.postCompleteChangeSets(parmsComplete, null);
        } catch (TeamRepositoryException e) {
        	throw StatusHelper.wrap("Complete failed", e, new IndentingPrintStream(config.getContext().stderr()), "");
        }
        
        config.getContext().stdout().println("Completed the new change set");
        
        ParmsSuspendChangeSets parmsSuspend = new ParmsSuspendChangeSets();
        parmsSuspend.workspace = parmsWorkspace;
        parmsSuspend.changeSetItemIds = new String[] {cs2Id};
        
        try {
			client.postSuspendChangeSets(parmsSuspend, null);
		} catch (TeamRepositoryException e) {
			throw StatusHelper.wrap("Suspend failed", e, new IndentingPrintStream(config.getContext().stderr()), "");
		}
		
		config.getContext().stdout().println("Suspended the new change set");
		
		// Make new changes and check them into a new change set
		try {
			PrintStream os = new PrintStream(new File(folder, "file.txt"));
			
			writeFileContents(os, false, true);
			
			os.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace(config.getContext().stderr());
			config.getContext().stderr().println("ERROR: Could not write file changes to disk. Aborting...");
			return;
		}
		
		parmsCheckIn = new ParmsCheckInChanges();
		parmsCIShareable = new ParmsCheckInShareablesRequest();
		parmsCIShareable.sandboxPath = sandbox.getAbsolutePath();
		parmsCIShareable.resourcesToCheckIn = new String[] { "folder/file.txt" };
		parmsCheckIn.paths = new ParmsCheckInShareablesRequest[] {parmsCIShareable};
        parmsCheckIn.preoperationRefresh = new ParmsPreOperationRefresh();
        parmsCheckIn.preoperationRefresh.refreshMode = IFilesystemRestClient.PREOP_REFRESH_COMPLETE;
		
		checkInResult = null;
		try {
			checkInResult = client.postCheckInChanges(parmsCheckIn, null);
        } catch (TeamRepositoryException e) {
        	throw StatusHelper.wrap("Checkin failed", e, new IndentingPrintStream(config.getContext().stderr()), "");
        }
        
        config.getContext().stdout().println("Checked in new changes to the file");
        
        // Resume the other change set creating an auto-resolvable conflict
        ParmsResumeChangeSets parmsResume = new ParmsResumeChangeSets();
        parmsResume.changeSetItemIds = new String[] {cs2Id};
        parmsResume.workspace = parmsWorkspace;
        
        try {
			client.postResumeChangeSets(parmsResume, null);
		} catch (TeamRepositoryException e) {
			throw StatusHelper.wrap("Resume failed", e, new IndentingPrintStream(config.getContext().stderr()), "");
		}
		
		config.getContext().stdout().println("Resumed existing change set");
		
		// Attempt auto-resolve of the conflict
		ParmsGetWorkspaceDetails parmsWsInfo = new ParmsGetWorkspaceDetails();
		parmsWsInfo.workspaces = new ParmsGetWorkspace[1];
		parmsWsInfo.workspaces[0] = new ParmsGetWorkspace();
		parmsWsInfo.workspaces[0].refresh = true;
		parmsWsInfo.workspaces[0].includeComponents = true;
		parmsWsInfo.workspaces[0].workspace = parmsWorkspace;
		
		GetWorkspaceDetailsResultDTO wsInfo = null;
		try {
			wsInfo = client.getWorkspaceDetails(parmsWsInfo, null);
		} catch (TeamRepositoryException e) {
			throw StatusHelper.wrap("Workspace lookup failed.", e, new IndentingPrintStream(config.getContext().stderr()), "");
		}
		
		String componentId = ((WorkspaceComponentDTO)((WorkspaceDetailsDTO)wsInfo.getWorkspaceDetails().get(0)).getComponents().get(0)).getItemId();

		try {
			client.getRefreshSyncView(null);
			client.getJoinSyncView(null);
        } catch (TeamRepositoryException e) {
			throw StatusHelper.wrap("Sync view retrieval failed", e, new IndentingPrintStream(config.getContext().stderr()), "");
		}
		
        SyncViewDTO syncView = null;
        try {
        	syncView = client.getJoinSyncView(null);
        } catch (TeamRepositoryException e) {
			throw StatusHelper.wrap("Sync view retrieval failed", e, new IndentingPrintStream(config.getContext().stderr()), "");
		}
        
        ParmsConflictsToResolve parmsConflictToResolve = null;
        for (WorkspaceSyncDTO workspaceView : (List<WorkspaceSyncDTO>) syncView.getWorkspaces()) {
            if (workspaceView.getWorkspaceItemId().equals(parmsWorkspace.workspaceItemId)) {
                for (ComponentSyncDTO componentView : (List<ComponentSyncDTO>) workspaceView.getComponents()) {
                    for (UnresolvedFolderSyncDTO folderView : (List<UnresolvedFolderSyncDTO>) componentView.getUnresolved()) {
                       ConflictSyncDTO conflict = (ConflictSyncDTO) folderView.getConflicts().get(0);
                       
                       parmsConflictToResolve = new ParmsConflictsToResolve();
                       
                       parmsConflictToResolve.componentItemId = componentId;
                       parmsConflictToResolve.versionableItemId = conflict.getVersionableItemId();
                       parmsConflictToResolve.kind = conflict.getKind();
                       if (IFilesystemRestClient.CONFLICT.equals(conflict.getKind())) {
                    	   parmsConflictToResolve.conflictType = conflict.getConflictType();
                       }
                    }
                }
            } else {
            	config.getContext().stdout().println("That's weird, there are other workspaces in this sync view.");
            }
        }
        
        if (parmsConflictToResolve == null) {
        	config.getContext().stderr().println("ERROR: Could not find the conflict in the sync view in order to auto-resolve. Aborting...");
        	return;
        }
		
		ParmsAutoMerge parmsAutoMerge = new ParmsAutoMerge();
		parmsAutoMerge.workspace = parmsWorkspace;
		parmsAutoMerge.conflictsToResolve = new ParmsConflictsToResolve[] {parmsConflictToResolve};
		parmsAutoMerge.pendingChangesDilemmaHandler = new ParmsPendingChangesDilemmaHandler();
		parmsAutoMerge.pendingChangesDilemmaHandler.pendingChangesInstruction = IFilesystemRestClient.CONTINUE;
		
		try {
			client.postAutoResolve(parmsAutoMerge, null);
		} catch (TeamRepositoryException e) {
			e.printStackTrace(config.getContext().stderr());
		}
		
		config.getContext().stdout().println("Auto-merged the conflict");
		
		// Check file on disk and make sure it has all of the lines
		try {
			FileInputStream fis = new FileInputStream(new File(folder, "file.txt"));
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			int b = fis.read();
			
			while (b != -1) {
				baos.write(b);
				b = fis.read();
			}
			baos.close();
			fis.close();
			
			String contents = new String(baos.toByteArray());
			
			if (!contents.contains("Line9") || !contents.contains("Line0")) {
				config.getContext().stderr().println("ERROR: File was not auto-merged. Verify that SWT libraries are available for this platform.");
			}
		} catch (FileNotFoundException e) {
			e.printStackTrace(config.getContext().stderr());
		} catch (IOException e) {
			e.printStackTrace(config.getContext().stderr());
		}
		
		config.getContext().stdout().println("Finished checking file contents");
		
		// Check the symbolic link and make sure it refers to the text file
		try {
			FileInputStream fis = new FileInputStream(new File(folder, "link"));
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			int b = fis.read();
			
			while (b != -1) {
				baos.write(b);
				b = fis.read();
			}
			baos.close();
			fis.close();
			
			String contents = new String(baos.toByteArray());
			
			if (!contents.contains("Line1")) {
				config.getContext().stderr().println("ERROR: Symbolic link was not created. Verify that Eclipse Filesystem libraries are available for this platform.");
			}
			
			// Delete the file and make sure we can't read anything from the symbolic link
			new File(folder, "file.txt").delete();
			
			try {
				fis = new FileInputStream(new File(folder, "link"));
				fis.read();
				config.getContext().stderr().println("ERROR: Symbolic link is not a symbolic link. Verify that Eclipse Filesystem libraries are available for this platform.");
			} catch (IOException e) {
				// This is expected.
			}
		} catch (FileNotFoundException e) {
			if (linkCreated) {
				e.printStackTrace(config.getContext().stderr());
				config.getContext().stderr().println("ERROR: Link could not be read because it was not found. Aborting...");
			}
		} catch (IOException e) {
			if (linkCreated) {
				e.printStackTrace(config.getContext().stderr());
				config.getContext().stderr().println("ERROR: Link could not be read. Aborting...");
			}
		}
		
		config.getContext().stdout().println("Finished checking symbolic link");
	}

	private void writeFileContents(PrintStream os, boolean prepend, boolean append) {
		if (prepend) {
			os.println("Line0");
		}
		os.println("Line1");
		os.println("Line2");
		os.println("Line3");
		os.println("Line4");
		os.println("Line5");
		os.println("Line6");
		os.println("Line7");
		os.println("Line8");
		if (append) {
			os.println("Line9");
		}
	}

}
