/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2013. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.rtc.cli.infrastructure.internal.core;

import java.util.Collection;


import java.util.List;

import org.eclipse.core.runtime.IConfigurationElement;

import com.ibm.team.rtc.cli.infrastructure.internal.core.SubcommandLoader;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.Options;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.ConflictingOptionException;

/**
 * Internal configuration of a subcommand.
 */
public interface ISubcommandDefinition {

    /** Thrown when a subcommand cannot be loaded. */
    static final public class LoadSubcommandException extends Exception {
        private static final long serialVersionUID = 5460487952425245289L;

        final private String subcommand;

        public LoadSubcommandException(String subcommand, Throwable reason) {
            super(reason);
            this.subcommand = subcommand;
        }

        public String getSubcommand() {
            return subcommand;
        }
    }

    /** Thrown when options cannot be loaded. */
    static final public class LoadOptionsException extends Exception {
        private static final long serialVersionUID = 5460487952425245289L;

        final private String subcommand;

        public LoadOptionsException(String subcommand, Throwable reason) {
            super(reason);
            this.subcommand = subcommand;
        }

        public String getSubcommand() {
            return subcommand;
        }
    }
    
    /** Get the subcommand configuration element. */
    public IConfigurationElement getConfigurationElement();

    /** Get the attribute of the subcommand. */
    public String getAttr(String attrName);

    /** Get the unique commandline definition of the subcommand. */
    public String getFullSubcommandName();

    /**
     * Get the list of possible aliases for the subcommand.
     * 
     * @return An array of aliases. If there are no aliases, an empty array will
     *         be returned.
     */
    public String[] getAliases();

    /**
     * Get the short help name of the subcommand. It should only contain one
     * line of text, and be no more than 70 characters long.
     */
    public String getShortHelp();

    /**
     * Provide the full help message. This may be any length.
     */
    // public String getLongHelp();

    /**
     * Get the associated subcommand. May be <code>null</code> if the subcommand
     * is a parent.
     */
    public ISubcommand loadSubcommand() throws LoadSubcommandException, IllegalArgumentException;

    /**
     * Provide the options associated with the given subcommand.
     * 
     * @throws ConflictingOptionException
     */
    public Options getOptions() throws LoadOptionsException, ConflictingOptionException;

    /**
     * Return the module that the item is contained in. This is implementation
     * specific, but should contain enough information for the implementing
     * framework to provide version information.
     */
    public String getModuleName();

    /**
     * Return the ID of the parent subcommand.
     * 
     * @return Null if the subcommand doesn't have a parent.
     */
    public String getParentId();

    public ISubcommandDefinition getParent();

    /**
     * Get the list of ancestors of the receiver. The list is empty if there are
     * no ancestors. The list never includes the receiver. The top-level
     * ancestor is first in the list.
     */
    public List<ISubcommandDefinition> getAncestors();

    /**
     * Get the ID of the subcommand.
     * 
     * @return May be null.
     */
    public String getId();

    /**
     * @return <code>true</code> if this subcommand is a parent, ie, it returns
     *         null for {@link #loadSubcommand()}.
     */
    public boolean isParent();
    
    /**
     * @return <code>true</code> if this subcommand has a class specified.
     */
    public boolean isRunnable();

    /**
     * Return a the child subcommand with the given name.
     * 
     * @return A child with the given name or alias.
     * 
     * @throws IllegalArgumentException
     *             if the receiver does not define an ID.
     */
    public ISubcommandDefinition getChild(String childName);

    /**
     * @return a list of children.
     * @throws IllegalArgumentException
     *             if the receiver is not a parent.
     */
    public Collection<ISubcommandDefinition> getChildren();

    /**
     * Used by the help subcommand to determine if the receiver should be shown
     * in the listing of subcommands.
     * 
     * @return <code>true</code> if the receiver should be shown in the listing
     *         of subcommands shown by 'help', <code>false</code> otherwise.
     * 
     * @deprecated For use only during the 0.6 release.
     */
    public boolean showInHelpListing();

    /**
     * @return The SubcommandLoader that loaded this subcommand.
     */
    public SubcommandLoader getLoader();
    
    /**
     * @return Returns whether the specified command is deprecated.
     */
    public boolean isDeprecated();
}
