/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2013. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.rtc.cli.infrastructure.internal.core;


import java.io.PrintStream;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Pattern;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.osgi.util.NLS;

import com.ibm.team.rtc.cli.infrastructure.internal.core.Constants;
import com.ibm.team.rtc.cli.infrastructure.internal.Messages;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IOptionSource;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommand;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommandDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.Options;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.ConflictingOptionException;

class SubcommandDefinition implements ISubcommandDefinition {

    IConfigurationElement ext;

    private final SubcommandLoader loader;

    private String name = null;

    private String shortHelp = null;

    private PrintStream printStream;

    public SubcommandDefinition(SubcommandLoader loader, IConfigurationElement config,
            PrintStream printStream) {
        this.loader = loader;
        this.ext = config;
        this.printStream = printStream;
    }
    
    public IConfigurationElement getConfigurationElement() {
        return ext;
    }

    public String getAttr(String attrName) {
        return getAttr(attrName, false);
    }

    private String getAttr(String attrName, boolean required) {
        String attr = ext.getAttribute(attrName);

        if (required && attr == null) {
            printStream.println(NLS.bind(Messages.SubcommandDefinition_0, attrName));
        }

        return attr;
    }

    public String getFullSubcommandName() {
        if (name == null) {
            name = getAttr(Constants.PT_ATTR_NAME, true);
        }

        return name;
    }

    public String getShortHelp() {
        if (shortHelp == null) {
            shortHelp = getAttr(Constants.PT_ATTR_SHORT_HELP, true);
        }

        return shortHelp;
    }

    public String getLongHelp() {
        return ext.getAttribute(Constants.PT_ATTR_LONG_HELP);
    }

    public String getOptionsClass() {
        return ext.getAttribute(Constants.PT_ATTR_OPTIONS);
    }

    public Options getOptions() throws LoadOptionsException, ConflictingOptionException {
        Options opts = null;

        boolean mustSet = true;

        if (null == getOptionsClass()) {
            opts = new Options(false);
        } else {
            IOptionSource src;

            try {
                src = (IOptionSource) ext.createExecutableExtension(Constants.PT_ATTR_OPTIONS);
            } catch (CoreException e) {
                throw new LoadOptionsException(name, e);
            }

            opts = src.getOptions();
            mustSet = false;
        }

        // Set the long help (if needed)
        String longHelp = getLongHelp();

        assert mustSet == false || (longHelp != null) : "Extension point for subcommand \"" + getId() + "\" does not set options class or long help."; //$NON-NLS-1$ //$NON-NLS-2$

        if (longHelp != null) {
            opts.setLongHelp(longHelp);
        }

        return opts;
    }

    @SuppressWarnings("unchecked")
    public ISubcommand loadSubcommand() throws LoadSubcommandException {
        try {
            if (null == getSubcommandName()) {
                return null;
            }

            return (ISubcommand) ext.createExecutableExtension(Constants.PT_ATTR_CLASS);
        } catch (CoreException e) {
            throw new LoadSubcommandException(name, e);
        }
    }

    public String getSubcommandName() {
        return ext.getAttribute(Constants.PT_ATTR_CLASS);
    }

    public String[] getAliases() {
        String aliasString = getAttr(Constants.PT_ATTR_ALIASES, false);

        if (aliasString == null) {
            return new String[0];
        }

        // Parse the alias list
        return aliasString.split("\\s+"); //$NON-NLS-1$
    }

    public String getModuleName() {
        return ext.getContributor().getName();
    }

    public String getId() {
        return getAttr(Constants.PT_ATTR_ID, false);
    }

    public String getParentId() {
        return getAttr(Constants.PT_ATTR_PARENT, false);
    }

    public ISubcommandDefinition getParent() {
        String parentId = getParentId();

        if (parentId == null) {
            return null;
        }

        Collection<ISubcommandDefinition> all = loader.getAllDefinitions();

        if (parentId != null) {
            for (ISubcommandDefinition defn : all) {
                if (parentId.equals(defn.getId())) {
                    return defn;
                }
            }
        }

        throw new IllegalArgumentException(NLS.bind(Messages.SubcommandDefinition_11, parentId));
    }

    public List<ISubcommandDefinition> getAncestors() {
        LinkedList<ISubcommandDefinition> ancestors = new LinkedList<ISubcommandDefinition>();

        ISubcommandDefinition parent = this.getParent();
        while (parent != null) {
            ancestors.addFirst(parent);

            parent = parent.getParent();
        }

        return ancestors;
    }

    public boolean isParent() {
        Collection<ISubcommandDefinition> children = loader.getChildSubcommands(this);
        return (children.size() > 0);
    }
    
    public boolean isRunnable() {
        return null != ext.getAttribute(Constants.PT_ATTR_CLASS);
    }

    public ISubcommandDefinition getChild(String childName) {
        Collection<ISubcommandDefinition> children = loader.getChildSubcommands(this);

        // Check the children to see if any have this canonical name
        for (ISubcommandDefinition child : children) {
            if (childName.equals(child.getFullSubcommandName())) {
                return child;
            }
        }

        // Check the children for nicknames
        for (ISubcommandDefinition child : children) {
            for (String alias : child.getAliases()) {
                if (childName.equals(alias)) {
                    return child;
                }
            }
        }

        throw new IllegalArgumentException(NLS.bind(Messages.SubcommandDefinition_12, childName));
    }

    public Collection<ISubcommandDefinition> getChildren() {
        return loader.getChildSubcommands(this);
    }

    public boolean showInHelpListing() {
        return ext.getAttribute(Constants.PT_ATTR_HIDE) == null;
    }

    /**
     * Pattern to split scopes with. We make it volatile so that it can be set
     * without locking.
     */
    private volatile static Pattern SCOPE_SPLITTER = null;

    /**
     * Check to see if the config element is in the given group.
     */
    public static boolean inGroup(IConfigurationElement config, String requestedGroup) {
        Collection<String> scopes = getScopes(config);

        if (scopes.contains(requestedGroup)) {
            return true;
        }

        return false;
    }

    /**
     * @return The scopes that the given element is part of.
     */
    public static Collection<String> getScopes(IConfigurationElement config) {
        String actual = config.getAttribute(Constants.PT_ATTR_SCOPE);

        if (actual == null || actual.trim().length() == 0) {
            return Collections.singleton(Constants.PT_ATTR_SCOPE_DEFAULT);
        }

        Pattern splitter = SCOPE_SPLITTER;
        if (SCOPE_SPLITTER == null) {
            splitter = SCOPE_SPLITTER = Pattern.compile("\\s+"); //$NON-NLS-1$
        }

        // Check to see if one of the groups matches our request
        ArrayList<String> scopes = new ArrayList<String>(5);
        for (String group : splitter.split(actual)) {
            scopes.add(group);
        }

        return scopes;
    }

    public SubcommandLoader getLoader() {
        return loader;
    }

    public int getOverrideValue() {
        int overrideValue = 1; // default override value
        String override = getAttr(Constants.PT_ATTR_OVERRIDE, false);
        if (override != null) {
            try {
                overrideValue = Integer.parseInt(override);
            } catch (NumberFormatException e) {
                ; // default value
            }
        }
        
        return overrideValue;
    }
    
    public boolean isDeprecated() {
        return (null != ext.getAttribute(Constants.PT_ATTR_DEPRECATED));
    }
}
