/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.rtc.cli.infrastructure.internal.core;


import static com.ibm.team.rtc.cli.infrastructure.internal.util.SubcommandUtil.printDebugTiming;


import java.io.IOException;
import java.io.PrintStream;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.RegistryFactory;
import org.eclipse.equinox.app.IApplicationContext;
import org.eclipse.osgi.util.NLS;

import com.ibm.team.rtc.cli.infrastructure.internal.Activator;
import com.ibm.team.rtc.cli.infrastructure.internal.core.Constants;
import com.ibm.team.rtc.cli.infrastructure.internal.Messages;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IClientConfiguration;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommand;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommandDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommandDefinition.LoadOptionsException;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommandDefinition.LoadSubcommandException;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.CLIParser;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.HaltingPositionalOptionDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.ICommandLine;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.INamedOptionDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.IOptionKey;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.NamedOptionDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.OptionKey;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.Options;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.ArgumentOverflowException;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.ArgumentUnderflowException;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.ConflictingOptionException;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.MalformedArgumentException;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.MalformedCommandLineException;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.MalformedOptException;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.UnknownOptionException;
import com.ibm.team.rtc.cli.infrastructure.internal.util.OptionGroupUtil;
import com.ibm.team.rtc.cli.infrastructure.internal.util.StatusHelper;
import com.ibm.team.rtc.cli.infrastructure.internal.util.StringUtil;
import com.ibm.team.rtc.cli.infrastructure.internal.util.SubcommandUtil;
import com.ibm.team.rtc.cli.infrastructure.internal.core.CLIClientException;

/**
 * Launcher for CLI. Responsible for determining subcommand and launching it.
 */
public class SubcommandLauncher {
    
    /** The name of the "help" subcommand. */
    private static final String SUBCOMMAND_HELP = "help"; //$NON-NLS-1$

    /**
     * Request help.
     */
    public static NamedOptionDefinition OPT_HELP = new NamedOptionDefinition("h", "help", 0); //$NON-NLS-1$//$NON-NLS-2$
    
    public static NamedOptionDefinition OPT_SYMBOLIC_HELP = new NamedOptionDefinition("?", "?", 0); //$NON-NLS-1$//$NON-NLS-2$

    /** The name of the "version" subcommand. */
    private static final String SUBCOMMAND_VERSION = "version"; //$NON-NLS-1$

    /**
     * Request the version number. 
     */
    public static final  NamedOptionDefinition OPT_VERSION = new NamedOptionDefinition("v", "version", 0); //$NON-NLS-1$ //$NON-NLS-2$

    public static HaltingPositionalOptionDefinition OPT_SUBCOMMAND_NAME =
            new HaltingPositionalOptionDefinition("command", 0, 1); //$NON-NLS-1$

    public static final int UNINITIALIZED_RETURN_CODE = -1;

    /** The group of subcommands we're supposed to load. */
    private final String subcommandGroup;

    private SubcommandLoader loader;
    private static PrintStream err;

    @SuppressWarnings("static-access")
    public SubcommandLauncher(String subcommandGroup, PrintStream err) {
        this.subcommandGroup = subcommandGroup;
        this.err = err;
    }

    /**
     * Run the subcommand specified in the given set of command line arguments.
     * // populate our logs in the same log as long as it's the // same old
     * daemon and same old config path. private static Logger logger = null;
     * private static Object obj = new Object();
     * 
     * 
     * @param env
     *            Context the subcommand should use to interact with the outside
     *            world.
     * 
     * @param appContext
     *            Deprecated. Should be null. Used to set the
     *            {@link IApplicationContext} for legacy apps. See
     *            IClientConfiguration#getApplicationContext().
     * 
     * @param progressMonitor
     *            The progress monitor that should be used to detect
     *            cancellation. May not be null.
     * 
     * @return An integer return value.
     * 
     * @throws Exception
     */
    public int run(IExecutionContext env, IApplicationContext appContext,
            IProgressMonitor progressMonitor) throws Exception {
        if (IClientConfiguration.DEBUG_TIMING) {
            printDebugTiming(env, "Entering Application.start()", env.stderr()); //$NON-NLS-1$
        }

        try {
            doStart(env, appContext, progressMonitor);

            assert returnValue != UNINITIALIZED_RETURN_CODE;
            return returnValue;
        } catch (InternalErrorException e) {
            return returnValue;
        } catch (Exception e) {
            err.println(Messages.Application_0);
            e.printStackTrace(env.stderr());
            throw e;
        } catch (Throwable t) {
            err.println(Messages.Application_0);
            t.printStackTrace(env.stderr());
            throw new RuntimeException(t);
        } finally {
            /**
             * Debugging to get around launcher problems. We now rely on the
             * launcher to show our error messages. If we're running without a
             * launcher, we have to explicitly display our error output.
             */
            if (IClientConfiguration.DEBUG_OUTPUT) { //$NON-NLS-1$ //$NON-NLS-2$
                String prop = System.getProperty("eclipse.exitdata"); //$NON-NLS-1$
                if (prop != null) {
                    err.println(prop);
                }
            }
        }
    }

    private int returnValue = -1;
    enum CommandName { MISSING, UNKNOWN, VALID }; 

    public void doStart(IExecutionContext ctx, IApplicationContext legacyContext,
            IProgressMonitor progressMonitor) throws InternalErrorException, CLIClientException {

        // Parse the command-line
        ICommandLine args = null;
        String cmdName = null;

        loader = SubcommandLoader.getScope(subcommandGroup, ctx.stderr());

        try {
            args = parseCommandLine(ctx);

            cmdName = args.getOption(OPT_SUBCOMMAND_NAME, null);
        } catch (UnknownOptionException e) {
            throw complain(
                    ctx,
                    Constants.STATUS_ARGUMENT_SYNTAX_ERROR,
                    Messages.Application_36,
                    NLS.bind(Messages.Application_CONCATENATE,
                            NLS.bind(Messages.Application_4, e.getOption()),
                            NLS.bind(Messages.Application_1, ctx.getAppName())));
        } catch (ArgumentUnderflowException e) {
            List<?> missing = e.getMissingArguments();
            String msg = null;
            if (missing.size() > 0) {
                msg = NLS.bind(Messages.Application_58, missing.iterator().next());
            } else {
                msg = Messages.Application_28;
            }
            
            throw complain(ctx, Constants.STATUS_ARGUMENT_SYNTAX_ERROR, Messages.Application_28,
                    msg);
        } catch (MalformedOptException e) {
            throw complain(ctx, Constants.STATUS_ARGUMENT_SYNTAX_ERROR,
                    Messages.Application_29, NLS.bind(Messages.Application_CONCATENATE,
                            NLS.bind(Messages.Application_6, e.getOption()),
                            NLS.bind(Messages.Application_1, ctx.getAppName())));
        } catch (MalformedArgumentException e) {
            throw complain(ctx, Constants.STATUS_ARGUMENT_SYNTAX_ERROR,
                    Messages.Application_MALFORMED_ARGUMENT,
                    NLS.bind(Messages.Application_CONCATENATE,
                            NLS.bind(Messages.Application_INVALID_SELECTOR, e.getArgument()),
                            NLS.bind(Messages.Application_1, ctx.getAppName())));
        } catch (MalformedCommandLineException e) {
            // This is a bad command-line, but we should have caught it.
            // Treat as an internal error.
            e.printStackTrace(ctx.stderr());
            throw complain(ctx, Constants.STATUS_INTERNAL_ERROR, Messages.Application_30,
                    e.getLocalizedMessage());
        } catch (ConflictingOptionException e) {
            e.printStackTrace(ctx.stderr());
            throw complain(ctx, Constants.STATUS_INTERNAL_ERROR, Messages.Application_31,
                    e.getLocalizedMessage());
        } catch (IOException e) {
            e.printStackTrace(ctx.stderr());
            throw complain(ctx, Constants.STATUS_INTERNAL_ERROR, Messages.Application_57,
                    e.getLocalizedMessage());
        }

        // Transmute options to commands (if appropriate)
        CommandName commandName = CommandName.VALID;
        if (cmdName == null) {
            if (args.hasOption(OPT_VERSION)) {
                cmdName = SUBCOMMAND_VERSION;            
            } else if (args.hasOption(OPT_HELP) || args.hasOption(OPT_SYMBOLIC_HELP)) {
                cmdName = SUBCOMMAND_HELP;
            } else {
                // show help
                cmdName = SUBCOMMAND_HELP;
                commandName = CommandName.MISSING;
            }
        }

        // Get the command
        ISubcommandDefinition defn = loader.loadToplevel(cmdName);
        if (defn == null) {
            // show help
            defn = loader.loadToplevel(SUBCOMMAND_HELP);
            commandName = CommandName.UNKNOWN;
        }

        LinkedList<ISubcommandDefinition> parents = new LinkedList<ISubcommandDefinition>();
        ICommandLine subargs = null;
        if (commandName == CommandName.VALID) {
            int index = 0;
            List<String> unconsumed = args.getUnconsumedArgs();
            while (defn.isParent()) {
                if  (index == unconsumed.size()) {
                    if (!defn.isRunnable()) {
                        CharSequence subcommands = SubcommandUtil.getExecutionString(defn);
                        throw complain(
                                ctx,
                                Constants.STATUS_ARGUMENT_SYNTAX_ERROR,
                                NLS.bind(Messages.Application_49, subcommands),
                                NLS.bind(Messages.Application_CONCATENATE, Messages.Application_50,
                                        NLS.bind(Messages.Application_2, ctx.getAppName(), subcommands)));
                    }
                    break;
                }
                
                parents.add(defn);
                String childName = unconsumed.get(index);

                try {
                    defn = defn.getChild(childName);
                    index++;
                } catch (IllegalArgumentException e) {
                    ; // maybe the unconsumed arg wasn't meant to be a child command
                    defn = parents.removeLast();
                    if (!defn.isRunnable()) {
                        throw complain(
                                ctx,
                                Constants.STATUS_ARGUMENT_SYNTAX_ERROR,
                                Messages.Application_52,
                                NLS.bind(
                                        Messages.Application_CONCATENATE,
                                        NLS.bind(Messages.Application_51, childName),
                                        NLS.bind(Messages.Application_2, ctx.getAppName(),
                                                SubcommandUtil.getExecutionString(defn))));
                    }
                    break;
                }
            }
            
            // remove from the args that were consumed as commands
            for (int count = index; count > 0; count--) {
                unconsumed.remove(0);
            }            
            
            // Parse the subcommand command-line
            if (args != null) {
                // args == null on error.
                try {
                    subargs = parseSubCommandLine(defn, args.getUnconsumedArgs());
                } catch (LoadOptionsException e) {
                    throw complain(ctx, Constants.STATUS_CONFIG_FAILURE,
                            NLS.bind(Messages.Application_35, SubcommandUtil.getExecutionString(defn)),
                            e.getCause().getLocalizedMessage());
                } catch (ArgumentOverflowException e) {
                    throw complain(
                            ctx,
                            defn,
                            Constants.STATUS_ARGUMENT_SYNTAX_ERROR,
                            Messages.Application_36,
                            NLS.bind(
                                    Messages.Application_CONCATENATE,
                                    NLS.bind(Messages.Application_11,
                                            SubcommandUtil.getExecutionString(defn),
                                            e.getExtraArgument()),
                                    NLS.bind(Messages.Application_2, ctx.getAppName(),
                                            SubcommandUtil.getExecutionString(defn))));
                } catch (ArgumentUnderflowException e) {
                    throw complain(
                            ctx,
                            defn,
                            Constants.STATUS_ARGUMENT_SYNTAX_ERROR,
                            Messages.Application_36,
                            NLS.bind(
                                    Messages.Application_CONCATENATE,
                                    NLS.bind(Messages.Application_12,
                                            SubcommandUtil.getExecutionString(defn),
                                            StringUtil.listToString(e.getMissingArguments())),
                                    NLS.bind(Messages.Application_2, ctx.getAppName(),
                                            SubcommandUtil.getExecutionString(defn))));
                } catch (UnknownOptionException e) {
                	StringBuffer parentCommandName = new StringBuffer();
                	boolean applicableToParent = optionApplicableToParentCommand(e.getOption(), ctx, parentCommandName); 
                    String message1 = NLS.bind(Messages.Application_13,
                            SubcommandUtil.getExecutionString(defn), e.getOption());
                    String message2= NLS.bind(Messages.Application_2, ctx.getAppName(),
                            SubcommandUtil.getExecutionString(defn));
                    String message3 = NLS.bind(Messages.Application_60, e.getOption(), parentCommandName);
                	throw complain(ctx, defn, Constants.STATUS_ARGUMENT_SYNTAX_ERROR,
                            Messages.Application_36, applicableToParent ?
                           		(message1 + Messages.HelpCmd_24 + message3 + Messages.HelpCmd_24 + message2):
                                   NLS.bind(Messages.Application_CONCATENATE,message1, message2));
                } catch (MalformedOptException e) {
                	String message = NLS.bind(Messages.Application_CONCATENATE,
                            NLS.bind(Messages.Application_14, e.getOption()), e.getProblem()); 
                    throw complain(
                            ctx,
                            defn,
                            Constants.STATUS_ARGUMENT_SYNTAX_ERROR,
                            Messages.Application_36,
                            NLS.bind(Messages.Application_CONCATENATE, message, 
                            		NLS.bind(Messages.Application_2, ctx.getAppName(),
                                    SubcommandUtil.getExecutionString(defn))));
                } catch (MalformedCommandLineException e) {
                    throw complain(
                            ctx,
                            defn,
                            Constants.STATUS_ARGUMENT_SYNTAX_ERROR,
                            Messages.Application_36,
                            NLS.bind(
                                    Messages.Application_CONCATENATE,
                                    e.getLocalizedMessage(),
                                    NLS.bind(Messages.Application_2, ctx.getAppName(),
                                            SubcommandUtil.getExecutionString(defn))));
                } catch (ConflictingOptionException e) {
                    e.printStackTrace(ctx.stderr());
                    throw complain(ctx, Constants.STATUS_INTERNAL_ERROR,
                            NLS.bind(Messages.Application_41, SubcommandUtil.getExecutionString(defn)),
                            e.getLocalizedMessage());
                }
            }
        }
        
        // Run the subcommand
        run(ctx, legacyContext, defn, parents, args, subargs);

        // If the help was shown on an error, display the appropriate error
        if (commandName == commandName.MISSING) {
            throw complain(ctx, Constants.STATUS_ARGUMENT_SYNTAX_ERROR,
                    Messages.Application_32, NLS.bind(Messages.Application_7,
                            OPT_VERSION.getName(), OPT_HELP.getName()));
        } else if (commandName == commandName.UNKNOWN) {
            throw complain(ctx, Constants.STATUS_UNKNOWN_SUBCOMMAND, Messages.Application_33,
                    NLS.bind(Messages.Application_8, cmdName));
        }
    }
    
    private InternalErrorException complain(IExecutionContext ctx, ISubcommandDefinition defn, int errorCode, String errorTitle,
            String errorDescription) {
        returnValue = errorCode;

        String formattedHelp = null;
        try {
        	if (defn != null && defn.getOptions().getSupportsNewHelp()) {
        		formattedHelp = SubcommandUtil.getFormattedShortUsageHelp(ctx, defn);
        	}
        }
        catch (LoadOptionsException e) { 
        	// Ignore for now, not a major problem
        	// but log it
            e.printStackTrace(ctx.stderr());
        	
        }
        catch (ConflictingOptionException e) {
        	// Ignore for now, not a major problem
        	// but log it
            e.printStackTrace(ctx.stderr());
        }
        
        	
        if (errorDescription != null) {
        	if (formattedHelp != null) {
        		errorDescription = NLS.bind(Messages.Application_JOIN, errorDescription, formattedHelp);
        	}
          
    		if (ctx.isLocal()) {
    			// This is required to prevent the Eclipse from printing a verbose
                // Java exited with an error code message when no terminating error message is provided
                ctx.setTerminatingError("", "");  //$NON-NLS-1$
    		    ctx.stderr().println(NLS.bind(Messages.ERROR_TITLE1, errorTitle));
                ctx.stderr().println(errorDescription);
    		}
    		else {
    		    ctx.setTerminatingError(errorTitle, errorDescription);
    		}            
        }

        return new InternalErrorException();
    }

    private InternalErrorException complain(IExecutionContext env, int errorCode, String errorTitle,
            String errorDescription) {
       return complain(env, null, errorCode, errorTitle, errorDescription);
    }
    
    private boolean optionApplicableToParentCommand(String optString, IExecutionContext ctx, StringBuffer parentCommandName)
    {
    	boolean applicableToParentCommand = false ; //$NON-NLS$1
        String PREFIX_LONGOPT = CLIParser.PREFIX_LONGOPT;
        String PREFIX_SHORTOPT = CLIParser.PREFIX_SHORTOPT;
        String optName = "";
        if (optString.startsWith(PREFIX_LONGOPT ) && optString.length() > PREFIX_LONGOPT.length()) { 
            optName = optString.substring(PREFIX_LONGOPT.length());
        }
        else if (optString.startsWith(PREFIX_SHORTOPT) && optString.length() > PREFIX_SHORTOPT.length()) {
            optName = optString.substring(PREFIX_SHORTOPT.length());
        }
        else {
            optName = optString;
        }
        try {
	        Options appOptions = getAppOptions();
            for (INamedOptionDefinition namedOption : appOptions.getNamedOptions()) {
                if(optName.equals(namedOption.getShortOpt()) || 
                      optName.equals(namedOption.getLongOpt())) {
                    parentCommandName.append(ctx.getAppName());
                	applicableToParentCommand = true;
                    break;
                }
            }
        }
        catch(ConflictingOptionException exp)
        {
        	// Ignore this exception
        }
        return applicableToParentCommand;
    }

    private InternalErrorException returnValue(IExecutionContext env, int errorCode) {
        returnValue = errorCode;
        // This is required to prevent the Eclipse Launcher from printing a verbose
        // Java exited with an error code message when no terminating error message is provided
        env.setTerminatingError("", "");  //$NON-NLS-1$//$NON-NLS-2$

        return new InternalErrorException();
    }

    /**
     * An exception indicating that the error message has already been shown,
     * and that the launcher should quietly return without showing a further
     * error message.
     */
    private static class InternalErrorException extends Exception {
        private static final long serialVersionUID = 7057110785038430374L;
    }

    /** Not used. */
    public void stop() {
    }
    
    // TODO: Needed for backwards compatibility for legacy scm cli options. Clean up in the next code refactoring.
    public static final IOptionKey OPT_DISALLOW_INTERACTION = new OptionKey("non-interactive"); //$NON-NLS-1$
    public static final IOptionKey OPT_CONFIG_LOCATION = new OptionKey("config"); //$NON-NLS-1$
    public static final IOptionKey OPT_DRY_RUN = new OptionKey("dry-run"); //$NON-NLS-1$
    public static final IOptionKey OPT_NO_MASK = new OptionKey("no-mask"); //$NON-NLS-1$
    public static final IOptionKey OPT_SHOW_UUID = new OptionKey("show-uuid"); //$NON-NLS-1$
    public static final IOptionKey OPT_SHOW_ALIAS = new OptionKey("show-alias"); //$NON-NLS-1$
    private static String[] ALWAYS_VALUES = new String[] {"y", "yes"};  //$NON-NLS-1$ //$NON-NLS-2$
    private static String[] NEVER_VALUES = new String[] {"n", "no"};  //$NON-NLS-1$ //$NON-NLS-2$

    /**
     * Parse the command-line for the app. This does not handle the
     * subcommand-specific arguments, just those intended for the app.
     * 
     * @param context
     * @throws IOException
     * @throws CLIFileSystemClientException
     */
    private ICommandLine parseCommandLine(IExecutionContext context)
            throws MalformedCommandLineException, ConflictingOptionException, IOException,
            CLIClientException {
        // Create the command-line
    	Options appOptions = getAppOptions();
        // Parse
        String[] args = context.arguments();

        Map<String, String> env = context.environment();

        // This code needs to be removed once the (bug 51096 - Windows Build
        // server to run SCM tests) is fixed.
        String forceGlob = env.get(Constants.JAZZSCM_FORCE_GLOBBING);

        boolean forceGlobFlag = "1".equals(forceGlob); //$NON-NLS-1$

        // expand wild card chars only for windows
        if (System.getProperty("os.name").toLowerCase().contains("windows") //$NON-NLS-1$ //$NON-NLS-2$
                || forceGlobFlag == true) {
            String globIgnore = env.get("GLOBIGNORE"); //$NON-NLS-1$
            String[] ignoreStrs = null;
            if (globIgnore != null) {
                ignoreStrs = globIgnore.split(";"); //$NON-NLS-1$
            }
            
            List<String> argList = CLIParser.expandWildCardArgs(args, ignoreStrs, context);
            args = (String[]) argList.toArray(new String[argList.size()]);
        }

        CLIParser parser = new CLIParser(appOptions, args);

        return parser.parse();
    }

    /**
     * Parse the command-line for the given subcommand.
     * 
     * @param args
     *            The args passed to the subcommand
     * 
     * @return
     * @throws MalformedCommandLineException
     *             , LoadOptionsException, ConflictingOptionException
     * @throws CLIFileSystemClientException
     */
    private ICommandLine parseSubCommandLine(ISubcommandDefinition defn, List<String> subargs)
            throws MalformedCommandLineException, LoadOptionsException, ConflictingOptionException,
            CLIClientException {
        Options options = defn.getOptions();

        return new CLIParser(options, subargs).parse();
    }

    private void run(IExecutionContext ctx, IApplicationContext legacyContext,
            ISubcommandDefinition defn, LinkedList<ISubcommandDefinition> parents,
            ICommandLine args, ICommandLine subargs)
            throws InternalErrorException, CLIClientException {
        // Generate the configuration
        IClientConfiguration config;
        IConfigurationElement[] configs = RegistryFactory.getRegistry()
                .getConfigurationElementsFor(Activator.PLUGIN_ID, Constants.PT_CLIENTCONFIG);
        // use the default client configuration if no configuration was specified
        if (configs.length == 0) {
            config = new ClientConfiguration();
        } else if (configs.length == 1) {
            try {
                config = (IClientConfiguration) configs[0].createExecutableExtension(
                        Constants.PT_ATTR_CLASS);
            } catch (CoreException e) {
                throw complain(ctx, Constants.STATUS_CONFIG_FAILURE, "Client Configuration",
                        "Cannot load client configuration");
            }
        } else {
            throw complain(ctx, Constants.STATUS_CONFIG_FAILURE, "Client Configuration",
                    "Found more than one client configuration");
        }
        
        try {
            // Initialize the configuration
            config.initialize(loader, ctx, legacyContext, defn, parents, args, subargs);
            
            // Load the subcommand
            ISubcommand subcmd;
            try {
                subcmd = defn.loadSubcommand();
            } catch (LoadSubcommandException e) {
                throw complain(ctx, Constants.STATUS_CONFIG_FAILURE,
                        NLS.bind(Messages.Application_43, SubcommandUtil.getExecutionString(defn)),
                        e.getLocalizedMessage());
            }

            // Timing debug output pre-subcommand
            if (IClientConfiguration.DEBUG_TIMING) {
                printDebugTiming(ctx, NLS.bind(Messages.Application_16, ctx.getAppName()), ctx.stderr());
            }

            // Run the subcommand
            returnValue = subcmd.run(config);

            // Timing debug. Show regardless of result.
            if (IClientConfiguration.DEBUG_TIMING) {
                printDebugTiming(ctx, NLS.bind(Messages.Application_17, ctx.getAppName()), ctx.stderr());
            }
        } catch (CLIClientException e) {
            IStatus status = e.getStatus();
            if (status == null) {
                throw e;
            }

            if (Constants.showStackTraceFor(status.getCode())) {
                StatusHelper.showStatusException(status, config.getContext().stderr());
            }
            
            if (status.getSeverity() == IStatus.INFO) {
                throw returnValue(ctx, status.getCode());
            } else {
                throw complain(ctx, status.getCode(), NLS.bind(Messages.Application_44,
                        SubcommandUtil.getExecutionString(defn)), status.getMessage());
            }
        } finally {
            config.deInitialize();
        }
    }
    
    private Options getAppOptions() throws ConflictingOptionException
    {
    	Options opt = new Options(true); 
    	opt.addOption(OPT_SUBCOMMAND_NAME, Messages.Application_21);
    	opt.addOption(OPT_HELP, Messages.Application_22);
        OPT_SYMBOLIC_HELP.hideOption();
    	opt.addOption(OPT_SYMBOLIC_HELP, Messages.Application_22);


        // TODO: Needed for backwards compatibility of legacy scm cli options. Clean up in the next code refactoring.
    	opt.addOption(OPT_DISALLOW_INTERACTION, null, "non-interactive", Messages.Application_19, 0); //$NON-NLS-1$
    	opt.addOption(OPT_CONFIG_LOCATION, null, "config", Messages.Application_20, 1); //$NON-NLS-1$
    	opt.addOption(OPT_NO_MASK, null, "no-mask", Messages.Application_NO_MASK_HELP, 0); //$NON-NLS-1$

    	Object[] yesNoList = new Object[] {ALWAYS_VALUES[0], ALWAYS_VALUES[1], NEVER_VALUES[0], NEVER_VALUES[1]};
    	opt.addOption(OPT_SHOW_UUID, "u", "show-uuid", NLS.bind(Messages.Application_55, yesNoList), 1); //$NON-NLS-1$ //$NON-NLS-2$ 
    	opt.addOption(OPT_SHOW_ALIAS, "a", "show-alias", NLS.bind(Messages.Application_56, yesNoList), 1); //$NON-NLS-1$ //$NON-NLS-2$
        
        opt.addOption(OPT_VERSION, Messages.CommonOptions_11);
        
        return opt;
    }
}
