/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2013. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.rtc.cli.infrastructure.internal.core;

import static com.ibm.team.rtc.cli.infrastructure.internal.core.SubcommandDefinition.inGroup;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;

import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.RegistryFactory;
import org.eclipse.osgi.util.NLS;

import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommandDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.Activator;
import com.ibm.team.rtc.cli.infrastructure.internal.core.Constants;
import com.ibm.team.rtc.cli.infrastructure.internal.Messages;

/**
 * Responsible for loading subcommands.
 */
public class SubcommandLoader {

    private static HashMap<String, SubcommandLoader> loaders;
    private static PrintStream printStream;

    public static synchronized SubcommandLoader getScope(String groupName, PrintStream ps) {
        if (groupName == null) {
            groupName = ""; //$NON-NLS-1$
        }

        printStream = ps;

        if (loaders == null) {
            loaders = new HashMap<String, SubcommandLoader>();
        }

        SubcommandLoader loader = loaders.get(groupName);
        if (loader == null) {
            loader = new SubcommandLoader(groupName);
            loaders.put(groupName, loader);
        }

        return loader;
    }

    /**
     * @return A listing of all scope names
     */
    public static Collection<String> getAllScopeNames() {
        SortedSet<String> allScopes = new TreeSet<String>();

        IConfigurationElement[] configs = RegistryFactory.getRegistry()
                .getConfigurationElementsFor(Activator.PLUGIN_ID, Constants.PT_SUBCOMMAND);

        for (IConfigurationElement config : configs) {
            Collection<String> scopes = SubcommandDefinition.getScopes(config);

            allScopes.addAll(scopes);
        }

        return allScopes;
    }

    List<ISubcommandDefinition> subcommands;

    /** The grouping that this loader provides. */
    private final String group;

    public SubcommandLoader(String group) {
        this.group = group;
    }

    /** Grab the subcommands out of the definition. */
    private void loadSubcommands() {
        if (subcommands != null) {
            return;
        }

        IConfigurationElement[] configs = RegistryFactory.getRegistry()
                .getConfigurationElementsFor(Activator.PLUGIN_ID, Constants.PT_SUBCOMMAND);
        
        Map<String, SubcommandDefinition> idToDefn = new HashMap<String, SubcommandDefinition>();
        Map<String, List<String>> nameToId = new HashMap<String, List<String>>();
        List<String> defnsToIgnore = new ArrayList<String>();

        for (IConfigurationElement config : configs) {
            if (!inGroup(config, group)) {
                continue;
            }

            // get the definition
            SubcommandDefinition defn = new SubcommandDefinition(this, config, printStream);
            
            // check if we already have a command definition with the same command name
            List<String> defnIds = nameToId.get(defn.getFullSubcommandName());
            
            // add to list if it is a new definition or if the parent of new definition and the
            // matched definitions is not common
            SubcommandDefinition matchedDefn = null;
            if (defnIds == null ||
                    (matchedDefn = matchCommonParentDefinition(defn, defnIds, idToDefn)) == null) {
                idToDefn.put(defn.getId(), defn);
                if (defnIds == null) {
                    defnIds = new ArrayList<String>();
                    nameToId.put(defn.getFullSubcommandName(), defnIds);
                }
                defnIds.add(defn.getId());
                continue;
            }

            // The new definition has a common parent with an existing definition,
            // we will take the definition that has the largest override value.
            // If the override values are the same then error out.
            if (defn.getOverrideValue() > matchedDefn.getOverrideValue()) {
                idToDefn.remove(matchedDefn.getId());
                idToDefn.put(defn.getId(), defn);
                defnIds.remove(matchedDefn.getId());
                defnIds.add(defn.getId());
                
                // If the definition was a parent, we should remove all its children.
                // For now add it to the ignore list because we do not know the order in which
                // the definitions are read.
                if (matchedDefn.getParentId() == null) {
                    defnsToIgnore.add(matchedDefn.getId());
                }
            } else if (defn.getOverrideValue() == matchedDefn.getOverrideValue()) {
                // TODO: Change this to a cli exception. For now keeping it this way
                // as this changes IClientConfiguration signature and it affects downstream clients.
                throw new IllegalArgumentException(NLS.bind(Messages.SubcommandLoader_MultipleExtensions,
                        new String[] {defn.getFullSubcommandName(), defn.getId(), matchedDefn.getId()}));
            } else {
                // If the definition was a parent, we should remove all its children.
                // For now add it to the ignore list because we do not know the order in which
                // the definitions are read.
                if (defn.getParentId() == null) {
                    defnsToIgnore.add(defn.getId());
                }
            }
        }

        subcommands = new ArrayList<ISubcommandDefinition>(idToDefn.values().size());
        subcommands.addAll(idToDefn.values());
        
        // If we have ignored parent definitions, remove the child items for those definitions
        Iterator<ISubcommandDefinition> it = subcommands.iterator();
        while (it.hasNext()) {
            ISubcommandDefinition defn = it.next();
            if (defn.getParentId() == null) {
                continue;
            }
            
            for (String ignoreDefnId : defnsToIgnore) {
                if (defn.getParentId().equals(ignoreDefnId)) {
                    it.remove();
                    break;
                }
            }
        }
    }

    private SubcommandDefinition matchCommonParentDefinition(SubcommandDefinition defn,
            List<String> defnIds, Map<String, SubcommandDefinition> idToDefn) {
        for (String defnId : defnIds) {
            SubcommandDefinition defn2 = idToDefn.get(defnId);
            if ((defn.getParentId() == null && defn2.getParentId() == null) ||                
                    (defn.getParentId() != null && defn2.getParentId() != null &&
                    defn.getParentId().equals(defn2.getParentId()))) {
                return defn2;
            }
        }
        
        return null;
    }

    /**
     * Ensure that the passed definition is sane.
     */
    public boolean validate() {
        loadSubcommands();

        // Make sure we have an ID
        HashSet<String> ids = new HashSet<String>();
        for (ISubcommandDefinition d : subcommands) {
            SubcommandDefinition defn = (SubcommandDefinition) d;

            // Ensure we have a name
            if (defn.getFullSubcommandName() == null) {
                throw new AssertionError(NLS.bind(Messages.SubcommandLoader_0, defn.getId()));
            }

            // Ensure we have an id.
            String id = defn.getId();
            if (id == null) {
                throw new AssertionError(NLS.bind(Messages.SubcommandLoader_1,
                        defn.getFullSubcommandName()));
            }

            // Ensure we have a unique ID
            if (ids.contains(id)) {
                throw new AssertionError(NLS.bind(Messages.SubcommandLoader_2, id));
            }
            ids.add(id);

            // Ensure we have options or a longHelp
            if (defn.getOptionsClass() == null) {
                if (defn.getLongHelp() == null) {
                    throw new AssertionError(NLS.bind(Messages.SubcommandLoader_3,
                            defn.getFullSubcommandName(), id));
                }
            }

            // Look for parent cycles.
            if (defn.getParentId() != null) {
                HashSet<String> ancestors = new HashSet<String>();

                ISubcommandDefinition parent = defn.getParent();
                do {
                    if (ancestors.contains(parent.getId())) {
                        throw new AssertionError(NLS.bind(Messages.SubcommandLoader_4,
                                defn.getId(), parent.getId()));
                    }

                    ancestors.add(parent.getId());
                    parent = parent.getParent();
                } while (parent != null);
            }
        }

        return true;
    }

    /**
     * Find the top level (ie, unparented) subcommand for the given name.
     * 
     * @return <code>null</code> if there is no such command.
     */
    public ISubcommandDefinition loadToplevel(String name) {
        loadSubcommands();

        // Check the canonical name
        for (ISubcommandDefinition subcommand : subcommands) {

            if (name.equals(subcommand.getFullSubcommandName()) && subcommand.getParent() == null) {
                return subcommand;
            }
        }

        // Check the aliases
        for (ISubcommandDefinition subcommand : subcommands) {

            for (String alias : subcommand.getAliases()) {
                if (name.equals(alias) && subcommand.getParent() == null) {
                    return subcommand;
                }
            }
        }

        return null;
    }

    /**
     * Return the known definitions.
     */
    public Collection<ISubcommandDefinition> getAllDefinitions() {
        loadSubcommands();

        return subcommands;
    }

    /**
     * Return all child subcommands for the given definition.
     * 
     * @throws IllegalArgumentException
     *             If the given subcommand doesn't have an id.
     */
    public Collection<ISubcommandDefinition> getChildSubcommands(ISubcommandDefinition defn) {
        String id = defn.getId();
        if (id == null) {
            throw new IllegalArgumentException(Messages.SubcommandLoader_5);
        }

        loadSubcommands();

        // Walk the list of subcommands and gather those that declare a parent
        // id
        ArrayList<ISubcommandDefinition> children = new ArrayList<ISubcommandDefinition>(
                subcommands.size());

        // Check the canonical name
        for (ISubcommandDefinition child : subcommands) {
            String parentId = child.getParentId();
            if (parentId == null) {
                continue;
            }

            if (id.equals(parentId)) {
                children.add(child);
            }
        }

        return children;
    }
}
