/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2005, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.rtc.cli.infrastructure.internal.parser;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.osgi.util.NLS;

import com.ibm.team.rtc.cli.infrastructure.internal.core.IExecutionContext;
import com.ibm.team.rtc.cli.infrastructure.internal.Messages;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.CommandLine.OptValue;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.ArgumentOverflowException;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.ArgumentUnderflowException;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.MalformedCommandLineException;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.MalformedOptException;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.UnknownOptionException;
import com.ibm.team.rtc.cli.infrastructure.internal.util.StringMatcher;

public class CLIParser {

	public static final String PREFIX_LONGOPT = "--"; //$NON-NLS-1$
    public static final String PREFIX_SHORTOPT = "-"; //$NON-NLS-1$

    private final Options options;
    private final String[] commandLine;

    /**
     * Causes the parser to stop parsing after the current argument has been
     * completed.
     */
    public void halt() {

    }

    public CLIParser(Options opts, String[] commandLine) {
        this.options = opts;
        this.commandLine = commandLine;
    }

    public CLIParser(Options opts, List<String> list) {
        this.options = opts;
        this.commandLine = list.toArray(new String[0]);
    }

    /**
     * Check and expand wildcards '*','?' in the command-line.
     * 
     * @param args
     * @param ignoreStrs
     * @param context
     * @return
     * @throws IOException
     */
    public static List<String> expandWildCardArgs(String args[], String[] ignoreStrs,
            IExecutionContext context) throws IOException {

        ArrayList<String> finalArgList = new ArrayList<String>();
        ArrayList<String> expandedArgList = new ArrayList<String>();
        ArrayList<StringMatcher> ignoreList = new ArrayList<StringMatcher>();

        if (ignoreStrs != null) {

            for (String ignoreStr : ignoreStrs) {
                ignoreList.add(new StringMatcher(ignoreStr, false, false));
            }
        }

        String wd = context.getCurrentWorkingDirectory();

        // a check to see if path specified in -d is valid
        IPath currDir1 = new Path(wd);
        File dir = currDir1.toFile();
        if (!dir.exists()) {
            finalArgList.addAll(0, Arrays.asList(args));
            return finalArgList;
        }

        Map<String, String> env = context.environment();

        for (String arg : args) {

            IPath pathStr = new Path(arg);

            // for paths such as C:IBM*\file.txt, get the CWD for the
            // corresponding drive
            // and append the path IBM*\file.txt to it
            if (pathStr.getDevice() != null && !pathStr.isAbsolute()) {
                String deviceId = pathStr.getDevice();
                // device id is a char followed by ':' . eg- C:
                // if there's a device id with more than one char, then it's not
                // a device in windows. add and continue. see bug 104643
                if (deviceId.length() > 2) {
                    finalArgList.add(arg);
                    continue;
                }
                String key = "=" + deviceId; //$NON-NLS-1$
                pathStr = new Path(env.get(key) + File.separator + arg.substring(2, arg.length()));
                arg = pathStr.toOSString();
            }

            // if no wildcard pattern in the argument, then continue
            if (!arg.contains("*") && !arg.contains("?")) { //$NON-NLS-1$ //$NON-NLS-2$
                finalArgList.add(arg);
                continue;
            } else if (checkEscaped(arg, finalArgList)) {
            	// check if '*' or '? are escaped
           		continue;
            }

            int idx = 0;
            IPath currDir = currDir1;
            String[] paths = pathStr.segments();

            // If the path is an absolute path, the current directory will be
            // set to the directory prior to the wildcard sub path
            // eg: C:\somepath\file*, then make C:\somepath as the CWD
            if (pathStr.isAbsolute()) {
                IPath cwd = null;
                // check to see if the user path contains a device id or
                if (pathStr.getDevice() != null) {
                    cwd = new Path(pathStr.getDevice() + File.separator);
                } else {
                    cwd = new Path(File.separator);
                }

                // for the pattern such as C:\somepath\dir*\file*, move till
                // the segment containing C:\somepath and make it as CWD
                for (String str : paths) {
                    if (!str.contains("*") && !str.contains("?")) { //$NON-NLS-1$ //$NON-NLS-2$
                        cwd = cwd.append(str);
                    } else {
                        break;
                    }
                    ++idx;
                }

                currDir = cwd;
            }

            // if path contains a trailing separator eg: C:\somepath\dir*\d\ ,
            // in
            // this case the exact match should be a directory at the end.
            // if C:\somepath\dir2\d , a file, matches but is rejected.

            boolean hasTrailingSeparator = pathStr.hasTrailingSeparator() ? true : false;

            expandedArgList.add(currDir.toString());
            for (int i = idx; i < pathStr.segmentCount(); ++i) {

                String pathSgmt = pathStr.segment(i);

                // try to expand each segment within the path. expandedArgList
                // will contain newly expanded segments.
                // finalArgList stores all the resulting candidates at the end.
                // ignoreList contains the ignoreStrs.
                if (!expandPath(currDir, pathSgmt, expandedArgList, finalArgList, ignoreList,
                        i == idx, i == paths.length - 1, hasTrailingSeparator)) {

                    // if we are unable to expand the path further. clear the
                    // list.
                    expandedArgList.clear();

                    // add arg to the final list and continue with next arg.
                    finalArgList.add(arg);

                    break;
                }
            }
        }
        return finalArgList;
    }

    private static boolean checkEscaped(String arg,
			ArrayList<String> finalArgList) {
    	StringBuilder builder = new StringBuilder();
    	boolean escaped = false;
    	boolean handledEscaped = false;
    	for (char c : arg.toCharArray()) {
    		switch(c) {
    		case '\\' :
    			escaped = true;
    			break;
    		case '*' :
    			if (escaped) {
    				builder.append('*');
    				escaped = false;
    				handledEscaped = true;
    			}
    			break;
    		case '?' :
    			if (escaped) {
    				builder.append('?');
    				escaped = false;
    				handledEscaped = true;
    			}
    			break;
    		default:
    			if (escaped) {
    				builder.append('\\');
    				escaped = false;
    			}
    			builder.append(c);
    			break;
    		}
    	}
    	if (handledEscaped) {
    		finalArgList.add(builder.toString());
    	}
		return handledEscaped;
	}

	/**
     * expands all the segments of a path
     * 
     * @param currDir
     *            current directory
     * @param pattern
     * @param pathList
     *            list of parent paths which are expanded
     * @param argList
     *            final list of expanded file paths which do not contain any "*"
     *            or "?"
     * @param ignoreStrs
     *            list of ignore strings
     * @param isFirstSgmt
     *            true when the call to this func is made with first segment of
     *            the path str
     * @param isLastSgmt
     *            true when the call to this func is made with last segment of
     *            the path str
     * @param isNotAPattern
     *            true when the current segment does not contain a wild card
     *            char
     * @param hasTrailingSeparator
     *            true when the path contains \ at the end of the path
     */
    private static boolean expandPath(IPath currDir, String pattern,
            ArrayList<String> expandedArgList, ArrayList<String> finalArgList,
            List<StringMatcher> ignoreStrs, boolean isFirstSgmt, boolean isLastSgmt,
            boolean hasTrailingSeparator) throws IOException {

        // we need two lists to process the expansion of the patterns
        // suppose user enters 2 patterns to a ci command. lscm -d c:\work ci
        // dir\a*\file*.txt C:\myWS\dir*\t*.txt
        // considering the first pattern dir\a*\file*.txt, the expansion happens
        // in the following way
        // two lists are maintained and a final argument list.
        // One of the list is expandedArgList. This list would contain previous
        // expansions if any were processed.
        // The second is the newlyExpandedArgList which would store the incoming
        // expansions.
        // Inorder to get a newly expanded str, for eg: dir*/a*/b*.txt
        // First dir* is processed which might expand to C:\work\dir1 ,
        // C:\work\dir2 which is stored in expandedArgList.
        // All the children underneath each of these are fetched, then matched
        // against the pattern a*.
        // The matched candidates are stored in the newlyExpandedArgList. eg:
        // c:\work\dir1\a1 , c:\work\dir1\a2, c:\work\dir2\b1, c:\work\dir2\b2
        // etc.
        // This process repeats until the leaf of the each arg is reached. all
        // these paths are stored in the finalArgList.
        // FinalArgList stores the resulting expanded strings.
        // On reaching the last segment in the pattern, b*.txt, we iterate over
        // the list of last expanded patterns c:\work\dir1\a1 , c:\work\dir1\a2
        // ..
        // then fetch the children underneath each one of the above.
        // match and store the resulting candidates in the final arg list.

        boolean matchFound = false;

        // if path contains a*/b/c*, in this case we need to only append 'b'.
        // isAPattern allows us to do that.
        boolean isAPattern = pattern.contains("*") || pattern.contains("?"); //$NON-NLS-1$ //$NON-NLS-2$

        // stores the newly about to be, expanded child nodes.
        // eg: a*/b*, if a* was already expanded in the previous call
        // this list would now store a1/b1, a1/b2, a2/b1
        ArrayList<String> newlyExpandedArgList = new ArrayList<String>();

        // for a list of paths already expanded, iterate over each one of them.
        // Eg: dir* matched dir1, dir2, dir3...
        for (String pathStr : expandedArgList) {

            IPath cwd = null;

            // Make CWD + pathStr as current cwd
            // incase of first segment, CWD is in the array
            if (isFirstSgmt) {
                cwd = new Path(currDir.toString());
            } else {
                cwd = new Path(pathStr);
            }

            // if the current path segment was plain text eg: a*\b\c*, without a
            // wildcard, eg: b
            // then for each segment already expanded, for a1, a2, a3 ..., we
            // carry out further expansion by adding
            // 'b' to each a1, a2, a3... and then get further results.
            if (!isAPattern) {

                IPath filePath = new Path(cwd.toOSString() + File.separator + pattern);
                File file = filePath.toFile();

                // we add the pattern 'b' to a1, then check if this file exists,
                // is a dir, incase we have not yet
                // reached the last segment, as we match files only in the last
                // segment.
                if (file.exists() && (!isLastSgmt && file.isDirectory()) || (isLastSgmt)) {

                    // if it's the last segment, and user had specified pattern
                    // like a\b\c\, then we need to match
                    // only dir at the last segment
                    if (isLastSgmt && hasTrailingSeparator && !file.isDirectory()) {
                        continue;
                    }

                    if (ignoreStrs.size() > 0
                            && shouldIgnoreFile(ignoreStrs, filePath.toOSString(), file.getName())) {
                        continue;
                    }

                    String result = filePath.toOSString();
                    matchFound = true;

                    // if we are at the last segment of the pathStr, which means
                    // the entire expansion is over
                    // we need to add this to the finalArgList. add it to the
                    // final list if it's already not present
                    if (isLastSgmt && !finalArgList.contains(result)) {
                        finalArgList.add(result);

                    } else { // we are still at the middle of expansion of
                             // certain pattern
                        newlyExpandedArgList.add(result);
                    }
                }

                continue;
            }

            // Further processing only for wildcard patterns

            File dir = cwd.toFile();
            File[] children = dir.listFiles();
            if (children != null) {
	            StringMatcher strMatcher = new StringMatcher(pattern, false, false);
	
	            for (File file : children) {
	
	                String fileName = file.getName();
	                IPath filePath = new Path(cwd.toOSString() + File.separator + fileName);
	
	                if (strMatcher.match(fileName)) {
	
	                    // if this is not the last segment and we have a matched a
	                    // file, then continue.
	                    // we match files only in the last segment of a pathStr
	                    if (!isLastSgmt && !file.isDirectory()) {
	                        continue;
	                    }
	
	                    // if it's the last segment, and user had specified pattern
	                    // like a\b\c\, then we need to match
	                    // only dir at the last segment. if its not a dir, then
	                    // continue
	                    if (isLastSgmt && hasTrailingSeparator && !file.isDirectory()) {
	                        continue;
	                    }
	
	                    // if GLOBIGNORE env var is set
	                    if (ignoreStrs.size() > 0
	                            && shouldIgnoreFile(ignoreStrs, filePath.toOSString(), fileName)) {
	
	                        continue;
	                    }
	
	                    // We found our match
	                    matchFound = true;
	                    String result = filePath.toOSString();
	
	                    // if we are at the last segment of the pathStr, which means
	                    // the entire expansion is over
	                    // we need to add this to the finalArgList. add it to the
	                    // final list if it's already not present
	                    if (isLastSgmt) {
	                        if (!finalArgList.contains(result)) {
	                            finalArgList.add(result);
	                        }
	
	                    } else { // we are still at the middle of expansion of
	                             // certain pattern
	                        newlyExpandedArgList.add(result);
	                    }
	
	                }
	            }
            }
        }
        // assign the newly expanded child nodes to pathList
        // i.e for a pattern like dir*/a*, after expansion of a*, dir1/a1,
        // dir1/a2. make this list as the current list.
        expandedArgList.clear();
        expandedArgList.addAll(newlyExpandedArgList);

        return matchFound;
    }

    /**
     * Ignore the file if pattern specified as a part of GLOBIGNORE
     * 
     * @param ignoreStrs
     *            The ignore strings(or string patterns) present in GLOBIGNORE
     * @param filePath
     *            The file path including filename
     * @param filename
     *            The filename
     * @return
     */
    private static boolean shouldIgnoreFile(List<StringMatcher> ignoreStrs, String filePath,
            String fileName) {

        for (StringMatcher ignoreStr : ignoreStrs) {

            // try matching the file path to a ignore pattern
            if (ignoreStr.match(filePath)) {
                return true;
            } else {

                // try matching only the filename to a pattern
                if (ignoreStr.match(fileName)) {
                    return true;
                }
            }
        }
        return false;
    }

    /** Logic for gathering up positional arguments. */
    private class PositionalArgumentAccumulator {

        int totalArgsReceived = 0;

        protected LinkedList<IPositionalOptionDefinition> defs;

        /** The accumulated values. */
        protected Map<IOptionKey, OptValue> positionalOpts = new HashMap<IOptionKey, OptValue>();

        PositionalArgumentAccumulator() {
            this.defs = new LinkedList<IPositionalOptionDefinition>(options.getPositionalOptions());
        }

        void addPositionalArg(String newArg, ParseController parseController)
                throws ArgumentOverflowException {
            if (defs.isEmpty()) {
                throw new ArgumentOverflowException(-1, newArg);
            }

            IPositionalOptionDefinition def = defs.getFirst();

            def.matchArgument(parseController);

            // Get the value list
            IOptionKey id = def.getId();

            OptValue values = positionalOpts.get(id);

            if (values == null) {
                List<String> list = new LinkedList<String>();
                values = new OptValue(list);

                positionalOpts.put(id, values);
            }

            // Add the value
            values.args.add(newArg);
            totalArgsReceived++;

            if (values.args.size() >= def.getMaxCount()
                    && def.getMaxCount() != Options.ARG_INFINITE) {
                defs.removeFirst();
            }
        }

        int getExpectedArgCount() {
            if (options.positionalOptionsAreInfinite()) {
                return Options.ARG_INFINITE;
            }

            int expected = 0;
            for (IPositionalOptionDefinition def : defs) {
                expected += def.getMaxCount();
            }

            return expected;
        }

        /**
         * Throw an exception if there aren't enough positional arguments.
         * 
         * @throws ArgumentUnderflowException
         */
        public void ensureSatisfied() throws ArgumentUnderflowException {
            // See if any of our remaining defs require arguments
            for (int i = 0; i < defs.size(); i++) {
                IPositionalOptionDefinition def = defs.get(i);

                IOptionKey id = def.getId();

                OptValue values = positionalOpts.get(id);

                if (values == null) {
                    if (def.getMinCount() != 0) {
                        // Missing required args
                        throw new ArgumentUnderflowException(getExpectedArgCount(),
                                totalArgsReceived, defs);
                    }
                } else {
                    if (values.args.size() < def.getMinCount()) {
                        throw new ArgumentUnderflowException(getExpectedArgCount(),
                                totalArgsReceived, defs);
                    }
                }
            }
        }

        /**
         * @return The positional arguments.
         */
        public Map<IOptionKey, OptValue> getArgs() {
            return positionalOpts;
        }
    }

    private class ParseController implements IParseController {
        boolean haltAfterOption = false;

        public void haltAfterOption() {
            haltAfterOption = true;
        }

        /**
         * @return <code>true</code> if the parse should continue.
         */
        public boolean carryOn() {
            return !haltAfterOption;
        }
    }

    @SuppressWarnings("null")
    public ICommandLine parse() throws MalformedCommandLineException {
        INamedOptionDefinition currentOpt = null;
        LinkedList<String> currentOptArgs = null;

        Map<IOptionKey, OptValue> namedOpts = new HashMap<IOptionKey, OptValue>();

        PositionalArgumentAccumulator positionalArgs = new PositionalArgumentAccumulator();

        ParseController parseController = new ParseController();

        // Parse the command line
        int i;
        for (i = 0; i < commandLine.length && parseController.carryOn(); i++) {
            String item = commandLine[i];

            INamedOptionDefinition opt = null;

            try {
                opt = asOpt(item);
            } catch (UnknownOptionException e) {
                if (options.shouldStopOnUnknownValue()) {
                    break;
                }

                throw e;
            }

            if (opt == null) {
                // This is an argument
                if (currentOpt == null) {
                    // It's a positional argument for the whole cmdline
                    try {
                        positionalArgs.addPositionalArg(item, parseController);

                    } catch (ArgumentOverflowException e) {

                        if (options.shouldStopOnUnknownValue()) {
                            break;
                        }

                        throw e;
                    }

                } else {
                    // It's a positional argument for the current opt.
                    if (currentOpt.getArgCount() == Options.ARG_INFINITE
                            || currentOpt.getArgCount() > currentOptArgs.size()) {

                        currentOpt.matchArgument(parseController);

                        currentOptArgs.add(item);

                        if (currentOpt.getArgCount() == currentOptArgs.size()) {
                            // We've filled our quota
                            namedOpts.put(currentOpt.getId(), new OptValue(currentOptArgs));

                            currentOpt = null;
                            currentOptArgs = null;
                        }
                    }
                    // If we just had a currentOpt which takes no args
                    // we would have added it to namedOpts and set currentOpt = null
                }

                continue;
            }

          	// Check whether the opt is already present in namedOpts
            if(namedOpts.containsKey(opt.getId())){
            	throw new MalformedOptException(item, Messages.RepeatedNamedOption);
            }

            if (currentOpt == null) {
                // This is a new opt

                if (opt.getArgCount() == 0) {
                    namedOpts.put(opt.getId(), new OptValue(currentOptArgs));
                } else {
                    currentOpt = opt;
                    currentOptArgs = new LinkedList<String>();
                }

                continue;
            } else {
                // There's a new opt, even though we're accumulating args for
                // an existing op
                if ((currentOpt.getArgCount() == Options.ARG_INFINITE && currentOpt.getMinArgCount() == 0)
                        || currentOpt.getArgCount() == 0
                        || (currentOptArgs.size() >= currentOpt.getMinArgCount())) {
                    // We've filled our quota
                    namedOpts.put(currentOpt.getId(), new OptValue(currentOptArgs));

                    if (opt.getArgCount() == 0) {
                        namedOpts.put(opt.getId(), new OptValue(currentOptArgs));
                        // reset currentOpt
                        // We didn't have to do it in the if condition because 
                        // the condition matches only if currentOpt == null
                        currentOpt = null;
                        currentOptArgs = null;
                    }
                    else
                    {
	                    currentOpt = opt;
	                    currentOptArgs = new LinkedList<String>();
                    }
                    ;
                } else {
                    // Underflow
                    throw new ArgumentUnderflowException(currentOpt.getArgCount(),
                            currentOptArgs.size(), Collections.singletonList(currentOpt));
                }
            }
        }

        // Gather up the unconsumed arguments
        LinkedList<String> unconsumed = new LinkedList<String>();

        for (int j = i; j < commandLine.length; j++) {
            unconsumed.add(commandLine[j]);
        }

        // Handle unfinished options
        if (currentOpt != null) {
            if (currentOpt.getArgCount() == currentOptArgs.size()
                    || (currentOpt.getArgCount() == Options.ARG_INFINITE && currentOpt.getMinArgCount() == 0)
                    || (currentOptArgs.size() >= currentOpt.getMinArgCount())) {
                namedOpts.put(currentOpt.getId(), new OptValue(currentOptArgs));

                currentOpt = null;
                currentOptArgs = null;
            } else {
                throw new ArgumentUnderflowException(currentOpt.getArgCount(),
                        currentOptArgs.size(), Collections.singletonList(currentOpt));
            }
        }

        // Verify arguments
        positionalArgs.ensureSatisfied();

        return new CommandLine(options, namedOpts, positionalArgs.getArgs(), unconsumed);
    }

   /**
     * Checks to see if the given element is a known option.
     * 
     * @param item
     * 
     * @throws MalformedCommandLineException
     */
    protected INamedOptionDefinition asOpt(String item) throws MalformedCommandLineException {

        // Is it a long opt?
        if (item.startsWith(PREFIX_LONGOPT) && item.length() >= PREFIX_LONGOPT.length()) {
            String optName = item.substring(PREFIX_LONGOPT.length());

            for (INamedOptionDefinition candidate : options.getNamedOptions()) {
                if (optName.equals(candidate.getLongOpt())) {
                    return candidate;
                }
            }

            // Let it fall through to the other if condition
            if(!item.equals(PREFIX_LONGOPT)) {
                throw new UnknownOptionException(item);            	
            }
        }

        // Is it a short opt?
        if (item.startsWith(PREFIX_SHORTOPT)) {

            if (item.length() == PREFIX_SHORTOPT.length()) {
                // We have a "-". Explicitly allowed.
                return null;
            }

            if (item.length() != (PREFIX_SHORTOPT.length() + 1)) {
                throw new MalformedOptException(item, Messages.CLIParser_1);
            }

            String optName = item.substring(PREFIX_SHORTOPT.length());

            // If the list of named options does not contain an empty longOpt,
            // we would have come here. Now optName will be equal to - and 
            // it will encounter this condition. We will now treat item as an
            // argument
            if (PREFIX_SHORTOPT.equals(optName)) {
            	// We have a "--". Explicitly disallowed as an argument.
            	return null;
            }

            // It's a single character opt.
            for (INamedOptionDefinition candidate : options.getNamedOptions()) {
                if (optName.equals(candidate.getShortOpt())) {
                    return candidate;
                }
            }

            throw new UnknownOptionException(item);
        }

        // It's neither.
        return null;
    }
}
