/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.rtc.cli.infrastructure.internal.parser;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.Map.Entry;

import org.eclipse.osgi.util.NLS;

import com.ibm.team.rtc.cli.infrastructure.internal.core.Constants;
import com.ibm.team.rtc.cli.infrastructure.internal.Messages;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.MalformedArgumentException;

public class CommandLine implements ICommandLine {

    static class OptValue {
        final List<String> args;

        OptValue(List<String> args) {
            this.args = args;
        }
    }

    static class OptValueArg {
        final List<ICommandLineArgument> args;

        OptValueArg(List<ICommandLineArgument> args) {
            this.args = args;
        }
    }

    /** The unconsumed arguments. */
    final List<String> unconsumed;

    final Options options;

    private final Map<IOptionKey, OptValueArg> namedOpts;

    private final Map<IOptionKey, OptValueArg> positionalOpts;

    public CommandLine(Options options, Map<IOptionKey, OptValue> namedOpts,
            Map<IOptionKey, OptValue> posOpts, LinkedList<String> unconsumed)
            throws MalformedArgumentException {
        this.options = options;
        this.namedOpts = getOptArgs(namedOpts);
        this.positionalOpts = getOptArgs(posOpts);
        this.unconsumed = unconsumed;
    }

    public boolean hasOption(IOptionKey id) {
        return namedOpts.containsKey(id) || positionalOpts.containsKey(id);
    }

    public boolean hasOption(IOptionKey id, Map<IOptionKey, OptValue> opts) {
        if (namedOpts != null) {
            return namedOpts.containsKey(id);
        }
        return opts.containsKey(id);
    }

    public String getOption(IOptionKey argName) {
        ICommandLineArgument value = getOptionValue(argName);
        return (value != null) ? value.getValue() : null;
    }

    public List<String> getOptions(IOptionKey argName) {
        List<ICommandLineArgument> argList = getOptionValues(argName);
        List<String> selectors = new ArrayList<String>(argList.size());
        for (ICommandLineArgument arg : argList) {
            selectors.add(arg.getValue());
        }
        
        return selectors;
    }

    public List<String> getUnconsumedArgs() {
        return unconsumed;
    }

    public Collection<IOptionKey> getDefinedOpts() {
        List<IOptionKey> allKeys = new ArrayList<IOptionKey>(namedOpts.size()
                + positionalOpts.size());

        allKeys.addAll(namedOpts.keySet());
        allKeys.addAll(positionalOpts.keySet());

        return allKeys;
    }

    public String getOption(IOptionKey argName, String defaultValue) {
    	ICommandLineArgument defaultArg = new CommandLineArgument(argName, defaultValue);
        ICommandLineArgument value = getOptionValue(argName, defaultArg);

        return value.getValue();
    }

    public Options getDefinition() {
        return options;
    }

    /**
     * @param opts
     * @return
     * @throws Exception
     */
    private Map<IOptionKey, OptValueArg> getOptArgs(Map<IOptionKey, OptValue> opts)
            throws MalformedArgumentException {

        Map<IOptionKey, OptValueArg> optValueMap = new HashMap<IOptionKey, OptValueArg>();

        for (Map.Entry<IOptionKey, OptValue> optEntry : opts.entrySet()) {

            // value can be null if it's a switch with no args
            OptValueArg valueArg = constructCommandLineArg(optEntry.getKey(), optEntry.getValue());

            optValueMap.put(optEntry.getKey(), valueArg);
        }

        return optValueMap;
    }

    private OptValueArg getOptionValue(IOptionKey id, boolean isRequired) {
        OptValueArg toReturn;

        toReturn = namedOpts.get(id);
        if (toReturn != null) {
            return toReturn;
        }

        toReturn = positionalOpts.get(id);
        if (toReturn != null) {
            return toReturn;
        }
        
        if (toReturn == null && !isRequired) {
        	return null;
        }

        throw new IllegalArgumentException(NLS.bind(Messages.CommandLine_3, id.toString()));
    }

    public ICommandLineArgument getOptionValue(IOptionKey argName) {
    	IOptionDefinition optDef = options.getOption(argName);
        OptValueArg value = getOptionValue(argName, optDef.isRequired());

        // Ensure the arg is the proper size
        if (!optDef.isRequired() && (value == null || value.args == null || value.args.size() == 0)) {
        	return null;
        }
        
        if (value == null || value.args == null || value.args.size() == 0) {
            throw new IllegalArgumentException(NLS.bind(Messages.CommandLine_0, argName));
        }

        if (value.args.size() > 1) {
            throw new IllegalArgumentException(NLS.bind(Messages.CommandLine_1, argName));
        }

        return value.args.get(0);
    }

    public ICommandLineArgument getOptionValue(IOptionKey argName, ICommandLineArgument defaultValue) {
        ICommandLineArgument value = null;

        if (hasOption(argName)) {
            value = getOptionValue(argName);
        }
        
        if (value == null) {
            value = defaultValue;
        }

        return value;
    }

    public List<ICommandLineArgument> getOptionValues(IOptionKey argName) {
        if (!hasOption(argName)) {
            throw new NoSuchElementException(argName.toString());
        }
        
        IOptionDefinition optDef = options.getOption(argName);
        List<ICommandLineArgument> selectors;
        OptValueArg value = getOptionValue(argName, optDef.isRequired());
        if (value != null) {
        	selectors = value.args;
        } else {
        	selectors = Collections.emptyList();
        }
        
        return selectors;
    }

    /**
     * Parses the value token based on the semantics of the IOptionKey
     * @param option the key that may have a delimiter that makes token parsing complex
     * @param value user type string value to be parsed
     * @return a semantically well formed token. May return null if the value is empty. i.e. the existence of the key is the only requirement to parse.
     * @throws MalformedArgumentException if the parsing fails
     */
    private OptValueArg constructCommandLineArg(IOptionKey option, OptValue value) throws MalformedArgumentException {
        List<ICommandLineArgument> cmdLineArgs = new ArrayList<ICommandLineArgument>();


        // value.args can be null if it's a switch with no args
        if (value.args == null) {
            return null;
        }

        String delimiter = option.getDelimiter();
        int delimLen = (delimiter != null) ? delimiter.length() : 0;

        for (String str : value.args) {
            boolean foundSelector = false;

            if (delimiter != null) {
                // if str starts with the delimiter, error out as there will not be any prefix
                if (str.startsWith(delimiter)) {
                    throw new MalformedArgumentException(str);
                }

                String prefix = null, suffix = null;
                int index = 0;

                do {
                    index = str.indexOf(delimiter, index);
                    if (index != -1) {
                        // if the occurrence of delimiter is not escaped, then we found the prefix
                        // and suffix.
                        if (str.charAt(index - 1) != '\\') {
                            prefix = str.substring(0, index);
                            suffix = (index + delimLen >= str.length()) ? null :
                                str.substring(index + delimLen);

                            if (suffix == null) {
                                throw new MalformedArgumentException(str);
                            }
                            
                            String nonEscapedPrefix = stripEscapedChars(prefix, delimiter);
                            cmdLineArgs.add(new CommandLineArgument(option, str, nonEscapedPrefix,
                                    suffix));
                            foundSelector = true;
                            break;
                        }

                        // move the index past the delimiter
                        index += delimLen;
                    }
                } while (index != -1);
            }

            if (!foundSelector) {
                String nonEscapedStr = (delimiter == null) ? str : stripEscapedChars(str, delimiter);
                cmdLineArgs.add(new CommandLineArgument(option, nonEscapedStr, nonEscapedStr, null));
            }
        } // for loop ends

        return new OptValueArg(cmdLineArgs);
    }

    private String stripEscapedChars(String selector, String delimiter) {
        StringBuffer selectorBuf = new StringBuffer();

        assert (selector != null);
        int delimLen = delimiter.length();

        for (int i = 0; i < selector.length();) {
            // strip backslash char from the selector
            boolean strip = false;
            if (selector.charAt(i) == '\\' && (i + delimLen) <= selector.length()) {
                int index = selector.indexOf(delimiter, i + delimLen);
                if (index != -1 && index == i + delimLen) {
                    strip = true;
                }
            }
            
            if (!strip){
                selectorBuf.append(selector.charAt(i));
                ++i;
            } else {
                // move the index past the delimiter
                i += delimLen;
            }
        }
        
        return selectorBuf.toString();
    }
}