/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2013. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.rtc.cli.infrastructure.internal.parser;

import java.util.Collection;
import java.util.List;
import java.util.NoSuchElementException;

/**
 * A representation of data parsed from the commandline.
 */
public interface ICommandLine {
    /**
     * Check if the given argument was passed in
     */
    public boolean hasOption(IOptionKey opt);

    /**
     * Convenience method for:
     * 
     * <pre>
     * String value = null;
     * if (args.hasOption(argName)) {
     *     value = args.getOption(argName);
     * } else {
     *     value = defaultValue;
     * }
     * </pre>
     * 
     * @param opt
     *            The value to find.
     * 
     * @param defaultValue
     *            The default value to return if the argument is not defined.
     * 
     * @return The value, as defined on the commandline, or the value of
     *         <code>defaultValue</code>
     */
    public String getOption(IOptionKey opt, String defaultValue);

    /**
     * Convenience method for:
     * 
     * <p>
     * <code>getOptions(optName).get(0)</code>
     * 
     * <p>
     * Throws an exception if the option doesn't have exactly one argument.
     */
    public String getOption(IOptionKey opt);

    /**
     * Return the list of arguments for the given option, or null if there are
     * no arguments.
     * 
     * @throws NoSuchElementException
     *             if the given key is not defined.
     */
    public List<String> getOptions(IOptionKey argName);

    /**
     * Get the list of unconsumed arguments. If there are no arguments, an empty
     * list is returned.
     */
    public List<String> getUnconsumedArgs();

    /**
     * Returns a collection of the IDs of the defined options. If none were
     * defined, an empty list is returned.
     */
    public Collection<IOptionKey> getDefinedOpts();

    /**
     * Returns the Options object that was used as a reference to create this
     * commandline.
     */
    public Options getDefinition();

    /**
     * Return the argument as ICommandLineArgument for the given option, or null
     * if there are no arguments.
     * 
     * @throws NoSuchElementException
     *             if the given key is not defined.
     */
    public ICommandLineArgument getOptionValue(IOptionKey argName);

    /**
     * Return the argument as ICommandLineArgument for the given option, or
     * default value if there are no arguments.
     * 
     */
    public ICommandLineArgument getOptionValue(IOptionKey argName, ICommandLineArgument defaultValue);

    /**
     * Return the list of arguments as ICommandLineArgument for the given
     * option, or null if there are no arguments.
     * 
     * @throws NoSuchElementException
     *             if the given key is not defined.
     */
    public List<ICommandLineArgument> getOptionValues(IOptionKey argName);
}
