/*******************************************************************************
 * Licensed Materials - Property of IBM
 * © Copyright IBM Corporation 2005, 2015. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.rtc.cli.infrastructure.internal.parser;

import com.ibm.team.rtc.cli.infrastructure.internal.Messages;

/**
 * Create a new option for the commandline.
 */
public class NamedOptionDefinition extends AbstractOptionDefinition implements
        INamedOptionDefinition {

    private final String shortOpt;
    private final String longOpt;
    private final int argCount;
    private final int minArgCount;
    private boolean hideOption = false;

    public NamedOptionDefinition(IOptionKey id, String shortOpt, String longOpt, int argCount) {
        this(id, longOpt, shortOpt, longOpt, argCount, argCount, null);
    }

    public NamedOptionDefinition(IOptionKey id, String shortOpt, String longOpt, int argCount,
            String delimiter) {
        this(id, longOpt, shortOpt, longOpt, argCount, argCount, delimiter);
    }   
    
    public NamedOptionDefinition(IOptionKey id,  String shortOpt, String longOpt, int minArgCount, 
    		int argCount) {
        this(id, longOpt, shortOpt, longOpt, minArgCount, argCount, null);
    }
    
    public NamedOptionDefinition(IOptionKey id, String shortOpt, String longOpt, int minArgCount, int argCount,
            String delimiter) {
        this(id, longOpt, shortOpt, longOpt, minArgCount, argCount, delimiter);
    }

    public NamedOptionDefinition(IOptionKey id, String name, String shortOpt, String longOpt,
            int argCount) {
        this(id, name, shortOpt, longOpt, argCount, argCount, null);
    }
    
    public NamedOptionDefinition(IOptionKey id, String name, String shortOpt, String longOpt,
            int argCount, String delimiter) {
    	this(id, name, shortOpt, longOpt, argCount, argCount, delimiter);
    }

    /**
     * @param id an option key ID or the USE_SELF_AS_ID sentinal
     * @param name the name of the option
     * @param shortOpt may be {@code null}; must be 1 character long. (why isn't this a Character object?)
     * @param longOpt may be {@code null}; must not begin with "--";
     * @param minArgCount the minimum numer of arguments for this option. 0 means it will be optional
     * @param argCount the maximum number of expected options for this argument. Options.ARG_INFINITE for infinite
     * @param delimiter CommonOptions.delimiter to specify @ sign to specify a repo
     */
    public NamedOptionDefinition(IOptionKey id, String name, String shortOpt, String longOpt,
            int minArgCount, int argCount, String delimiter) {
        super(id, name, delimiter);

        if (shortOpt != null && shortOpt.length() != 1) {
            throw new IllegalArgumentException(Messages.NamedOptionDefinition_0);
        }

        if (longOpt != null && longOpt.startsWith(CLIParser.PREFIX_LONGOPT)) {
            throw new IllegalArgumentException(Messages.NamedOptionDefinition_1);
        }

        this.shortOpt = shortOpt;
        this.longOpt = longOpt;
        this.argCount = argCount;
        this.minArgCount = minArgCount;
    }

    public NamedOptionDefinition(String shortOpt, String longOpt, int argCount) {
        this(USE_SELF_AS_ID, shortOpt, longOpt, argCount, null);
    }
    
    public NamedOptionDefinition(String shortOpt, String longOpt, int argCount, String delimiter) {
        this(USE_SELF_AS_ID, shortOpt, longOpt, argCount, delimiter);
    }
    
    public NamedOptionDefinition(String shortOpt, String longOpt, int minArgCount, int argCount, String delimiter) {
        this(USE_SELF_AS_ID, longOpt, shortOpt, longOpt, minArgCount, argCount, delimiter);
    }

    /**
     * Get the short (one character) definition of this argument. Does not
     * include a '-'.
     */
    public String getShortOpt() {
        return shortOpt;
    }

    /**
     * Get the long argument name. May be null.
     */
    public String getLongOpt() {
        return longOpt;
    }

    /** Get the number of expected arguments. */
    public int getArgCount() {
        return argCount;
    }

    /** Get the least number of expected arguments. */
    @Override
    public int getMinArgCount() {
        return minArgCount;
    }

    public String toString() {
        StringBuffer buf = new StringBuffer();

        if (shortOpt != null) {
            buf.append("-"); //$NON-NLS-1$
            buf.append(shortOpt);
        }

        if (longOpt != null) {
            if (buf.length() > 0) {
                buf.append("/"); //$NON-NLS-1$
            }

            buf.append("--"); //$NON-NLS-1$
            buf.append(longOpt);
        }

        return buf.toString();
    }
    
    public void hideOption() {
        hideOption = true;
    }

    @Override
    public boolean isHidden() {
        return hideOption;
    }
    
    @Override
    public boolean isRequired() {
    	return getMinArgCount() > 0;
    }
}
