/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.rtc.cli.infrastructure.internal.parser;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.osgi.util.NLS;

import com.ibm.team.rtc.cli.infrastructure.internal.Messages;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.ConflictingOptionException;
import com.ibm.team.rtc.cli.infrastructure.internal.util.Pair;

/**
 * Configuration for a commandline parser. Iterating over instances of this
 * class walks the list of non-positional options.
 */
public class Options {

    /**
     * Set to true if the parser should stop when it hits a token it doesn't
     * have a definition for.
     */
    final private boolean stopOnUnknownValue;

    /** The ordered set of positional options. Note that these are top level options.
     *  Inner levels can be described using Simple, MutuallyExclusiveGroup and ContinuousGroup.
     */ 
    final private LinkedList<IPositionalOptionDefinition> positionalOptions = new LinkedList<IPositionalOptionDefinition>();

    /** The set of named options. Not that these contain all named options added
     *  This includes : top level named options, named options added through inner levels.
     *  Inner levels can be described using Simple, MutuallyExclusiveGroup and ContinuousGroup.
     */     
    final private List<INamedOptionDefinition> namedOptions = new LinkedList<INamedOptionDefinition>();
    
    /** The set of top level named options with flag specifying whether they are required or not. 
     * We maintain similar but all inclusive list in namedOptions. 
     */     
    final private HashMap<INamedOptionDefinition, Boolean> topLevelNamedOptions = new HashMap<INamedOptionDefinition, Boolean>();
    
    /** The set of options used for printing. 
     * Since we will be storing different object hierarchies, the type of this
     * list will be object 
     */
    final private List<Object> printOptions = new LinkedList<Object>();
    
    /** The set of top level mutually exclusive option groups */
    List<MutuallyExclusiveGroup> mutuallyExclusiveOptions = new LinkedList<MutuallyExclusiveGroup>();
    
    /** The set of top level continuous option groups */
    //final private List<ContinuousGroup> continuousOptions = new LinkedList<ContinuousGroup>();
    
    /** Whether this options instance supports new style of short help */
    private boolean supportsNewHelpFormat = false;
    
    /** The long help for this set of options. */
    String longHelp = null;

    /***
    /**
     * @see Options#Options(boolean, int).
     */
    public static final int ARG_INFINITE = -1;

    /**
     * @param stopOnUnknownValue
     *            Set to <code>true</code> if the parser should gracefully stop
     *            when it hits an unexpected value (either an opt without a
     *            definition, or an unexpected positional argument). If set to
     *            <code>false</code>, then an unknown value is considered an
     *            error.
     * 
     */
    public Options(boolean stopOnUnknownValue) {
    	this(stopOnUnknownValue, false);
    }
    
    /**
     * 
     */
    public Options(boolean stopOnUnknownValue, boolean supportsNewHelpFormat) {
        this.stopOnUnknownValue = stopOnUnknownValue;
        this.supportsNewHelpFormat = supportsNewHelpFormat;
    }

    /**
     * Add the given option.
     * 
     * @param key
     *            A unique object used to refer to the option in the resulting
     *            {@link ICommandLine}.
     * 
     * @param shortOpt
     *            The short name of the option. Should include a preceding dash.
     *            May only be one character long.
     * 
     * @param longOpt
     *            The long name of the option. Should include two preceding
     *            dashes. May be any length.
     * 
     * @param help
     *            Human-meaningful help text.
     * 
     * @param argCount
     *            The number of arguments this option requires. May be an
     *            integer (indicating that exactly that number of arguments are
     *            required), or ARG_INFINITE (indicating that zero or more
     *            arguments are allowed).
     * 
     * @return The receiver.
     */
    public Options addOption(IOptionKey key, String shortOpt, String longOpt, String help,
            int argCount) throws ConflictingOptionException {
        INamedOptionDefinition opt = new NamedOptionDefinition(key, shortOpt, longOpt, argCount);
        return addOption(opt, help, false);
    }

    /**
     * Add the given option.
     * 
     * @param key
     *            A unique object used to refer to the option in the resulting
     *            {@link ICommandLine}.
     * 
     * @param shortOpt
     *            The short name of the option. Should include a preceding dash.
     *            May only be one character long.
     * 
     * @param longOpt
     *            The long name of the option. Should include two preceding
     *            dashes. May be any length.
     * 
     * @param help
     *            Human-meaningful help text.
     * 
     * @param argCount
     *            The number of arguments this option requires. May be an
     *            integer (indicating that exactly that number of arguments are
     *            required), or ARG_INFINITE (indicating that zero or more
     *            arguments are allowed).
     * 
     * @param delimiter
     *            The delimiter to split the argument. It will split the argument
     *            into prefix and suffix after the first match of the delimiter.
     *
     * @return The receiver.
     */
    public Options addOption(IOptionKey key, String shortOpt, String longOpt, String help,
            int argCount, String delimiter) throws ConflictingOptionException {
        INamedOptionDefinition opt = new NamedOptionDefinition(key, shortOpt, longOpt, argCount,
                delimiter);
        return addOption(opt, help, false);
    }

    /**
     * Check to see if an object with this key already exists.
     */
    public boolean hasOption(IOptionKey key) {
        return getOption(key) != null;
    }

    /**
     * Return the option with the given key, or <code>null</code> if the key is
     * not registered.
     */
    public IOptionDefinition getOption(IOptionKey key) {
        for (INamedOptionDefinition opt : namedOptions) {
            if (key.equals(opt.getId())) {
                return opt;
            }
        }

        for (IPositionalOptionDefinition opt : positionalOptions) {
            if (key.equals(opt.getId())) {
                return opt;
            }
        }

        return null;
    }

    /** Return the list of positional options. */
    public LinkedList<IPositionalOptionDefinition> getPositionalOptions() {
    	// TODO Flatten the list from mutually exclusive group and then append
    	// it to the list that we have

    	// TODO Flatten the list from continuous group and then append
    	// it to the list that we have
        return positionalOptions;
    }

    public Options addOption(IPositionalOptionDefinition opt, String helpText)
            throws ConflictingOptionException {
        // Make sure that we don't have an option with an infinite option list
        // preceeding us.
        if (positionalOptionsAreInfinite()) {
            IPositionalOptionDefinition last = positionalOptions.getLast();

            throw new IllegalArgumentException(
                    NLS.bind(Messages.Options_0, last.getId().toString()));
        }

        ensureNonConflicting(opt);

        positionalOptions.add(opt);
        printOptions.add(opt);
        optionHelp.add(new HelpPair(opt.getId(), helpText));

        return this;
    }

    /**
     * Check to see if the positional options consume an infinite number of
     * arguments.
     * 
     * @return
     */
    public boolean positionalOptionsAreInfinite() {
        if (positionalOptions.size() > 0) {
            IPositionalOptionDefinition last = positionalOptions.getLast();

            return last.getMaxCount() == ARG_INFINITE;
        }

        return false;
    }
    
    public Options addOption(INamedOptionDefinition opt,  String helpString)
            throws ConflictingOptionException {
    	return addOption(opt, helpString, false);
    }
    
    public Options addOption(INamedOptionDefinition opt,  String helpString, boolean isRequired)
            throws ConflictingOptionException {
    	if (opt.getId() == null) {
            throw new IllegalArgumentException(Messages.Options_1);
        }

    	ensureNonConflicting(opt);
    	
        namedOptions.add(opt);
        topLevelNamedOptions.put(opt, isRequired);
        printOptions.add(opt);
        optionHelp.add(new HelpPair(opt.getId(), helpString));

        return this;
    }

    public Options addOption(IOptionGroup grp) throws ConflictingOptionException {
    	// Add named and positional options from this group
    	addNamedOptionsFromGroup(grp);
    	addPositionalOptionsFromGroup(grp);
    	// Add this group to the list used for printing.
    	printOptions.add(grp);
    	return this;
    }
    

    private void addNamedOptionsFromGroup(IOptionGroup grp) {
    	Collection<Pair<INamedOptionDefinition, String>> namedOptionsFromGrp = grp.getNamedOptionsWithHelp();
    	for (Pair<INamedOptionDefinition, String> p : namedOptionsFromGrp) {
		    INamedOptionDefinition opt = p.getFirst();
	    	if (opt.getId() == null) {
	            throw new IllegalArgumentException(Messages.Options_1);
	        }

    		try {
    			ensureNonConflicting(opt);
    		}
    		catch(ConflictingOptionException exp) {
    			// Since there is already an option, we don't have to process it again
    			continue;
    		}

	        namedOptions.add(opt);
	        optionHelp.add(new HelpPair(opt.getId(), p.getSecond()));
    	}
    }
    
    private void addPositionalOptionsFromGroup(IOptionGroup grp) throws ConflictingOptionException {
    	Collection<Pair<IPositionalOptionDefinition, String>> positionalOptionsWithHelp = grp.getPositionalOptionsWithHelp();
    	for (Pair<IPositionalOptionDefinition, String> positionalOpt : positionalOptionsWithHelp) {
    		IPositionalOptionDefinition positionalDef = positionalOpt.getFirst();
	    	if (positionalDef.getId() == null) {
	            throw new IllegalArgumentException(Messages.Options_1);
	        }

	    	try {
	            ensureNonConflicting(positionalDef);
            }
            catch(ConflictingOptionException exp) {
                // Since there is already an option, we don't have to process it again
                continue;
            }

	        positionalOptions.add(positionalDef);
	        optionHelp.add(new HelpPair(positionalDef.getId(), positionalOpt.getSecond()));
    	}
    }

    /**
     * Ensure that the passed option does not conflict with any of the current
     * options. If a conflict would happen, an exception is thrown.
     */
    private void ensureNonConflicting(IOptionDefinition opt) throws ConflictingOptionException {
        // Ensure the ID doesn't conflict
        if (hasOption(opt.getId())) {
            throw new ConflictingOptionException(opt, null, "id", opt.getId().getName()); //$NON-NLS-1$
        }

        // Ensure that the name doesn't conflict
        for (IOptionDefinition other : namedOptions) {
            if (other.getName().equals(opt.getName())) {
                throw new ConflictingOptionException(opt, other, "name", other.getName()); //$NON-NLS-1$
            }
        }

        for (IOptionDefinition other : positionalOptions) {
            if (other.getName().equals(opt.getName())) {
                throw new ConflictingOptionException(opt, other, "name", other.getName()); //$NON-NLS-1$
            }
        }

        if (opt instanceof INamedOptionDefinition) {
            INamedOptionDefinition candidate = (INamedOptionDefinition) opt;

            for (INamedOptionDefinition current : namedOptions) {
                // Ensure that short opts don't conflict
                String shortOpt = current.getShortOpt();
                if (shortOpt != null && shortOpt.equals(candidate.getShortOpt())) {
                    throw new ConflictingOptionException(opt, current,
                            "shortOpt", current.getShortOpt()); //$NON-NLS-1$
                }

                // Ensure that long opts don't conflict
                String longOpt = current.getLongOpt();
                if (longOpt != null && longOpt.equals(candidate.getLongOpt())) {
                    throw new ConflictingOptionException(opt, current,
                            "longOpt", current.getLongOpt()); //$NON-NLS-1$
                }

            }
        }

    }

    public List<INamedOptionDefinition> getNamedOptions() {
    	return namedOptions;
    }
    
    public HashMap<INamedOptionDefinition, Boolean> getTopLevelNamedOptions() {
    	return topLevelNamedOptions;
    }

    public boolean shouldStopOnUnknownValue() {
        return stopOnUnknownValue;
    }

    /** Set the long help message. */
    public void setLongHelp(String longHelp) {
        this.longHelp = longHelp;
    }

    public String getLongHelp() {
        return longHelp;
    }

    private static class HelpPair {
        final IOptionKey key;
        final String help;

        public HelpPair(IOptionKey optionKey, String helpText) {
            this.key = optionKey;
            this.help = helpText;
        }
    }

    private final List<HelpPair> optionHelp = new LinkedList<HelpPair>();

    /** Get the help string describing the given key. */
    public String getHelpFor(IOptionKey optionKey) {
        for (HelpPair pair : optionHelp) {
            if (pair.key == optionKey) {
                return pair.help;
            }
        }

        throw new IllegalArgumentException(Messages.Options_2);
    }
    
    /** Whether this options instance supports new format for short help */
    public void setSupportsNewHelp(boolean isSupported) {
    	supportsNewHelpFormat  = isSupported;
    }
    
    /** Whether this options instance supports new format for short help */
    public boolean getSupportsNewHelp() {
    	return supportsNewHelpFormat;
    }
    
    public Collection<Object> getAllOptions() {
    	return printOptions;
    }
}
