/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.rtc.cli.infrastructure.internal.subcommands;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.osgi.util.NLS;

import com.ibm.team.rtc.cli.infrastructure.internal.core.CLIClientException;
import com.ibm.team.rtc.cli.infrastructure.internal.core.Constants;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IClientConfiguration;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IOptionSource;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommand;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommandDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommandDefinition.LoadOptionsException;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.ICommandLine;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.INamedOptionDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.IPositionalOptionDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.NamedOptionDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.Options;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.PositionalOptionDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.ConflictingOptionException;
import com.ibm.team.rtc.cli.infrastructure.internal.util.IndentingPrintStream;
import com.ibm.team.rtc.cli.infrastructure.internal.util.OptionGroupUtil;
import com.ibm.team.rtc.cli.infrastructure.internal.util.StatusHelper;
import com.ibm.team.rtc.cli.infrastructure.internal.util.StringUtil;
import com.ibm.team.rtc.cli.infrastructure.internal.util.SubcommandNameCompator;
import com.ibm.team.rtc.cli.infrastructure.internal.util.SubcommandUtil;
import com.ibm.team.rtc.cli.infrastructure.internal.Messages;

/**
 * Provide online help: a listing of available subcommands and basic options.
 */
public class HelpCmd implements ISubcommand, IOptionSource {

	/** String to display if there is no help. */
	public static final String NO_HELP_STRING = Messages.HelpCmd_0;
	
    public static PositionalOptionDefinition OPT_SUBCOMMAND_NAME = 
    	new PositionalOptionDefinition("subcmd", 0, Options.ARG_INFINITE); //$NON-NLS-1$

    public static NamedOptionDefinition OPT_ALL =
    	new NamedOptionDefinition(null, "all", 0); //$NON-NLS-1$
    
	private static final class NamedOptionComparator implements
			Comparator<INamedOptionDefinition> {
		public int compare(INamedOptionDefinition o1, INamedOptionDefinition o2) {
			String s1 = o1.getShortOpt();
			String s2 = o2.getShortOpt();
			if (s1 == null && s2 == null) {
				s1 = o1.getLongOpt();
				s2 = o2.getLongOpt();
			}
			if (s1 != null && s2 != null) {
				return s1.compareToIgnoreCase(s2);
			} else if (s1 != null) {
				return 1;
			} else if (s2 != null) {
				return -1;
			}
			return o1.getArgCount() - o2.getArgCount();
		}
	}

	private int termWidth;
	private boolean showDeprecatedHelp = false;

	public Options getOptions() throws ConflictingOptionException {
		Options options = new Options(false);
		
		options.setLongHelp(Messages.HelpCmd_1);
		
		options
			.addOption(OPT_SUBCOMMAND_NAME, Messages.HelpCmd_2)
			.addOption(OPT_ALL, Messages.HelpCmd_3);
		
		return options;
	}
	
	public void initialize(IClientConfiguration config) throws CLIClientException {
	}

	public int run(IClientConfiguration config) throws CLIClientException {
		termWidth = SubcommandUtil.getTerminalWidth(config);
		config.getContext().stdout().println();
		ICommandLine subargs = config.getSubcommandCommandLine();
		
        String strShowDeprecatedHelp = config.getContext().environment().
                get(Constants.SHOW_DEPRECATED_HELP);
        showDeprecatedHelp = (strShowDeprecatedHelp != null && 
                  strShowDeprecatedHelp.equalsIgnoreCase(Constants.SHOW_DEPRECATED_HELP_VALUE));

		if (subargs != null && subargs.hasOption(OPT_ALL)) {
			showAllSubcommands(config);
		}
		else if (subargs == null || !subargs.hasOption(OPT_SUBCOMMAND_NAME)) {
			showSubcommandListing(config);
		}
		else {
			showSubCommand(subargs.getOptions(OPT_SUBCOMMAND_NAME), config);
		}
		
		return 0;
	}
	
	
	/** Show the subcommands that are currently installed. */
	private void showSubcommandListing(IClientConfiguration config) {
		// Show the application args
		Options options = config.getCommandLine().getDefinition();
		
		config.getContext().stdout().print(NLS.bind(Messages.HelpCmd_4, config.getContext().getAppName()));
		showUsage(options, config, false);
		
		config.getContext().stdout().println();
		
		// List the subcommands
		List<ISubcommandDefinition> subcommands = new ArrayList<ISubcommandDefinition>(config.getSubcommandDefinitions());
		Collections.sort(subcommands, new SubcommandNameCompator());

		IndentingPrintStream out = new IndentingPrintStream(config.getContext().stdout());
		
		out.println(Messages.HelpCmd_5);
		
		out = out.indent();

		// Determine length of longest parent subcommand name
		int maxNameLength = 0;
		for (ISubcommandDefinition defn : subcommands) {
			boolean IsADeprecatedParent = defn.isDeprecated() &&  defn.isParent() && !hasVisibleChildren(defn);
			boolean IsADeprecatedNonParent = defn.isDeprecated() &&  !defn.isParent(); 
			if (!defn.showInHelpListing() || ((IsADeprecatedNonParent || IsADeprecatedParent) && !showDeprecatedHelp))
				continue;

			if (defn.getParentId() == null)
				maxNameLength = Math.max(maxNameLength,  defn.getFullSubcommandName().length());
		}
			
		for (ISubcommandDefinition defn : subcommands) {
			boolean IsADeprecatedParent = defn.isDeprecated() &&  defn.isParent() && !hasVisibleChildren(defn);
			boolean IsADeprecatedNonParent = defn.isDeprecated() &&  !defn.isParent(); 
			if (!defn.showInHelpListing() || ((IsADeprecatedNonParent || IsADeprecatedParent) && !showDeprecatedHelp))
				continue;

			if (defn.getParentId() == null) {
				String longLine = NLS.bind(Messages.HelpCmd_6, StringUtil.pad(defn.getFullSubcommandName(), maxNameLength), defn.getShortHelp());
				Collection<String> lines = StringUtil.wrapAndIndent(longLine, termWidth - out.getIndent());
				for (String line : lines) {
					out.println(line);
				}
			}
		}
	}
	
	private void showAllSubcommands(IClientConfiguration config) throws CLIClientException {
		Collection<ISubcommandDefinition> subcommands = config.getSubcommandDefinitions();

		List<CLIClientException> exceptions = new LinkedList<CLIClientException>();
		StringBuffer erroringNames = new StringBuffer();

		boolean isFirst = true;
		for (ISubcommandDefinition defn : subcommands) {
			if (!defn.showInHelpListing() || (defn.isDeprecated() && !showDeprecatedHelp))
				continue;
			
			if (!isFirst) {
				config.getContext().stdout().println();
				config.getContext().stdout().println();
			}
			isFirst = false;
			
			try {
				showSubcommand(config, Collections.singletonList(defn.getFullSubcommandName()), defn);
			} catch (CLIClientException e) {
				config.getContext().stderr().println(NLS.bind(Messages.HelpCmd_9, defn.getFullSubcommandName()));
				exceptions.add(e);
				
				if (erroringNames.length() > 0) {
					erroringNames.append(Messages.HelpCmd_10);
				}
				erroringNames.append(defn.getFullSubcommandName());
			}
			
		}
		
		if (exceptions.size() > 0) {
			throw StatusHelper.internalError(NLS.bind(Messages.HelpCmd_11, erroringNames.toString()), exceptions);
		}
	}

	/** Show subcmd-specific help information. */
	private void showSubCommand(List<String> cmdNames, IClientConfiguration config) throws CLIClientException {
		// Find the first subcommand
		String cmdName = cmdNames.get(0);
		ISubcommandDefinition myDef = config.getSubcommandDefinition();
		ISubcommandDefinition parent = myDef.getLoader().loadToplevel(cmdName);

		if (parent == null) {
			throw StatusHelper.argSyntax(NLS.bind(Messages.HelpCmd_12, cmdName), config);			
		}
		
		if (parent.getParentId() != null) {
			StringBuffer buf = new StringBuffer();
			
			for (ISubcommandDefinition anc : parent.getAncestors()) {
				buf.append(anc.getFullSubcommandName());
				buf.append(Messages.HelpCmd_13);
			}
			
			buf.append(cmdName);
			
			throw StatusHelper.argSyntax(NLS.bind(Messages.HelpCmd_14, cmdName), config);
		}
		
        // Walk the list of descendents
		ISubcommandDefinition child = parent;
		for (int i = 1; i < cmdNames.size() ; i++) {
			try {
				child = parent.getChild(cmdNames.get(i));
				parent = child;
			} catch (IllegalArgumentException e) {
				throw StatusHelper.argSyntax(NLS.bind(Messages.HelpCmd_15, cmdNames.get(i)), config);
			}
		}
		
        if (child.isDeprecated() && !showDeprecatedHelp && !hasVisibleChildren(child)) {
            showDeprecatedHelpHint(parent, config);
            return;
        }

        showSubcommand(config, cmdNames, child);
	}
	
	private void showDeprecatedHelpHint(ISubcommandDefinition defn, IClientConfiguration config)
	        throws CLIClientException {
        String deprecatedAlternative = defn.getAttr(Constants.PT_ATTR_DEPRECATED);
        if (deprecatedAlternative == null || deprecatedAlternative.isEmpty()) {
            throw StatusHelper.argSyntax(NLS.bind(Messages.HelpCmd_SHOW_DEPRECATED_GENERIC_MESSAGE,
                    Constants.SHOW_DEPRECATED_HELP, Constants.SHOW_DEPRECATED_HELP_VALUE), config);
        } else if (defn.isParent()) {
            String[] deprecatedAlternatives = deprecatedAlternative.split(" ");
            StringBuffer alternatives = new StringBuffer();
            for (String alternative : deprecatedAlternatives) {
                if (alternatives.length() != 0) {
                    alternatives.append(", ");
                }                           
                alternatives.append(alternative);
            }
            throw StatusHelper.argSyntax(NLS.bind(Messages.HelpCmd_SHOW_DEPRECATED_PARENT_MESSAGE,
                    new String[] {alternatives.toString(), Constants.SHOW_DEPRECATED_HELP,
                    Constants.SHOW_DEPRECATED_HELP_VALUE}), config);
        } else {
            throw StatusHelper.argSyntax(NLS.bind(Messages.HelpCmd_SHOW_DEPRECATED_COMMAND_MESSAGE,
                    new String[] {deprecatedAlternative, Constants.SHOW_DEPRECATED_HELP,
                    Constants.SHOW_DEPRECATED_HELP_VALUE}), config);
        }
	}
	

	/** Display an individual subcommand. 
	 * @param config */
	private void showSubcommand(IClientConfiguration config, List<String> cmdNames, ISubcommandDefinition defn)
			throws CLIClientException {
		String cmdName = cmdNames.get(cmdNames.size()-1);
		
		Options options;
		try {
			options = defn.getOptions();
		} catch (LoadOptionsException e) {
			throw StatusHelper.failure(NLS.bind(Messages.HelpCmd_16, cmdName,
					e.getCause().getMessage()), e);
			
		} catch (ConflictingOptionException e) {
			throw StatusHelper.failure(NLS.bind(Messages.HelpCmd_17, cmdName,
					e.getMessage()), e);
		}

		// Header
		{
			List<ISubcommandDefinition> ancestors = defn.getAncestors();
			config.getContext().stdout().print(Messages.HelpCmd_19);
			config.getContext().stdout().print(config.getContext().getAppName() + " "); //$NON-NLS-1$
			
			for (ISubcommandDefinition ancestor : ancestors) {
				config.getContext().stdout().print(ancestor.getFullSubcommandName() + " "); //$NON-NLS-1$
			}
			
			config.getContext().stdout().print(defn.getFullSubcommandName());
			
			config.getContext().stdout().println();
			config.getContext().stdout().println();
		}
		

		// Long help (if available)
		String longHelp = options.getLongHelp();
		if (longHelp == null) {
			config.getContext().stdout().println(NO_HELP_STRING);
		} else {
			String purposeLine1 = NLS.bind(Messages.HelpCmd_34, longHelp);
			if (purposeLine1.length() < termWidth) {
				config.getContext().stdout().println(purposeLine1);
			} else {
				purposeLine1 =  NLS.bind(Messages.HelpCmd_34, ""); //$NON-NLS-1$
				config.getContext().stdout().println(purposeLine1);
				
				Collection<String> lines = StringUtil.wrap(longHelp, termWidth);
				for (String line : lines) {
					config.getContext().stdout().print("  "); //$NON-NLS-1$
					config.getContext().stdout().println(line);
				}
			}
		}

		config.getContext().stdout().println();

		String[] aliases = defn.getAliases();
		if (aliases.length > 0) {
            StringBuffer parentCommands = new StringBuffer();
            {
    			boolean trailing = false;
    			for (int i = 0; i < cmdNames.size() - 1 ; i++) {
    				if(trailing)
    				{
    					parentCommands.append(Messages.HelpCmd_13);
    				}
    				trailing = true;
    				parentCommands.append(cmdNames.get(i));
    			}
            }
			config.getContext().stdout().print(Messages.HelpCmd_21);
			
			boolean trailing = false;
			for (String alias : aliases) {
				if (trailing) {
					config.getContext().stdout().print(Messages.HelpCmd_10);
				}
				trailing = true;
				
				config.getContext().stdout().print(parentCommands.toString().isEmpty()? alias : (parentCommands.toString() + Messages.HelpCmd_13 + alias));
			}
			
			config.getContext().stdout().println();
			config.getContext().stdout().println();
		}
		
		
		// Show the usage prefix ("Usage: scm <subcommand-path> ")
		StringBuffer prefixString = new StringBuffer();

		// Show usage for top level
        boolean helpShown = false;
        if (defn.isRunnable() && (!defn.isDeprecated() || showDeprecatedHelp)) {
            // Show the usage prefix ("Usage: scm <subcommand-path> ")
            showUsagePrefix(defn, config, prefixString, (options.getSupportsNewHelp() == false));

            if(options.getSupportsNewHelp()) {
                showNewUsage(options, prefixString, config, true);
            }
            else {
                showUsage(options, config, true);
            }
            helpShown = true;
        }
		
        // Show parent/child specific usage line
        if (defn.isParent() && hasVisibleChildren(defn)) {
            if (helpShown) {
                config.getContext().stdout().println();
                config.getContext().stdout().println();
            }
            
            // Show the usage prefix ("Usage: scm <subcommand-path> ")
            showUsagePrefix(defn, config, prefixString, (options.getSupportsNewHelp() == false));
            showChildCommands(defn, config);
        }
	}
	
    private boolean hasVisibleChildren(ISubcommandDefinition defn) {
        List<ISubcommandDefinition> children = new ArrayList<ISubcommandDefinition>(defn.getChildren());
        for (ISubcommandDefinition child : children) {
            if (!child.showInHelpListing() || (child.isDeprecated() && !showDeprecatedHelp))
                continue;
            
            return true;
        }
        return false;
    }

	private void showUsagePrefix(ISubcommandDefinition defn, IClientConfiguration config,
	        StringBuffer prefixString, boolean print) {
        prefixString.append(config.getContext().getAppName());
		prefixString.append(Messages.HelpCmd_24);
        if(print) {
            config.getContext().stdout().print(NLS.bind(Messages.HelpCmd_23, config.getContext().getAppName()));
        }

        List<ISubcommandDefinition> ancestors = defn.getAncestors();
        ancestors.add(defn);
        
        for (ISubcommandDefinition ancestor : ancestors) {
            prefixString.append(ancestor.getFullSubcommandName() + Messages.HelpCmd_24);
            if(print) {
                config.getContext().stdout().print(ancestor.getFullSubcommandName() + Messages.HelpCmd_24);
            }
        }
	}
	
	private void showChildCommands(ISubcommandDefinition defn, IClientConfiguration config) {
        List<ISubcommandDefinition> children = new ArrayList<ISubcommandDefinition>(defn.getChildren());
        Collections.sort(children, new SubcommandNameCompator());
        
        config.getContext().stdout().print(Messages.HelpCmd_35 + " "); //$NON-NLS-1$
        config.getContext().stdout().println(Messages.HelpCmd_25);
        
        String longLine = NLS.bind(Messages.HelpCmd_28, config.getContext().getAppName(), defn.getFullSubcommandName());
        Collection<String> lines = StringUtil.wrap(longLine, termWidth);
        for (String line : lines) {
            config.getContext().stdout().println(line);
        }
        config.getContext().stdout().println();

        config.getContext().stdout().println(Messages.HelpCmd_26);
        
        IndentingPrintStream out = new IndentingPrintStream(config.getContext().stdout()).indent();
        
        int maxNameLength = 0;
        for (ISubcommandDefinition child : children) {
            maxNameLength = Math.max(maxNameLength, child.getFullSubcommandName().length());
        }
        
        for (ISubcommandDefinition child : children) {
            if (!child.showInHelpListing() || (child.isDeprecated() && !showDeprecatedHelp))
                continue;

				longLine = NLS.bind(Messages.HelpCmd_27, StringUtil.pad(child.getFullSubcommandName(), maxNameLength), child.getShortHelp());
				lines = StringUtil.wrapAndIndent(longLine, termWidth - out.getIndent());
				for (String line : lines) {
					out.println(line);
			}
		}
	}

	/** 
	 * 
	 */
	private void showNewUsage(Options options, StringBuffer prefixString, IClientConfiguration config, boolean isLeaf) {
		/**
		 * Logic 
		 * First create a structure with top level options
		 * This should include named, mutually exclusive groups and positional options
		 * Get the usage help for this top level options
		 * First look for any continuous groups
		 *   If there are no continuous groups , then print the top level options as is
		 *   Else for each continuous group, get the usage from each of its inner group
		 *   or positional. Finally print this along with the top level options.  
		 *   Then print a new line.
		 */
		IndentingPrintStream out = new IndentingPrintStream(config.getContext().stdout());
		List<StringBuffer> resultOptStrings = OptionGroupUtil.getShortUsageHelp(options);

		if (resultOptStrings.size() > 1) {
			out.println(Messages.HelpCmd_41);
		}
		else {
			prefixString.insert(0, Messages.HelpCmd_41);
		}
			 
		boolean trailing = false;
		for(StringBuffer resultOptString : resultOptStrings) {
			if(trailing) {
				out.println();
			}
			trailing = true;
			resultOptString.insert(0, prefixString);
			Collection<String> lines = StringUtil.wrapAndIndent(resultOptString.toString(), prefixString.toString(), termWidth - out.getIndent());
			for (String line : lines) {					
				out.println(line);
			}
		}
		out.println();

		// Show positional definitions
		IndentingPrintStream pos = out.indent();
		int maxWidth = 0;
		for (IPositionalOptionDefinition opt : options.getPositionalOptions()) {
			maxWidth = Math.max(maxWidth, opt.getName().length());
		}
		
		for (IPositionalOptionDefinition opt : options.getPositionalOptions()) {
			String optionString = StringUtil.pad(opt.getName(), maxWidth) + 
					NLS.bind(Messages.HelpCmd_30, options.getHelpFor(opt.getId()));
			Collection<String> lines = StringUtil.wrapAndIndent(optionString, termWidth - pos.getIndent());
			for (String line : lines) {
				pos.println(line);
			}
		}

		// Show named options
		// Get NamedOptions in the order by which they were added to Options
		Collection<INamedOptionDefinition> namedOptions = options.getNamedOptions();
		if (namedOptions.size() > 0) {
			out.println();
			out.print(Messages.HelpCmd_31); // Prevent the options line from being indented
			out.println();
		}

		out = out.indent();
		
		// Find the width of the names.
		HashMap<INamedOptionDefinition, StringBuffer> names = new HashMap<INamedOptionDefinition, StringBuffer>();
		maxWidth = 0;
		for (INamedOptionDefinition opt : namedOptions) {

			if (opt.isHidden()) {
		        continue;
		    }
		    		    
			StringBuffer buf = new StringBuffer();
			
			// Show the shortopt
			boolean hasShortOpt = false;
			if (opt.getShortOpt() != null) {
				buf.append("-" + opt.getShortOpt()); //$NON-NLS-1$
				hasShortOpt = true;
			}

			// Show the longopt
			if (opt.getLongOpt() != null) {
				String commaSequence  = ", ";
				if (hasShortOpt) {
					buf.append(commaSequence); //$NON-NLS-1$
				}
				else { // Append ' ' multiplied by the number of for a shortOpt + commaSequence 
					int whiteSpaceLength = "-".length() + 1 + commaSequence.length();
					buf.append(StringUtil.pad(" ", whiteSpaceLength));
				}

				buf.append("--" + opt.getLongOpt()); //$NON-NLS-1$

			}

			// Show the args
			if (opt.getArgCount() == Options.ARG_INFINITE) {
				buf.append(Messages.HelpCmd_40);
			}
			else {
				int minArgCount = opt.getMinArgCount();
				int argCount = opt.getArgCount();
				for (int i = 0; i < argCount; i++) {
					if (i < minArgCount) {
						buf.append(Messages.HelpCmd_37);
					}
					else {
						buf.append(Messages.HelpCmd_38);
					}
				}
			}
			
			// Store for later display
			names.put(opt, buf);
			
			// Determine the width
			maxWidth = Math.max(maxWidth, buf.length());
		}		

		maxWidth += 2;
		for (INamedOptionDefinition opt : namedOptions) {
		    if (opt.isHidden()) {
		        continue;
		    }
		    
			String optionString = NLS.bind(Messages.HelpCmd_33,
			        StringUtil.pad(names.get(opt), maxWidth),
			        options.getHelpFor(opt.getId()));
			Collection<String> lines = StringUtil.wrapAndIndent(optionString, termWidth - out.getIndent());
			for (String line : lines) {
				out.println(line);
			}
		}		
	}

	/** Show the "usage" block for non-parent subcommands. */
	private void showUsage(Options options, IClientConfiguration config, boolean isLeaf) {
		IndentingPrintStream out = new IndentingPrintStream(config.getContext().stdout());
		
		if (options.getNamedOptions().size() > 0) {
			out.print(Messages.HelpCmd_29);
		}
		
		for (IPositionalOptionDefinition opt : options.getPositionalOptions()) {
			boolean isRequired = opt.getMinCount() > 0;

			if (isRequired) {
				out.print("<"); //$NON-NLS-1$
			} else {
				out.print("["); //$NON-NLS-1$
			}

			out.print(opt.getName());

			if (opt.getMaxCount() == Options.ARG_INFINITE) {
				out.print("..."); //$NON-NLS-1$
			}

			if (isRequired) {
				out.print(">"); //$NON-NLS-1$
			} else {
				out.print("]"); //$NON-NLS-1$
			}

			out.print(" "); //$NON-NLS-1$
		}

		out.println();
		out.println();
		
		if (!isLeaf) {
			out.println(NLS.bind(Messages.HelpCmd_36, config.getContext().getAppName()));
			out.println();
		}
		
		// Show positional definitions
		IndentingPrintStream pos = out.indent();
		int maxWidth = 0;
		for (IPositionalOptionDefinition opt : options.getPositionalOptions()) {
			maxWidth = Math.max(maxWidth, opt.getName().length());
		}
		
		for (IPositionalOptionDefinition opt : options.getPositionalOptions()) {
			String optionString = StringUtil.pad(opt.getName(), maxWidth) + 
					NLS.bind(Messages.HelpCmd_30, options.getHelpFor(opt.getId()));
			Collection<String> lines = StringUtil.wrapAndIndent(optionString, termWidth - pos.getIndent());
			for (String line : lines) {
				pos.println(line);
			}
		}
		

		// Show named options
		if (options.getNamedOptions().size() > 0) {
			out.println();
			out.print(Messages.HelpCmd_31); // Prevent the options line from being indented
			out.println();
		}

		out = out.indent();
		
		// Find the width of the names.
		HashMap<INamedOptionDefinition, StringBuffer> names = new HashMap<INamedOptionDefinition, StringBuffer>();
		
		maxWidth = 0;
		List<INamedOptionDefinition> namedOptions = new ArrayList<INamedOptionDefinition>(options.getNamedOptions());
		Collections.sort(namedOptions, new NamedOptionComparator());
		for (INamedOptionDefinition opt : namedOptions) {
		    if (opt.isHidden()) {
		        continue;
		    }
		    
			StringBuffer buf = new StringBuffer();
			
			// Show the shortopt
			boolean hasShortOpt = false;
			if (opt.getShortOpt() != null) {
				buf.append("-" + opt.getShortOpt()); //$NON-NLS-1$
				hasShortOpt = true;
			}

			// Show the longopt
			if (opt.getLongOpt() != null) {
				if (hasShortOpt) {
					buf.append(" "); //$NON-NLS-1$
				}

				if (hasShortOpt) {
					buf.append("["); //$NON-NLS-1$
				}

				buf.append("--" + opt.getLongOpt()); //$NON-NLS-1$

				if (hasShortOpt) {
					buf.append("]"); //$NON-NLS-1$
				}

			}

			// Show the args
			if (opt.getArgCount() == Options.ARG_INFINITE) {
				buf.append(" ..."); //$NON-NLS-1$
			}
			else {
				for (int i = opt.getArgCount(); i > 0; i--) {
					buf.append(Messages.HelpCmd_32);
				}
			}
			
			// Store for later display
			names.put(opt, buf);
			
			// Determine the width
			maxWidth = Math.max(maxWidth, buf.length());
		}		

		maxWidth += 2;
		
		
		for (INamedOptionDefinition opt : namedOptions) {
		    if (opt.isHidden()) {
		        continue;
		    }
		    
			String optionString = NLS.bind(Messages.HelpCmd_33,
			        StringUtil.pad(names.get(opt), maxWidth),
			        options.getHelpFor(opt.getId()));
			Collection<String> lines = StringUtil.wrapAndIndent(optionString, termWidth - out.getIndent());
			for (String line : lines) {
				out.println(line);
			}
		}		
	}
}

