/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2016. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.rtc.cli.infrastructure.internal.subcommands;

import java.util.ArrayList;

import java.util.Collection;
import java.util.Dictionary;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import org.eclipse.core.runtime.Platform;
import org.eclipse.osgi.util.NLS;
import org.osgi.framework.Bundle;

import com.ibm.team.rtc.cli.infrastructure.internal.Messages;
import com.ibm.team.rtc.cli.infrastructure.internal.core.CLIClientException;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IClientConfiguration;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IOptionSource;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommand;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommandDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.util.IndentingPrintStream;
import com.ibm.team.rtc.cli.infrastructure.internal.util.StringUtil;
import com.ibm.team.rtc.cli.infrastructure.internal.util.SubcommandUtil;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.Options;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.ConflictingOptionException;

/**
 * Display subcommand versions.
 */
public class VersionCmd implements ISubcommand, IOptionSource {
	//Map<Map<namespace, version>, sub-Commands>
	private Map<Map<String, String>, String> commandList = new HashMap<Map<String, String>, String>();
	public Options getOptions() throws ConflictingOptionException {
		Options options = new Options(false);
		
		options.setLongHelp(Messages.VersionCmd_0);
		
		return options;
	}

    public void initialize(IClientConfiguration config) throws CLIClientException {
    }

    public int run(IClientConfiguration config)
			throws CLIClientException {
    	//Get the commands and populate in "commandList"
    	getCommands(config);
    	
    	//Display the commands
    	displayCommands(config, commandList);
    	
    	return 0;
	}
    /**
     * Fetch the bundle name,version and corresponding sub-commands
     * @return Map (Map (bundleName -> bundleVersion) -> sub-commands)
     * @throws CLIClientException
     */
    public Map<Map<String, String>, String> getCommands(IClientConfiguration config) throws CLIClientException {
    	// Show the plugin versions
		Map<String, List<ISubcommandDefinition>> grouped = new HashMap<String, List<ISubcommandDefinition>>();
		
		for (ISubcommandDefinition defn : config.getSubcommandDefinitions()) {
			List<ISubcommandDefinition> list = grouped.get(defn.getModuleName());
			
			if (list == null) {
				list = new ArrayList<ISubcommandDefinition>();
				grouped.put(defn.getModuleName(), list);
			}
			
			list.add(defn);
		}
		
		for (Map.Entry<String, List<ISubcommandDefinition>> entry : grouped.entrySet()) {
			getVersion(entry.getKey(), entry.getValue(), config);
		}
		
		return commandList;
    }

	/**
	 * Populates "commandList" with the bundle name,version and corresponding sub-commands 
	 * @return
	 */
	@SuppressWarnings("unchecked")
	private void getVersion(String bundleName, List<ISubcommandDefinition> defns, IClientConfiguration config) {
		Bundle bundle = Platform.getBundle(bundleName);
		
		// Get the bundle version
		String version = Messages.VersionCmd_2;
		{
			Dictionary<String, String> headers = bundle.getHeaders();

			final String VERSION_KEY = "bundle-version"; // Already lowercase'd //$NON-NLS-1$

			for (Enumeration<String> i = headers.keys(); i.hasMoreElements();) {
				String candidate = i.nextElement();

				if (VERSION_KEY.equals(candidate.toLowerCase())) {
					version = headers.get(candidate);
					break;
				}
			}
		}
		
		String namespace = bundle.getSymbolicName();
		
		Map<String, String> key = new HashMap<String, String>();
		key.put(namespace, version);
		StringBuffer buf = new StringBuffer();
		
		for (ISubcommandDefinition defn : defns) {
			if (buf.length() > 0) {
				buf.append(", "); //$NON-NLS-1$
			}
			
			for (ISubcommandDefinition ancestor : defn.getAncestors()) {
				buf.append(ancestor.getFullSubcommandName());
				buf.append('/');
			}
			
			// Generate the list of parents
			buf.append(defn.getFullSubcommandName());
		}
		commandList.put(key, buf.toString());
		
	}
	
	/**
	 * Displays the bundle name,version and corresponding sub-commands
	 */	
	public void displayCommands(IClientConfiguration config, Map<Map<String, String>, String> commandList){
		for (Map.Entry<Map<String, String>, String> entry : commandList.entrySet()) {
			
			if (entry.getKey() == null || entry.getValue() == null) {
		    	continue;
		    }
			//Display the namespace and bundle version.
			for (Entry<String, String> entry1 : entry.getKey().entrySet()) {
					config.getContext().stdout().println(NLS.bind(Messages.VersionCmd_3, entry1.getKey(), entry1.getValue()));
			} 
	    	config.getContext().stdout().println(Messages.VersionCmd_5);
	    	//Displays the sub-commands.
	    	IndentingPrintStream out = new IndentingPrintStream(config.getContext().stdout(), 2);
			Collection<String> lines = StringUtil.wrap(entry.getValue().toString(), SubcommandUtil.getTerminalWidth(config) - out.getIndent());
			for (String line : lines) {
				out.println(line);
			}
			config.getContext().stdout().println();
    	}
	}
	
}
