/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.rtc.cli.infrastructure.internal.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import com.ibm.team.rtc.cli.infrastructure.internal.Messages;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.ContinuousGroup;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.INamedOptionDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.IOptionGroup;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.IPositionalOptionDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.MutuallyExclusiveGroup;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.Options;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.SimpleGroup;

public class OptionGroupUtil {
	
    public final static String HelpCmdHtmlArgsFinite = " &lt;arg&gt;"; //$NON-NLS-1$
    public final static String HelpCmdHtmlArgsInfinite = " &lt;arg ...&gt;"; //$NON-NLS-1$
    
	public static boolean AllAreRequired(SimpleGroup grp) {
		for (Pair<INamedOptionDefinition,Boolean> def : grp.getNamedOptionsWithFlag()) {
			Boolean isRequired = def.getSecond();
			if(!isRequired)
				return false;
		}
		for (IPositionalOptionDefinition positionalDef: grp.getPositionalOptions()) {
			if (!positionalDef.isRequired())
				return false;
		}
		return true;
	}
	
	public static boolean AllAreRequired(MutuallyExclusiveGroup grp) {
		for (SimpleGroup simpleGroup : grp.getSimpleGroups()) {
			if(!AllAreRequired(simpleGroup))
				return false;
		}
		return true;
	}
	
	public static boolean AllAreOptional(SimpleGroup grp) {
		return !grp.isRequired();
	}
	
	public static boolean AllAreOptional(MutuallyExclusiveGroup grp) {
		for (SimpleGroup simpleGroup : grp.getSimpleGroups()) {
			Boolean optional = AllAreOptional(simpleGroup);
			if(!optional)
				return false;
		}
		return true;
	}
	
	public static boolean AllInnerGroupsAreSingleton(MutuallyExclusiveGroup grp) {
		for (SimpleGroup simpleGroup : grp.getSimpleGroups()) {
			if (simpleGroup.size() > 1)
				return false;
		}
		return true;
	}

	public static void continuousGroupPrinter(ContinuousGroup grp, boolean escapeHtml, StringBuffer buf) {
		Collection<IOptionGroup> optionGroups = grp.getOptions();
		
		String openingTag = "";
		String closingTag = "";
		// If the continuousGroup is required, then print out ()
		if (grp.isRequired()) {
			openingTag = "(";
			closingTag = ")";
		}
		
		// Append the opening tag to start with
		buf.append(openingTag);

		// Print the groups in the continuous group
		boolean trailing = false;
		for (IOptionGroup group : optionGroups) {
			if(trailing)
				buf.append(Messages.HelpCmd_24);
			trailing = true;
			
			if (group instanceof MutuallyExclusiveGroup) {
			    mutuallyExclusiveGroupPrinter((MutuallyExclusiveGroup)group, escapeHtml,  buf);
			} else if (group instanceof SimpleGroup) {
			    SimpleGroup simpleGroup = (SimpleGroup) group;
			    boolean printOuterTags = (simpleGroup.size() > 1);
	            simpleGroupPrinter(simpleGroup, false, printOuterTags, escapeHtml, buf);			    
			}
		}
		
		// Append the closing tag before leaving.
		buf.append(closingTag);
	}

	public static void mutuallyExclusiveGroupPrinter(MutuallyExclusiveGroup grp, boolean escapeHtml, StringBuffer buf) {
		Collection<SimpleGroup> simpleGroups = grp.getSimpleGroups();
		// Decide whether to show [] or ()
		boolean trailing = false;
		String openingTag = OptionGroupUtil.AllAreOptional(grp)? "[" : "(";
		String closingTag = OptionGroupUtil.AllAreOptional(grp)? "]" : ")";
		buf.append(openingTag);
		boolean canSkipOptionalTag = OptionGroupUtil.AllInnerGroupsAreSingleton(grp);
		for (SimpleGroup simpleGroup : simpleGroups) {
			if(trailing)
				buf.append(" | ");
			trailing = true;
			simpleGroupPrinter(simpleGroup, canSkipOptionalTag, false, escapeHtml, buf);
		}
		buf.append(closingTag);
	}
	
	public static void simpleGroupPrinter(SimpleGroup simpleGroup, boolean canSkipInnerOptionalTag, boolean printOuterTags, boolean escapeHtml, StringBuffer buf) {
		Collection<Pair<INamedOptionDefinition, Boolean>> namedOptions = simpleGroup.getNamedOptionsWithFlag();
		Collection<IPositionalOptionDefinition> positionalOptions = simpleGroup.getPositionalOptions(); 
		String openingTag = (printOuterTags)?(!simpleGroup.isRequired()? "[" :"("):"";
		String closingTag = (printOuterTags)?(!simpleGroup.isRequired()?"]":")"):"";
		buf.append(openingTag);
		boolean trailing = false;
		for (Pair<INamedOptionDefinition, Boolean> namedOptWithFlag : namedOptions)
		{
			if(trailing)
				buf.append(Messages.HelpCmd_24);
			trailing = true;
			
			INamedOptionDefinition def = namedOptWithFlag.getFirst();
			Boolean isRequired = namedOptWithFlag.getSecond();
			
			optionPrinter(def, isRequired, canSkipInnerOptionalTag, escapeHtml, buf);			
		}
		
		if (namedOptions.size() > 0 && positionalOptions.size() > 0) {
		   buf.append(Messages.HelpCmd_24);
		}
		
		trailing = false;
		for (IPositionalOptionDefinition positionalOption : positionalOptions)
		{
			if(trailing)
				buf.append(Messages.HelpCmd_24);
			trailing = true;

			optionPrinter(positionalOption, buf,  escapeHtml);
		}
		buf.append(closingTag);
	}

	public static void optionPrinter(INamedOptionDefinition opt, boolean isRequired, boolean canSkipOptionalTag, boolean escapeHtml, StringBuffer buf)
	{
	    if (opt.isHidden()) {
	        return;
	    }
	    
		// Decide whether to print [] around the option
	    String openingTag = "";
	    String closingTag = "";
	    if (!isRequired && !canSkipOptionalTag) {
	    	openingTag = "[";
	    	closingTag = "]";
	    }
	    	
	    buf.append(openingTag);

	    // Show the shortopt
		boolean hasShortOpt = false;
		if (opt.getShortOpt() != null) {
			buf.append("-" + opt.getShortOpt()); //$NON-NLS-1$
			hasShortOpt = true;
		}

		// Show the longopt
		if (opt.getLongOpt() != null) {
			if (hasShortOpt) {
				buf.append("/"); //$NON-NLS-1$
			}
			buf.append("--" + opt.getLongOpt()); //$NON-NLS-1$
		}

		// Show the args
		if (opt.getArgCount() == Options.ARG_INFINITE) {
			if (escapeHtml) {
				buf.append(HelpCmdHtmlArgsInfinite);
			}
			else {
				buf.append(Messages.HelpCmd_40); 
			}
		}
		else {
			int minArgCount = opt.getMinArgCount();
			int argCount = opt.getArgCount();
			for (int i = 0; i < argCount; i++) {
				if (i < minArgCount) {
					if (escapeHtml) {
						buf.append(HelpCmdHtmlArgsFinite);
					}
					else {
						buf.append(Messages.HelpCmd_37);
					}
				}
				else {
					buf.append(Messages.HelpCmd_38);
				}
			}
		}
		
	    buf.append(closingTag);
	}
	
	public static void optionPrinter(IPositionalOptionDefinition opt, StringBuffer buf, boolean escapeHtml)
	{
		boolean isRequired = (opt.getMinCount() > 0) || (opt.isShowAsRequired());
		
		String openingTag = (isRequired)? ((escapeHtml)?"&lt;":"<") : "[";
		String closingTag = (isRequired) ? ((escapeHtml)? "&gt;" : ">"): "]";
		
		buf.append(openingTag);
		buf.append(opt.getName());

		if (opt.getMaxCount() == Options.ARG_INFINITE) {
			buf.append("..."); //$NON-NLS-1$
		}
		
		buf.append(closingTag);
	}	
	
	public static List<StringBuffer> getShortUsageHelp(Options options) {
		return getShortUsageHelp(options.getAllOptions(), options.getTopLevelNamedOptions(), false);
	}
	
	public static List<StringBuffer> getShortUsageHelpForHtml(Options options) {
		return getShortUsageHelp(options.getAllOptions(), options.getTopLevelNamedOptions(), true);
	}
	
	public static List<StringBuffer> getShortUsageHelp(Collection<Object> allOptionsForPrint, HashMap<INamedOptionDefinition, Boolean> topLevelNamedOptions, boolean escapeHtml) {
		List<ContinuousGroup> continuousGrps = new LinkedList<ContinuousGroup>();
		List<StringBuffer> resultHelpTextList = new ArrayList<StringBuffer>();
		StringBuffer topLevelOptHelpText = new StringBuffer();
		boolean trailing = false;
		for (Object opt : allOptionsForPrint) {
			if (opt instanceof INamedOptionDefinition) {
				if (trailing)
					topLevelOptHelpText.append(Messages.HelpCmd_24);
				boolean isRequired = false;
				INamedOptionDefinition namedOpt = (INamedOptionDefinition) opt;
				if (topLevelNamedOptions.containsKey(namedOpt)) {
					isRequired = topLevelNamedOptions.get(namedOpt);
				}
				OptionGroupUtil.optionPrinter(namedOpt, isRequired, false, escapeHtml, topLevelOptHelpText);
			}
			else if (opt instanceof IPositionalOptionDefinition) {
				if (trailing)
					topLevelOptHelpText.append(Messages.HelpCmd_24);
				OptionGroupUtil.optionPrinter((IPositionalOptionDefinition)opt, topLevelOptHelpText, escapeHtml);
			}
			else if (opt instanceof SimpleGroup) {
				if (trailing)
					topLevelOptHelpText.append(Messages.HelpCmd_24);
				// Get the string to be printed for this and add this to the running buffer for top level options
				SimpleGroup simpleGroup = (SimpleGroup) opt;
				OptionGroupUtil.simpleGroupPrinter(simpleGroup, false, true, escapeHtml, topLevelOptHelpText);
			}
			else if (opt instanceof MutuallyExclusiveGroup) {
				if (trailing)
					topLevelOptHelpText.append(Messages.HelpCmd_24);
				// Get the string to be printed for this and add this to the running buffer for top level options
				MutuallyExclusiveGroup mutuallyExclusiveGroup = (MutuallyExclusiveGroup) opt;
				OptionGroupUtil.mutuallyExclusiveGroupPrinter(mutuallyExclusiveGroup, escapeHtml, topLevelOptHelpText);
			}
			else if (opt instanceof ContinuousGroup) {
				continuousGrps.add((ContinuousGroup) opt);
			}
			trailing = true;
		}
		
		if (continuousGrps.size() == 0) {
			resultHelpTextList.add(topLevelOptHelpText);
		}
		else {
			// Once we have the continuous group, print it and append topLevelString to it.
			// Finally add it to resultOptStrings.
			for (ContinuousGroup grp : continuousGrps) {
				StringBuffer contGroupString = new StringBuffer();
				continuousGroupPrinter(grp, escapeHtml, contGroupString);
				if (!contGroupString.toString().isEmpty()) {
					contGroupString.append(Messages.HelpCmd_24);
				}
				contGroupString.append(topLevelOptHelpText);
				resultHelpTextList.add(contGroupString);
			}
		}
		return resultHelpTextList;
	}
}
