/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2013. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.rtc.cli.infrastructure.internal.util;

import java.io.PrintStream;

import java.util.Collection;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.MultiStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.osgi.util.NLS;

import com.ibm.team.rtc.cli.infrastructure.internal.Activator;
import com.ibm.team.rtc.cli.infrastructure.internal.core.CLIClientException;
import com.ibm.team.rtc.cli.infrastructure.internal.core.Constants;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IClientConfiguration;
import com.ibm.team.rtc.cli.infrastructure.internal.Messages;

public class StatusHelper {
	/** 
	 * Create a status indicating that the command-line was bogus.  
	 */
	public static CLIClientException argSyntax(String message, IClientConfiguration config) {
		CharSequence cmdName = SubcommandUtil.getExecutionString(config.getSubcommandDefinition());
		if (cmdName.toString().equals("help")) { //$NON-NLS-1$
			return createException(NLS.bind(Messages.Application_CONCATENATE, message, NLS.bind(Messages.Application_1, config.getContext().getAppName())), Constants.STATUS_ARGUMENT_SYNTAX_ERROR, null);
		} else {
			return createException(NLS.bind(Messages.Application_CONCATENATE, message, NLS.bind(Messages.Application_2, config.getContext().getAppName(), cmdName)), Constants.STATUS_ARGUMENT_SYNTAX_ERROR, null);
		}
	}
	
	/**
	 * Create a status indicating a failure for an unknown reason, likely due to an  
	 * internal problem. 
	 */
	public static CLIClientException failure(String message, Throwable t) {
		return createException(message, Constants.STATUS_FAILURE, t);		
	}

	/**
	 * Indicate that the client has gotten into a confused state. Some internal 
	 * assumptions were invalid. 
	 */
	public static CLIClientException internalError(String message) {
		return createException(message, Constants.STATUS_INTERNAL_ERROR, null);
	}
	
	/**
	 * Indicate that the client has gotten into a confused state. Some internal 
	 * assumptions were invalid. 
	 */
	public static CLIClientException internalError(String message, Exception e) {
		return createException(message, Constants.STATUS_INTERNAL_ERROR, e);
	}
	
	/**
     * Return a number of errors. 
     */
    public static CLIClientException internalError(String message,
            Collection<CLIClientException> exceptions) {
        
        MultiStatus multi = new MultiStatus(Activator.PLUGIN_ID, Constants.STATUS_INTERNAL_ERROR, message, null);
        
        for (CLIClientException exception : exceptions) {
            multi.add(new Status(IStatus.ERROR, Activator.PLUGIN_ID, 
                    Constants.STATUS_INTERNAL_ERROR, exception.getMessage(), 
                    exception));
        }
        
        return new CLIClientException(multi);
    }

    
	public static CLIClientException createException(String message, int errorCode, Throwable t) {
	    return new CLIClientException(
	            new Status(IStatus.ERROR, Activator.PLUGIN_ID, 
	                    errorCode, message, t)
	            );
	}
	
	public static void showStatusException(IStatus status, PrintStream err) {
        // Display a stack trace, since something's gone unexpectedly wrong.
        if (status.isMultiStatus()) {
            if (status.getException() != null) {
                err.println(Messages.Application_24);
                status.getException().printStackTrace(err);
            }
            
            // Gussy up the exception
            MultiStatus multi = (MultiStatus)status;
            if (multi.getChildren().length > 1) {
                err.print(Messages.Application_25);
            }
            err.println(status.getMessage());
            
            if (multi.getException() != null) {
                multi.getException().printStackTrace(err);
            }

            writeChildStatuses(multi, err);
        }
        else {
            if (status.getException() != null) {
                status.getException().printStackTrace(err);
            }
        }
	}

    private static void writeChildStatuses(MultiStatus multi, PrintStream err) {
        // Write the children
        for (IStatus child : multi.getChildren()) {
            if (child.getException() == null) {
                err.println(NLS.bind(Messages.Application_26, child.getMessage()));
            } else {
                child.getException().printStackTrace(err);
            }
            
            err.println();
            err.println();
            
            if (child.isMultiStatus()) {
                writeChildStatuses((MultiStatus)child, err);
            }
        }
    }
}
