/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/

package com.ibm.team.rtc.cli.infrastructure.internal.util;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import org.eclipse.osgi.util.NLS;

import com.ibm.team.rtc.cli.infrastructure.internal.Messages;

/**
 * 
 */
public class StringUtil {

    // Minimum size we will wrap to.
    public static final int MIN_WRAP_LIMIT = 10;

    public static enum PaddingInstruction {
        LEFT_WITH_SPACE, RIGHT_WITH_SPACE, LEFT_WITH_ALTERNATING_DOTS
    }

    public static void printTable(IndentingPrintStream out, PaddingInstruction[] alignment,
            CharSequence... strings) {
        final int numColumns = alignment.length;

        CharSequence[] padded = new CharSequence[strings.length];
        int[] columnWidth = new int[numColumns];
        for (int i = 0; i < strings.length; ++i) {
            columnWidth[i % numColumns] = Math
                    .max(columnWidth[i % numColumns], strings[i].length());
        }

        for (int i = 0; i < strings.length; ++i) {
            PaddingInstruction pi = alignment[i % numColumns];

            switch (pi) {
            case LEFT_WITH_SPACE:
                // Left align
                if ((i % numColumns) != (numColumns - 1)) {
                    padded[i] = pad(strings[i], columnWidth[i % numColumns]);
                } else {
                    padded[i] = strings[i];
                }
                break;

            case RIGHT_WITH_SPACE:
                // Right align
                padded[i] = padLeft(strings[i], columnWidth[i % numColumns], ' ');
                break;

            case LEFT_WITH_ALTERNATING_DOTS:
                if ((i % numColumns) != (numColumns - 1)) {
                    char padChar = ' ';
                    if ((i % 2) == 1) {
                        padChar = '.';
                    }
                    padded[i] = pad(strings[i], columnWidth[i % numColumns], padChar);
                } else {
                    padded[i] = strings[i];
                }
                break;

            default:
                throw new IllegalArgumentException("Unknown alignment instruction: " + pi); //$NON-NLS-1$
            }
        }

        printTable(out, numColumns, false, padded);
    }

    public static void printTable(IndentingPrintStream out, int numColumns, boolean pad,
            CharSequence... strings) {
        if (pad) {
            CharSequence[] padded = new CharSequence[strings.length];
            int[] columnWidth = new int[numColumns];
            for (int i = 0; i < strings.length; ++i) {
                columnWidth[i % numColumns] = Math.max(columnWidth[i % numColumns],
                        strings[i].length());
            }

            for (int i = 0; i < strings.length; ++i) {
                if ((i % numColumns) == numColumns - 1) {
                    padded[i] = strings[i];
                } else {
                    padded[i] = pad(strings[i], columnWidth[i % numColumns]);
                }

            }

            printTable(out, numColumns, false, padded);
        } else {
            for (int i = 0; i < strings.length; i += numColumns) {
                StringBuilder builder = new StringBuilder();
                for (int j = i; j < i + numColumns; ++j) {
                    builder.append(strings[j]);
                    builder.append(' ');
                }
                out.println(builder.toString());
            }
        }
    }

    /**
     * Returns a padded version of the given string
     */
    public static CharSequence pad(CharSequence str, int width, char c) {
        StringBuffer buf = new StringBuffer(str);

        for (int i = str.length(); i < width; i++) {
            buf.append(c);
        }

        return buf;
    }

    /**
     * Returns a padded version of the given string
     */
    public static CharSequence padLeft(CharSequence str, int width, char c) {
        StringBuffer buf = new StringBuffer();

        for (int i = str.length(); i < width; i++) {
            buf.append(c);
        }

        buf.append(str);

        return buf;
    }

    public static CharSequence pad(CharSequence str, int width) {
        return pad(str, width, ' ');
    }

    /**
     * Escape all instances of the given character in <code>toEscape</code> by
     * prefixing it with <code>escapeChar</code>.
     * 
     * @param str
     *            The string with characters to escape.
     * 
     * @param toEscape
     *            A mapping of items to escape with their escaped characters.
     *            The key is the value to escape (by prefixing the escapeChar),
     *            and the value is the item to replace it with.
     * 
     * @param escapeChar
     *            The character to prefix onto illegal characters.
     */
    public static String escape(String str, Map<Character, Character> toEscape, char escapeChar) {
        StringBuffer buf = new StringBuffer(str.length());

        for (char c : str.toCharArray()) {
            // Test to see if the character is in the list of items to escape
            Character item = toEscape.get(Character.valueOf(c));

            if (item == null) {
                buf.append(c);
            } else {
                buf.append(escapeChar);
                buf.append(item);
            }
        }

        return buf.toString();
    }

    /**
     * Remove the escape character from the given string.
     * 
     * @param str
     * @param escapeChar
     */
    public static String unescape(String str, Map<Character, Character> toEscape, char escapeChar) {
        StringBuffer buf = new StringBuffer(str.length());

        // Invert the toEscape table
        HashMap<Character, Character> escaped = new HashMap<Character, Character>();
        for (Map.Entry<Character, Character> entry : toEscape.entrySet()) {
            if (escaped.containsKey(entry.getValue())) {
                throw new IllegalArgumentException(
                        NLS.bind(Messages.StringUtil_0, entry.getValue()));
            }

            escaped.put(entry.getValue(), entry.getKey());
        }

        // Remove the escape
        boolean isEscaped = false;
        for (char c : str.toCharArray()) {
            Character item = Character.valueOf(c);

            if (isEscaped) {
                if (!escaped.containsKey(item)) {
                    throw new IllegalArgumentException(NLS.bind(Messages.StringUtil_1, item));
                }

                buf.append(escaped.get(item));
                isEscaped = false;

            } else {
                if (c == escapeChar) {
                    isEscaped = true;
                } else {
                    buf.append(c);
                }
            }

        }

        return buf.toString();
    }

    private static final Map<Character, Character> PATH_ESCAPE_MAPPING;

    static {
        PATH_ESCAPE_MAPPING = new HashMap<Character, Character>();

        PATH_ESCAPE_MAPPING.put('/', '/'); // Field terminator
    }

    /**
     * Split a path into segments.
     * 
     * @param str
     */
    public static String[] splitEscapedPath(String str) {
        List<String> segments = splitEscaped(str, PATH_ESCAPE_MAPPING, '/', '\\');

        // Collapse empty segments
        for (Iterator<String> i = segments.iterator(); i.hasNext();) {
            String segment = i.next();

            assert segment != null;

            if (segment.length() == 0) {
                i.remove();
            }
        }

        return segments.toArray(new String[] {});
    }

    /**
     * @param str
     * @param splitOn
     * @param escapeChar
     */
    public static List<String> splitEscaped(String str, Map<Character, Character> toEscape,
            char splitOn, char escapeChar) {
        LinkedList<String> toReturn = new LinkedList<String>();

        StringBuffer accumulator = new StringBuffer(str.length());

        boolean isEscaped = false;
        for (char c : str.toCharArray()) {

            if (isEscaped) {
                accumulator.append(c);
                isEscaped = false;
            } else {
                if (c == escapeChar) {
                    isEscaped = true;
                    accumulator.append(c);
                } else if (c == splitOn) {
                    toReturn.add(unescape(accumulator.toString(), toEscape, escapeChar));

                    accumulator = new StringBuffer(str.length());
                } else {
                    accumulator.append(c);
                }
            }
        }

        toReturn.add(accumulator.toString());

        return toReturn;
    }

    /**
     * Concatenates the elements in the string, providing a minimum of extra
     * text in the result. The elements are {@link T#toString()} method is used
     * to get the human-meaningful string representation.
     */
    public static <T> CharSequence listToString(List<T> items) {
        StringBuffer buf = new StringBuffer();

        for (T item : items) {
            if (buf.length() > 0) {
                buf.append(", "); //$NON-NLS-1$
            }

            buf.append(item.toString());
        }

        return buf;
    }

    public static final char PATH_SEPARATOR = '/';

    /**
     * Converts the given set of path segments into a string.
     * 
     * @param segments
     */
    public static String createPathString(String[] segments) {
        StringBuffer sb = new StringBuffer();

        for (String segment : segments) {
            sb.append(PATH_SEPARATOR);
            sb.append(segment);
        }

        return sb.toString();
    }

    public static String createPathString(List<String> segments) {
        String[] path = segments.toArray(new String[segments.size()]);
        return createPathString(path);
    }

    public static String join(String glue, String pieces[]) {
        if (pieces.length == 0) {
            return ""; //$NON-NLS-1$
        }

        StringBuffer buf = new StringBuffer();

        buf.append(pieces[0]);

        for (int i = 1; i < pieces.length; i++) {
            buf.append(glue);
            buf.append(pieces[i]);
        }

        return buf.toString();
    }

    /**
     * THIS MAY ONLY BE CALLED FROM LAUNCHERS!
     * 
     * Find the PWD that we're to work with. Normally, this is our user
     * directory, unless we're given a -d/--directory argument.
     * 
     * Do not call this from subcommands. It checks the system properties, so it
     * can only be called from launchers (either the OSGi launcher or the
     * minimal launcher).
     */
    public static String findEffectiveCwd(String[] args) {
        String path = getPathArgument(args, "-d", "--directory"); //$NON-NLS-1$//$NON-NLS-2$

        File result;
        if (path == null) {
            result = new File(System.getProperty("user.dir")); //$NON-NLS-1$
        } else {
            // Make sure the given path is absolute
            result = new File(path);

            if (!result.isAbsolute()) {
                result = new File(System.getProperty("user.dir"), path); //$NON-NLS-1$
            }
        }

        try {
            // TODO: Is canonicalizing the file here correct?
            // If the folder structure is:
            // /dir/target/subfolder
            // /folder/file.txt
            // /folder/symlink (-> dir/target)
            // and I pass in -d /folder/symlink/subfolder, and a command on
            // "../../file.txt" is expected to be resolved?
            result = result.getCanonicalFile();
        } catch (IOException e) {
        }

        return result.getAbsolutePath();
    }

    public static String getPathArgument(String[] args, String shortName, String longName) {
        boolean returnNextValue = false;

        for (String arg : args) {
            if (returnNextValue) {
                return arg;
            }

            if (arg.equals(shortName) || arg.equals(longName)) {
                returnNextValue = true;
            }
        }

        // We could have a malformed argument here (returnNextValue == true),
        // but we can't do much with that, so we'll just pretend the argument
        // doesn't exist. We'll end up connecting to a daemon that will complain
        // about it anyway.

        return null;
    }

    /*
     * Split the string into multiple lines with the provided delim and the
     * length of the line should not exceed the maximum specified length
     */
    public static String[] splitString(String str, String delim, int maxLength) {
        int noOfLines = (str.length() / maxLength) + 1;
        List<String> strArr = new ArrayList<String>(noOfLines);

        if (noOfLines == 1) {
            strArr.add(str);
        } else {
            StringTokenizer token = new StringTokenizer(str, delim);

            StringBuffer line = new StringBuffer();
            String word = null;
            while (token.hasMoreTokens()) {
                word = token.nextToken();

                if (line.length() + word.length() > maxLength) {
                    strArr.add(line.toString());
                    line.delete(0, line.length());
                }
                line.append(word);
                line.append(delim);
            }
            if (line.length() > 0) {
                strArr.add(line.toString());
            }
        }
        return strArr.toArray(new String[strArr.size()]);
    }

    /**
     * A convenience method that will wrap the input to a maximum line size of
     * DEFAULT_WRAP_LIMIT and indent subsequent lines to the " - " character.
     * 
     * @see Collection<String> wrapAndIndent(String text, String allignmentText,
     *      int wrapLimit)
     * @param text
     *            - text to wrap
     * @return A collection of strings wrapped and indented as described in the
     *         comment
     */
    public static Collection<String> wrapAndIndent(String text, int width) {
        return wrapAndIndent(text, " - ", width); //$NON-NLS-1$
    }

    /**
     * Wrap the input text to a particular line size, only breaking on white
     * space. In addition, the text is formatted such that each subsequent line
     * is indented to line up after a particular character sequence.
     * 
     * Tabs in the input are not supported Bidi text is not supported
     * 
     * If the indent string does not lie within the 1st line of text, then no
     * indentation is done. The text is just wrapped to the appropriate wrap
     * size.
     * 
     * Example:
     * 
     * If input is: Name - This will describe a name at great length
     * 
     * And allignemntText is " - ", and wrapLimit is 29
     * 
     * Then Output is: Name - This will describe a name at great length
     * 
     * It is suggested that allignmentText be more than a single character to
     * avoid false matches in the input string
     * 
     * @param text
     *            - The line to wrap and indent
     * @param allignmentText
     *            - The text denoting the start of the indenting
     * @param wrapLimit
     *            - Maximum size of a line
     * @return A collection of strings wrapped and indented as described in the
     *         comment
     */
    public static Collection<String> wrapAndIndent(String text, String allignmentText, int width) {

        // Pragmatic change: On Windows, when a line is displayed that is
        // exactly terminal width
        // windows displays a free new line character. This is NOT the case on
        // Mac and Unix.
        // This means, on windows a caller would have to know this, check the
        // length of the returned
        // lines and make a decision on whether to display it with println (for
        // sizes < width) or print
        // (for sizes == width)
        //
        // Although technically a breach of contract, by wrapping to n-1 we
        // shield the caller from
        // this minutia.
        //
        // This seems like a better choice than having the user pass in
        // terminalWidth - 1, or checking the length of the output

        width--;

        if (text == null || text.length() == 0)
            return new ArrayList<String>();

        if (allignmentText == null)
            return wrap(text, width);

        if (width < MIN_WRAP_LIMIT) {
            Collection<String> lines = new ArrayList<String>();
            lines.add(text);
            return lines;
        }

        Collection<String> lines = new ArrayList<String>();
        if (text.length() <= width) {
            lines.add(text);
            return lines;
        }

        int cursor = 0;
        String line = getWrapLine(text, 0, width);
        lines.add(line);
        cursor += (line.length());

        int indent = text.indexOf(allignmentText);
        if (indent == -1 || indent > width - allignmentText.length()) {
            indent = 0;
        } else {
            indent = indent + allignmentText.length();
        }

        char[] padChars = new char[indent];
        Arrays.fill(padChars, ' ');
        String padString = new String(padChars);

        while (cursor < text.length()) {
            line = getWrapLine(text, cursor, width - indent);
            lines.add(padString + line.trim());
            cursor += (line.length());
        }
        return lines;
    }

    // Return a substring from text that is at most length maxLength but must
    // end on whitespace or eol
    private static String getWrapLine(String text, int start, int maxLength) {
        int end = Math.min(text.length(), start + maxLength);
        if (end < text.length()) {
            while (!Character.isWhitespace(text.charAt(end)) && end > start) {
                end--;
            }
        }

        if (end == start) {
            end = Math.min(text.length(), start + maxLength);
        } else {
            // Make sure new line doesn't start with a space
            end = Math.min(end, text.length());
        }

        return text.substring(start, end);
    }

    /**
     * Wrap the input to lines of at most width, only breaking lines on
     * whitespace
     * 
     * @param text
     *            to wrap
     * @return a collection of appropriately sized lines
     */
    public static Collection<String> wrap(String text, int width) {
        return wrapAndIndent(text, "", width); //$NON-NLS-1$
    }
    
    /**
     * Wrap the input text to a particular line size, only breaking on white
     * space. In addition, the text is formatted such that each subsequent line
     * is indented with the specified number of spaces.
     * 
     * @param text
     *            - The line to wrap and indent
     * @param width
     *            - Maximum size of a line
     * @param indent
     *            - The number of spaces to indent with
     * @return A collection of strings wrapped and indented as described in the
     *         comment
     * @return
     */
    public static Collection<String> wrapAndIndent(String text, int width, int indent) {
        int cursor = 0;
        char[] padChars = new char[indent];
        Arrays.fill(padChars, ' ');
        String padString = new String(padChars);

        String line;
        Collection<String> lines = new ArrayList<String>();
        while (cursor < text.length()) {
            line = getWrapLine(text, cursor, width - indent);
            lines.add(padString + line.trim());
            cursor += (line.length());
        }
        return lines;
    }
}
