/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2005, 2014. All Rights Reserved.
 * 
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp. 
 *******************************************************************************/
package com.ibm.team.rtc.cli.infrastructure.internal.util;

import java.io.PrintStream;

import java.util.Collection;
import java.util.List;

import com.ibm.team.rtc.cli.infrastructure.internal.Messages;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IClientConfiguration;
import com.ibm.team.rtc.cli.infrastructure.internal.core.IExecutionContext;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommandDefinition;
import com.ibm.team.rtc.cli.infrastructure.internal.core.ISubcommandDefinition.LoadOptionsException;
import com.ibm.team.rtc.cli.infrastructure.internal.parser.exceptions.ConflictingOptionException;

/**
 * Utility functions for subcommands. Provides I/O methods, as well filesystem
 * traversal functions.  
 */
public abstract class SubcommandUtil {

    /**
	 * Prints debug timing information. Prints the elapsed time from
	 * application launch.
	 * @param label The label to give this timing information.
	 */
	public static void printDebugTiming(IExecutionContext ctx, String label, PrintStream ps) {
		ps.println(label + ": " + (System.currentTimeMillis() - ctx.getStartTime())); //$NON-NLS-1$
	}
	
	
	public static CharSequence getExecutionString(ISubcommandDefinition defn) {
		List<ISubcommandDefinition> ancestors = defn.getAncestors();
		StringBuffer buf = new StringBuffer();
		
		for (ISubcommandDefinition ancestor : ancestors) {
			buf.append(ancestor.getFullSubcommandName());
			
			buf.append(" "); //$NON-NLS-1$
		}
		
		buf.append(defn.getFullSubcommandName());
		
		return buf;
	}
	
    public static int getTerminalWidth(IClientConfiguration config) {
    	return getTerminalWidth(config.getContext());
    }
    
    public static int getTerminalWidth(IExecutionContext ctx) {
        int columns = 80;
        String columnsString = ctx.environment().get("COLUMNS"); //$NON-NLS-1$
        if (columnsString != null) {
            try {
                columns = Integer.parseInt(columnsString);
            } catch (NumberFormatException e) {
            }
        }
        return columns;
    }
    
	
	public static String getUsagePrefix(ISubcommandDefinition defn, String appName) {
		StringBuffer prefixString = new StringBuffer();
		// Append the word "Usage" before the command/subcommand name
		prefixString.append(Messages.HelpCmd_41);
	    prefixString.append(appName);
		prefixString.append(Messages.HelpCmd_24);	    
	
	    List<ISubcommandDefinition> ancestors = defn.getAncestors();
	    ancestors.add(defn);
	    
	    for (ISubcommandDefinition ancestor : ancestors) {
	        prefixString.append(ancestor.getFullSubcommandName() + Messages.HelpCmd_24);
	    }
	    return prefixString.toString();
	}
	
	
	/***
	 * 
	 * @param termWidth
	 * @param indent
	 * @param prefixString - 
	 * @param usageTextList - obtained from OptionGrouputil
	 * @return
	 * @throws ConflictingOptionException 
	 * @throws LoadOptionsException 
	 */
	public static String getFormattedShortUsageHelp(IExecutionContext ctx, ISubcommandDefinition defn) throws LoadOptionsException, ConflictingOptionException {
		StringBuffer outputString = new StringBuffer();
		String prefixString = getUsagePrefix(defn, ctx.getAppName());
		List<StringBuffer> usageTextList = OptionGroupUtil.getShortUsageHelp(defn.getOptions());
		int termWidth = getTerminalWidth(ctx);
		boolean trailing = false;
		for(StringBuffer usageText : usageTextList) {
			if(trailing) {
				outputString.append(Messages.HelpCmd_39);
			}
			trailing = true;
			usageText.insert(0, prefixString);
			Collection<String> lines = StringUtil.wrapAndIndent(usageText.toString(), prefixString.toString(), termWidth);
			addLinesToBuffer(lines, outputString);
		}
		return outputString.toString();
	}
	
	private static void addLinesToBuffer(Collection<String> lines, StringBuffer outputString) {
		boolean lineTrailing = false;
		for (String line : lines) {	
			if (lineTrailing) {
				outputString.append(Messages.HelpCmd_39);					
			}
			lineTrailing = true;
			outputString.append(line);
		}
	}
}
